/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-bool-array.c - Packed array of booleans.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-bool-array.h"
#include <string.h>         /* memcpy, memset */

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_BOOL_ARRAY_TYPE)

#define BYTES_FOR_BITS(n) (((n) + 7) >> 3)

FluxBoolArray *
flux_bool_array_new (void)
{
  FluxBoolArray *bool_array;

  bool_array = g_new0 (FluxBoolArray, 1);
  flux_bool_array_init (bool_array);

  return bool_array;
}

void
flux_bool_array_init (FluxBoolArray *bool_array)
{
  flux_object_init              (FLUX_OBJECT (bool_array));
  flux_object_set_type          (FLUX_OBJECT (bool_array), FLUX_BOOL_ARRAY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (bool_array), flux_bool_array_finalize);

  bool_array->data = NULL;
  bool_array->size = 0;
}

void
flux_bool_array_finalize (FluxObject *object)
{
  FluxBoolArray *bool_array = FLUX_BOOL_ARRAY (object);

  g_free (bool_array->data);
}

guint
flux_bool_array_get_size (FluxBoolArray *bool_array)
{
  assert_types (bool_array);

  return bool_array->size;
}

void
flux_bool_array_set_size (FluxBoolArray *bool_array, guint len)
{
  assert_types (bool_array);

  bool_array->data = g_realloc (bool_array->data, BYTES_FOR_BITS (len));

  if (BYTES_FOR_BITS (len) > BYTES_FOR_BITS (bool_array->size))
  {
    memset (bool_array->data + BYTES_FOR_BITS (bool_array->size), 0,
            BYTES_FOR_BITS (len) - BYTES_FOR_BITS (bool_array->size));
  }

  bool_array->size = len;
}

#define maybe_expand(bool_array) \
  G_STMT_START { \
    if (BYTES_FOR_BITS (bool_array->size) < BYTES_FOR_BITS (bool_array->size + 1)) \
    { \
      bool_array->data = g_realloc (bool_array->data, BYTES_FOR_BITS (bool_array->size + 1)); \
      *(bool_array->data + BYTES_FOR_BITS (bool_array->size + 1) - 1) = 0; \
    } \
  } G_STMT_END

#define maybe_contract(bool_array) \
  G_STMT_START { \
    if (BYTES_FOR_BITS (bool_array->size) < BYTES_FOR_BITS (bool_array->size + 1)) \
    { \
      bool_array->data = g_realloc (bool_array->data, BYTES_FOR_BITS (bool_array->size)); \
    } \
  } G_STMT_END

void
flux_bool_array_append (FluxBoolArray *bool_array, gboolean val)
{
  assert_types (bool_array);

  maybe_expand (bool_array);
  bool_array->size++;

  if (val)
    bool_array->data [BYTES_FOR_BITS (bool_array->size) - 1] |= 0x80 >> ((bool_array->size - 1) % 8);
  else
    bool_array->data [BYTES_FOR_BITS (bool_array->size) - 1] &= 0xff7f >> ((bool_array->size - 1) % 8);
}

static void
shift_array_right (FluxBoolArray *bool_array, guint insert_index)
{
  gboolean carry_bit;
  guchar   first_byte;
  guchar   first_byte_second_half;
  guchar   first_mask;
  gint     i;

  first_mask = 0xff >> (insert_index % 8);
  first_byte = bool_array->data [insert_index >> 3];
  carry_bit = first_byte & 0x01;

  first_byte_second_half = first_byte & first_mask;
  first_byte &= ~first_mask;
  first_byte_second_half >>= 1;
  first_byte |= first_byte_second_half;

  bool_array->data [insert_index >> 3] = first_byte;

  for (i = (insert_index >> 3) + 1; i < BYTES_FOR_BITS (bool_array->size); i++)
  {
    guchar new_carry_bit;

    new_carry_bit = bool_array->data [i] & 0x01;
    bool_array->data [i] = (bool_array->data [i] >> 1) | (carry_bit << 7);
    carry_bit = new_carry_bit;
  }
}

static void
shift_array_left (FluxBoolArray *bool_array, guint remove_index)
{
  gboolean carry_bit = FALSE;
  guchar   first_byte;
  guchar   first_byte_second_half;
  guchar   first_mask;
  gint     i;

  for (i = BYTES_FOR_BITS (bool_array->size) - 1; i > (remove_index >> 3); i--)
  {
    guchar new_carry_bit;

    new_carry_bit = bool_array->data [i] >> 7;
    bool_array->data [i] = (bool_array->data [i] << 1) | carry_bit;
    carry_bit = new_carry_bit;
  }

  first_mask = 0xff >> (remove_index % 8);
  first_byte = bool_array->data [remove_index >> 3];

  first_byte_second_half = ((first_byte << 1) & first_mask) | carry_bit;
  first_byte &= ~first_mask;
  first_byte |= first_byte_second_half;

  bool_array->data [remove_index >> 3] = first_byte;
}

void
flux_bool_array_prepend (FluxBoolArray *bool_array, gboolean val)
{
  assert_types (bool_array);

  maybe_expand (bool_array);
  bool_array->size++;

  shift_array_right (bool_array, 0);
  bool_array->data [0] |= (val << 7);
}

void
flux_bool_array_insert (FluxBoolArray *bool_array, guint index, gboolean val)
{
  assert_types (bool_array);
  g_return_if_fail (index <= bool_array->size);

  maybe_expand (bool_array);
  bool_array->size++;

  shift_array_right (bool_array, index);

  if (val)
    bool_array->data [index >> 3] |= 0x80 >> (index % 8);
}

void
flux_bool_array_remove (FluxBoolArray *bool_array, guint index)
{
  assert_types (bool_array);
  g_return_if_fail (index < bool_array->size);

  shift_array_left (bool_array, index);

  bool_array->size--;
  maybe_contract (bool_array);
}

gboolean
flux_bool_array_get_value (FluxBoolArray *bool_array, guint index)
{
  assert_types (bool_array);
  g_return_val_if_fail (index < bool_array->size, FALSE);

  if (bool_array->data [index >> 3] & (0x80 >> (index % 8)))
    return TRUE;

  return FALSE;
}

void
flux_bool_array_set_value (FluxBoolArray *bool_array, guint index, gboolean val)
{
  assert_types (bool_array);
  g_return_if_fail (index < bool_array->size);

  if (val)
    bool_array->data [index >> 3] |= (0x80 >> (index % 8));
  else
    bool_array->data [index >> 3] &= ~(0x80 >> (index % 8));
}

void
flux_bool_array_peek_data (FluxBoolArray *bool_array, gconstpointer *data, guint *byte_len)
{
  assert_types (bool_array);

  if (byte_len)
    *byte_len = BYTES_FOR_BITS (bool_array->size);
  if (data)
    *data = bool_array->data;
}

void
flux_bool_array_set_data (FluxBoolArray *bool_array, gconstpointer src, guint bit_len)
{
  assert_types (bool_array);

  g_free (bool_array->data);

  if (src && bit_len)
  {
    bool_array->data = g_malloc (BYTES_FOR_BITS (bit_len));
    memcpy (bool_array->data, src, BYTES_FOR_BITS (bit_len));
    bool_array->size = bit_len;
  }
  else
  {
    bool_array->data = NULL;
    bool_array->size = 0;
  }
}
