/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-cached-file.c - Simple, in-process wrapper cache for seekable fd.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-cached-file.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>        /* lseek, read, write */
#include <fcntl.h>
#include <errno.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>          /* lseek, read, write */
#else
# include <stdio.h>           /* SEEK_SET */
# include <io.h>              /* lseek, read, write */
# include "win32/win32-compat.h"
#endif

#define DEBUG(x)

#define IS_CACHED(cached_file, file_index, cache_index) \
  ((cached_file)->key [cache_index] == (file_index))

#define SET_CACHED(cached_file, file_index, cache_index) \
  ((cached_file)->key [cache_index] = (file_index))

#define SET_UNCACHED(cached_file, cache_index) \
  (cached_file)->key [cache_index] = ~0

#define CACHE_ADDRESS(cached_file, cache_index) \
  ((guchar *) cached_file->data + ((cache_index) * (cached_file)->elem_size))

#define READ_ELEM(cached_file, file_offset, cache_index) \
  safe_read (cached_file, CACHE_ADDRESS (cached_file, cache_index), \
             file_offset, (cached_file)->elem_size)

static void
make_illegal_keys (FluxCachedFile *cache)
{
  gint i;

  for (i = 0; i < cache->elem_count; i++)
    SET_UNCACHED (cache, i);
}

static gboolean
safe_seek (FluxCachedFile *cached_file, guint offset)
{
  if (offset == cached_file->actual_position)
    return TRUE;

  if (lseek (cached_file->fd, offset, SEEK_SET) < 0)
    return FALSE;

  cached_file->actual_position = offset;
  return TRUE;
}

static gboolean
safe_read (FluxCachedFile *cached_file, gpointer dest, guint offset, guint len)
{
  gint     bytes_done = 0;
  gboolean result     = TRUE;

  if (!safe_seek (cached_file, offset))
    return FALSE;

  while (bytes_done < len)
  {
    gint n;

    errno = 0;

    n = read (cached_file->fd, (guchar *) dest + bytes_done, len - bytes_done);
    if (n < 1)
    {
      if ((errno == 0 || errno == ENOENT || errno == EINVAL) && bytes_done > 0)
      {
        /* Allow partial reads at end of file */
        memset ((guchar *) dest + bytes_done, 0, len - bytes_done);
        break;
      }
      if (errno != EINTR)
      {
        DEBUG (g_print ("Read outside file, offset/len=%d/%d, errno=%d (%s)\n",
                        offset, len, errno, strerror (errno)));
        result = FALSE;
        break;
      }
    }
    if (n > 0)
      bytes_done += n;
  }

  cached_file->actual_position += bytes_done;
  return result;
}

static gboolean
safe_write (FluxCachedFile *cached_file, gconstpointer src, guint offset, guint len)
{
  gint     bytes_done = 0;
  gboolean result     = TRUE;

  if (!safe_seek (cached_file, offset))
    return FALSE;

  while (bytes_done < len)
  {
    gint n;

    errno = 0;

    n = write (cached_file->fd, (const guchar *) src + bytes_done, len - bytes_done);
    if (n < 1 && errno != EINTR)
    {
      result = FALSE;
      break;
    }
    if (n > 0)
      bytes_done += n;
  }

  cached_file->actual_position += bytes_done;
  return TRUE;
}

void
flux_cached_file_init (FluxCachedFile *cache, gint fd, guint elem_size, guint elem_count)
{
  flux_object_init              (FLUX_OBJECT (cache));
  flux_object_set_type          (FLUX_OBJECT (cache), FLUX_CACHED_FILE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (cache), flux_cached_file_finalize);

  cache->fd          = fd;

  cache->elem_size   = elem_size;
  cache->elem_count  = elem_count;

  cache->key         = g_new (guint, elem_count);
  cache->data        = g_malloc (elem_size * elem_count);

  cache->peek_buffer = NULL;

  cache->hits        = 0;
  cache->misses      = 0;

  make_illegal_keys (cache);

  lseek (fd, 0, SEEK_SET);
  cache->actual_position   = 0;
  cache->expected_position = 0;

#ifndef G_PLATFORM_WIN32
  fcntl (fd, F_SETFL, 0);
#endif
}

/**
 * flux_cached_file_new: Creates a new FluxCachedFile.
 *
 * @fd: Seekable file descriptor.
 * @elem_size: Bytes per element.
 * @elem_count: Number of elements in the cache.
 * 
 * Creates a new @FluxCachedFile from the file @fd with @elem_size
 * bytes per element and @elem_count elements.
 * 
 * Return value: A new @FluxCachedFile.
 **/
FluxCachedFile *
flux_cached_file_new (gint fd, guint elem_size, guint elem_num)
{
  FluxCachedFile *cache;

  cache = g_new0 (FluxCachedFile, 1);
  flux_cached_file_init (cache, fd, elem_size, elem_num);
  
  return cache;
}

void
flux_cached_file_finalize (FluxObject *cache)
{
  g_free (FLUX_CACHED_FILE (cache)->peek_buffer);
  g_free (FLUX_CACHED_FILE (cache)->data);
  g_free (FLUX_CACHED_FILE (cache)->key);
  close (FLUX_CACHED_FILE (cache)->fd);
}

guint
flux_cached_file_tell (FluxCachedFile *cached_file)
{
  return cached_file->expected_position;
}

gboolean
flux_cached_file_seek (FluxCachedFile *cached_file, guint offset)
{
  cached_file->expected_position = offset;
  return TRUE;
}

gconstpointer
flux_cached_file_peek (FluxCachedFile *cached_file, guint len)
{
  guint offset;
  guint first_offset;               /* First elem's offset into file */
  guint first_internal_offset;      /* Offset into first elem */
  guint first_key;                  /* File index of first elem */
  guint first_index;                /* Cache index of first elem */
  guint key, index, file_offset;
  guint read_len = 0;

  offset = cached_file->expected_position;

  first_internal_offset = (offset % cached_file->elem_size);
  first_key             = (offset / cached_file->elem_size);
  first_offset          = offset - first_internal_offset;
  first_index           = first_key % cached_file->elem_count;

  if (first_index + ((first_internal_offset + len - 1) / cached_file->elem_size) >= cached_file->elem_count)
  {
    /* Data won't be contiguous in cache; make an exception and read to peek_buffer */
    if (cached_file->peek_buffer)
      g_free (cached_file->peek_buffer);

    cached_file->peek_buffer = g_malloc (len);

    if (!flux_cached_file_seek (cached_file, offset))
      return NULL;

    if (!flux_cached_file_read (cached_file, cached_file->peek_buffer, len))
      return NULL;

    return cached_file->peek_buffer;
  }

  key         = first_key;
  index       = first_index;
  file_offset = first_offset;

  while (read_len < first_internal_offset + len)
  {
    if (!IS_CACHED (cached_file, key, index))
    {
      if (!READ_ELEM (cached_file, file_offset, index))
        return NULL;
      SET_CACHED (cached_file, key, index);
    }

    read_len += cached_file->elem_size;

    key++;
    index++;
    file_offset += cached_file->elem_size;
  }

  return (guchar *) cached_file->data + (first_index * cached_file->elem_size) +
    first_internal_offset;
}

gboolean
flux_cached_file_read (FluxCachedFile *cached_file, gpointer dest, guint len)
{
  guint offset;
  guint first_offset;               /* First elem's offset into file */
  guint first_internal_offset;      /* Offset into first elem */
  guint first_key;                  /* File index of first elem */
  guint last_key;                   /* File index of last elem */
  guint first_index;                /* Cache index of first elem */
  guint key, index, file_offset;
  guint copied_len = 0;

  offset = cached_file->expected_position;

  first_internal_offset = (offset % cached_file->elem_size);
  first_key             = (offset / cached_file->elem_size);
  last_key              = (offset + len - 1) / cached_file->elem_size;
  first_offset          = offset - first_internal_offset;
  first_index           = first_key % cached_file->elem_count;

  DEBUG (g_print ("read: ofs=%d len=%d first_key=%d last_key=%d int_ofs=%d\n",
                  offset, len, first_key, last_key, first_internal_offset));

  key         = first_key;
  file_offset = first_offset;

  while (copied_len < len)
  {
    guint avail_elems;
    guint todo_elems;
    guint tocopy_len;
    gint  i;

    avail_elems = (cached_file->elem_count - first_index);
    todo_elems  = last_key - first_key + 1;
    todo_elems  = MIN (todo_elems, avail_elems);

    index = first_index;

    for (i = 0; i < todo_elems; i++)
    {
      DEBUG (g_print ("cache: elem=%d index=%d file_offset=%d ", key, index, file_offset));

      if (!IS_CACHED (cached_file, key, index))
      {
        if (!READ_ELEM (cached_file, file_offset, index))
          return FALSE;
        SET_CACHED (cached_file, key, index);
        DEBUG (g_print ("MISS\n"));
      }
      else
        DEBUG (g_print ("HIT\n"));

      file_offset += cached_file->elem_size;
      key++;
      index++;
    }

    tocopy_len = (todo_elems * cached_file->elem_size) - first_internal_offset;
    tocopy_len = MIN (tocopy_len, len - copied_len);

    memcpy ((guchar *) dest + copied_len,
            (guchar *) cached_file->data + (first_index * cached_file->elem_size) + first_internal_offset,
            tocopy_len);

    copied_len += tocopy_len;

    first_key             = key;
    first_index           = 0;
    first_internal_offset = 0;
  }

  cached_file->expected_position += len;
  return TRUE;
}

gboolean
flux_cached_file_write (FluxCachedFile *cached_file, gconstpointer src, guint len)
{
  guint offset;
  guint first_offset;               /* First elem's offset into file */
  guint first_internal_offset;      /* Offset into first elem */
  guint first_key;                  /* File index of first elem */
  guint last_key;                   /* File index of last elem */
  guint first_index;                /* Cache index of first elem */
  guint key, index, file_offset;
  guint copied_len = 0;

  offset = cached_file->expected_position;

  first_internal_offset = (offset % cached_file->elem_size);
  first_key             = (offset / cached_file->elem_size);
  last_key              = (offset + len - 1) / cached_file->elem_size;
  first_offset          = offset - first_internal_offset;
  first_index           = first_key % cached_file->elem_count;

  DEBUG (g_print ("write: ofs=%d len=%d first_key=%d last_key=%d int_ofs=%d\n",
                  offset, len, first_key, last_key, first_internal_offset));

  key         = first_key;
  file_offset = first_offset;

  if (!safe_write (cached_file, src, offset, len))
    return FALSE;

  while (copied_len < len)
  {
    guint avail_elems;
    guint todo_elems;
    guint tocopy_len;
    gint  i;

    avail_elems = (cached_file->elem_count - first_index);
    todo_elems  = last_key - first_key + 1;
    todo_elems  = MIN (todo_elems, avail_elems);

    index = first_index;

    for (i = 0; i < todo_elems; i++)
    {
      DEBUG (g_print ("cache: elem=%d index=%d file_offset=%d ", key, index, file_offset));

      if (key == first_key || key == last_key)
      {
        if (!IS_CACHED (cached_file, key, index))
        {
          SET_UNCACHED (cached_file, index);
          DEBUG (g_print ("MISS\n"));
        }
      }
      else
      {
        SET_CACHED (cached_file, key, index);
        DEBUG (g_print ("HIT\n"));
      }

      file_offset += cached_file->elem_size;
      key++;
      index++;
    }

    tocopy_len = (todo_elems * cached_file->elem_size) - first_internal_offset;
    tocopy_len = MIN (tocopy_len, len - copied_len);

    memcpy ((guchar *) cached_file->data + (first_index * cached_file->elem_size) + first_internal_offset,
            (guchar *) src + copied_len,
            tocopy_len);

    copied_len += tocopy_len;

    first_key             = key;
    first_index           = 0;
    first_internal_offset = 0;
  }

  cached_file->expected_position += len;
  return TRUE;
}

gint
flux_cached_file_get_fd (FluxCachedFile *cached_file)
{
  return cached_file->fd;
}
