/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-cipher.c - Generic encryption/decryption context.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-types.h"
#include "flux-init.h"
#include "flux-marshaller.h"
#include "flux-elgamal.h"
#include "flux-twofish.h"
#include "flux-cipher.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define D(x)

#define CIPHER_VERSION                   0

#define CIPHER_INVITATION_TYPE_DEFAULT   0
#define CIPHER_CONFIRMATION_TYPE_DEFAULT 1

typedef struct
{
  guint16  version;
  guint16  type;

  guchar  *public_key;
  guint    public_key_len;
}
InvitationM;

typedef struct
{
  guint16  version;
  guint16  type;

  guchar  *encrypted_sym_key;
  guint    encrypted_sym_key_len;
}
ConfirmationM;

static FluxMarshaller *cipher_marshaller = NULL;

static FluxMarshalledStructEntry invitation_m_entries [] =
{
  FLUX_MARSHAL_UINT16      ("version",    InvitationM, version,                    TRUE, 0),
  FLUX_MARSHAL_UINT16      ("type",       InvitationM, type,                       TRUE, 0),
  FLUX_MARSHAL_RAW_DYNAMIC ("public_key", InvitationM, public_key, public_key_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStructEntry confirmation_m_entries [] =
{
  FLUX_MARSHAL_UINT16      ("version",           ConfirmationM, version, TRUE, 0),
  FLUX_MARSHAL_UINT16      ("type",              ConfirmationM, type,    TRUE, 0),
  FLUX_MARSHAL_RAW_DYNAMIC ("encrypted_sym_key", ConfirmationM, encrypted_sym_key, encrypted_sym_key_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStruct cipher_marshal_info [] =
{
  { "cipher-invitation",   sizeof (InvitationM),   FLUX_MARSHALLER_CLEAR, invitation_m_entries },
  { "cipher-confirmation", sizeof (ConfirmationM), FLUX_MARSHALLER_CLEAR, confirmation_m_entries },
  { NULL }
};

static void
init_marshaller (void)
{
  if (!cipher_marshaller)
  {
    cipher_marshaller = flux_marshaller_new ();
    flux_marshaller_register_many (cipher_marshaller, cipher_marshal_info);
  }
}

static void
unref_keys (FluxCipher *cip)
{
  if (cip->sym_key_encrypt)  flux_object_unref (cip->sym_key_encrypt);
  if (cip->sym_key_decrypt)  flux_object_unref (cip->sym_key_decrypt);
  if (cip->asym_key_encrypt) flux_object_unref (cip->asym_key_encrypt);
  if (cip->asym_key_decrypt) flux_object_unref (cip->asym_key_decrypt);
  if (cip->asym_key_sign)    flux_object_unref (cip->asym_key_sign);
  if (cip->asym_key_verify)  flux_object_unref (cip->asym_key_verify);
}

FluxCipher *
flux_cipher_new ()
{
  FluxCipher *cip;

  flux_init_mp ();

  cip = g_new0 (FluxCipher, 1);
  flux_object_init              (FLUX_OBJECT (cip));
  flux_object_set_type          (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (cip), flux_cipher_finalize);

  return cip;
}

void
flux_cipher_finalize (FluxObject *obj)
{
  unref_keys (FLUX_CIPHER (obj));
}

FluxCipher *
flux_cipher_dup (FluxCipher *old)
{
  FluxCipher *new;

  flux_object_check (FLUX_OBJECT (old), FLUX_CIPHER_TYPE);

  new = flux_cipher_new ();

  /* Copy flags */

  new->got_invitation    = old->got_invitation;
  new->sent_confirmation = old->sent_confirmation;
  new->got_confirmation  = old->got_confirmation;

  /* Symmetric keys need to be copied, since their vectors are modified by
   * operations (for all modes except ECB). */

  if (old->sym_key_encrypt)
  {
    if (flux_object_get_type (old->sym_key_encrypt) == FLUX_TWOFISH_KEY_TYPE)
      new->sym_key_encrypt = FLUX_OBJECT (flux_twofish_dup_key (FLUX_TWOFISH_KEY (old->sym_key_encrypt)));
  }

  if (old->sym_key_decrypt)
  {
    if (flux_object_get_type (old->sym_key_decrypt) == FLUX_TWOFISH_KEY_TYPE)
      new->sym_key_decrypt = FLUX_OBJECT (flux_twofish_dup_key (FLUX_TWOFISH_KEY (old->sym_key_decrypt)));
  }

  /* Asymmetric keys are constant (except potentially for their associated
   * RNGs, but RNG state doesn't matter), so we ref them. */

  new->asym_key_encrypt = old->asym_key_encrypt;
  new->asym_key_decrypt = old->asym_key_decrypt;
  new->asym_key_sign    = old->asym_key_sign;
  new->asym_key_verify  = old->asym_key_verify;

  if (new->asym_key_encrypt) flux_object_ref (new->asym_key_encrypt);
  if (new->asym_key_decrypt) flux_object_ref (new->asym_key_decrypt);
  if (new->asym_key_sign)    flux_object_ref (new->asym_key_sign);
  if (new->asym_key_verify)  flux_object_ref (new->asym_key_verify);

  return new;
}

void
flux_cipher_generate (FluxCipher *cip)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  unref_keys (cip);
  cip->got_invitation = cip->sent_confirmation = cip->got_confirmation = FALSE;

  cip->sym_key_encrypt  = FLUX_OBJECT (flux_twofish_generate_key (128));
  cip->sym_key_decrypt  = FLUX_OBJECT (flux_twofish_dup_key (FLUX_TWOFISH_KEY (cip->sym_key_encrypt)));

  D (g_print ("Generated twofish keys.\n"));

  flux_elgamal_generate_keypair (768,
                                 (FluxElgamalPublicKey **) &cip->asym_key_encrypt,
                                 (FluxElgamalSecretKey **) &cip->asym_key_decrypt);

  D (g_print ("Generated elgamal keys.\n"));

  cip->asym_key_sign    = cip->asym_key_decrypt;
  cip->asym_key_verify  = cip->asym_key_encrypt;

  flux_object_ref (cip->asym_key_sign);
  flux_object_ref (cip->asym_key_verify);
}

void
flux_cipher_randomize (FluxCipher *cip)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  if (cip->sym_key_encrypt)
    flux_object_unref (FLUX_OBJECT (cip->sym_key_encrypt));
  if (cip->sym_key_decrypt)
    flux_object_unref (FLUX_OBJECT (cip->sym_key_decrypt));

  cip->sym_key_encrypt  = FLUX_OBJECT (flux_twofish_generate_key (128));
  cip->sym_key_decrypt  = FLUX_OBJECT (flux_twofish_dup_key (FLUX_TWOFISH_KEY (cip->sym_key_encrypt)));
}

void
flux_cipher_export_invitation (FluxCipher *cip, guchar **out, guint *out_len)
{
  InvitationM invitation_m;

  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);
  /* Change this if/when we have more than one asymmetric cipher */
  flux_object_check (cip->asym_key_encrypt, FLUX_ELGAMAL_PUBLIC_KEY_TYPE);
  init_marshaller ();

  invitation_m.version = CIPHER_VERSION;
  invitation_m.type    = CIPHER_INVITATION_TYPE_DEFAULT;

  flux_elgamal_export_public_key (FLUX_ELGAMAL_PUBLIC_KEY (cip->asym_key_encrypt),
                                  &invitation_m.public_key, &invitation_m.public_key_len);

  flux_marshaller_encode_pack (cipher_marshaller, "cipher-invitation", &invitation_m,
                               (gpointer *) out, out_len);

  g_free (invitation_m.public_key);
}

gboolean
flux_cipher_import_invitation (FluxCipher *cip, const guchar *in, guint in_len)
{
  InvitationM *invitation_m = NULL;
  FluxObject  *key;
  gboolean     result = FALSE;

  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);
  g_return_val_if_fail (in != NULL, FALSE);
  g_return_val_if_fail (in_len > 0, FALSE);
  init_marshaller ();

  if (!flux_marshaller_decode_pack (cipher_marshaller, "cipher-invitation",
                                    in, in_len, (gpointer) &invitation_m))
    return FALSE;

  if (invitation_m->version    >  CIPHER_VERSION                 ||
      invitation_m->type       != CIPHER_INVITATION_TYPE_DEFAULT ||
      invitation_m->public_key == NULL)
    goto out;

  key = FLUX_OBJECT (flux_elgamal_import_public_key (invitation_m->public_key, invitation_m->public_key_len));
  if (!key)
    goto out;

  if (cip->asym_key_encrypt)
    flux_object_unref (cip->asym_key_encrypt);
  if (cip->asym_key_verify)
    flux_object_unref (cip->asym_key_verify);
  
  cip->asym_key_encrypt = cip->asym_key_verify = key;
  flux_object_ref (key);  /* Will have 2 refs, which is correct */

  cip->got_invitation = TRUE;
  result = TRUE;

out:
  if (invitation_m)
  {
    if (invitation_m->public_key)
      g_free (invitation_m->public_key);

    g_free (invitation_m);
  }

  return result;
}

void
flux_cipher_export_confirmation (FluxCipher *cip, guchar **out, guint *out_len)
{
  ConfirmationM  confirmation_m;
  guchar        *tf_key,       *tf_iv,     *tf_key_iv, *message;
  guint          tf_key_len,    tf_iv_len,              message_len;
  guint16        tf_key_len_pi, tf_iv_len_pi;

  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);
  g_return_if_fail (cip->got_invitation == TRUE);
  init_marshaller ();

  /* Change this if/when we have more than one pair of ciphers */
  flux_object_check (cip->asym_key_encrypt, FLUX_ELGAMAL_PUBLIC_KEY_TYPE);
  flux_object_check (cip->sym_key_encrypt,  FLUX_TWOFISH_KEY_TYPE);

  /* Export twofish elements */

  flux_twofish_export_key (FLUX_TWOFISH_KEY (cip->sym_key_encrypt), &tf_key, &tf_key_len);
  flux_twofish_export_iv  (FLUX_TWOFISH_KEY (cip->sym_key_encrypt), &tf_iv,  &tf_iv_len);

  /* Combine elements in single block, with header indicating sizes */

  tf_key_iv = g_malloc (sizeof (tf_key_len_pi) + sizeof (tf_iv_len_pi) + tf_key_len + tf_iv_len);

  tf_key_len_pi = tf_key_len;
  tf_iv_len_pi  = tf_iv_len;
  tf_iv_len_pi  = GUINT16_TO_BE (tf_iv_len_pi);
  tf_key_len_pi = GUINT16_TO_BE (tf_key_len_pi);

  memcpy (tf_key_iv,                          &tf_key_len_pi, sizeof (tf_key_len_pi));
  memcpy (tf_key_iv + sizeof (tf_key_len_pi), &tf_iv_len_pi,  sizeof (tf_iv_len_pi));

  memcpy (tf_key_iv + sizeof (tf_key_len_pi) + sizeof (tf_iv_len_pi),              tf_key, tf_key_len);
  memcpy (tf_key_iv + sizeof (tf_key_len_pi) + sizeof (tf_iv_len_pi) + tf_key_len, tf_iv,  tf_iv_len);

  g_free (tf_key);
  g_free (tf_iv);

  /* Encrypt block to produce message */

  flux_elgamal_encrypt (FLUX_ELGAMAL_PUBLIC_KEY (cip->asym_key_encrypt),
                        &message, &message_len,
                        tf_key_iv, sizeof (tf_key_len_pi) + sizeof (tf_iv_len_pi) + tf_key_len + tf_iv_len);

  g_free (tf_key_iv);

  /* Produce final output, with header */

  confirmation_m.version               = CIPHER_VERSION;
  confirmation_m.type                  = CIPHER_CONFIRMATION_TYPE_DEFAULT;
  confirmation_m.encrypted_sym_key     = message;
  confirmation_m.encrypted_sym_key_len = message_len;

  flux_marshaller_encode_pack (cipher_marshaller, "cipher-confirmation", &confirmation_m,
                               (gpointer *) out, out_len);

  /* Permute encryption key's IV */

  if (cip->sym_key_decrypt)
    flux_object_unref (cip->sym_key_decrypt);
  cip->sym_key_decrypt = FLUX_OBJECT (flux_twofish_dup_key (FLUX_TWOFISH_KEY (cip->sym_key_encrypt)));

  flux_twofish_permute_iv (FLUX_TWOFISH_KEY (cip->sym_key_encrypt));

  g_free (message);
  cip->sent_confirmation = TRUE;
}

gboolean
flux_cipher_import_confirmation (FluxCipher *cip, const guchar *in, guint in_len)
{
  ConfirmationM *confirmation_m = NULL;
  guchar        *message = NULL;
  guint          message_len;
  guint16        tf_key_len, tf_iv_len;
  gboolean       result = FALSE;

  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);
  g_return_val_if_fail (in != NULL, FALSE);
  g_return_val_if_fail (in_len > 0, FALSE);
  init_marshaller ();

  if (!flux_marshaller_decode_pack (cipher_marshaller, "cipher-confirmation",
                                    in, in_len, (gpointer) &confirmation_m))
    return FALSE;

  if (confirmation_m->version          >  CIPHER_VERSION                   ||
      confirmation_m->type             != CIPHER_CONFIRMATION_TYPE_DEFAULT ||
      confirmation_m->encrypted_sym_key == NULL)
    goto out;

  /* Decrypt message block */

  flux_elgamal_decrypt (FLUX_ELGAMAL_SECRET_KEY (cip->asym_key_decrypt),
                        &message, &message_len,
                        confirmation_m->encrypted_sym_key, confirmation_m->encrypted_sym_key_len);
  if (!message)
    goto out;

  /* Import key/IV lengths */

  memcpy (&tf_key_len, message,                       sizeof (tf_key_len));
  memcpy (&tf_iv_len,  message + sizeof (tf_key_len), sizeof (tf_iv_len));

  tf_key_len = GUINT16_FROM_BE (tf_key_len);
  tf_iv_len  = GUINT16_FROM_BE (tf_iv_len);

  /* Import encryption key/IV */

  if (cip->sym_key_encrypt)
    flux_object_unref (cip->sym_key_encrypt);

  cip->sym_key_encrypt = FLUX_OBJECT (
    flux_twofish_import_key (message     + sizeof (tf_key_len) + sizeof (tf_iv_len),
                             message_len - sizeof (tf_key_len) - sizeof (tf_iv_len) - tf_iv_len));

  flux_twofish_import_iv (FLUX_TWOFISH_KEY (cip->sym_key_encrypt),
                          message + sizeof (tf_key_len) + sizeof (tf_iv_len) + tf_key_len);

  /* Decryption key's IV is a permutation of the encryption key's */

  if (cip->sym_key_decrypt)
    flux_object_unref (cip->sym_key_decrypt);

  cip->sym_key_decrypt = FLUX_OBJECT (flux_twofish_dup_key (FLUX_TWOFISH_KEY (cip->sym_key_encrypt)));
  flux_twofish_permute_iv (FLUX_TWOFISH_KEY (cip->sym_key_decrypt));

  result = TRUE;

out:
  if (message)
    g_free (message);

  if (confirmation_m)
  {
    if (confirmation_m->encrypted_sym_key)
      g_free (confirmation_m->encrypted_sym_key);

    g_free (confirmation_m);
  }

  return result;
}

void
flux_cipher_encrypt_buffer (FluxCipher *cip, guchar *buf, guint len)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  flux_twofish_encrypt_ofb (FLUX_TWOFISH_KEY (cip->sym_key_encrypt), buf, buf, len);
}

void
flux_cipher_encrypt_buffer_copy (FluxCipher *cip, guchar *out, const guchar *in, guint len)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  flux_twofish_encrypt_ofb (FLUX_TWOFISH_KEY (cip->sym_key_encrypt), out, in, len);
}

void
flux_cipher_decrypt_buffer (FluxCipher *cip, guchar *buf, guint len)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  flux_twofish_decrypt_ofb (FLUX_TWOFISH_KEY (cip->sym_key_decrypt), buf, buf, len);
}

void
flux_cipher_decrypt_buffer_copy (FluxCipher *cip, guchar *out, const guchar *in, guint len)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  flux_twofish_decrypt_ofb (FLUX_TWOFISH_KEY (cip->sym_key_decrypt), out, in, len);
}

gboolean
flux_cipher_is_ready (FluxCipher *cip)
{
  flux_object_check (FLUX_OBJECT (cip), FLUX_CIPHER_TYPE);

  if (cip->sym_key_encrypt && cip->sym_key_decrypt)
    return TRUE;

  return FALSE;
}


/* --- */


#ifdef DEBUG_CIPHER_MAIN

#define CRYPT_SIZE 10240
#define CRYPT_ITERATIONS 100

int main (int argc, char *argv [])
{
  FluxCipher *a, *b;
  guchar *buf1, buf2 [CRYPT_SIZE], buf3 [CRYPT_SIZE];
  guint buf1_len;
  int i;

  for (;;)
  {
    a = flux_cipher_new ();
    b = flux_cipher_new ();

    flux_cipher_generate (a);
    printf (":"); fflush (stdout);
    flux_cipher_generate (b);
    printf ("|"); fflush (stdout);

    flux_cipher_export_invitation (a, &buf1, &buf1_len);
    flux_cipher_import_invitation (b, buf1, buf1_len);
    g_free (buf1);

    flux_cipher_export_confirmation (b, &buf1, &buf1_len);
    flux_cipher_import_confirmation (a, buf1, buf1_len);
    g_free (buf1);

    printf (":"); fflush (stdout);

    for (i = 0; i < CRYPT_ITERATIONS; i++)
    {
      memset (buf2, 'a', CRYPT_SIZE);
      memcpy (buf3, buf2, CRYPT_SIZE);

      flux_cipher_encrypt_buffer (a, buf2, CRYPT_SIZE);
      flux_cipher_decrypt_buffer (b, buf2, CRYPT_SIZE);

      if (memcmp (buf2, buf3, CRYPT_SIZE))
        g_error ("Encrypt/decrypt failed.");
    }

    flux_object_unref (FLUX_OBJECT (a));
    flux_object_unref (FLUX_OBJECT (b));

    printf ("."); fflush (stdout);
  }

  exit (0);
}

#endif
