/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-comm.c - N-ary tree-based protocol layer for FluxIO.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/time.h>
#else
# include <io.h>
# include <winsock2.h>
# include <time.h>
#endif

#include "flux-io.h"
#include "flux-comm.h"

#define D(x)

/* Block format
 * ============
 *
 * Block Header
 * ------------
 *
 * 4 bytes   Total length of all data in block (LE)
 * 2 bytes   flags (LE)
 * `- bit 0  TRUE = Breadth first serialization, FALSE = Infix
 * 
 * For each node
 * -------------
 *
 * 4 bytes   Data size (LE)
 * 4 bytes   Number of children (LE)
 * N bytes   Data (N = [0 .. 2^32 - 1], specified by 'Data size' field)
 * 
 * The root node is the first node transmitted, and is always present.
 * Depending on the serialization order, you may receive the children
 * breadth-first or infix. */

#define BLOCK_NODES_MAX_DEF    256
#define BLOCK_SIZE_MAX_DEF  256000
#define BLOCK_DEPTH_MAX_DEF    256

#define BLOCK_HEADER_SIZE (sizeof (guint32) + sizeof (guint16))
#define NODE_HEADER_SIZE  (sizeof (guint32) + sizeof (guint32))

typedef enum
{
  FLUX_COMM_STATE_NONE,
  FLUX_COMM_STATE_BLOCK_HEADER,
  FLUX_COMM_STATE_NODE_HEADER,
  FLUX_COMM_STATE_NODE_DATA
}
FluxCommState;

typedef struct flux_comm_block_header
{
  guint32 bytes;
  guint16 flags;
}
FluxCommBlockHeader;

typedef struct flux_comm_node_header
{
  guint32 size;
  guint32 children;
}
FluxCommNodeHeader;

static void      enqueue_to      (FluxComm *comm, FluxTree *queue, FluxTree *block, guint16 flags,
                                  FluxCommSentFunc *func, gpointer func_data);
static FluxTree *dequeue_from    (FluxComm *comm, FluxTree *queue, guint16 *flags,
                                  FluxCommSentFunc **func, gpointer *func_data);
static void      read_from_io_cb (FluxIO *io, guint min_len, gpointer data);
static void      write_to_io_cb  (FluxIO *io, guint min_len, gpointer data);

void
flux_comm_init (FluxComm *comm, FluxIO *io)
{
  flux_object_init              (FLUX_OBJECT (comm));
  flux_object_set_type          (FLUX_OBJECT (comm), FLUX_COMM_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (comm), flux_comm_finalize);

  comm->io             = io;
  comm->queue_in       = flux_tree_new ();
  comm->queue_out      = flux_tree_new ();
  comm->sent_func_tree = flux_tree_new ();

  flux_object_ref (FLUX_OBJECT (comm->io));
  flux_io_set_read_func  (comm->io, read_from_io_cb, comm);
  flux_io_set_write_func (comm->io, write_to_io_cb, comm);
  flux_comm_set_limits (comm, BLOCK_NODES_MAX_DEF, BLOCK_SIZE_MAX_DEF, BLOCK_DEPTH_MAX_DEF);
  flux_comm_set_auto_flush (comm, TRUE);
}

FluxComm *
flux_comm_new (FluxIO *io)
{
  FluxComm *comm;

  comm = g_new0 (FluxComm, 1);
  flux_comm_init (comm, io);

  return comm;
}

void
flux_comm_finalize (FluxObject *obj)
{
  FluxComm *comm;

  comm = FLUX_COMM (obj);

  flux_object_unref (FLUX_OBJECT (comm->io));

  if (comm->root_in)  flux_tree_destroy (comm->root_in);   /* NOTE: Set to zero when queued */
  if (comm->root_out) flux_tree_destroy (comm->root_out);  /* NOTE: Set to zero when sent */

  flux_tree_destroy (comm->sent_func_tree);
  flux_tree_destroy (comm->queue_in);
  flux_tree_destroy (comm->queue_out);
}

void
flux_comm_enqueue_full (FluxComm *comm, FluxTree *tree, FluxCommOrder order,
                        FluxCommSentFunc *func, gpointer func_data)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);
  enqueue_to (comm, comm->queue_out, tree, order, func, func_data);

  write_to_io_cb (comm->io, 0, comm);
}

void
flux_comm_enqueue (FluxComm *comm, FluxTree *tree)
{
  flux_comm_enqueue_full (comm, tree, FLUX_COMM_ORDER_BREADTH_FIRST, NULL, NULL);
}

FluxTree *
flux_comm_dequeue (FluxComm *comm)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);
  return dequeue_from (comm, comm->queue_in, NULL, NULL, NULL);
}

void
flux_comm_reset (FluxComm *comm)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);

  if (comm->root_in)
  {
    flux_tree_destroy (comm->root_in);
    comm->root_in = NULL;
  }

  if (comm->root_out)
  {
    flux_tree_destroy (comm->root_out);
    comm->root_out = NULL;
  }

  comm->node_in   = NULL;
  comm->node_out  = NULL;

  comm->state_in  = FLUX_COMM_STATE_NONE;
  comm->state_out = FLUX_COMM_STATE_NONE;
}

FluxIO *
flux_comm_get_io (FluxComm *comm)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);

  return comm->io;
}

void
flux_comm_set_dequeue_func (FluxComm *comm, FluxCommDequeueFunc *func, gpointer data)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);

  comm->dequeue_func      = func;
  comm->dequeue_func_data = data;
}

void
flux_comm_set_limits (FluxComm *comm, gulong nodes, gulong size, gulong depth)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);

  comm->block_nodes_max = nodes;
  comm->block_size_max  = size;
  comm->block_depth_max = depth;
}

void
flux_comm_get_limits (FluxComm *comm, gulong *nodes, gulong *size, gulong *depth)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);

  *nodes = comm->block_nodes_max;
  *size  = comm->block_size_max;
  *depth = comm->block_depth_max;
}

void
flux_comm_set_auto_flush (FluxComm *comm, gboolean auto_flush)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);
  comm->do_auto_flush = auto_flush;
}

void
flux_comm_flush (FluxComm *comm)
{
  flux_object_check (FLUX_OBJECT (comm), FLUX_COMM_TYPE);

  flux_dual_fifo_flush_output (FLUX_DUAL_FIFO (comm->io));
  write_to_io_cb (comm->io, 0, comm);
}

static void
enqueue_to (FluxComm *comm, FluxTree *queue, FluxTree *block, guint16 flags,
            FluxCommSentFunc *func, gpointer func_data)
{
  FluxTree *elem;

  elem = flux_tree_new_with_data (&flags, sizeof (flags));
  flux_tree_add_as_last_child (queue, elem);
  flux_tree_add_as_last_child (elem, block);
  if (func)
  {
    elem = flux_tree_new_with_parent_and_data (elem, &func, sizeof (func));
    if (func_data)
      flux_tree_new_with_parent_and_data (elem, &func_data, sizeof (gpointer));
  }
}

static void
add_sent_func (FluxComm *comm, FluxCommSentFunc *func, gpointer func_data)
{
  FluxTree *elem;

  elem = flux_tree_new_with_parent_and_data (comm->sent_func_tree, &func, sizeof (func));
  if (func_data)
    flux_tree_new_with_parent_and_data (elem, &func_data, sizeof (gpointer));
}

static void
dispatch_sent_funcs (FluxComm *comm)
{
  FluxCommSentFunc *func;
  gpointer func_data;
  FluxTree *t0, *t1;

  for (t0 = flux_tree_get_first_child (comm->sent_func_tree); t0; t0 = flux_tree_get_next (t0))
  {
    flux_tree_data_get_bytes (t0, &func, 0, sizeof (FluxCommSentFunc *));

    t1 = flux_tree_get_first_child (t0);
    if (t1)
      flux_tree_data_get_bytes (t1, &func_data, 0, sizeof (gpointer));
    else
      func_data = NULL;

    comm->recursing = TRUE;
    func (comm, NULL, func_data);
    comm->recursing = FALSE;
  }

  for (t0 = flux_tree_get_first_child (comm->sent_func_tree); t0;
       t0 = flux_tree_get_first_child (comm->sent_func_tree))
  {
    flux_tree_destroy (t0);
  }
}

static FluxTree *
dequeue_from (FluxComm *comm, FluxTree *queue, guint16 *flags,
              FluxCommSentFunc **func, gpointer *func_data)
{
  FluxTree *elem0, *block, *func_node;

  elem0 = flux_tree_get_first_child (queue);
  if (elem0 == NULL)
    return NULL;

  if (flags)
    flux_tree_data_get_bytes (elem0, flags, 0, sizeof (*flags));
  
  block     = flux_tree_get_first_child (elem0);
  func_node = flux_tree_get_last_child  (elem0);

  if (func && func_node != block)
  {
    FluxTree *func_data_node;

    flux_tree_data_get_bytes (func_node, func, 0, sizeof (FluxCommSentFunc *));
    func_data_node = flux_tree_get_first_child (func_node);
    if (func_data && func_data_node)
      flux_tree_data_get_bytes (func_data_node, func_data, 0, sizeof (gpointer));
  }

  flux_tree_detach (block);
  flux_tree_destroy (elem0);
  return block;
}

static FluxTree *
get_next_node_in (FluxComm *comm)
{
  FluxTree *tt0;

  switch (comm->order_in)
  {
    case FLUX_COMM_ORDER_BREADTH_FIRST:
      tt0 = flux_tree_get_next_in_breadth_with_level (comm->node_in, comm->block_in_level,
                                                      comm->block_in_level);
      if (!tt0)
      {
        comm->block_in_level++;
        tt0 = flux_tree_get_next_in_breadth_with_level (comm->root_in, comm->block_in_level, 0);
      }
      break;

    case FLUX_COMM_ORDER_DEPTH_FIRST:
    default:
      tt0 = flux_tree_get_next_infix (comm->node_in, comm->root_in);
      break;
  }

  return tt0;
}

static FluxTree *
get_next_node_out (FluxComm *comm)
{
  FluxTree *tt0;

  switch (comm->order_out)
  {
    case FLUX_COMM_ORDER_BREADTH_FIRST:
      tt0 = flux_tree_get_next_in_breadth_with_level (comm->node_out, comm->block_out_level,
                                                      comm->block_out_level);
      if (!tt0)
      {
        comm->block_out_level++;
        tt0 = flux_tree_get_next_in_breadth_with_level (comm->root_out, comm->block_out_level, 0);
      }
      break;

    case FLUX_COMM_ORDER_DEPTH_FIRST:
    default:
      tt0 = flux_tree_get_next_infix (comm->node_out, comm->root_out);
      break;
  }

  return tt0;
}

static gboolean
read_block_header (FluxComm *comm, guint *data_len)
{
  FluxCommBlockHeader cbh;

  D (g_print ("Trying to read block header. data_len == %u\n", *data_len));

  if (*data_len < BLOCK_HEADER_SIZE)
    return FALSE;

  D (g_print ("Reading block header.\n"));

  /* --- Get the header and set up block --- */

  comm->root_in = comm->node_in = flux_tree_new ();  /* Make root node */
  flux_dual_fifo_read (FLUX_DUAL_FIFO (comm->io), &cbh.bytes, sizeof (cbh.bytes));
  flux_dual_fifo_read (FLUX_DUAL_FIFO (comm->io), &cbh.flags, sizeof (cbh.flags));

  comm->block_in_size = comm->block_in_bytes_left = GUINT32_FROM_LE (cbh.bytes);
  cbh.flags = GUINT16_FROM_LE (cbh.flags);
  comm->order_in = cbh.flags & FLUX_COMM_ORDER_BREADTH_FIRST;
  comm->block_in_level = 0;

  *data_len -= BLOCK_HEADER_SIZE;
  flux_io_set_read_threshold (comm->io, NODE_HEADER_SIZE);
  comm->state_in = FLUX_COMM_STATE_NODE_HEADER;
  return TRUE;
}

static gboolean
read_node_header (FluxComm *comm, guint *data_len)
{
  FluxCommNodeHeader cnh;
  gint i;

  if (*data_len < NODE_HEADER_SIZE)
    return FALSE;

  /* --- Get the header and set up node --- */

  flux_dual_fifo_read (FLUX_DUAL_FIFO (comm->io), &cnh.size, sizeof (cnh.size));
  flux_dual_fifo_read (FLUX_DUAL_FIFO (comm->io), &cnh.children, sizeof (cnh.children));
  comm->node_in_bytes_left = GUINT32_FROM_LE (cnh.size);
  cnh.children = GUINT32_FROM_LE (cnh.children);

  /* Make child nodes, to be filled with data later */
  for (i = 0; i < cnh.children; i++)
    flux_tree_add (comm->node_in, flux_tree_new ());

  *data_len -= NODE_HEADER_SIZE;
  flux_io_set_read_threshold (comm->io, 1);
  comm->state_in = FLUX_COMM_STATE_NODE_DATA;  /* len != 0? */
  return TRUE;
}

static gboolean
read_node_data (FluxComm *comm, guint *data_len)
{
  guint grab_len;
  guchar *data;

  if (*data_len == 0)
    return FALSE;

  /* --- Shuffle data chunk --- */

  grab_len = MIN (comm->node_in_bytes_left, *data_len);

  if (grab_len)
  {
    data = g_malloc (grab_len);
    flux_dual_fifo_read (FLUX_DUAL_FIFO (comm->io), data, grab_len);
    flux_tree_data_append_bytes (comm->node_in, data, grab_len);
    g_free (data);

    comm->node_in_bytes_left -= grab_len;
    comm->block_in_bytes_left -= grab_len;
  }

  if (!comm->node_in_bytes_left)
  {
    comm->node_in = get_next_node_in (comm);

    if (comm->node_in)
    {
      flux_io_set_read_threshold (comm->io, NODE_HEADER_SIZE);
      comm->state_in = FLUX_COMM_STATE_NODE_HEADER;
    }
    else
    {
      /* Queue this block and expect another one */
      enqueue_to (comm, comm->queue_in, comm->root_in, 0, NULL, NULL);
      comm->root_in = NULL;
      flux_io_set_read_threshold (comm->io, BLOCK_HEADER_SIZE);
      comm->state_in = FLUX_COMM_STATE_BLOCK_HEADER;
    }
  }

  *data_len -= grab_len;
  return TRUE;
}

static void
read_from_io_cb (FluxIO *io, guint min_len, gpointer data)
{
  FluxComm *comm;
  guint data_len;
  gboolean success = TRUE;
  FluxTree *tree;

  comm = FLUX_COMM (data);
  data_len = flux_dual_fifo_get_max_read (FLUX_DUAL_FIFO (comm->io));

  while (success)
  {
    switch (comm->state_in)
    {
      case FLUX_COMM_STATE_BLOCK_HEADER:
        success = read_block_header (comm, &data_len);
        break;

      case FLUX_COMM_STATE_NODE_HEADER:
        success = read_node_header (comm, &data_len);
        break;

      case FLUX_COMM_STATE_NODE_DATA:
        success = read_node_data (comm, &data_len);
        break;

      case FLUX_COMM_STATE_NONE:
        comm->state_in = FLUX_COMM_STATE_BLOCK_HEADER;
        break;

      default:
        g_error ("Unknown FluxComm internal state in read callback.");
    }
  }

  if (comm->dequeue_func)
  {
    while ((tree = flux_comm_dequeue (comm)))
      comm->dequeue_func (comm, tree, comm->dequeue_func_data);
  }
}

static gboolean
write_block_header (FluxComm *comm, guint *avail_len)
{
  FluxCommBlockHeader cbh;

  if (*avail_len < BLOCK_HEADER_SIZE)
    return FALSE;

  /* Select next tree from the out queue */
  comm->root_out = comm->node_out =
    dequeue_from (comm, comm->queue_out, &(cbh.flags), &comm->sent_out_func,
                  &comm->sent_out_func_data);

  if (comm->root_out == NULL)
    return FALSE;

  cbh.bytes = comm->block_out_size = comm->block_out_bytes_left =
    flux_tree_data_get_size_all (comm->root_out) + flux_tree_data_get_size (comm->root_out);
  cbh.bytes = GUINT32_TO_LE (cbh.bytes);

  comm->order_out = cbh.flags & FLUX_COMM_ORDER_BREADTH_FIRST;
  cbh.flags = GUINT16_TO_LE (cbh.flags);

  flux_dual_fifo_write (FLUX_DUAL_FIFO (comm->io), &cbh.bytes, sizeof (cbh.bytes));
  flux_dual_fifo_write (FLUX_DUAL_FIFO (comm->io), &cbh.flags, sizeof (cbh.flags));

  comm->block_out_level = 0;
  *avail_len -= BLOCK_HEADER_SIZE;
  flux_io_set_write_threshold (comm->io, NODE_HEADER_SIZE);
  comm->state_out = FLUX_COMM_STATE_NODE_HEADER;
  return TRUE;
}

static gboolean
write_node_header (FluxComm *comm, guint *avail_len)
{
  FluxCommNodeHeader cnh;

  if (*avail_len < NODE_HEADER_SIZE)
    return FALSE;

  /* --- Send the node header --- */

  comm->node_out_bytes_left = flux_tree_data_get_size (comm->node_out);
  cnh.size = GUINT32_TO_LE (comm->node_out_bytes_left);
  cnh.children = GUINT32_TO_LE (flux_tree_count_children (comm->node_out));

  flux_dual_fifo_write (FLUX_DUAL_FIFO (comm->io), &cnh.size, sizeof (cnh.size));
  flux_dual_fifo_write (FLUX_DUAL_FIFO (comm->io), &cnh.children, sizeof (cnh.children));

  *avail_len -= NODE_HEADER_SIZE;
  flux_io_set_write_threshold (comm->io, 1);
  comm->state_out = FLUX_COMM_STATE_NODE_DATA;
  return TRUE;
}

static gboolean
write_node_data (FluxComm *comm, guint *avail_len)
{
  gboolean ret = TRUE;
  guint push_len;
  guchar *data;

  if (*avail_len == 0)
    return FALSE;

  /* --- Shuffle data chunk --- */

  push_len = MIN (comm->node_out_bytes_left, *avail_len);

  data = g_malloc (push_len);
  flux_tree_data_get_bytes (comm->node_out,
                          data,
                          flux_tree_data_get_size (comm->node_out) - comm->node_out_bytes_left,
                          push_len);
  flux_dual_fifo_write (FLUX_DUAL_FIFO (comm->io), data, push_len);
  g_free (data);

  comm->node_out_bytes_left -= push_len;
  comm->block_out_bytes_left -= push_len;
  
  if (!comm->node_out_bytes_left)
  {
    comm->node_out = get_next_node_out (comm);

    if (comm->node_out)
    {
      flux_io_set_write_threshold (comm->io, NODE_HEADER_SIZE);
      comm->state_out = FLUX_COMM_STATE_NODE_HEADER;
    }
    else
    {
      /* Block is finished; prepare to send next */
      flux_io_set_write_threshold (comm->io, BLOCK_HEADER_SIZE);
      comm->state_out = FLUX_COMM_STATE_BLOCK_HEADER;

      if (comm->root_out)
      {
        if (comm->sent_out_func)
        {
          add_sent_func (comm, comm->sent_out_func, comm->sent_out_func_data);
          comm->sent_out_func = NULL;
        }

        flux_tree_destroy (comm->root_out);
        comm->root_out = NULL;
      }
    }
  }

  if (flux_tree_count_children (comm->queue_out) == 0 && comm->do_auto_flush)
    flux_dual_fifo_flush_output (FLUX_DUAL_FIFO (comm->io));

  *avail_len -= push_len;
  return ret;
}

static void
write_to_io_cb (FluxIO *io, guint min_len, gpointer data)
{
  FluxComm *comm;
  guint avail_len;
  gboolean success = TRUE;

  comm = FLUX_COMM (data);
  if (!flux_io_is_connected (comm->io))
    return;

  avail_len = flux_dual_fifo_get_max_write (FLUX_DUAL_FIFO (comm->io));

  while (success)
  {
    switch (comm->state_out)
    {
      case FLUX_COMM_STATE_BLOCK_HEADER:
        success = write_block_header (comm, &avail_len);
        break;

      case FLUX_COMM_STATE_NODE_HEADER:
        success = write_node_header (comm, &avail_len);
        break;

      case FLUX_COMM_STATE_NODE_DATA:
        success = write_node_data (comm, &avail_len);
        break;

      case FLUX_COMM_STATE_NONE:
        comm->state_out = FLUX_COMM_STATE_BLOCK_HEADER;
        break;

      default:
        g_error ("Unknown FluxComm internal state in write callback.");
    }
  }

  if (!comm->recursing)
    dispatch_sent_funcs (comm);
}
