/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-data-cache.c - Simple, flat cache for random access data.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-data-cache.h"
#include <stdlib.h>
#include <string.h>

static void
make_illegal_keys (FluxDataCache *cache)
{
  gint i;

  for (i = 0; i < cache->elem_num; i++)
    cache->key [i] = i + 1;
}

void
flux_data_cache_init (FluxDataCache *cache, guint elem_size, guint elem_num)
{
  flux_object_init              (FLUX_OBJECT (cache));
  flux_object_set_type          (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (cache), flux_data_cache_finalize);

  cache->elem_size = elem_size;
  cache->elem_num  = elem_num;
  cache->key       = g_new    (guint, elem_num);
  cache->data      = g_malloc (elem_size * elem_num);

  make_illegal_keys (cache);
}

/**
 * flux_data_cache_new: Creates a new FluxDataCache.
 * 
 * @elem_size: Bytes per element.
 * @elem_num: Number of elements in the cache.
 * 
 * Creates a new @FluxDataCache with @elem_size bytes per element and
 * @elem_num elements.
 * 
 * Return value: A new @FluxDataCache.
 **/
FluxDataCache *
flux_data_cache_new (guint elem_size, guint elem_num)
{
  FluxDataCache *cache;

  cache = g_new0 (FluxDataCache, 1);
  flux_data_cache_init (cache, elem_size, elem_num);
  
  return cache;
}

void
flux_data_cache_finalize (FluxObject *cache)
{
  g_free (FLUX_DATA_CACHE (cache)->data);
  g_free (FLUX_DATA_CACHE (cache)->key);
}

void
flux_data_cache_flush (FluxDataCache *cache)
{
  flux_object_check (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);

  make_illegal_keys (cache);
  /* #warning FIXME: No need to clear the cache data, I think. */
  memset (cache->data, 0, cache->elem_size * cache->elem_num);
}

gconstpointer
flux_data_cache_insert (FluxDataCache *cache, guint key, gconstpointer data)
{
  guint line;
  gpointer inserted;
  
  flux_object_check (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);

  line = key % cache->elem_num;
  cache->key [line] = key;
  inserted = (guchar *) cache->data + (line * cache->elem_size);

  memcpy (inserted, data, cache->elem_size);
  return inserted;
}

gconstpointer
flux_data_cache_get_ptr (FluxDataCache *cache, guint key)
{
  guint line;

  flux_object_check (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);

  line = key % cache->elem_num;
  if (cache->key [line] == key) 
  {
    cache->hits++;
    return ((guchar *) cache->data + (line * cache->elem_size));
  }

  cache->misses++;
  return NULL;
}

gpointer
flux_data_cache_get_insert_ptr (FluxDataCache *cache, guint key)
{
  guint line;
  gpointer inserted;

  flux_object_check (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);

  line = key % cache->elem_num;
  cache->key [line] = key;
  inserted = (guchar *) cache->data + (line * cache->elem_size);

  return inserted;
}

gboolean
flux_data_cache_get_data (FluxDataCache *cache, guint key, gpointer dest)
{
  guint line;

  flux_object_check (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);

  line = key % cache->elem_num;
  if (cache->key [line] == key)
  {
    cache->hits++;
    memcpy (dest, (guchar *) cache->data + (line * cache->elem_size), cache->elem_size);
    return TRUE;
  }

  cache->misses++;
  return FALSE;
}

void
flux_data_cache_invalidate (FluxDataCache *cache, guint key)
{
  guint line;

  flux_object_check (FLUX_OBJECT (cache), FLUX_DATA_CACHE_TYPE);

  line = key % cache->elem_num;
  cache->key [line] = line + 1;
}
