/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-db-header.c - DB file header.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-db-header.h"
#include <sys/types.h>       /* read, write, lseek */
#include <string.h>          /* memcpy */
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>         /* read, write, lseek */
#else
# include <stdio.h>          /* SEEK_SET */
# include <io.h>             /* read, write, lseek */
# include "win32/win32-compat.h"
#endif

typedef struct
{
  guint8  a;
  guint16 b;
  guint32 c;
  guint64 d;
  gchar   e;
  gfloat  f;
  gdouble g;
  gchar   h [3];
}
PadTest;

FluxDBHeader *
flux_db_header_create (const gchar *type, guint32 version)
{
  FluxDBHeader *header;

  g_return_val_if_fail (type != NULL, NULL);

  header = g_new0 (FluxDBHeader, 1);
  g_strlcpy (header->type, type, FLUX_DB_HEADER_TYPE_SIZE);
  header->version       = version;
  header->byte_order    = G_BYTE_ORDER;
  header->int_size      = sizeof (gint);
  header->long_size     = sizeof (glong);
  header->short_size    = sizeof (gshort);
  header->float_size    = sizeof (gfloat);
  header->double_size   = sizeof (gdouble);
  header->pad_test_size = sizeof (PadTest);

  return header;
}

static void
to_pi (FluxDBHeader *header)
{
  header->version       = GUINT32_TO_BE (header->version);
  header->byte_order    = GUINT32_TO_BE (header->byte_order);
  header->pad_test_size = GUINT16_TO_BE (header->pad_test_size);
}

static void
from_pi (FluxDBHeader *header)
{
  header->version       = GUINT32_FROM_BE (header->version);
  header->byte_order    = GUINT32_FROM_BE (header->byte_order);
  header->pad_test_size = GUINT16_FROM_BE (header->pad_test_size);
}

FluxDBHeader *
flux_db_header_read (gint fd)
{
  FluxDBHeader *header;

  g_return_val_if_fail (fd >= 0, NULL);

  header = g_new0 (FluxDBHeader, 1);

  /* FIXME: Error checking */
  lseek (fd, 0, SEEK_SET);
  read  (fd, header->type,           FLUX_DB_HEADER_TYPE_SIZE);
  read  (fd, &header->version,       sizeof (guint32));
  read  (fd, &header->byte_order,    sizeof (guint32));
  read  (fd, &header->int_size,      sizeof (guint8));
  read  (fd, &header->long_size,     sizeof (guint8));
  read  (fd, &header->short_size,    sizeof (guint8));
  read  (fd, &header->float_size,    sizeof (guint8));
  read  (fd, &header->double_size,   sizeof (guint8));
  read  (fd, &header->pad_test_size, sizeof (guint16));
  read  (fd, header->user,           FLUX_DB_HEADER_USER_SIZE);

  from_pi (header);
  return header;
}

void
flux_db_header_write (FluxDBHeader *header, gint fd)
{
  FluxDBHeader *temp_header;

  g_return_if_fail (header != NULL);
  g_return_if_fail (fd >= 0);

  temp_header = g_new (FluxDBHeader, 1);
  memcpy (temp_header, header, sizeof (FluxDBHeader));
  to_pi (temp_header);

  /* FIXME: Error checking */
  lseek (fd, 0, SEEK_SET);
  write (fd, temp_header->type,           FLUX_DB_HEADER_TYPE_SIZE);
  write (fd, &temp_header->version,       sizeof (guint32));
  write (fd, &temp_header->byte_order,    sizeof (guint32));
  write (fd, &temp_header->int_size,      sizeof (guint8));
  write (fd, &temp_header->long_size,     sizeof (guint8));
  write (fd, &temp_header->short_size,    sizeof (guint8));
  write (fd, &temp_header->float_size,    sizeof (guint8));
  write (fd, &temp_header->double_size,   sizeof (guint8));
  write (fd, &temp_header->pad_test_size, sizeof (guint16));
  write (fd, temp_header->user,           FLUX_DB_HEADER_USER_SIZE);

  g_free (temp_header);
}

void
flux_db_header_free (FluxDBHeader *header)
{
  g_free (header);
}
