/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-disk-chunk.c - Low-overhead same-size disk block allocator.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-disk-chunk.h"
#include <sys/types.h>        /* lseek, read, write, stat */
#include <sys/stat.h>         /* stat */
#include <string.h>           /* memcpy */
#include <errno.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>           /* lseek, read, write, stat */
#else
# include <stdio.h>            /* SEEK_SET */
# include <io.h>               /* lseek, read, write */
# include "win32/win32-compat.h"
#endif

#define DEBUG(x)

#define DB_NAME                "FluxDiskChunk"
#define DB_VERSION             2

#define CACHE_SIZE             (256 * 1024)
#define ROOT_TAG_ARENA_SIZE    4096

#define FIRST_FREE_ATOM_OFFSET FLUX_DB_HEADER_SIZE
#define NO_FREE_ATOMS          0xffffffff

#define ROOT_TAGS_OFFSET       (FIRST_FREE_ATOM_OFFSET + sizeof (FluxItem))
#define DATA_OFFSET            (ROOT_TAGS_OFFSET + ROOT_TAG_ARENA_SIZE)

#define ROOT_TAG_MAX_LEN       255

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_DISK_CHUNK_TYPE)

#define ATOM_OFFSET(disk_chunk, atom) \
  (DATA_OFFSET + ((atom) * (disk_chunk)->atomsize))

#define is_valid_atom(disk_chunk, atom) \
  ((gint) ATOM_OFFSET (disk_chunk, atom) > \
   (gint) (disk_chunk)->file_size - (gint) (disk_chunk)->atomsize ? FALSE : TRUE)

#define invalidate_atom(disk_chunk, atom) \
  flux_data_cache_invalidate (disk_chunk->data_cache, atom);

typedef struct
{
  guchar   tag [ROOT_TAG_MAX_LEN + 1];  /* Tag */
  FluxItem root;                        /* Value */

  gint     arena_offset;                /* Location */
}
RootTag;

struct _DiskChunkFreeNode
{
  FluxItem next;
};

static gboolean
safe_read (gint fd, gpointer dest, gint len)
{
  gint bytes_done = 0;

  while (bytes_done < len)
  {
    gint n;

    n = read (fd, (guchar *) dest + bytes_done, len - bytes_done);
    if (n < 1 && errno != EINTR)
      return FALSE;
    if (n > 0)
      bytes_done += n;
  }

  return TRUE;
}

static gboolean
safe_write (gint fd, gconstpointer src, gint len)
{
  gint bytes_done = 0;

  while (bytes_done < len)
  {
    gint n;

    n = write (fd, (const guchar *) src + bytes_done, len - bytes_done);
    if (n < 1 && errno != EINTR)
      return FALSE;
    if (n > 0)
      bytes_done += n;
  }

  return TRUE;
}

static FluxItem
read_first_free_atom (FluxDiskChunk *disk_chunk)
{
  FluxItem atom;

  lseek (disk_chunk->fd, FIRST_FREE_ATOM_OFFSET, SEEK_SET);
  if (safe_read (disk_chunk->fd, &atom, sizeof (FluxItem)))
    return atom;

  return FLUX_ITEM_NULL;
}

static FluxItem
get_first_free_atom (FluxDiskChunk *disk_chunk)
{
  return disk_chunk->first_free_atom;
}

static gboolean
set_first_free_atom (FluxDiskChunk *disk_chunk, FluxItem atom)
{
  lseek (disk_chunk->fd, FIRST_FREE_ATOM_OFFSET, SEEK_SET);
  if (safe_write (disk_chunk->fd, &atom, sizeof (FluxItem)))
  {
    disk_chunk->first_free_atom = atom;
    return TRUE;
  }

  return FALSE;
}

static gboolean
load_root_tags (FluxDiskChunk *disk_chunk)
{
  guchar arena [ROOT_TAG_ARENA_SIZE];
  gint   i, j;

  lseek (disk_chunk->fd, ROOT_TAGS_OFFSET, SEEK_SET);

  if (!safe_read (disk_chunk->fd, arena, ROOT_TAG_ARENA_SIZE))
    return FALSE;

  if (disk_chunk->tag_table)
    g_hash_table_destroy (disk_chunk->tag_table);

  disk_chunk->tag_table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

  for (i = j = 0; j < ROOT_TAG_ARENA_SIZE - sizeof (FluxItem) - 1 && arena [i]; j++)
  {
    RootTag *root_tag;

    if (arena [j] == '\0')
    {
      if (j - i <= ROOT_TAG_MAX_LEN)
      {
        root_tag = g_new0 (RootTag, 1);
        root_tag->arena_offset = i;
        strcpy (root_tag->tag, &arena [i]);
        memcpy (&root_tag->root, &arena [j + 1], sizeof (FluxItem));

        g_hash_table_insert (disk_chunk->tag_table, g_strdup (root_tag->tag), root_tag);
      }

      j += 1 + sizeof (FluxItem);
      i = j;
    }
  }

  disk_chunk->tag_arena_end = i;
  return TRUE;
}

static gboolean
add_root_tag (FluxDiskChunk *disk_chunk, const gchar *tag, FluxItem root)
{
  RootTag *root_tag;
  guint    len;

  len = strlen (tag);

  if (len > ROOT_TAG_MAX_LEN)
    return FALSE;

  if (disk_chunk->tag_arena_end + len + 1 + sizeof (FluxItem) > ROOT_TAG_ARENA_SIZE)
    return FALSE;

  lseek (disk_chunk->fd, ROOT_TAGS_OFFSET + disk_chunk->tag_arena_end, SEEK_SET);

  if (!safe_write (disk_chunk->fd, tag, len + 1))
    return FALSE;

  if (!safe_write (disk_chunk->fd, &root, sizeof (FluxItem)))
    return FALSE;

  root_tag = g_new0 (RootTag, 1);
  root_tag->arena_offset = disk_chunk->tag_arena_end;
  strcpy (root_tag->tag, tag);
  root_tag->root = root;

  disk_chunk->tag_arena_end += len + 1 + sizeof (FluxItem);

  g_hash_table_insert (disk_chunk->tag_table, g_strdup (root_tag->tag), root_tag);
  return TRUE;
}

static gboolean
set_root_tag (FluxDiskChunk *disk_chunk, const gchar *tag, FluxItem root)
{
  RootTag *root_tag;

  root_tag = g_hash_table_lookup (disk_chunk->tag_table, tag);
  if (!root_tag)
    return add_root_tag (disk_chunk, tag, root);

  if (root_tag->root == root)
    return TRUE;

  lseek (disk_chunk->fd, ROOT_TAGS_OFFSET + root_tag->arena_offset + strlen (tag) + 1, SEEK_SET);
  if (!safe_write (disk_chunk->fd, &root, sizeof (FluxItem)))
    return FALSE;

  root_tag->root = root;
  return TRUE;
}

static FluxItem
get_root_tag (FluxDiskChunk *disk_chunk, const gchar *tag)
{
  RootTag *root_tag;

  root_tag = g_hash_table_lookup (disk_chunk->tag_table, tag);
  if (!root_tag)
    return FLUX_ITEM_NULL;

  return root_tag->root;
}

static gboolean
init_root_tags (FluxDiskChunk *disk_chunk)
{
  guchar arena [ROOT_TAG_ARENA_SIZE];

  memset (arena, 0, ROOT_TAG_ARENA_SIZE);
  lseek (disk_chunk->fd, ROOT_TAGS_OFFSET, SEEK_SET);
  return safe_write (disk_chunk->fd, arena, ROOT_TAG_ARENA_SIZE);
}

static gboolean
seek_to_atom (FluxDiskChunk *disk_chunk, FluxItem atom)
{
  gint offset;

  offset = ATOM_OFFSET (disk_chunk, atom);

  if (offset > (gint) disk_chunk->file_size - (gint) disk_chunk->atomsize)
    return FALSE;

  lseek (disk_chunk->fd, offset, SEEK_SET);
  return TRUE;
}

static gboolean
read_atom (FluxDiskChunk *disk_chunk, gpointer dest)
{
  return safe_read (disk_chunk->fd, dest, disk_chunk->atomsize);
}

static gboolean
write_atom (FluxDiskChunk *disk_chunk, gconstpointer src)
{
  return safe_write (disk_chunk->fd, src, disk_chunk->atomsize);
}

static FluxItem
append_new_atom (FluxDiskChunk *disk_chunk)
{
  FluxItem atom;
  guchar   null = 255;

  atom = (disk_chunk->file_size - DATA_OFFSET + disk_chunk->atomsize - 1) /
    disk_chunk->atomsize;

  /* Poke last byte in atom, forcing the file system to allocate some space.
   * This lets us detect disk-full errors earlier. */
  lseek (disk_chunk->fd, ATOM_OFFSET (disk_chunk, atom) + disk_chunk->atomsize - 1, SEEK_SET);
  if (!safe_write (disk_chunk->fd, &null, 1))
    return FLUX_ITEM_NULL;

  disk_chunk->file_size += disk_chunk->atomsize;
  return atom;
}

DEBUG (

static gboolean
find_in_free_list (FluxDiskChunk *disk_chunk, FluxItem find_atom)
{
  FluxItem atom;
  FluxItem first_atom;

  atom = get_first_free_atom (disk_chunk);
  while (atom != NO_FREE_ATOMS)
  {
    DiskChunkFreeNode free_node;

    if (atom == find_atom)
      return TRUE;

    seek_to_atom (disk_chunk, atom);  /* FIXME: Check retval */
    safe_read (disk_chunk->fd, &free_node, sizeof (DiskChunkFreeNode));
    atom = free_node.next;
  }

  return FALSE;
}

)

static guint
dc_get_arena_size (FluxStorage *storage)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);

  assert_types (disk_chunk);
  return disk_chunk->file_size;
}

static guint
dc_get_used_size (FluxStorage *storage)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);

  assert_types (disk_chunk);
  return disk_chunk->file_size - disk_chunk->free_size;
}

static FluxItem
dc_get_root (FluxStorage *storage, const gchar *tag)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);

  assert_types (disk_chunk);

  return get_root_tag (disk_chunk, tag);
}

static gboolean
dc_set_root (FluxStorage *storage, const gchar *tag, FluxItem item)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);

  assert_types (disk_chunk);

  return set_root_tag (disk_chunk, tag, item);
}

static gboolean
dc_free (FluxStorage *storage, FluxItem item)
{
  FluxDiskChunk     *disk_chunk = FLUX_DISK_CHUNK (storage);
  FluxItem           first_free_atom;
  DiskChunkFreeNode  free_node;

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  first_free_atom = get_first_free_atom (disk_chunk);

  if (!seek_to_atom (disk_chunk, item))
    return FALSE;

  free_node.next = first_free_atom;

  if (safe_write (disk_chunk->fd, &free_node, sizeof (DiskChunkFreeNode)))
  {
    set_first_free_atom (disk_chunk, item);
    disk_chunk->free_size += disk_chunk->atomsize;
  }
  else
    return FALSE;

  return TRUE;
}

static FluxItem
dc_alloc (FluxStorage *storage, guint len)
{
  FluxDiskChunk     *disk_chunk = FLUX_DISK_CHUNK (storage);
  FluxItem           first_free_atom;
  DiskChunkFreeNode  free_node;

  assert_types (disk_chunk);
  g_return_val_if_fail (len != disk_chunk->atomsize, FLUX_ITEM_NULL);

  first_free_atom = get_first_free_atom (disk_chunk);
  if (first_free_atom == NO_FREE_ATOMS)
  {
    return append_new_atom (disk_chunk);
  }

  seek_to_atom (disk_chunk, first_free_atom);
  if (!safe_read (disk_chunk->fd, &free_node, sizeof (DiskChunkFreeNode)))
    return FLUX_ITEM_NULL;

  set_first_free_atom (disk_chunk, free_node.next);
  disk_chunk->free_size -= disk_chunk->atomsize;
  return first_free_atom;
}

static FluxItem
dc_alloc0 (FluxStorage *storage, guint len)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);
  FluxItem       atom;
  gpointer       data;
  gboolean       success;

  assert_types (disk_chunk);
  g_return_val_if_fail (len != disk_chunk->atomsize, FLUX_ITEM_NULL);

  atom = dc_alloc (storage, len);
  if (atom == FLUX_ITEM_NULL)
    return FLUX_ITEM_NULL;

  data = g_malloc0 (disk_chunk->atomsize);
  seek_to_atom (disk_chunk, atom);
  success = write_atom (disk_chunk, data);
  g_free (data);

  if (!success)
  {
    dc_free (storage, atom);
    return FLUX_ITEM_NULL;
  }

  return atom;
}

static gboolean
dc_read (FluxStorage *storage, FluxItem item, gpointer dest)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);
  gboolean       success;

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  if (!is_valid_atom (disk_chunk, item))
    return FALSE;

  if (!dest)
    return TRUE;

  if (flux_data_cache_get_data (disk_chunk->data_cache, item, dest))
  {
    success = TRUE;
  }
  else
  {
    if (!seek_to_atom (disk_chunk, item))
      return FALSE;

    success = read_atom (disk_chunk, dest);
    if (success)
      flux_data_cache_insert (disk_chunk->data_cache, item, dest);
  }

  return success;
}

static gboolean
dc_read_full (FluxStorage *storage, FluxItem item, gpointer dest, guint offset, guint len)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);
  gpointer       block_ptr;

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (offset + len <= disk_chunk->atomsize, FALSE);

  if (!is_valid_atom (disk_chunk, item))
    return FALSE;

  if (!dest || len == 0)
    return TRUE;

  block_ptr = (gpointer) flux_data_cache_get_ptr (disk_chunk->data_cache, item);
  if (!block_ptr)
  {
    if (!seek_to_atom (disk_chunk, item))
      return FALSE;

    block_ptr = flux_data_cache_get_insert_ptr (disk_chunk->data_cache, item);
    if (!block_ptr || !read_atom (disk_chunk, block_ptr))
    {
      invalidate_atom (disk_chunk, item);
      return FALSE;
    }
  }

  memcpy (dest, (guchar *) block_ptr + offset, len);
  return TRUE;
}

static gconstpointer
dc_peek (FluxStorage *storage, FluxItem item, guint offset, guint len)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);
  gpointer       block_ptr;

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, NULL);
  g_return_val_if_fail (offset + len <= disk_chunk->atomsize, NULL);

  if (!is_valid_atom (disk_chunk, item))
    return NULL;

  block_ptr = (gpointer) flux_data_cache_get_ptr (disk_chunk->data_cache, item);
  if (!block_ptr)
  {
    if (!seek_to_atom (disk_chunk, item))
      return NULL;

    block_ptr = flux_data_cache_get_insert_ptr (disk_chunk->data_cache, item);
    if (!block_ptr || !read_atom (disk_chunk, block_ptr))
      return NULL;
  }

  return block_ptr;
}

static gboolean
dc_write (FluxStorage *storage, FluxItem item, gconstpointer src)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);
  gboolean       success;

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (src != NULL, FALSE);

  if (!seek_to_atom (disk_chunk, item))
    return FALSE;

  /* If the user passed a NULL source, don't actually write */
  if (!src)
    return TRUE;

  success = write_atom (disk_chunk, src);
  if (success)
    flux_data_cache_insert (disk_chunk->data_cache, item, src);

  return success;
}

static gboolean
dc_write_full (FluxStorage *storage, FluxItem item, gconstpointer src, guint offset, guint len)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);
  gpointer       block_ptr;

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (offset + len <= disk_chunk->atomsize, FALSE);

  if (!is_valid_atom (disk_chunk, item))
    return FALSE;

  if (!src || len == 0)
    return TRUE;

  block_ptr = (gpointer) flux_data_cache_get_ptr (disk_chunk->data_cache, item);
  if (!block_ptr)
  {
    if (!seek_to_atom (disk_chunk, item))
      return FALSE;

    block_ptr = flux_data_cache_get_insert_ptr (disk_chunk->data_cache, item);
    if (!block_ptr || !read_atom (disk_chunk, block_ptr))
    {
      invalidate_atom (disk_chunk, item);
      return FALSE;
    }
  }

  memcpy ((guchar *) block_ptr + offset, src, len);
  seek_to_atom (disk_chunk, item);
  if (!write_atom (disk_chunk, block_ptr))
  {
    invalidate_atom (disk_chunk, item);
    return FALSE;
  }

  return TRUE;
}

static guint
dc_get_item_len (FluxStorage *storage, FluxItem item)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);

  assert_types (disk_chunk);

  if (item == FLUX_ITEM_NULL)
    return disk_chunk->atomsize;

  if (!is_valid_atom (disk_chunk, item))
    return 0;

  return disk_chunk->atomsize;
}

static FluxItem
dc_realloc (FluxStorage *storage, FluxItem item, guint len)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (storage);

  assert_types (disk_chunk);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (len != disk_chunk->atomsize, FLUX_ITEM_NULL);

  /* Fixed-size allocator, element size never changes */
  return item;
}

static guint
count_free_atoms (FluxDiskChunk *disk_chunk)
{
  FluxItem atom, first_atom;
  guint    n_free_atoms = 0;

  atom = first_atom = get_first_free_atom (disk_chunk);
  while (atom != NO_FREE_ATOMS)
  {
    DiskChunkFreeNode free_node;

    n_free_atoms++;
    seek_to_atom (disk_chunk, atom);
    if (!safe_read (disk_chunk->fd, &free_node, sizeof (DiskChunkFreeNode)))
      break;

    atom = free_node.next;

    if (atom == first_atom)
      g_error ("Looped free list!");

    if (n_free_atoms * disk_chunk->atomsize > disk_chunk->file_size)
      g_error ("Looped free list!");
  }

  return n_free_atoms;
}

static gboolean
init_file (FluxDiskChunk *disk_chunk, FluxDBHeader *header)
{
  FluxItem no_free_atoms = NO_FREE_ATOMS;

  flux_db_header_write (header, disk_chunk->fd);
  lseek (disk_chunk->fd, FLUX_DB_HEADER_SIZE, SEEK_SET);
  if (!safe_write (disk_chunk->fd, &no_free_atoms, sizeof (FluxItem)))
    return FALSE;

  if (!init_root_tags (disk_chunk))
    return FALSE;

  ftruncate (disk_chunk->fd, DATA_OFFSET);
  return TRUE;
}

static gboolean
maybe_init_file (FluxDiskChunk *disk_chunk)
{
  struct stat   sbuf;
  FluxDBHeader *expected_header;
  gboolean      success;

  /* Construct expected header */
  expected_header = flux_db_header_create (DB_NAME, DB_VERSION);
  memcpy (expected_header->user, &disk_chunk->atomsize,
          MIN (sizeof (disk_chunk->atomsize), FLUX_DB_HEADER_USER_SIZE));

  fstat (disk_chunk->fd, &sbuf);
  if (sbuf.st_size < FLUX_DB_HEADER_SIZE + sizeof (FluxItem))
  {
    success = init_file (disk_chunk, expected_header);
  }
  else
  {
    FluxDBHeader *real_header = flux_db_header_read (disk_chunk->fd);

    if (memcmp (expected_header, real_header, sizeof (FluxDBHeader)))
    {
      success = init_file (disk_chunk, expected_header);
    }
    else
      success = TRUE;

    flux_db_header_free (real_header);
  }

  flux_db_header_free (expected_header);
  return success;
}

gboolean
flux_disk_chunk_init (FluxDiskChunk *disk_chunk, gint fd, guint atomsize)
{
  FluxStorage *storage;
  struct stat  sbuf;

  g_return_val_if_fail (fd >= 0, FALSE);
  g_return_val_if_fail (atomsize > 0, FALSE);

  flux_storage_init             (FLUX_STORAGE (disk_chunk));
  flux_object_set_type          (FLUX_OBJECT (disk_chunk), FLUX_DISK_CHUNK_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (disk_chunk), flux_disk_chunk_finalize);

  disk_chunk->fd       = fd;
  disk_chunk->atomsize = MAX (atomsize, sizeof (DiskChunkFreeNode));

  if (!maybe_init_file (disk_chunk))
    return FALSE;

  disk_chunk->first_free_atom = read_first_free_atom (disk_chunk);

  if (!load_root_tags (disk_chunk))
    return FALSE;

  fstat (fd, &sbuf);
  disk_chunk->file_size = sbuf.st_size;
  disk_chunk->free_size = count_free_atoms (disk_chunk) * atomsize;

  disk_chunk->data_cache = flux_data_cache_new (atomsize, CACHE_SIZE / atomsize);

  /* Init storage */

  storage = FLUX_STORAGE (disk_chunk);

  storage->capabilities   = FLUX_STORAGE_PERSISTENT;

  storage->get_arena_size = dc_get_arena_size;
  storage->get_used_size  = dc_get_used_size;

  storage->get_root       = dc_get_root;
  storage->set_root       = dc_set_root;

  storage->alloc          = dc_alloc;
  storage->alloc0         = dc_alloc0;
  storage->realloc        = dc_realloc;
  storage->free           = dc_free;

  storage->peek           = dc_peek;
  storage->do_read        = dc_read;
  storage->do_read_full   = dc_read_full;
  storage->do_write       = dc_write;
  storage->do_write_full  = dc_write_full;

  storage->get_item_len   = dc_get_item_len;

  return TRUE;
}

FluxDiskChunk *
flux_disk_chunk_new (gint fd, guint atomsize)
{
  FluxDiskChunk *disk_chunk;

  disk_chunk = g_new0 (FluxDiskChunk, 1);
  if (!flux_disk_chunk_init (disk_chunk, fd, atomsize))
  {
    g_free (disk_chunk);
    return NULL;
  }

  return disk_chunk;
}

void
flux_disk_chunk_finalize (FluxObject *object)
{
  FluxDiskChunk *disk_chunk = FLUX_DISK_CHUNK (object);

  assert_types (disk_chunk);

  g_hash_table_destroy (disk_chunk->tag_table);

  close (disk_chunk->fd);
  flux_object_unref (FLUX_OBJECT (disk_chunk->data_cache));

  flux_storage_finalize (object);
}
