/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-disk-heap.c - Low-overhead any-size disk allocator.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-storage.h"
#include "flux-disk-heap.h"
#include <sys/types.h>        /* lseek, read, write, stat */
#include <sys/stat.h>         /* stat */
#include <string.h>           /* memcpy */
#include <errno.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>          /* lseek, read, write, stat */
#else
# include <io.h>              /* lseek, read, write */
# include "win32/win32-compat.h"
#endif

#define DEBUG(x)
#define DEBUG_FUNCS 1

#define DB_NAME           "FluxDiskHeap"
#define DB_VERSION        1

#define DEFAULT_CACHE_ELEM_SIZE  64
#define DEFAULT_CACHE_ELEM_COUNT 1024

#define MIN_SIZE          8

#define ROOT_TAG_MAX_LEN  255

#define FREE_COUNT_OFFSET (FLUX_DB_HEADER_SIZE)
#define FIRST_TAG_OFFSET  (FREE_COUNT_OFFSET + sizeof (gint))
#define FREE_TABLE_OFFSET (FIRST_TAG_OFFSET + sizeof (FluxItem))
#define ARENA_OFFSET      (FREE_TABLE_OFFSET + get_block_free_buckets () * sizeof (FluxItem))

#define assert_types(o)        flux_object_check (FLUX_OBJECT (o), FLUX_DISK_HEAP_TYPE)
#define assert_cursor_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_DISK_HEAP_CURSOR_TYPE)

typedef struct
{
  gboolean in_use;
  guint    len;
}
BlockHead;

typedef struct
{
  FluxItem next;
  FluxItem root;                        /* Value */
  gchar    tag [ROOT_TAG_MAX_LEN + 1];  /* Key */
}
RootTag;

static FluxDiskHeapCursor *cursor_new (FluxDiskHeap *disk_heap);

static const gint block_increment [] =
{
  /* Start at MIN_SIZE */

   8, /* Inc     1: [      4,    5, ..      12] */
  10, /* Inc     2: [     12,   14, ..      32] */
  24, /* Inc     4: [     32,   36, ..     128] */
  16, /* Inc     8: [    128,  136, ..     256] */
  16, /* Inc    16: [    256,  272, ..     512] */
  16, /* Inc    32: [    512,  544, ..    1024] */
  16, /* Inc    64: [   1024, 1088, ..    2048] */
  16, /* Inc   128: [   2048,       ..    4096] */
  16, /* Inc   256: [   4096,       ..    8192] */
  16, /* Inc   512: [   8192,       ..   16384] */
  16, /* Inc  1024: [  16384,       ..   32768] */
  16, /* Inc  2048: [  32768,       ..   65536] */
  16, /* Inc  4096: [  65536,       ..  131972] */
  16, /* Inc  8192: [ 131972,       ..  262144] */
  16, /* Inc 16384: [ 262144,       ..  524288] */
  16, /* Inc 32768: [ 524288,       .. 1048576] */
  16, /* Inc 65536: [1048576,       .. 2097152] */

  /* 266 buckets */

  -1
};

#if 1

#define safe_peek(disk_heap, len)       flux_cached_file_peek  (disk_heap->cached_file, len)
#define safe_read(disk_heap, dest, len) flux_cached_file_read  (disk_heap->cached_file, dest, len)
#define safe_write(disk_heap, src, len) flux_cached_file_write (disk_heap->cached_file, src, len)
#define safe_seek(disk_heap, offset)    flux_cached_file_seek  (disk_heap->cached_file, offset)
#define safe_seek_rel(disk_heap, offset) \
  flux_cached_file_seek (disk_heap->cached_file, flux_cached_file_tell (disk_heap->cached_file) + (offset))

#else

static gboolean
safe_read (FluxDiskHeap *disk_heap, gpointer dest, gint len)
{
  gint bytes_done = 0;

  while (bytes_done < len)
  {
    gint n;

    n = read (flux_cached_file_get_fd (disk_heap->cached_file),
              (guchar *) dest + bytes_done, len - bytes_done);
    if (n < 1 && errno != EINTR)
      return FALSE;
    if (n > 0)
      bytes_done += n;
  }

  return TRUE;
}

/* NOTE: Not for production use! */
static gconstpointer
safe_peek (FluxDiskHeap *disk_heap, gint len)
{
  static gchar *peek_buffer = NULL;

  if (peek_buffer)
    g_free (peek_buffer);

  peek_buffer = g_malloc (len);
  safe_read (disk_heap, peek_buffer, len);
  return peek_buffer;
}

static gboolean
safe_write (FluxDiskHeap *disk_heap, gconstpointer src, gint len)
{
  gint bytes_done = 0;

  while (bytes_done < len)
  {
    gint n;

    n = write (flux_cached_file_get_fd (disk_heap->cached_file),
               (const guchar *) src + bytes_done, len - bytes_done);
    if (n < 1 && errno != EINTR)
      return FALSE;
    if (n > 0)
      bytes_done += n;
  }

  return TRUE;
}

static gboolean
safe_seek (FluxDiskHeap *disk_heap, guint offset)
{
  if (lseek (flux_cached_file_get_fd (disk_heap->cached_file), offset, SEEK_SET) < 0)
    return FALSE;

  return TRUE;
}

static gboolean
safe_seek_rel (FluxDiskHeap *disk_heap, gint offset)
{
  if (lseek (flux_cached_file_get_fd (disk_heap->cached_file), offset, SEEK_CUR) < 0)
    return FALSE;

  return TRUE;
}

#endif

static gboolean
read_free_count (FluxDiskHeap *disk_heap)
{
  if (!safe_seek (disk_heap, FREE_COUNT_OFFSET))
    return FALSE;

  return safe_read (disk_heap, &disk_heap->free_size, sizeof (gint));
}

static gboolean
set_free_count (FluxDiskHeap *disk_heap, gint count)
{
  gboolean success;

  if (!safe_seek (disk_heap, FREE_COUNT_OFFSET))
    return FALSE;

  success = safe_write (disk_heap, &count, sizeof (gint));
  if (success)
    disk_heap->free_size = count;

  return success;
}

static FluxItem
get_first_root_tag (FluxDiskHeap *disk_heap)
{
  FluxItem item;

  if (!safe_seek (disk_heap, FIRST_TAG_OFFSET))
    return FLUX_ITEM_NULL;

  if (!safe_read (disk_heap, &item, sizeof (FluxItem)))
    return FLUX_ITEM_NULL;

  return item;
}

static gboolean
set_first_root_tag (FluxDiskHeap *disk_heap, FluxItem root_item)
{
  if (!safe_seek (disk_heap, FIRST_TAG_OFFSET))
    return FALSE;

  return safe_write (disk_heap, &root_item, sizeof (FluxItem));
}

static gint
get_block_free_buckets (void)
{
  static gint block_free_buckets = -1;

  if (block_free_buckets == -1)
  {
    gint i;

    block_free_buckets = 0;

    for (i = 0; block_increment [i] > -1; i++)
      block_free_buckets += block_increment [i];
  }

  return block_free_buckets;
}

static FluxItem *
alloc_free_table (void)
{
  FluxItem *table;
  gint      i;

  table = g_new (FluxItem, get_block_free_buckets ());

  for (i = 0; i < get_block_free_buckets (); i++)
    table [i] = FLUX_ITEM_NULL;

  return table;
}

static FluxItem *
read_free_table (FluxDiskHeap *disk_heap)
{
  FluxItem *table;

  if (!safe_seek (disk_heap, FREE_TABLE_OFFSET))
    return NULL;

  table = alloc_free_table ();
  if (!safe_read (disk_heap, table, get_block_free_buckets () * sizeof (FluxItem)))
  {
    g_free (table);
    return NULL;
  }

  return table;
}

static gboolean
write_free_bucket (FluxDiskHeap *disk_heap, FluxItem *table, gint bucket)
{
  if (!safe_seek (disk_heap, FREE_TABLE_OFFSET + bucket * sizeof (FluxItem)))
    return FALSE;

  return safe_write (disk_heap, &table [bucket], sizeof (FluxItem));
}

/* We always alloc from a bigger or same-size bucket */
static gint
find_fit_bucket_for_alloc (gint size)
{
  gint  bucket = 0;
  guint step   = 1;
  gint  accum  = MIN_SIZE;
  gint  i;

  for (i = 0; block_increment [i] > -1; step <<= 1, i++)
  {
    if (block_increment [i] == 0)
      continue;

    if (size <= accum + (block_increment [i] - 1) * step)
    {
      if (accum >= size)
        return bucket;

      return bucket + ((size - accum + step - 1) / step);
    }

    bucket += block_increment [i];
    accum  += block_increment [i] * step;
  }

  g_assert_not_reached ();
  return -1;
}

static gint
find_valid_bucket_for_alloc (FluxItem *free_table, gint size)
{
  gint bucket;

  for (bucket = find_fit_bucket_for_alloc (size); bucket < get_block_free_buckets (); bucket++)
  {
    if (free_table [bucket] != FLUX_ITEM_NULL)
      break;
  }

  if (bucket == get_block_free_buckets ())
    return -1;

  return bucket;
}

/* We always put a block into a same-size or smaller bucket */
static gint
find_fit_bucket_for_free (gint size)
{
  gint  bucket = 0;
  guint step   = 1;
  gint  accum  = MIN_SIZE;
  gint  i;

  for (i = 0; block_increment [i] > -1; step <<= 1, i++)
  {
    if (block_increment [i] == 0)
      continue;

    if (size < accum + block_increment [i] * step)
    {
      /* Hm. Probably can't happen? */
      if (accum >= size)
        return bucket;

      return bucket + ((size - accum) / step);
    }

    bucket += block_increment [i];
    accum  += block_increment [i] * step;
  }

  g_assert_not_reached ();
  return -1;
}

static gint
read_block_head (FluxDiskHeap *disk_heap, FluxItem ptr, BlockHead *head)
{
  const guchar *c;
  guint         len;
  gint          head_len = 1;
  gint          buf_len;

  if (!safe_seek (disk_heap, ptr))
    return -1;

  for (buf_len = 4; buf_len; buf_len--)
  {
    if ((c = safe_peek (disk_heap, buf_len)))
      break;
  }

  if (!c)
    return -1;

  if (c [0] & 0x80)
    head->in_use = TRUE;
  else
    head->in_use = FALSE;

  len = c [0] & 0x3f;

  if (c [0] & 0x40)
  {
    if (buf_len < 2)
      return -1;

    len |= ((c [1] << 6) & (0x7f << 6));
    head_len++;

    if (c [1] & 0x80)
    {
      if (buf_len < 3)
        return -1;

      len |= (c [2] << (6 + 7)) & (0x7f << (6 + 7));
      head_len++;

      if (c [2] & 0x80)
      {
        if (buf_len < 4)
          return -1;

        len |= (c [3] << (6 + 7 + 7)) & (0xff << (6 + 7 + 7));
        head_len++;
      }
    }
  }

  safe_seek (disk_heap, (FluxItem) ((guchar *) ptr + head_len));

  head->len = len;
  return head_len;
}

static gint
write_block_head (FluxDiskHeap *disk_heap, FluxItem ptr, BlockHead *head)
{
  guchar buf [4];
  guint  len;
  gint   head_len = 1;

  *((guint32 *) buf) = 0x00000000;
  len = head->len;

  if (!safe_seek (disk_heap, ptr))
    return -1;

  if (head->in_use)
    buf [0] = 0x80;

  buf [0] |= len & 0x3f;
  len >>= 6;

  if (len)
  {
    buf [0] |= 0x40;
    buf [1] = len & 0x7f;
    len >>= 7;
    head_len++;
  }

  if (len)
  {
    buf [1] |= 0x80;
    buf [2] = len & 0x7f;
    len >>= 7;
    head_len++;
  }

  if (len)
  {
    buf [2] |= 0x80;
    buf [3] = len & 0x7f;
    head_len++;
  }

  if (!safe_write (disk_heap, buf, head_len))
    return -1;

  return head_len;
}

static gint
calc_block_head_len (BlockHead *head)
{
  if (!(head->len >> 6))
    return 1;
  if (!(head->len >> (6 + 7)))
    return 2;
  if (!(head->len >> (6 + 7 + 7)))
    return 3;

  return 4;
}

static FluxItem
append_new_entry (FluxDiskHeap *disk_heap, gint len)
{
  FluxItem  ptr;
  BlockHead head;
  gint      head_len;
  guchar    null = 255;

  ptr = disk_heap->file_size;

  head.in_use = TRUE;
  head.len    = len;

  head_len = write_block_head (disk_heap, ptr, &head);
  if (head_len < 0)
    return FLUX_ITEM_NULL;

  if (len > 0)
  {
    /* Poke last byte in entry, forcing the file system to allocate some space.
     * This lets us detect disk-full errors earlier. */
    safe_seek (disk_heap, ptr + head_len + len - 1);

    if (!safe_write (disk_heap, &null, 1))
    {
      ftruncate (flux_cached_file_get_fd (disk_heap->cached_file), disk_heap->file_size);
      return FLUX_ITEM_NULL;
    }
  }

  disk_heap->file_size += head_len;
  disk_heap->file_size += len;

  return ptr;
}

static gboolean
unlink_free_block (FluxDiskHeap *disk_heap, FluxItem ptr)
{
  BlockHead unlinked_head;
  FluxItem  prev_ptr, next_ptr;

  if (read_block_head (disk_heap, ptr, &unlinked_head) < 0)
    return FALSE;

  if (unlinked_head.len < sizeof (FluxItem) * 2)
    return TRUE;  /* Block isn't linked at all */

  /* Block has next ptr */
  if (!safe_read (disk_heap, &next_ptr, sizeof (FluxItem)))
    return FALSE;

  /* Block has prev ptr */
  if (!safe_read (disk_heap, &prev_ptr, sizeof (FluxItem)))
    return FALSE;

  if (prev_ptr == FLUX_ITEM_NULL)
  {
    gint     bucket;
    FluxItem old_bucket_value;

    /* Prev pointer is NULL. Either we're first on the free list, or
     * we didn't store a prev pointer (in which case we might still be
     * first on the free list, but probably aren't).
     *
     * If we aren't first on the free list, we can't be unlinked. */

    bucket = find_fit_bucket_for_free (unlinked_head.len);
    if (disk_heap->free_table [bucket] != ptr)
      return FALSE;  /* Someone's pointing to us, but we can't find out who */

    old_bucket_value = disk_heap->free_table [bucket];
    disk_heap->free_table [bucket] = next_ptr;
    if (!write_free_bucket (disk_heap, disk_heap->free_table, bucket))
    {
      disk_heap->free_table [bucket] = old_bucket_value;
      return FALSE;
    }
  }
  else
  {
    BlockHead prev_head;

    /* There's a block before us on the free list. Set its next ptr to
     * point to our next. Since it's on a list, it must have a next ptr. */

    if (read_block_head (disk_heap, prev_ptr, &prev_head) < 0)
      return FALSE;

    if (!safe_write (disk_heap, &next_ptr, sizeof (FluxItem)))
      return FALSE;
  }

  if (next_ptr != FLUX_ITEM_NULL)
  {
    BlockHead next_head;

    /* There's a block behind us on the free list. Set its prev ptr
     * to point to our prev. */

    if (read_block_head (disk_heap, next_ptr, &next_head) < 0)
      return FALSE;

    if (!safe_seek_rel (disk_heap, sizeof (FluxItem)))
      return FALSE;
    if (!safe_write (disk_heap, &prev_ptr, sizeof (FluxItem)))
      return FALSE;
  }

  return TRUE;
}

/* Warning: fd must be positioned at the data for the block we're freeing */
static gboolean
add_block_to_free_list (FluxDiskHeap *disk_heap, FluxItem ptr, gint len)
{
  FluxItem null_ptr = FLUX_ITEM_NULL;
  FluxItem next_ptr;
  gint     bucket;
  FluxItem old_bucket_value;

  /* No space in block for free list linkage */
  if (len < sizeof (FluxItem) * 2)
    return TRUE;

  bucket = find_fit_bucket_for_free (len);
  next_ptr = disk_heap->free_table [bucket];

  DEBUG (g_print (" -> bucket %d", bucket));

  /* Write pointer to next */
  if (!safe_write (disk_heap, &next_ptr, sizeof (FluxItem)))
    return FALSE;

  /* Write a NULL prev pointer */
  if (!safe_write (disk_heap, &null_ptr, sizeof (FluxItem)))
    return FALSE;

  /* Update next's prev ptr to point to us - if we have a next. */
  if (next_ptr != FLUX_ITEM_NULL)
  {
    BlockHead next_head;

    if (read_block_head (disk_heap, next_ptr, &next_head) < 0)
      return FALSE;

    if (!safe_seek_rel (disk_heap, sizeof (FluxItem)))
      return FALSE;

    if (!safe_write (disk_heap, &ptr, sizeof (FluxItem)))
      return FALSE;
  }

  old_bucket_value = disk_heap->free_table [bucket];
  disk_heap->free_table [bucket] = ptr;
  if (!write_free_bucket (disk_heap, disk_heap->free_table, bucket))
  {
    disk_heap->free_table [bucket] = old_bucket_value;
    return FALSE;
  }

  return TRUE;
}

static FluxItem
next_used_entry (FluxDiskHeap *disk_heap, FluxItem ptr)
{
  BlockHead head;

  if (ptr == FLUX_ITEM_NULL)
  {
    ptr = ARENA_OFFSET;
  }
  else if (ptr >= disk_heap->file_size)
  {
    /* This will only happen if the current entry was freed from
     * the end of the file, and it was truncated. */
    return FLUX_ITEM_NULL;
  }
  else
  {
    if (read_block_head (disk_heap, ptr, &head) < 0)
      return FLUX_ITEM_NULL;

    ptr += calc_block_head_len (&head) + head.len;
  }

  for ( ; ptr < disk_heap->file_size; )
  {
    if (read_block_head (disk_heap, ptr, &head) < 0)
      return FLUX_ITEM_NULL;

    if (head.in_use)
      return ptr;

    ptr += calc_block_head_len (&head) + head.len;
  }

  return FLUX_ITEM_NULL;
}

static gboolean
free_entry (FluxDiskHeap *disk_heap, FluxItem ptr)
{
  BlockHead first_head, head;
  FluxItem  next_ptr;
  gint      free_len, first_head_len;

  /* Read head of the block we're freeing, and mark it unused */
  if (read_block_head (disk_heap, ptr, &first_head) < 0)
    return FALSE;

  /* Prevent double free */
  if (!first_head.in_use)
    return FALSE;

  first_head.in_use = FALSE;

  DEBUG (g_print ("Free %d", first_head.len));

  if (!set_free_count (disk_heap, disk_heap->free_size + calc_block_head_len (&first_head) + first_head.len))
    return FALSE;

  /* Try to merge in following free blocks */
  memcpy (&head, &first_head, sizeof (BlockHead));
  next_ptr = ptr;

  for (;;)
  {
    next_ptr += calc_block_head_len (&head) + head.len;
    if (next_ptr >= disk_heap->file_size)
      break;

    if (read_block_head (disk_heap, next_ptr, &head) < 0)
      break;

    if (head.in_use)
      break;

    if (!unlink_free_block (disk_heap, next_ptr))
      break;
  }

  free_len = next_ptr - ptr;
  DEBUG (g_print (", after merge: %d", free_len));

  if (free_len == (1 << (6))         + 1 ||
      free_len == (1 << (6 + 7))     + 2 ||
      free_len == (1 << (6 + 7 + 7)) + 3)
  {
    BlockHead pad_head;

    /* This is a nasty special case; when the space taken by the header is
     * subtracted from free_len, the actual data len may fit in a smaller
     * header, then if we use a smaller header, the increased data length
     * may not fit into that. We have to make an extra stepping-stone header
     * with 0 bytes data len in order to solve this deadlock.
     *
     * 65 ->      1/2 bytes head (-2 = 63, +1 = 64, -1 = 63, etc),
     * 8194 ->    2/3 bytes head (-3 = 8191, +1 = 8192, -1 = 8191, +1 = 8192, etc),
     * 1048579 -> 3/4 bytes head (-4 = 1048575, +1 = 1048576, -1 = 1048575, etc). */

    pad_head.in_use = FALSE;
    pad_head.len    = 0;

    if (write_block_head (disk_heap, ptr, &pad_head) < 0)
      return FALSE;

    ptr++;
    free_len--;

    DEBUG (g_print (" (!)"));
  }

  first_head.len = free_len;
  first_head_len = calc_block_head_len (&first_head);

  first_head.len -= first_head_len;
  first_head.len += first_head_len - calc_block_head_len (&first_head);

  DEBUG (g_print (" actual: %d", first_head.len));

  if (write_block_head (disk_heap, ptr, &first_head) < 0)
    return FALSE;

  if (!add_block_to_free_list (disk_heap, ptr, first_head.len))
    return FALSE;

  DEBUG (g_print ("\n"));
  return TRUE;
}

#define CLEAR_BUF_LEN 1024

static gboolean
clear_entry (FluxDiskHeap *disk_heap, FluxItem ptr)
{
  guchar    buf [CLEAR_BUF_LEN];
  BlockHead head;
  gint      bytes_written = 0;

  /* Read head of the block we're freeing, and mark it unused */
  if (read_block_head (disk_heap, ptr, &head) < 0)
    return FALSE;

  /* Prevent clearing free blocks */
  if (!head.in_use)
    return FALSE;

  memset (buf, 0, CLEAR_BUF_LEN);

  while (head.len - bytes_written > 0)
  {
    gint bytes_to_write = MIN (CLEAR_BUF_LEN, (head.len - bytes_written));

    if (!safe_write (disk_heap, buf, bytes_to_write))
      return FALSE;

    bytes_written += bytes_to_write;
  }

  return TRUE;
}

static FluxItem
alloc_entry (FluxDiskHeap *disk_heap, gint len)
{
  FluxItem  ptr = FLUX_ITEM_NULL, next_free;
  BlockHead new_head, old_head;
  gint      bucket;
  FluxItem  old_bucket_value;
  gint      remaining_len;

  DEBUG (g_print ("Alloc %d: ", len));

  /* Try bigger blocks */
  bucket = find_valid_bucket_for_alloc (disk_heap->free_table, len);

  if (bucket == -1)
  {
    /* No? Then append. */
    DEBUG (g_print ("append\n"));
    return append_new_entry (disk_heap, len);
  }

  DEBUG (g_print ("<- bucket %d, ", bucket));

  /* Read our header, update free list head. Note that we can't have
   * a prev, since we're first on the list. */
  ptr = disk_heap->free_table [bucket];
  if (read_block_head (disk_heap, ptr, &old_head) < 0)
    return FLUX_ITEM_NULL;

  if (!safe_read (disk_heap, &next_free, sizeof (FluxItem)))
    return FLUX_ITEM_NULL;

  old_bucket_value = disk_heap->free_table [bucket];
  disk_heap->free_table [bucket] = next_free;
  if (!write_free_bucket (disk_heap, disk_heap->free_table, bucket))
  {
    disk_heap->free_table [bucket] = old_bucket_value;
    return FLUX_ITEM_NULL;
  }

  /* If we have a next, it might have a prev ptr back to us. If so,
   * set it to NULL. */
  if (next_free != FLUX_ITEM_NULL)
  {
    BlockHead next_head;
    FluxItem  null_ptr = FLUX_ITEM_NULL;

    if (read_block_head (disk_heap, next_free, &next_head) < 0)
      return FLUX_ITEM_NULL;

    if (!safe_seek_rel (disk_heap, sizeof (FluxItem)))
      return FLUX_ITEM_NULL;

    if (!safe_write (disk_heap, &null_ptr, sizeof (FluxItem)))
      return FLUX_ITEM_NULL;
  }

  new_head.in_use = TRUE;
  new_head.len    = len;

  remaining_len =
    calc_block_head_len (&old_head) + old_head.len -
    calc_block_head_len (&new_head) - len;

  g_assert (remaining_len >= 0);

  if (!set_free_count (disk_heap, disk_heap->free_size - calc_block_head_len (&new_head) - len))
    return FLUX_ITEM_NULL;

  DEBUG (g_print ("in block, %d remain", remaining_len));

  if (remaining_len > 0)
  {
    FluxItem  remaining_ptr;
    BlockHead remaining_head;
    gint      remaining_head_len;

    /* We're leaving a gap between end of this block and start of next -
     * make it into a block of its own */

    remaining_ptr = ptr + calc_block_head_len (&new_head) + len;

    if (remaining_len == (1 << (6))         + 1 ||
        remaining_len == (1 << (6 + 7))     + 2 ||
        remaining_len == (1 << (6 + 7 + 7)) + 3)
    {
      /* This is a nasty special case; when the space taken by the header is
       * subtracted from remaining_len, the actual data len may fit in a smaller
       * header, then if we use a smaller header, the increased data length
       * may not fit into that. We have to make an extra stepping-stone header
       * with 0 bytes data len in order to solve this deadlock.
       *
       * 65 ->      1/2 bytes head (-2 = 63, +1 = 64, -1 = 63, etc),
       * 8194 ->    2/3 bytes head (-3 = 8191, +1 = 8192, -1 = 8191, +1 = 8192, etc),
       * 1048579 -> 3/4 bytes head (-4 = 1048575, +1 = 1048576, -1 = 1048575, etc). */

      remaining_head.in_use = FALSE;
      remaining_head.len    = 0;

      if (write_block_head (disk_heap, remaining_ptr, &remaining_head) < 0)
        return FLUX_ITEM_NULL;

      remaining_ptr++;
      remaining_len--;

      DEBUG (g_print (" (!)"));
    }

    remaining_head.in_use = FALSE;
    remaining_head.len    = remaining_len;
    remaining_head_len    = calc_block_head_len (&remaining_head);

    remaining_head.len -= remaining_head_len;
    remaining_head.len += remaining_head_len - calc_block_head_len (&remaining_head);

    DEBUG (g_print (", %d data", remaining_head.len));

    if (write_block_head (disk_heap, remaining_ptr, &remaining_head) < 0)
      return FLUX_ITEM_NULL;

    if (!add_block_to_free_list (disk_heap, remaining_ptr, remaining_head.len))
      return FLUX_ITEM_NULL;
  }

  if (write_block_head (disk_heap, ptr, &new_head) < 0)
    return FLUX_ITEM_NULL;

  DEBUG (g_print ("\n"));
  return ptr;
}

static gboolean
read_root_tag (FluxDiskHeap *disk_heap, FluxItem item, RootTag *root_tag)
{
  BlockHead head;

  if (!read_block_head (disk_heap, item, &head))
    return FALSE;

  if (!head.in_use)
    return FALSE;

  if (head.len < sizeof (FluxItem) * 2 + 1 ||
      head.len > sizeof (FluxItem) * 2 + ROOT_TAG_MAX_LEN)
    return FALSE;

  if (!safe_read (disk_heap, &root_tag->next, sizeof (FluxItem)))
    return FALSE;
  if (!safe_read (disk_heap, &root_tag->root, sizeof (FluxItem)))
    return FALSE;
  if (!safe_read (disk_heap, root_tag->tag, head.len - sizeof (FluxItem) * 2))
    return FALSE;

  root_tag->tag [head.len - sizeof (FluxItem) * 2] = '\0';
  return TRUE;
}

static gboolean
write_root_tag (FluxDiskHeap *disk_heap, FluxItem item, RootTag *root_tag)
{
  BlockHead head;

  if (!read_block_head (disk_heap, item, &head))
    return FALSE;

  if (!head.in_use)
    return FALSE;

  if (head.len != sizeof (FluxItem) * 2 + strlen (root_tag->tag))
    return FALSE;

  if (!safe_write (disk_heap, &root_tag->next, sizeof (FluxItem)))
    return FALSE;
  if (!safe_write (disk_heap, &root_tag->root, sizeof (FluxItem)))
    return FALSE;
  if (!safe_write (disk_heap, root_tag->tag, strlen (root_tag->tag)))
    return FALSE;

  return TRUE;
}

static FluxItem
alloc_root_tag (FluxDiskHeap *disk_heap, RootTag *root_tag)
{
  return alloc_entry (disk_heap, sizeof (FluxItem) * 2 + strlen (root_tag->tag));
}

#ifdef DEBUG_FUNCS

#define DUMP_TABLE_LEN 8192

void flux_disk_heap_dump (FluxDiskHeap *disk_heap);

void
flux_disk_heap_dump (FluxDiskHeap *disk_heap)
{
  FluxItem ptr;
  guint    free_count [DUMP_TABLE_LEN];
  guint    use_count  [DUMP_TABLE_LEN];
  gint     i;

  for (i = 0; i < DUMP_TABLE_LEN; i++)
  {
    free_count [i] = 0;
    use_count  [i] = 0;
  }

  ptr = ARENA_OFFSET;

  for ( ; ptr < disk_heap->file_size; )
  {
    BlockHead head;

    read_block_head (disk_heap, ptr, &head);
    if (head.len >= DUMP_TABLE_LEN)
      head.len = DUMP_TABLE_LEN - 1;

    if (head.in_use)
      use_count [head.len]++;
    else
      free_count [head.len]++;

    ptr += calc_block_head_len (&head) + head.len;
  }

  g_print ("Used blocks\n-----------\n\n");

  for (i = 0; i < DUMP_TABLE_LEN - 1; i++)
  {
    if (use_count [i] > 0)
    {
      g_print ("%5d: %7d\n", i, use_count [i]);
    }
  }

  if (use_count [DUMP_TABLE_LEN - 1])
    g_print (">%d: %7d\n", DUMP_TABLE_LEN - 2, use_count [DUMP_TABLE_LEN - 1]);

  g_print ("Free blocks\n-----------\n\n");

  for (i = 0; i < DUMP_TABLE_LEN - 1; i++)
  {
    if (free_count [i] > 0)
    {
      g_print ("%5d: %7d\n", i, free_count [i]);
    }
  }

  if (free_count [DUMP_TABLE_LEN - 1])
    g_print (">%d: %7d\n", DUMP_TABLE_LEN - 2, free_count [DUMP_TABLE_LEN - 1]);
}

#endif

FluxDiskHeapCursor *
flux_disk_heap_traverse (FluxDiskHeap *disk_heap)
{
  FluxDiskHeapCursor *cursor;

  assert_types (disk_heap);

  cursor = cursor_new (disk_heap);
  cursor->current_ptr = next_used_entry (disk_heap, FLUX_ITEM_NULL);

  return cursor;
}

static guint
dh_get_arena_size (FluxStorage *storage)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);

  assert_types (disk_heap);

  return disk_heap->file_size;
}

static guint
dh_get_used_size (FluxStorage *storage)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);

  assert_types (disk_heap);

  return disk_heap->file_size - disk_heap->free_size;
}

static FluxItem
dh_get_root (FluxStorage *storage, const gchar *tag)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  RootTag       root_tag;
  FluxItem      root_item;

  assert_types (disk_heap);
  g_return_val_if_fail (tag != NULL, FLUX_ITEM_NULL);
  g_return_val_if_fail (strlen (tag) <= ROOT_TAG_MAX_LEN, FLUX_ITEM_NULL);

  root_item = get_first_root_tag (disk_heap);
  while (root_item != FLUX_ITEM_NULL)
  {
    if (!read_root_tag (disk_heap, root_item, &root_tag))
      break;

    if (!strcmp (tag, root_tag.tag))
      return root_tag.root;

    root_item = root_tag.next;
  }

  return FLUX_ITEM_NULL;
}

static gboolean
dh_set_root (FluxStorage *storage, const gchar *tag, FluxItem item)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  RootTag       root_tag;
  FluxItem      root_item;

  assert_types (disk_heap);
  g_return_val_if_fail (tag != NULL, FALSE);
  g_return_val_if_fail (strlen (tag) <= ROOT_TAG_MAX_LEN, FALSE);

  root_item = get_first_root_tag (disk_heap);
  while (root_item != FLUX_ITEM_NULL)
  {
    if (!read_root_tag (disk_heap, root_item, &root_tag))
      break;

    if (!strcmp (tag, root_tag.tag))
    {
      /* Found record, just update its item */
      root_tag.root = item;
      return write_root_tag (disk_heap, root_item, &root_tag);
    }

    root_item = root_tag.next;
  }

  /* Prepend a new record */

  root_tag.next = get_first_root_tag (disk_heap);
  root_tag.root = item;
  strcpy (root_tag.tag, tag);

  root_item = alloc_root_tag (disk_heap, &root_tag);
  if (root_item == FLUX_ITEM_NULL)
    return FALSE;

  if (!write_root_tag (disk_heap, root_item, &root_tag))
  {
    free_entry (disk_heap, root_item);
    return FALSE;
  }

  return set_first_root_tag (disk_heap, root_item);
}

static FluxItem
dh_alloc (FluxStorage *storage, guint len)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);

  assert_types (disk_heap);
  g_return_val_if_fail (len > 0, FLUX_ITEM_NULL);

  return alloc_entry (disk_heap, len);
}

static FluxItem
dh_alloc0 (FluxStorage *storage, guint len)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  FluxItem      item;

  assert_types (disk_heap);
  g_return_val_if_fail (len > 0, FLUX_ITEM_NULL);

  item = alloc_entry (disk_heap, len);
  if (item == FLUX_ITEM_NULL)
    return FLUX_ITEM_NULL;

  if (!clear_entry (disk_heap, item))
  {
    free_entry (disk_heap, item);  /* Don't care if this fails */
    return FLUX_ITEM_NULL;
  }

  return item;
}

static gboolean
dh_free (FluxStorage *storage, FluxItem item)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  return free_entry (disk_heap, item);
}

static gboolean
dh_read (FluxStorage *storage, FluxItem item, gpointer dest)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  BlockHead     head;

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  if (read_block_head (disk_heap, item, &head) < 0)
    return FALSE;

  /* Prevent read from freed block */
  if (!head.in_use)
    return FALSE;

  if (!safe_read (disk_heap, dest, head.len))
    return FALSE;

  return TRUE;
}

static gboolean
dh_read_full (FluxStorage *storage, FluxItem item, gpointer dest, guint offset, guint len)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  BlockHead     head;

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (dest != NULL, FALSE);

  if (len == 0)
    return TRUE;

  if (read_block_head (disk_heap, item, &head) < 0)
    return FALSE;

  /* Prevent read from freed block */
  if (!head.in_use)
    return FALSE;

  /* Prevent read outside block */
  if (offset + len > head.len)
    return FALSE;

  if (offset)
    safe_seek_rel (disk_heap, offset);

  if (!safe_read (disk_heap, dest, len))
    return FALSE;

  return TRUE;
}

static gconstpointer
dh_peek (FluxStorage *storage, FluxItem item, guint offset, guint len)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  BlockHead     head;

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  if (len == 0)
    return NULL;

  if (read_block_head (disk_heap, item, &head) < 0)
    return NULL;

  /* Prevent read from freed block */
  if (!head.in_use)
    return NULL;

  /* Prevent read outside block */
  if (offset + len > head.len)
    return NULL;

  if (offset)
    safe_seek_rel (disk_heap, offset);

  return safe_peek (disk_heap, len);
}

static gboolean
dh_write (FluxStorage *storage, FluxItem item, gconstpointer src)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  BlockHead     head;

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (src != NULL, FALSE);

  if (read_block_head (disk_heap, item, &head) < 0)
    return FALSE;

  /* Prevent write to freed block */
  if (!head.in_use)
    return FALSE;

  if (!safe_write (disk_heap, src, head.len))
    return FALSE;

  return TRUE;
}

static gboolean
dh_write_full (FluxStorage *storage, FluxItem item, gconstpointer src, guint offset, guint len)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  BlockHead     head;

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (src != NULL, FALSE);

  if (len == 0)
    return TRUE;

  if (read_block_head (disk_heap, item, &head) < 0)
    return FALSE;

  /* Prevent write to freed block */
  if (!head.in_use)
    return FALSE;

  /* Prevent write outside block */
  if (offset + len > head.len)
    return FALSE;

  if (offset)
    safe_seek_rel (disk_heap, offset);

  if (!safe_write (disk_heap, src, len))
    return FALSE;

  return TRUE;
}

static guint
dh_get_item_len (FluxStorage *storage, FluxItem item)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  BlockHead     head;

  assert_types (disk_heap);
  g_return_val_if_fail (item != FLUX_ITEM_NULL, 0);

  if (read_block_head (disk_heap, item, &head) < 0)
    return 0;

  /* Prevent read from freed block */
  if (!head.in_use)
    return 0;

  return head.len;
}

static FluxItem
dh_realloc (FluxStorage *storage, FluxItem item, guint len)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (storage);
  FluxItem      new_item;
  gpointer      old_data;
  guint         old_len;

  assert_types (disk_heap);

  /* NULL item -> allocate new entry */
  if (item == FLUX_ITEM_NULL)
  {
    if (len > 0)
      return dh_alloc (storage, len);

    return FLUX_ITEM_NULL;
  }

  /* Zero length -> free entry */
  if (len == 0)
  {
    if (item != FLUX_ITEM_NULL)
      dh_free (storage, item);

    return FLUX_ITEM_NULL;
  }

  /* Old entry didn't exist -> allocate new entry */
  old_len = dh_get_item_len (storage, item);
  if (old_len == 0)
    return dh_alloc (storage, len);

  /* Allocate new entry, migrate old data */
  old_data = g_malloc (old_len);
  dh_read (storage, item, old_data);

  new_item = dh_alloc (storage, len);
  if (new_item == FLUX_ITEM_NULL)
  {
    g_free (old_data);
    return FLUX_ITEM_NULL;
  }

  if (!dh_write_full (storage, new_item, old_data, 0, MIN (len, old_len)))
  {
    g_free (old_data);
    return FLUX_ITEM_NULL;
  }

  g_free (old_data);
  return new_item;
}

static gboolean
init_file (FluxDiskHeap *disk_heap, FluxDBHeader *header)
{
  FluxItem *free_table;
  gint      free_count = 0;
  FluxItem  first_tag  = FLUX_ITEM_NULL;
  gboolean  result     = FALSE;

  free_table = alloc_free_table ();

  flux_db_header_write (header, flux_cached_file_get_fd (disk_heap->cached_file));

  if (!safe_seek (disk_heap, FREE_COUNT_OFFSET))
    goto out;

  if (!safe_write (disk_heap, &free_count, sizeof (gint)))
    goto out;

  if (!safe_write (disk_heap, &first_tag, sizeof (FluxItem)))
    goto out;

  if (!safe_seek (disk_heap, FREE_TABLE_OFFSET))
    goto out;

  if (!safe_write (disk_heap, free_table, get_block_free_buckets () * sizeof (FluxItem)))
    goto out;

  ftruncate (flux_cached_file_get_fd (disk_heap->cached_file), ARENA_OFFSET);
  result = TRUE;

out:
  g_free (free_table);
  return result;
}

static gboolean
maybe_init_file (FluxDiskHeap *disk_heap)
{
  struct stat   sbuf;
  FluxDBHeader *expected_header;
  gboolean      success;

  /* Construct expected header */
  expected_header = flux_db_header_create (DB_NAME, DB_VERSION);

  fstat (flux_cached_file_get_fd (disk_heap->cached_file), &sbuf);
  if (sbuf.st_size < ARENA_OFFSET)
  {
    success = init_file (disk_heap, expected_header);
  }
  else
  {
    FluxDBHeader *real_header = flux_db_header_read (flux_cached_file_get_fd (disk_heap->cached_file));

    if (memcmp (expected_header, real_header, sizeof (FluxDBHeader)))
    {
      success = init_file (disk_heap, expected_header);
    }
    else
      success = TRUE;

    flux_db_header_free (real_header);
  }

  flux_db_header_free (expected_header);
  return success;
}

gboolean
flux_disk_heap_init (FluxDiskHeap *disk_heap, gint fd)
{
  FluxStorage *storage;
  struct stat  sbuf;

  g_return_val_if_fail (fd >= 0, FALSE);

  flux_storage_init             (FLUX_STORAGE (disk_heap));
  flux_object_set_type          (FLUX_OBJECT (disk_heap), FLUX_DISK_HEAP_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (disk_heap), flux_disk_heap_finalize);

  disk_heap->cached_file = flux_cached_file_new (fd, DEFAULT_CACHE_ELEM_SIZE, DEFAULT_CACHE_ELEM_COUNT);

  if (!maybe_init_file (disk_heap))
    return FALSE;

  fstat (fd, &sbuf);
  disk_heap->file_size = sbuf.st_size;

  if (!read_free_count (disk_heap))
    return FALSE;

  disk_heap->free_table = read_free_table (disk_heap);
  if (!disk_heap->free_table)
    return FALSE;

  /* Init storage */

  storage = FLUX_STORAGE (disk_heap);

  storage->capabilities   = FLUX_STORAGE_VARIABLE_LENGTH | FLUX_STORAGE_PERSISTENT;

  storage->get_arena_size = dh_get_arena_size;
  storage->get_used_size  = dh_get_used_size;

  storage->get_root       = dh_get_root;
  storage->set_root       = dh_set_root;

  storage->alloc          = dh_alloc;
  storage->alloc0         = dh_alloc0;
  storage->realloc        = dh_realloc;
  storage->free           = dh_free;

  storage->peek           = dh_peek;
  storage->do_read        = dh_read;
  storage->do_read_full   = dh_read_full;
  storage->do_write       = dh_write;
  storage->do_write_full  = dh_write_full;

  storage->get_item_len   = dh_get_item_len;

  return TRUE;
}

FluxDiskHeap *
flux_disk_heap_new (gint fd)
{
  FluxDiskHeap *disk_heap;

  disk_heap = g_new (FluxDiskHeap, 1);
  if (!flux_disk_heap_init (disk_heap, fd))
  {
    g_free (disk_heap);
    return NULL;
  }

  return disk_heap;
}

void
flux_disk_heap_finalize (FluxObject *object)
{
  FluxDiskHeap *disk_heap = FLUX_DISK_HEAP (object);

  assert_types (disk_heap);

  flux_object_unref (FLUX_OBJECT (disk_heap->cached_file));
  g_free (disk_heap->free_table);

  flux_storage_finalize (object);
}

static FluxDiskHeapCursor *
cursor_new (FluxDiskHeap *disk_heap)
{
  FluxDiskHeapCursor *cursor;

  cursor = g_new0 (FluxDiskHeapCursor, 1);
  flux_disk_heap_cursor_init (cursor, disk_heap);

  return cursor;
}

static gboolean
cursor_next (FluxCursor *base_cursor)
{
  FluxDiskHeapCursor *cursor = FLUX_DISK_HEAP_CURSOR (base_cursor);

  assert_cursor_types (cursor);

  if (cursor->current_ptr == FLUX_ITEM_NULL)
    return FALSE;

  cursor->current_ptr = next_used_entry (cursor->disk_heap, cursor->current_ptr);
  if (cursor->current_ptr == FLUX_ITEM_NULL)
    return FALSE;

  return TRUE;
}

void
flux_disk_heap_cursor_init (FluxDiskHeapCursor *disk_heap_cursor, FluxDiskHeap *disk_heap)
{
  flux_cursor_init              (FLUX_CURSOR (disk_heap_cursor));
  flux_object_set_type          (FLUX_OBJECT (disk_heap_cursor), FLUX_DISK_HEAP_CURSOR_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (disk_heap_cursor), flux_disk_heap_cursor_finalize);

  flux_object_ref (FLUX_OBJECT (disk_heap));
  disk_heap_cursor->disk_heap = disk_heap;

  FLUX_CURSOR (disk_heap_cursor)->next = cursor_next;
}

void
flux_disk_heap_cursor_finalize (FluxObject *object)
{
  FluxDiskHeapCursor *disk_heap_cursor = FLUX_DISK_HEAP_CURSOR (object);

  assert_cursor_types (disk_heap_cursor);

  flux_object_unref (FLUX_OBJECT (disk_heap_cursor->disk_heap));

  flux_cursor_finalize (object);
}

FluxItem
flux_disk_heap_cursor_get_pos (FluxDiskHeapCursor *cursor)
{
  assert_cursor_types (cursor);

  return cursor->current_ptr;
}
