/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-dual-fifo.c - Full-duplex byte FIFO with optional compression and ciphering.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-types.h"
#include "flux-cipher.h"
#include "flux-fifo.h"
#include "flux-dual-fifo.h"
#include <stdlib.h>
#include <string.h>

enum
{
  READ_FIFO,
  WRITE_FIFO
};

void
flux_dual_fifo_init (FluxDualFifo *dfifo, gint buffers_min, gint buffers_max, gint buf_size)
{
  flux_object_init              (FLUX_OBJECT (dfifo));
  flux_object_set_type          (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (dfifo), flux_dual_fifo_finalize);

  dfifo->fifo [READ_FIFO]  = flux_fifo_new (buffers_min, buffers_max, buf_size);
  dfifo->fifo [WRITE_FIFO] = flux_fifo_new (buffers_min, buffers_max, buf_size);

  flux_fifo_set_mode (dfifo->fifo [READ_FIFO],  FLUX_FIFO_MODE_DECODE);
  flux_fifo_set_mode (dfifo->fifo [WRITE_FIFO], FLUX_FIFO_MODE_ENCODE);
}

FluxDualFifo *
flux_dual_fifo_new (gint buffers_min, gint buffers_max, gint buf_size)
{
  FluxDualFifo *dfifo;

  dfifo = g_new0 (FluxDualFifo, 1);
  flux_dual_fifo_init (dfifo, buffers_min, buffers_max, buf_size);

  return dfifo;
}

void
flux_dual_fifo_finalize (FluxObject *dfifo)
{
  flux_object_unref (FLUX_OBJECT (FLUX_DUAL_FIFO (dfifo)->fifo [READ_FIFO]));
  flux_object_unref (FLUX_OBJECT (FLUX_DUAL_FIFO (dfifo)->fifo [WRITE_FIFO]));
}

void
flux_dual_fifo_write (FluxDualFifo *dfifo, gconstpointer source, gint len)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  flux_fifo_enqueue (dfifo->fifo [WRITE_FIFO], source, len);

  if (len && dfifo->write_hook)
    dfifo->write_hook (dfifo->write_hook_data);
}

gboolean
flux_dual_fifo_read (FluxDualFifo *dfifo, gpointer dest, gint len)
{
  gboolean success = TRUE;

  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);

  if (flux_fifo_get_used (dfifo->fifo [READ_FIFO]) < len)
    return FALSE;

  if (flux_fifo_dequeue (dfifo->fifo [READ_FIFO], dest, len) < len)
    success = FALSE;  /* Can't happen right now */

  if (len && dfifo->read_hook)
    dfifo->read_hook (dfifo->read_hook_data);

  return TRUE;
}

void
flux_dual_fifo_flush_output (FluxDualFifo *dfifo)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  flux_fifo_flush (dfifo->fifo [WRITE_FIFO]);

  if (dfifo->write_hook)
    dfifo->write_hook (dfifo->write_hook_data);
}

void
flux_dual_fifo_reset (FluxDualFifo *dfifo)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);

#if 1
  flux_fifo_set_compression (dfifo->fifo [READ_FIFO],  0);
  flux_fifo_set_compression (dfifo->fifo [WRITE_FIFO], 0);
#endif
  flux_fifo_reset (dfifo->fifo [READ_FIFO]);
  flux_fifo_reset (dfifo->fifo [WRITE_FIFO]);
}

void
flux_dual_fifo_set_compression (FluxDualFifo *dfifo, gint read_level, gint write_level)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);

  flux_fifo_set_compression (dfifo->fifo [READ_FIFO],  read_level);
  flux_fifo_set_compression (dfifo->fifo [WRITE_FIFO], write_level);
}

FluxFifo *
flux_dual_fifo_get_write_fifo (FluxDualFifo *dfifo)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  return dfifo->fifo [WRITE_FIFO];
}

FluxFifo *
flux_dual_fifo_get_read_fifo (FluxDualFifo *dfifo)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  return dfifo->fifo [READ_FIFO];
}

gint
flux_dual_fifo_get_max_write (FluxDualFifo *dfifo)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  return flux_fifo_get_free_max (dfifo->fifo [WRITE_FIFO]);
}

gint
flux_dual_fifo_get_max_read (FluxDualFifo *dfifo)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);
  return flux_fifo_get_used (dfifo->fifo [READ_FIFO]);
}

void
flux_dual_fifo_set_cipher (FluxDualFifo *dfifo, FluxCipher *cipher)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);

  if (cipher)
  {
    FluxCipher *read_cipher, *write_cipher;

    read_cipher  = flux_cipher_dup (cipher);
    write_cipher = flux_cipher_dup (cipher);

    flux_fifo_set_cipher (dfifo->fifo [WRITE_FIFO], write_cipher);
    flux_fifo_set_cipher (dfifo->fifo [READ_FIFO],  read_cipher);

    flux_object_unref (FLUX_OBJECT (read_cipher));
    flux_object_unref (FLUX_OBJECT (write_cipher));
  }
  else
  {
    flux_fifo_set_cipher (dfifo->fifo [WRITE_FIFO], NULL);
    flux_fifo_set_cipher (dfifo->fifo [READ_FIFO],  NULL);
  }
}

void
flux_dual_fifo_set_read_hook (FluxDualFifo *dfifo, FluxReadHook *hook, gpointer data)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);

  dfifo->read_hook      = hook;
  dfifo->read_hook_data = data;
}

void
flux_dual_fifo_set_write_hook (FluxDualFifo *dfifo, FluxReadHook *hook, gpointer data)
{
  flux_object_check (FLUX_OBJECT (dfifo), FLUX_DUAL_FIFO_TYPE);

  dfifo->write_hook      = hook;
  dfifo->write_hook_data = data;
}
