/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-ec.c - The elliptic curve asymmetric cipher.
 *
 * Copyright (C) 2003 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>          /* time */
#include <glib.h>
#include <gmp.h>
#include "flux-init.h"
#include "flux-marshaller.h"
#include "flux-rng.h"
#include "flux-ec.h"

#define DEBUG(x)

#define bits_in_mpz(n)   (mpz_sizeinbase (n, 2))
#define bits_to_bytes(b) (((b) + 7) >> 3)
#define bytes_to_bits(b) ((b) << 3)

typedef struct
{
  mpz_t x;
  mpz_t y;
}
ECPoint;

typedef struct
{
  mpz_t a2;
  mpz_t a6;
}
ECCurve;

typedef struct
{
  guint   m;    /* Positive integer specifying field GF(2^m) */

  guint   k1;   /* x^m + x^k3 + x^k2 + x^k1 + 1 */
  guint   k2;   /* x^m + x^k3 + x^k2 + x^k1 + 1 */
  guint   k3;   /* x^m + x^k3 + x^k2 + x^k1 + 1 */

  mpz_t   a;    /* An element of GF(2^m), E: y^2 + xy = x^3 + ax^2 + b */
  mpz_t   b;    /* An element of GF(2^m), E: y^2 + xy = x^3 + ax^2 + b */

  mpz_t   r;    /* Positive prime dividing the number of points on E */

  ECPoint g;    /* A point on E of order r */

  mpz_t   k;    /* Positive prime, k = #E/r */

  mpz_t   mod;  /* (1 << m) + (1 << k1) + (1 << k2) + (1 << k3) + 1 */
}
ECParameters;

typedef struct
{
  guint  m;

  guint  k1;
  guint  k2;
  guint  k3;

  gchar *a;
  gchar *b;

  gchar *r;

  gchar *g_x;
  gchar *g_y;

  gchar *k;
}
ECParametersInit;

static ECParameters *ec_param [FLUX_EC_CURVE_MAX] =
{
  NULL, NULL, NULL, NULL
};

static ECParametersInit ec_param_init [FLUX_EC_CURVE_MAX] =
{
  {
    163, 7, 6, 3,
    "1", "20a601907b8c953ca1481eb10512f78744a3205fd",
    "5846006549323611672814742442876390689256843201587",
    "3f0eba16286a2d57ea0991168d4994637e8343e36",
    "0d51fbc6c71a0094fa2cdd545b11c5c0c797324f1", "2"
  },
  {
    233, 74, 0, 0,
    "1", "066647ede6c332c7f8c0923bb58213b333b20e9ce4281fe115f7d8f90ad",
    "6901746346790563787434755862277025555839812737345013555379383634485463",
    "0fac9dfcbac8313bb2139f1bb755fef65bc391f8b36f8f8eb7371fd558b",
    "1006a08a41903350678e58528bebf8a0beff867a7ca36716f7e01f81052", "2"
  },
  {
    283, 12, 7, 5,
    "1", "27b680ac8b8596da5a4af8a19a0303fca97fd7645309fa2a581485af6263e313b79a2f5",
    "7770675568902916283677847627294075626569625924376904889109196526770044277787378692871",
    "5f939258db7dd90e1934f8c70b0dfec2eed25b8557eac9c80e2e198f8cdbecd86b12053",
    "3676854fe24141cb98fe6d4b20d02b4516ff702350eddb0826779c813f0df45be8112f4", "2"
  },
  {
    409, 87, 0, 0,
    "1",
    "021a5c2c8ee9feb5c4b9a753b7b476b7fd6422ef1f3dd674761fa99d6ac27c8a9a197b272822f6cd57a55aa4f50ae317b13545f",
    "661055968790248598951915308032771039828404682964281219284648798304157774827374805208143723762179110965979867288366567526771",
    "15d4860d088ddb3496b0c6064756260441cde4af1771d4db01ffe5b34e59703dc255a868a1180515603aeab60794e54bb7996a7",
    "061b1cfab6be5f32bbfa78324ed106a7636b9c5a7bd198d0158aa4f5488d08f38514f1fdf4b4f40d2181b3681c364ba0273c706",
    "2"
  }
};

static void
add_one_lshift_n (mpz_t a, guint n)
{
  mpz_t b;

  mpz_init (b);

  mpz_set_ui (b, 1);
  mpz_mul_2exp (b, b, n);
  mpz_add (a, a, b);

  mpz_clear (b);
}

static ECParameters *
get_params (FluxECCurveType curve_type)
{
  ECParameters     *ecp;
  ECParametersInit *ecpi;

  if (ec_param [curve_type])
    return ec_param [curve_type];

  ecpi = &ec_param_init [curve_type];

  ecp = g_new (ECParameters, 1);
  ecp->m = ecpi->m;

  ecp->k1 = ecpi->k1;
  ecp->k2 = ecpi->k2;
  ecp->k3 = ecpi->k3;

  mpz_init_set_str (ecp->a, ecpi->a, 16);
  mpz_init_set_str (ecp->b, ecpi->b, 16);

  mpz_init_set_str (ecp->r, ecpi->r, 10);

  mpz_init_set_str (ecp->g.x, ecpi->g_x, 16);
  mpz_init_set_str (ecp->g.y, ecpi->g_y, 16);

  mpz_init_set_str (ecp->k, ecpi->k, 16);

  /* Set up the modulus */

  mpz_init (ecp->mod);
  add_one_lshift_n (ecp->mod, ecp->m);
  add_one_lshift_n (ecp->mod, ecp->k1);
  if (ecp->k2)
    add_one_lshift_n (ecp->mod, ecp->k2);
  if (ecp->k3)
    add_one_lshift_n (ecp->mod, ecp->k3);

  mpz_add_ui (ecp->mod, ecp->mod, 1);

  ec_param [curve_type] = ecp;
  return ecp;
}

#if 0

typedef struct
{
  guchar *n;
  guint   n_len;

  guchar *e;
  guint   e_len;
}
PublicKeyM;

typedef struct
{
  guchar *n;
  guint   n_len;

  guchar *d;
  guint   d_len;
}
SecretKeyM;

static FluxMarshaller *key_marshaller = NULL;

static FluxMarshalledStructEntry public_key_m_entries [] =
{
  FLUX_MARSHAL_RAW_DYNAMIC ("n", PublicKeyM, n, n_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("e", PublicKeyM, e, e_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStructEntry secret_key_m_entries [] =
{
  FLUX_MARSHAL_RAW_DYNAMIC ("n", SecretKeyM, n, n_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("d", SecretKeyM, d, d_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStruct key_marshal_info [] =
{
  { "rsa-public-key", sizeof (PublicKeyM), FLUX_MARSHALLER_CLEAR, public_key_m_entries },
  { "rsa-secret-key", sizeof (SecretKeyM), FLUX_MARSHALLER_CLEAR, secret_key_m_entries },
  { NULL }
};

#endif

static guint
hex_to_int (char hex)
{
  if (hex >= '0' && hex <= '9')
    return (hex - '0');

  if (hex >= 'a' && hex <= 'f')
    return (hex - 'a' + 0x0a);

  if (hex >= 'A' && hex <= 'F')
    return (hex - 'A' + 0x0a);

  g_error ("Got a digit not in base16.");
  return (-1);  /* Never executed */
}

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void 
mpz_to_raw (mpz_t n, guchar *raw, guint len)
{
  char  *mpz_b16;
  guint  mpz_b16_len, mpz_b256_len;
  guint  i, j;

  if (mpz_sgn (n) < 0)
    g_error ("Got negative mpz.");

  /* Get base16 string for conversion to raw */

  mpz_b16 = mpz_get_str (NULL, 16, n);
  g_assert (mpz_b16);

  mpz_b16_len  = strlen (mpz_b16);
  mpz_b256_len = (mpz_b16_len + 1) / 2;

  if (len < mpz_b256_len)
    g_error ("Destination too small to fit mpz: %u vs %u.", len, mpz_b256_len);

  /* Convert base16 string to raw */

  memset (raw, 0, len);
  j = len - mpz_b256_len;
  i = 0;

  if (mpz_b16_len % 2)
    raw [j++] = hex_to_int (mpz_b16 [i++]);

  for ( ; mpz_b16 [i]; i += 2, j++)
  {
    raw [j] = (hex_to_int (mpz_b16 [i]) << 4) +
              (hex_to_int (mpz_b16 [i + 1]));
  }

  g_free (mpz_b16);
}

static void
mpz_to_raw_alloc (mpz_t n, guchar **out, guint *out_len)
{
  *out_len = bits_to_bytes (bits_in_mpz (n));
  *out = g_malloc (*out_len);

  mpz_to_raw (n, *out, *out_len);
}

static void
raw_to_mpz (mpz_t n, const guchar *raw, guint len)
{
  gchar *mpz_b16;
  guint  i;

  /* Create a base16 string to import with mpz_set_str () */

  mpz_b16 = g_malloc ((len * 2) + 1);

  for (i = 0; i < len; i++)
  {
    mpz_b16 [i * 2]       = int_to_hex ((raw [i] & 0xf0) >> 4);
    mpz_b16 [(i * 2) + 1] = int_to_hex (raw [i] & 0x0f);
  }

  mpz_b16 [i * 2] = '\0';

  /* Import string */

  mpz_set_str (n, mpz_b16, 16);
  g_free (mpz_b16);
}

static void
init_marshaller (void)
{
#if 0
  if (!key_marshaller)
  {
    key_marshaller = flux_marshaller_new ();
    flux_marshaller_register_many (key_marshaller, key_marshal_info);
  }
#endif
}

static FluxECPublicKey *
new_public (void)
{
  FluxECPublicKey *pkey;

  pkey = g_new0 (FluxECPublicKey, 1);
  flux_object_init              (FLUX_OBJECT (pkey));
  flux_object_set_type          (FLUX_OBJECT (pkey), FLUX_EC_PUBLIC_KEY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (pkey), (FluxFinalizeFunc *) flux_ec_finalize_public_key);

  init_marshaller ();

  return pkey;
}

static FluxECSecretKey *
new_secret (void)
{
  FluxECSecretKey *skey;

  skey = g_new0 (FluxECSecretKey, 1);
  flux_object_init              (FLUX_OBJECT (skey));
  flux_object_set_type          (FLUX_OBJECT (skey), FLUX_EC_SECRET_KEY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (skey), (FluxFinalizeFunc *) flux_ec_finalize_secret_key);

  init_marshaller ();

  return skey;
}

static FluxECPublicKey *
dup_public (FluxECPublicKey *old)
{
  FluxECPublicKey *new;

  new = new_public ();

  FLUX_EC_KEY (new)->curve_type = FLUX_EC_KEY (old)->curve_type;
  mpz_init_set (new->point_x, old->point_x);
  mpz_init_set (new->point_y, old->point_y);

  return new;
}

static FluxECSecretKey *
dup_secret (FluxECSecretKey *old)
{
  FluxECSecretKey *new;

  new = new_secret ();

  FLUX_EC_KEY (new)->curve_type = FLUX_EC_KEY (old)->curve_type;
  mpz_init_set (new->s, old->s);

  return new;
}

static void
randomize_mpz_strong (mpz_t n, guint bits_num, FluxRng *rng)
{
  guchar *buf;

  buf = g_malloc (bits_to_bytes (bits_num));

  flux_rng_get_bytes (rng, buf, bits_to_bytes (bits_num));
  buf [0] &= 0xff >> (8 - (bits_num % 8));  /* Trim MSB */
  raw_to_mpz         (n,  buf, bits_to_bytes (bits_num));

  g_free (buf);
}

static void
poly_inverse (mpz_t b, mpz_t a, mpz_t f)
{
  mpz_t c, u, v, tmp;
  gint  j;

  mpz_init (c);
  mpz_init (u);
  mpz_init (v);
  mpz_init (tmp);

  mpz_set (u, a); /* u <- a */
  mpz_set (v, f); /* v <- f */
  mpz_set_ui (b, 1); /* b <- 1 */

  while (mpz_sizeinbase (u, 2) != 1) /* deg(u) != 0 */
  {
    j = (mpz_sizeinbase (u, 2) - 1) - (mpz_sizeinbase (v, 2) - 1);  /* j <- deg(u) - deg(v) */

	  if (j < 0)
    {
      mpz_swap (u, v); /* u <-> v */
      mpz_swap (b, c); /* b <-> c */
		  j = -j; /* j <- -j */
	  }

	  /* u <- u + x^jv */
	  mpz_set (tmp, v);
    mpz_mul_2exp (tmp, tmp, j);
	  mpz_xor (u, u, tmp);

	  /* b <- b + x^jc */
    mpz_set (tmp, c);
    mpz_mul_2exp (tmp, tmp, j);
	  mpz_xor (b, b, tmp);
  }

  mpz_clear (tmp);
  mpz_clear (v);
  mpz_clear (u);
  mpz_clear (c);
}

static void
update_size (mpz_t n)
{
  mp_size_t  size = n->_mp_alloc;
  mp_limb_t *d    = n->_mp_d + size - 1;

  while (size > 1 && !*d--)
    size--;

  if (n->_mp_size >= 0)
    n->_mp_size = size;
  else
    n->_mp_size = -size;
}

static void
clear_limbs (mpz_t n)
{
  memset (n->_mp_d, 0, n->_mp_alloc * sizeof (mp_limb_t));
}

static void
poly_multiply (mpz_t c, mpz_t a, mpz_t b)
{
  mp_limb_t  mask = 1;
  mp_limb_t *pa, *pb, *cj;
  mp_size_t  i, j, t;
  mpz_t      b_tmp, c_tmp;
  gint       k;

  mpz_init_set (b_tmp, b);
  mpz_init2 (c_tmp, mpz_sizeinbase (a, 2) + mpz_sizeinbase (b, 2) + 1);
  clear_limbs (c_tmp);

  t = abs (a->_mp_size);
  pa = a->_mp_d;

  for (k = 0; k < GMP_NUMB_BITS; k++)
  {
    cj = c_tmp->_mp_d;

    for (j = 0; j < t; j++, *cj++)
    {
      if (pa [j] & mask)
      {
        pb = b_tmp->_mp_d;

        for (i = 0; i < abs (b_tmp->_mp_size); i++)
        {
          cj [i] ^= pb [i];
          cj [i] &= GMP_NUMB_MASK;
        }
      }
    }

    if (k != GMP_NUMB_BITS - 1)
    {
      mpz_mul_2exp (b_tmp, b_tmp, 1);
      mask <<= 1;
    }
  }

  update_size (c_tmp);
  mpz_swap (c_tmp, c);

  mpz_clear (c_tmp);
  mpz_clear (b_tmp);
}

static void
poly_divide_real (mpz_t a, mpz_t b)
{
  mpz_t quot, rem, t, tb;
  gint  msb_r, msb_b;
  gint  bit_count;

  mpz_init (quot);
  mpz_init (t);
  mpz_init (tb);
  mpz_init_set (rem, a);

  msb_r = mpz_sizeinbase (rem, 2) - 1;
  msb_b = mpz_sizeinbase (b, 2) - 1;
  bit_count = msb_r - msb_b;

  while (msb_r >= msb_b)
  {
    mpz_set (tb, b);

    mpz_mul_2exp (tb, tb, msb_r - msb_b);
    mpz_mul_2exp (t, t, msb_r - msb_b);
    mpz_ior (quot, quot, t);
    mpz_xor (rem, rem, tb);

    msb_r = mpz_sizeinbase (rem, 2) - 1;
  }

  update_size (quot);
  update_size (rem);

  mpz_swap (quot, a);
  mpz_swap (rem, b);

  mpz_clear (rem);
  mpz_clear (tb);
  mpz_clear (t);
  mpz_clear (quot);
}

static gint
get_sign (mpz_t n)
{
  return mpz_sgn (n) >= 0 ? 1 : -1;
}

static void
set_sign (mpz_t n, gint sign)
{
  if (sign == 0)
    sign = 1;

  if (get_sign (n) != sign)
    mpz_neg (n, n);
}

static void
poly_divide (mpz_t q, mpz_t r, mpz_t a, mpz_t b)
{
  mpz_t qtmp, rtmp;
  gint  cmp;

  cmp = mpz_cmp (a, b);
  if (cmp < 0)
  {
    if (r)
      mpz_set (r, a);
    if (q)
      mpz_set_ui (q, 0);

    return;
  }
  else if (cmp == 0)
  {
    if (r)
      mpz_set_ui (r, 0);
    if (q)
      mpz_set_ui (q, 1);

    return;
  }

  mpz_init_set (qtmp, a);
  mpz_init_set (rtmp, b);

  poly_divide_real (qtmp, rtmp);

  set_sign (rtmp, get_sign (a));

  if (get_sign (a) == get_sign (b))
    set_sign (qtmp, 1);
  else
    set_sign (qtmp, -1);

  /* NOTE: Not sure if this is needed, singe we use mpz_neg() to change sign */
  if (!mpz_cmp_ui (qtmp, 0))
    set_sign (qtmp, 1);
  if (!mpz_cmp_ui (rtmp, 0))
    set_sign (rtmp, 1);

  if (q)
    mpz_swap (qtmp, q);
  if (r)
    mpz_swap (rtmp, r);

  mpz_clear (rtmp);
  mpz_clear (qtmp);
}

static void
poly_modulo (mpz_t c, mpz_t a, mpz_t m)
{
  if (mpz_cmp (a, m) > 0)
    poly_divide (NULL, c, a, m);
  else
    mpz_set (c, a);
}

static void
poly_mulmod (mpz_t c, mpz_t a, mpz_t b, mpz_t m)
{
  poly_multiply (c, a, b);
  poly_modulo (c, c, m);
}

/* FIXME: Nails > 1 will fuck us over, but I'd rather not deal with carry just now. */

static const mp_limb_t poly_square_table[] =
{
  0x0000, 0x0001, 0x0004, 0x0005, 0x0010, 0x0011, 0x0014, 0x0015, 
  0x0040, 0x0041, 0x0044, 0x0045, 0x0050, 0x0051, 0x0054, 0x0055, 
  0x0100, 0x0101, 0x0104, 0x0105, 0x0110, 0x0111, 0x0114, 0x0115, 
  0x0140, 0x0141, 0x0144, 0x0145, 0x0150, 0x0151, 0x0154, 0x0155, 
  0x0400, 0x0401, 0x0404, 0x0405, 0x0410, 0x0411, 0x0414, 0x0415, 
  0x0440, 0x0441, 0x0444, 0x0445, 0x0450, 0x0451, 0x0454, 0x0455, 
  0x0500, 0x0501, 0x0504, 0x0505, 0x0510, 0x0511, 0x0514, 0x0515, 
  0x0540, 0x0541, 0x0544, 0x0545, 0x0550, 0x0551, 0x0554, 0x0555, 
  0x1000, 0x1001, 0x1004, 0x1005, 0x1010, 0x1011, 0x1014, 0x1015, 
  0x1040, 0x1041, 0x1044, 0x1045, 0x1050, 0x1051, 0x1054, 0x1055, 
  0x1100, 0x1101, 0x1104, 0x1105, 0x1110, 0x1111, 0x1114, 0x1115, 
  0x1140, 0x1141, 0x1144, 0x1145, 0x1150, 0x1151, 0x1154, 0x1155, 
  0x1400, 0x1401, 0x1404, 0x1405, 0x1410, 0x1411, 0x1414, 0x1415, 
  0x1440, 0x1441, 0x1444, 0x1445, 0x1450, 0x1451, 0x1454, 0x1455, 
  0x1500, 0x1501, 0x1504, 0x1505, 0x1510, 0x1511, 0x1514, 0x1515, 
  0x1540, 0x1541, 0x1544, 0x1545, 0x1550, 0x1551, 0x1554, 0x1555, 
  0x4000, 0x4001, 0x4004, 0x4005, 0x4010, 0x4011, 0x4014, 0x4015, 
  0x4040, 0x4041, 0x4044, 0x4045, 0x4050, 0x4051, 0x4054, 0x4055, 
  0x4100, 0x4101, 0x4104, 0x4105, 0x4110, 0x4111, 0x4114, 0x4115, 
  0x4140, 0x4141, 0x4144, 0x4145, 0x4150, 0x4151, 0x4154, 0x4155, 
  0x4400, 0x4401, 0x4404, 0x4405, 0x4410, 0x4411, 0x4414, 0x4415, 
  0x4440, 0x4441, 0x4444, 0x4445, 0x4450, 0x4451, 0x4454, 0x4455, 
  0x4500, 0x4501, 0x4504, 0x4505, 0x4510, 0x4511, 0x4514, 0x4515, 
  0x4540, 0x4541, 0x4544, 0x4545, 0x4550, 0x4551, 0x4554, 0x4555, 
  0x5000, 0x5001, 0x5004, 0x5005, 0x5010, 0x5011, 0x5014, 0x5015, 
  0x5040, 0x5041, 0x5044, 0x5045, 0x5050, 0x5051, 0x5054, 0x5055, 
  0x5100, 0x5101, 0x5104, 0x5105, 0x5110, 0x5111, 0x5114, 0x5115, 
  0x5140, 0x5141, 0x5144, 0x5145, 0x5150, 0x5151, 0x5154, 0x5155, 
  0x5400, 0x5401, 0x5404, 0x5405, 0x5410, 0x5411, 0x5414, 0x5415, 
  0x5440, 0x5441, 0x5444, 0x5445, 0x5450, 0x5451, 0x5454, 0x5455, 
  0x5500, 0x5501, 0x5504, 0x5505, 0x5510, 0x5511, 0x5514, 0x5515, 
  0x5540, 0x5541, 0x5544, 0x5545, 0x5550, 0x5551, 0x5554, 0x5555
};

static void
poly_squaremod (mpz_t b, mpz_t a, mpz_t m)
{
  mp_limb_t *pa, *pb;
  mp_size_t  i, t;

  t = abs (a->_mp_size);
  mpz_realloc (b, 2 * t);

  pa = a->_mp_d;
  pb = b->_mp_d;

#if GMP_LIMB_BITS == 8
  for (i = 0; i < t; i++)
  {
    if (!(i & 1))
      pb [i] = poly_square_table [pa [i]] & 0xff;
    else
      pb [i] = poly_square_table [pa [i - 1]] >> 8;
  }
#endif

#if GMP_LIMB_BITS == 16
  for (i = 0; i < t; i++)
  {
    pb [2 * i]      = poly_square_table [pa [i]        & 0xff];
    pb [2 * i + 1]  = poly_square_table [(pa [i] >> 8) & 0xff];
  }
#endif

#if GMP_LIMB_BITS == 32
  for (i = 0; i < t; i++)
  {
    pb [2 * i]      = poly_square_table [pa [i]         & 0xff];
    pb [2 * i]     |= poly_square_table [(pa [i] >> 8)  & 0xff] << 16;
    pb [2 * i + 1]  = poly_square_table [(pa [i] >> 16) & 0xff];
    pb [2 * i + 1] |= poly_square_table [(pa [i] >> 24) & 0xff] << 16;
  }
#endif

  /* NOTE: 48? Possible, but uncommon. */

#if GMP_LIMB_BITS == 64
  for (i = 0; i < t; i++)
  {
    pb [2 * i]      = poly_square_table [pa [i]         & 0xff];
    pb [2 * i]     |= poly_square_table [(pa [i] >> 8)  & 0xff] << 16;
    pb [2 * i]     |= poly_square_table [(pa [i] >> 16) & 0xff] << 32;
    pb [2 * i]     |= poly_square_table [(pa [i] >> 24) & 0xff] << 48;

    pb [2 * i + 1]  = poly_square_table [(pa [i] >> 32) & 0xff];
    pb [2 * i + 1] |= poly_square_table [(pa [i] >> 40) & 0xff] << 16;
    pb [2 * i + 1] |= poly_square_table [(pa [i] >> 48) & 0xff] << 32;
    pb [2 * i + 1] |= poly_square_table [(pa [i] >> 56) & 0xff] << 48;
  }
#endif

  update_size (b);
  poly_modulo (b, b, m);
}

static void
poly_to_octets (guchar **out, guint *out_len, mpz_t poly, ECParameters *param)
{
  guchar *buf;
  guint   poly_len_bits;
  guint   buf_len;
  gint    i;

  poly_len_bits = mpz_sizeinbase (poly, 2);
  buf_len = mpz_sizeinbase (param->mod, 256);
  buf = g_malloc0 (buf_len);

  for (i = 0; i < poly_len_bits; i++)
  {
    buf [buf_len - 1 - i / 8] |= mpz_tstbit (poly, i) << (i % 8);
  }

  *out     = buf;
  *out_len = buf_len;
}

static void
octets_to_mpz (mpz_t n, guchar *in, guint in_len)
{
  gint i;

  mpz_set_ui (n, 0);

  for (i = in_len - 1; i >= 0; i--)
  {
    mpz_mul_2exp (n, n, 8);
    mpz_add_ui (n, n, in [i]);
  }
}

static void
point_init (ECPoint *p)
{
  mpz_init (p->x);
  mpz_init (p->y);
}

static void
point_clear (ECPoint *p)
{
  mpz_clear (p->x);
  mpz_clear (p->y);
}

static void
point_add (ECPoint *p1, ECPoint *p2, ECParameters *param)
{
  mpz_t theta;
  mpz_t x1, y1;
  mpz_t x2, y2;

  mpz_init (theta);
  mpz_init (x1);
  mpz_init (y1);
  mpz_init (x2);
  mpz_init (y2);

  if (!mpz_cmp_ui (p1->x, 0) && !mpz_cmp_ui (p1->y, 0))
  {
    mpz_set (p1->x, p2->x);
    mpz_set (p1->y, p2->y);
    goto out;
  }

  if (!mpz_cmp_ui (p2->x, 0) && !mpz_cmp_ui (p2->y, 0))
    goto out;

  mpz_xor (x1, p1->x, p2->x);
  mpz_xor (y1, p1->y, p2->y);

  if (!mpz_cmp_ui (x1, 0))
  {
    mpz_set_ui (p1->x, 0);
    mpz_set_ui (p1->y, 0);
    goto out;
  }

  /* x1^-1 * y1 -> theta */
  poly_inverse (x1, x1, param->mod);
  poly_mulmod (theta, x1, y1, param->mod);

  /* theta^2 + theta + p1.x + p2.x + a2 -> x2 */
  poly_squaremod (x2, theta, param->mod);
  mpz_xor (x2, x2, theta);
  mpz_xor (x2, x2, p1->x);
  mpz_xor (x2, x2, p2->x);
  mpz_xor (x2, x2, param->a);

  /* (p1.x + x2) * theta + x2 + p1.y -> y2 */
  mpz_xor (y2, p1->x, x2);
  poly_mulmod (y2, y2, theta, param->mod);
  mpz_xor (y2, y2, x2);
  mpz_xor (y2, y2, p1->y);

  /* x2 -> p1.x, y2 -> p1.y */
  mpz_set (p1->x, x2);
  mpz_set (p1->y, y2);

out:
  mpz_clear (y2);
  mpz_clear (x2);
  mpz_clear (y1);
  mpz_clear (x1);
  mpz_clear (theta);
}

static void
point_dbl (ECPoint *p, ECParameters *param)
{
  mpz_t theta;
  mpz_t t, u;

  if (!mpz_cmp_ui (p->x, 0))
    return;

  mpz_init (theta);
  mpz_init (t);
  mpz_init (u);

  /* p.x + (p.x^-1 * p.y) -> theta */
  poly_inverse (t, p->x, param->mod);
  poly_mulmod (t, t, p->y, param->mod);
  mpz_xor (theta, t, p->x);

  /* theta^2 + theta + a2 -> t */
  poly_squaremod (t, theta, param->mod);
  mpz_xor (t, t, theta);
  mpz_xor (t, t, param->a);

  /* p.x^2 + (theta + 1) * t -> p.y */
  poly_mulmod (p->y, theta, t, param->mod);
  mpz_xor (p->y, p->y, t);
  poly_squaremod (u, p->x, param->mod);
  mpz_xor (p->y, p->y, u);

  /* t -> p.x */
  mpz_set (p->x, t);

  mpz_clear (u);
  mpz_clear (t);
  mpz_clear (theta);
}

static void
point_mul (ECPoint *p, mpz_t k, ECParameters *param)
{
  ECPoint q;
  gint    j;

  point_init (&q);

  for (j = mpz_sizeinbase (k, 2) - 1; j >= 0; j--)
  {
    point_dbl (&q, param);

    if (mpz_tstbit (k, j))
    {
      point_add (&q, p, param);
    }
  }

  mpz_set (p->x, q.x);
  mpz_set (p->y, q.y);

  point_clear (&q);
}

static gboolean
ec_sign_ecnr (FluxECSecretKey *secret_key,
              guchar       **out, guint *out_len,
              const guchar  *in,  guint  in_len)
{
  ECParameters *param;
  FluxRng      *rng;
  ECPoint       p;
  mpz_t         d, k, r, s, t;
  gchar        *buf;
  guint         buf_len;
  guchar       *temp;
  guint         temp_len;

  param = get_params (secret_key->key.curve_type);
  rng = flux_rng_new ();

  mpz_init (d);
  mpz_init (k);
  mpz_init (r);
  mpz_init (s);
  mpz_init (t);
  point_init (&p);

  do
  {
    /* Pick a random integer */

    randomize_mpz_strong (k, param->m, rng);
    mpz_mod (k, k, param->r);

    /* Compute corresponding point */

    mpz_set (p.x, param->g.x);
    mpz_set (p.y, param->g.y);

    point_mul (&p, k, param);

    /* Convert point to integer */

    /* mpz_set (t, p.x) works too, but the encoding will be incompatible with other
     * implementations. */

    poly_to_octets (&temp, &temp_len, p.x, param);
    octets_to_mpz (t, temp, temp_len);
    g_free (temp);

    mpz_mod (t, t, param->r);

    /* r <- d + t */

    raw_to_mpz (d, in, in_len);
    mpz_add (r, d, t);
    mpz_mod (r, r, param->r);

    DEBUG (g_print ("Signing data: %s\n", mpz_get_str (NULL, 16, d)));

    /* s <- k - sr */

    mpz_mul (t, secret_key->s, r);

    mpz_sub (s, k, t);
    mpz_mod (s, s, param->r);
  }
  while (!mpz_cmp_ui (s, 0) || !mpz_cmp_ui (r, 0));

  /* Jam r and s into one big mpz */

  DEBUG (g_print ("r, s before encoding: %s, %s\n",
                  mpz_get_str (NULL, 16, r), mpz_get_str (NULL, 16, s)));

  mpz_mul_2exp (r, r, param->m);
  mpz_ior (r, r, s);

  /* Output in octet form */

  buf_len = bits_to_bytes (bits_in_mpz (r));
  buf = g_malloc (buf_len);

  mpz_to_raw (r, buf, buf_len);

  *out     = buf;
  *out_len = buf_len;

  point_clear (&p);
  mpz_clear (t);
  mpz_clear (s);
  mpz_clear (r);
  mpz_clear (k);
  mpz_clear (d);

  /* TODO: Check that data fits, generate hash */
  return TRUE;
}

static gboolean
ec_verify_ecnr (FluxECPublicKey *public_key,
                guchar       **out, guint *out_len,
                const guchar  *in,  guint  in_len)
{
  ECParameters *param;
  ECPoint       p, q;
  mpz_t         d, r, s, t;
  gchar        *buf;
  guint         buf_len;
  guchar       *temp;
  guint         temp_len;

  param = get_params (public_key->key.curve_type);

  mpz_init (d);
  mpz_init (r);
  mpz_init (s);
  mpz_init (t);

  raw_to_mpz (r, in, in_len);

  /* Pull r and s out of concatenated mpz */

  mpz_tdiv_r_2exp (s, r, param->m);
  mpz_tdiv_q_2exp (r, r, param->m);

  DEBUG (g_print ("r, s after decoding: %s, %s\n", mpz_get_str (NULL, 16, r), mpz_get_str (NULL, 16, s)));

  /* P = (s * G) + (r * P) */

  mpz_init_set (p.x, param->g.x);
  mpz_init_set (p.y, param->g.y);

  mpz_init_set (q.x, public_key->point_x);
  mpz_init_set (q.y, public_key->point_y);

  point_mul (&p, s, param);
  point_mul (&q, r, param);

  point_add (&p, &q, param);

  /* t = P */

  /* mpz_set (t, p.x) works too, but the encoding will be incompatible with other
   * implementations. */

  poly_to_octets (&temp, &temp_len, p.x, param);
  octets_to_mpz (t, temp, temp_len);
  g_free (temp);

  mpz_mod (t, t, param->r);

  /* d = r - t */

  mpz_sub (d, r, t);
  mpz_mod (d, d, param->r);

  /* d now holds the data */

  DEBUG (g_print ("Recovered data: %s\n", mpz_get_str (NULL, 16, d)));

  buf_len = bits_to_bytes (bits_in_mpz (d));
  buf = g_malloc (buf_len);

  mpz_to_raw (d, buf, buf_len);

  *out     = buf;
  *out_len = buf_len;

  /* TODO: Verify hash */
  return TRUE;
}

static gboolean
ec_sign_dsa (FluxECSecretKey *secret_key,
             guchar       **signature, guint *signature_len,
             const guchar  *data,      guint  data_len)
{
  ECParameters *param;
  FluxRng      *rng;
  ECPoint       p;
  mpz_t         d, k, r, s, t, u;
  gchar        *buf;
  guint         buf_len;

  param = get_params (secret_key->key.curve_type);
  rng = flux_rng_new ();

  mpz_init (d);
  mpz_init (k);
  mpz_init (r);
  mpz_init (s);
  mpz_init (t);
  mpz_init (u);
  point_init (&p);

  do
  {
    guchar *temp;
    guint   temp_len;

    /* Pick a random integer */

    randomize_mpz_strong (k, param->m, rng);
    mpz_mod (k, k, param->r);

    /* Compute corresponding point */

    mpz_set (p.x, param->g.x);
    mpz_set (p.y, param->g.y);

    point_mul (&p, k, param);

    /* Convert point to integer */

    DEBUG (g_print ("P.x: %s\n", mpz_get_str (NULL, 16, p.x)));

    /* mpz_set (t, p.x) works too, but the encoding will be incompatible with other
     * implementations. */

    poly_to_octets (&temp, &temp_len, p.x, param);
    octets_to_mpz (t, temp, temp_len);
    g_free (temp);

    mpz_mod (r, t, param->r);

    DEBUG (g_print ("i:       %s\n", mpz_get_str (NULL, 16, t)));
    DEBUG (g_print ("r:       %s\n", mpz_get_str (NULL, 16, param->r)));
    DEBUG (g_print ("i mod r: %s\n", mpz_get_str (NULL, 16, r)));

    /* uinv = InvMod (u.s, sk.dp.r); */

    mpz_invert (t, k, param->r);
    mpz_mod (t, t, param->r);

    /* temp = f + (sk.s * sig.c); */

    mpz_mul (u, secret_key->s, r);
    raw_to_mpz (d, data, data_len);
    mpz_add (u, u, d);

    DEBUG (g_print ("Signing data: %s\n", mpz_get_str (NULL, 16, d)));

    /* sig.d = MulMod (uinv, temp, sk.dp.r); */

    mpz_mul (s, t, u);
    mpz_mod (s, s, param->r);
  }
  while (!mpz_cmp_ui (s, 0) || !mpz_cmp_ui (r, 0));

  /* Jam r and s into one big mpz */

  DEBUG (g_print ("r, s before encoding: %s, %s\n",
                  mpz_get_str (NULL, 16, r), mpz_get_str (NULL, 16, s)));

  mpz_mul_2exp (r, r, param->m);
  mpz_ior (r, r, s);

  /* Output in octet form */

  buf_len = bits_to_bytes (bits_in_mpz (r));
  buf = g_malloc (buf_len);

  mpz_to_raw (r, buf, buf_len);

  *signature     = buf;
  *signature_len = buf_len;

  point_clear (&p);
  mpz_clear (u);
  mpz_clear (t);
  mpz_clear (s);
  mpz_clear (r);
  mpz_clear (k);
  mpz_clear (d);

  /* TODO: Check that data fits, generate hash */
  return TRUE;
}

static gboolean
ec_verify_dsa (FluxECPublicKey *public_key,
               const guchar *data,      guint data_len,
               const guchar *signature, guint signature_len)
{
  ECParameters *param;
  ECPoint       p, q;
  mpz_t         d, r, s, t, u, v;
  guchar       *temp;
  guint         temp_len;

  param = get_params (public_key->key.curve_type);

  mpz_init (d);
  mpz_init (r);
  mpz_init (s);
  mpz_init (t);
  mpz_init (u);
  mpz_init (v);
  point_init (&p);
  point_init (&q);

  raw_to_mpz (d, data, data_len);
  raw_to_mpz (r, signature, signature_len);

  /* Pull r and s out of concatenated mpz */

  mpz_tdiv_r_2exp (s, r, param->m);
  mpz_tdiv_q_2exp (r, r, param->m);

  DEBUG (g_print ("r, s after decoding: %s, %s\n", mpz_get_str (NULL, 16, r), mpz_get_str (NULL, 16, s)));

  /* */

  mpz_invert (t, s, param->r);
  mpz_mul (u, d, t);
  mpz_mod (u, u, param->r);
  mpz_mul (v, r, t);
  mpz_mod (v, v, param->r);

  DEBUG (g_print ("h: %s\nh1: %s\n h2: %s\n", mpz_get_str (NULL, 16, t), mpz_get_str (NULL, 16, u),
                  mpz_get_str (NULL, 16, v)));

  /* */

  mpz_set (q.x, param->g.x);
  mpz_set (q.y, param->g.y);

  mpz_set (p.x, public_key->point_x);
  mpz_set (p.y, public_key->point_y);

  point_mul (&q, u, param);
  point_mul (&p, v, param);

  point_add (&p, &q, param);

  /* mpz_set (t, p.x) works too, but the encoding will be incompatible with other
   * implementations. */

  poly_to_octets (&temp, &temp_len, p.x, param);
  octets_to_mpz (t, temp, temp_len);
  g_free (temp);

  mpz_mod (t, t, param->r);

  DEBUG (g_print ("sig.c: %s\n   ci: %s\n", mpz_get_str (NULL, 16, r), mpz_get_str (NULL, 16, t)));
  DEBUG (g_print ("sig.c: %s\n   ci: %s\n", mpz_get_str (NULL, 2, r), mpz_get_str (NULL, 2, t)));

  point_clear (&q);
  point_clear (&p);
  mpz_clear (v);
  mpz_clear (u);
  mpz_clear (t);
  mpz_clear (s);
  mpz_clear (r);
  mpz_clear (d);

  /* TODO: Verify */
  return TRUE;
}

gboolean
flux_ec_generate_keypair (FluxECCurveType curve_type,
                          FluxECPublicKey **public_key, FluxECSecretKey **secret_key)
{
  ECParameters *param;
  FluxRng      *rng;
  ECPoint       point;
  mpz_t         s;

  g_return_val_if_fail (curve_type < FLUX_EC_CURVE_MAX, FALSE);
  g_return_val_if_fail (public_key != NULL, FALSE);
  g_return_val_if_fail (secret_key != NULL, FALSE);

  flux_init_mp ();

  param = get_params (curve_type);

  *public_key = new_public ();
  *secret_key = new_secret ();

  rng = flux_rng_new ();

  /* Secret */

  FLUX_EC_KEY (*secret_key)->curve_type = curve_type;

  mpz_init (s);
  randomize_mpz_strong (s, param->m, rng);
  mpz_mod (s, s, param->r);

  mpz_init_set ((*secret_key)->s, s);

  /* Public */

  FLUX_EC_KEY (*public_key)->curve_type = curve_type;

  mpz_init_set (point.x, param->g.x);
  mpz_init_set (point.y, param->g.y);

  point_mul (&point, s, param);

  mpz_init_set ((*public_key)->point_x, point.x);
  mpz_init_set ((*public_key)->point_y, point.y);

  mpz_clear (s);
  point_clear (&point);
  flux_object_unref (FLUX_OBJECT (rng));
  return TRUE;
}

FluxECKey *
flux_ec_dup_key (FluxECKey *old)
{
  FluxECKey *new;

  flux_object_check (FLUX_OBJECT (old), FLUX_EC_KEY_TYPE);

  if (flux_object_get_type (FLUX_OBJECT (old)) == FLUX_EC_PUBLIC_KEY_TYPE)
    new = FLUX_EC_KEY (dup_public (FLUX_EC_PUBLIC_KEY (old)));
  else
    new = FLUX_EC_KEY (dup_secret (FLUX_EC_SECRET_KEY (old)));

  return new;
}

void
flux_ec_finalize_public_key (FluxECPublicKey *public_key)
{
  mpz_clear (public_key->point_x);
  mpz_clear (public_key->point_y);
}

void
flux_ec_finalize_secret_key (FluxECSecretKey *secret_key)
{
  mpz_clear (secret_key->s);
}






#if 0

gint
main (gint argc, gchar *argv [])
{
  FluxECPublicKey *public_key;
  FluxECSecretKey *secret_key;
  guchar  in [] = { 0x55, 0xa8, 0x2b, 0x4e, 0x96, 0x39, 0x7e, 0x47, 0x1a, 0x24, 0xce, 0x38, 0x6a, 0x0b, 0x97, 0x87, 0x4c, 0x8a, 0xba, 0x76 };
  guchar *out, *out2;
  guint   out_len, out2_len;
  gint i;

  flux_ec_generate_keypair (FLUX_EC_CURVE_163, &public_key, &secret_key);
  ec_sign_ecnr (secret_key, &out, &out_len, in, 20);

  g_print ("Public key:\n%s\n%s\n\nPrivate key:\n%s\n",
           mpz_get_str (NULL, 16, public_key->point_x),
           mpz_get_str (NULL, 16, public_key->point_y),
           mpz_get_str (NULL, 16, secret_key->s));

#if 1
  for (i = 0; i < 100000; i++)
  {
    ec_sign_ecnr (secret_key, &out, &out_len, in, 20);
    g_free (out);
    /* ec_verify_ecnr (public_key, &out2, &out2_len, out, out_len);
     * g_free (out2); */
    g_print (".");
    fflush (stdout);
  }
#endif

#if 0
  ec_sign_dsa (secret_key, &out, &out_len, in, 20);
  ec_verify_dsa (public_key, in, 20, out, out_len);
#endif

  g_print ("Input len: %d  Signature len: %d  Recovered message len: %d\n",
           strlen (in), out_len, out2_len);
}

#endif













#if 0

guint
flux_rsa_get_key_length (FluxRSAKey *key)
{
  flux_object_check (FLUX_OBJECT (key), FLUX_RSA_KEY_TYPE);

  return bits_in_mpz (key->n);
}

static guint
get_ciphertext_block_size (mpz_t n)
{
  return bits_to_bytes (bits_in_mpz (n));
}

static guint
get_plaintext_block_size (mpz_t n)
{
  /* Reserve 1 bit for initial 0-byte guard, and 1 bit to make sure
   * input is lower than modulo */
  return (bits_in_mpz (n) - 2) >> 3;
}

guint
flux_rsa_get_block_size (FluxRSAKey *key)
{
  return get_plaintext_block_size (key->n);
}

static void
rsa_crypt (mpz_t n, mpz_t exp,
           guchar       **out, guint *out_len,
           const guchar  *in,  guint  in_len,
           gboolean plaintext_in)
{
  mpz_t   in_m, out_m;
  guchar *buf;
  gint    buf_len;

  mpz_init (in_m);
  mpz_init (out_m);

  raw_to_mpz (in_m, in, in_len);

  if (plaintext_in)
  {
    /* Set guard bit if plaintext in. This must be done before encryption,
     * so leading 0-bytes are preserved throughout the calculation. */
    mpz_setbit (in_m, bytes_to_bits (in_len));
  }

  mpz_powm (out_m, in_m, exp, n);

  if (plaintext_in)
  {
    /* Plaintext in, ciphertext out */
    buf_len = get_ciphertext_block_size (n);
  }
  else
  {
    /* Ciphertext in, plaintext out */
    buf_len = bits_to_bytes (bits_in_mpz (out_m) - 1);
    mpz_clrbit (out_m, bits_in_mpz (out_m) - 1);
  }

  buf = g_malloc (buf_len);
  mpz_to_raw (out_m, buf, buf_len);

  mpz_clear (in_m);
  mpz_clear (out_m);

  *out     = buf;
  *out_len = buf_len;
}

static void
rsa_crypt_multiple_blocks (mpz_t n, mpz_t exp,
                           guchar       **out, guint *out_len,
                           const guchar  *in,  guint  in_len,
                           gboolean plaintext_in)
{
  guchar *buf;
  guint   buf_len;
  gint    in_pos = 0, in_increment;

  if (plaintext_in)
    in_increment = get_plaintext_block_size (n);
  else
    in_increment = get_ciphertext_block_size (n);

  *out     = NULL;
  *out_len = 0;

  for ( ; in_pos < in_len; in_pos += in_increment)
  {
    rsa_crypt (n, exp,
               &buf,        &buf_len,
               in + in_pos, MIN (in_increment, in_len - in_pos),
               plaintext_in);

    if (*out == NULL)
    {
      *out     = buf;
      *out_len = buf_len;
    }
    else
    {
      *out = g_realloc (*out, *out_len + buf_len);
      memcpy (*out + *out_len, buf, buf_len);
      *out_len += buf_len;
      g_free (buf);
    }
  }
}

void
flux_rsa_sign (FluxRSASecretKey *secret_key,
               guchar       **signature, guint *signature_len,
               const guchar  *hash,      guint  hash_len)
{
  flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);

  rsa_crypt_multiple_blocks (secret_key->key.n, secret_key->d,
                             signature, signature_len,
                             hash,      hash_len,
                             TRUE);
}

gboolean
flux_rsa_verify (FluxRSAPublicKey *public_key,
                 const guchar *signature, guint signature_len,
                 const guchar *hash,      guint hash_len)
{
  guchar   *buf;
  gint      buf_len;
  gboolean  result = TRUE;

  flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);

  rsa_crypt_multiple_blocks (public_key->key.n, public_key->e,
                             &buf,       &buf_len,
                             signature,  signature_len,
                             FALSE);

  if (buf_len != hash_len)
    result = FALSE;

  if (memcmp (buf, hash, buf_len))
    result = FALSE;

  g_free (buf);
  return result;
}

void
flux_rsa_encrypt (FluxRSAPublicKey *public_key,
                  guchar       **ciphertext, guint *ciphertext_len,
                  const guchar  *plaintext,  guint  plaintext_len)
{
  flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);

  rsa_crypt_multiple_blocks (public_key->key.n, public_key->e,
                             ciphertext, ciphertext_len,
                             plaintext,  plaintext_len,
                             TRUE);
}

void
flux_rsa_decrypt (FluxRSASecretKey *secret_key,
                  guchar       **plaintext,  guint *plaintext_len,
                  const guchar  *ciphertext, guint  ciphertext_len)
{
  flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);

  rsa_crypt_multiple_blocks (secret_key->key.n, secret_key->d,
                             plaintext,  plaintext_len,
                             ciphertext, ciphertext_len,
                             FALSE);
}

FluxRSAPublicKey *
flux_rsa_import_public_key (const guchar *in, guint in_len)
{
  FluxRSAPublicKey *public_key;
  PublicKeyM       *public_key_m;
  mpz_t             temp_mpz;

  g_return_val_if_fail (in != NULL, NULL);
  g_return_val_if_fail (in_len > 1, NULL);

  public_key = new_public ();

  if (!flux_marshaller_decode_pack (key_marshaller, "rsa-public-key", in, in_len, (gpointer *) &public_key_m))
  {
    flux_object_unref (FLUX_OBJECT (public_key));
    return NULL;
  }

  if (public_key_m->n == NULL || public_key_m->e == NULL)
  {
    if (public_key_m->n)
      g_free (public_key_m->n);
    if (public_key_m->e)
      g_free (public_key_m->e);

    flux_object_unref (FLUX_OBJECT (public_key));
    g_free (public_key_m);
    return NULL;
  }

  mpz_init (temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->n, public_key_m->n_len);
  mpz_init_set (public_key->key.n, temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->e, public_key_m->e_len);
  mpz_init_set (public_key->e, temp_mpz);

  mpz_clear (temp_mpz);
  g_free (public_key_m->n);
  g_free (public_key_m->e);
  g_free (public_key_m);
  return public_key;
}

FluxRSASecretKey *
flux_rsa_import_secret_key (const guchar *in, guint in_len)
{
  FluxRSASecretKey *secret_key;
  SecretKeyM       *secret_key_m;
  mpz_t             temp_mpz;

  g_return_val_if_fail (in != NULL, NULL);
  g_return_val_if_fail (in_len > 1, NULL);

  secret_key = new_secret ();

  if (!flux_marshaller_decode_pack (key_marshaller, "rsa-secret-key", in, in_len, (gpointer *) &secret_key_m))
  {
    flux_object_unref (FLUX_OBJECT (secret_key));
    return NULL;
  }

  if (secret_key_m->n == NULL || secret_key_m->d == NULL)
  {
    if (secret_key_m->n)
      g_free (secret_key_m->n);
    if (secret_key_m->d)
      g_free (secret_key_m->d);

    flux_object_unref (FLUX_OBJECT (secret_key));
    g_free (secret_key_m);
    return NULL;
  }

  mpz_init (temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->n, secret_key_m->n_len);
  mpz_init_set (secret_key->key.n, temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->d, secret_key_m->d_len);
  mpz_init_set (secret_key->d, temp_mpz);

  mpz_clear (temp_mpz);
  g_free (secret_key_m->n);
  g_free (secret_key_m->d);
  g_free (secret_key_m);
  return secret_key;
}

void
flux_rsa_export_public_key (FluxRSAPublicKey *public_key, guchar **out, guint *out_len)
{
  PublicKeyM public_key_m;

  flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);
  g_return_if_fail (out != NULL);
  g_return_if_fail (out_len != NULL);

  mpz_to_raw_alloc (public_key->key.n, &public_key_m.n, &public_key_m.n_len);
  mpz_to_raw_alloc (public_key->e, &public_key_m.e, &public_key_m.e_len);

  flux_marshaller_encode_pack (key_marshaller, "rsa-public-key", &public_key_m, (gpointer *) out, out_len);

  g_free (public_key_m.n);
  g_free (public_key_m.e);
}

void
flux_rsa_export_secret_key (FluxRSASecretKey *secret_key, guchar **out, guint *out_len)
{
  SecretKeyM secret_key_m;

  flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);
  g_return_if_fail (out != NULL);
  g_return_if_fail (out_len != NULL);

  mpz_to_raw_alloc (secret_key->key.n, &secret_key_m.n, &secret_key_m.n_len);
  mpz_to_raw_alloc (secret_key->d, &secret_key_m.d, &secret_key_m.d_len);

  flux_marshaller_encode_pack (key_marshaller, "rsa-secret-key", &secret_key_m, (gpointer *) out, out_len);

  g_free (secret_key_m.n);
  g_free (secret_key_m.d);
}

#endif
