/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-ec.h - The Elliptic Curve asymmetric cipher.
 *
 * Copyright (C) 2004 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef _FLUX_EC_H
#define _FLUX_EC_H

#ifdef __FLUX_BUILD
# include "flux-types.h"
# include "flux-object.h"
#else
# include <flux/flux-types.h>
# include <flux/flux-object.h>
#endif

#include <gmp.h>

#define FLUX_EC_KEY(key)        ((FluxECKey *) key)
#define FLUX_EC_PUBLIC_KEY(key) ((FluxECPublicKey *) key)
#define FLUX_EC_SECRET_KEY(key) ((FluxECSecretKey *) key)

typedef enum
{
  FLUX_EC_CURVE_163,
  FLUX_EC_CURVE_233,
  FLUX_EC_CURVE_283,
  FLUX_EC_CURVE_409,
  FLUX_EC_CURVE_MAX
}
FluxECCurveType;

struct _FluxECKey
{
  FluxObject object;

  FluxECCurveType curve_type;
};

struct _FluxECPublicKey
{
  FluxECKey key;

  mpz_t point_x;
  mpz_t point_y;
};

struct _FluxECSecretKey
{
  FluxECKey key;

  mpz_t s;
};

/* --- */

gboolean          flux_ec_generate_keypair    (FluxECCurveType curve_type, FluxECPublicKey **public_key,
                                               FluxECSecretKey **secret_key);
FluxECKey        *flux_ec_dup_key             (FluxECKey *key);

void              flux_ec_finalize_public_key (FluxECPublicKey *pkey);
void              flux_ec_finalize_secret_key (FluxECSecretKey *skey);

FluxECPublicKey  *flux_ec_import_public_key   (const guchar *in, guint in_len);
FluxECSecretKey  *flux_ec_import_secret_key   (const guchar *in, guint in_len);

void              flux_ec_export_public_key   (FluxECPublicKey *public_key, guchar **out, guint *out_len);
void              flux_ec_export_secret_key   (FluxECSecretKey *secret_key, guchar **out, guint *out_len);

guint             flux_ec_get_key_length      (FluxECKey *key);
guint             flux_ec_get_block_size      (FluxECKey *key);

void              flux_ec_encrypt             (FluxECPublicKey *pkey,
                                               guchar      **ciphertext, guint *ciphertext_len,
                                               const guchar *plaintext,  guint  plaintext_len);
void              flux_ec_decrypt             (FluxECSecretKey *skey,
                                               guchar      **plaintext,  guint *plaintext_len,
                                               const guchar *ciphertext, guint  ciphertext_len);
void              flux_ec_sign                (FluxECSecretKey *skey,
                                               guchar      **signature,  guint *signature_len,
                                               const guchar *hash,       guint  hash_len);
gboolean          flux_ec_verify              (FluxECPublicKey *pkey,
                                               const guchar *signature,  guint  signature_len,
                                               const guchar *hash,       guint  hash_len);

#endif  /* _FLUX_EC_H */
