/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-elgamal.c - The Elgamal asymmetric cipher.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: The GNUPG team
 *          Hans Petter Jansson <hpj@copyleft.no>
 *
 * Original code taken from GNUPG and adapted to link dynamically with
 * the GMP (GNU Multiple Precision) bignum library instead of including
 * an obsolete version of its sources, at the price of not having secure
 * memory (this can now be remedied by specifying your own allocator, but
 * we don't do that yet). Also adapted to conform to Flux semantics.
 */

/* For a description of the algorithm, see:
 * Bruce Schneier: Applied Cryptography. John Wiley & Sons, 1996.
 * ISBN 0-471-11709-9. Pages 476 ff.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>          /* time */
#include <glib.h>
#include <gmp.h>
#include "flux-init.h"
#include "flux-rng.h"
#include "flux-marshaller.h"
#include "flux-elgamal.h"

#define D(x)

#define bits_in_mpz(n)   (mpz_sizeinbase (n, 2))
#define bits_to_bytes(b) (((b) + 7) >> 3)
#define bytes_to_bits(b) ((b) << 3)

typedef struct
{
  guchar *p;
  guint   p_len;

  guchar *g;
  guint   g_len;

  guchar *y;
  guint   y_len;
}
PublicKeyM;

typedef struct
{
  guchar *p;
  guint   p_len;

  guchar *g;
  guint   g_len;

  guchar *y;
  guint   y_len;

  guchar *x;
  guint   x_len;
}
SecretKeyM;

static FluxMarshaller *key_marshaller = NULL;

static FluxMarshalledStructEntry public_key_m_entries [] =
{
  FLUX_MARSHAL_RAW_DYNAMIC ("p", PublicKeyM, p, p_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("g", PublicKeyM, g, g_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("y", PublicKeyM, y, y_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStructEntry secret_key_m_entries [] =
{
  FLUX_MARSHAL_RAW_DYNAMIC ("p", SecretKeyM, p, p_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("g", SecretKeyM, g, g_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("y", SecretKeyM, y, y_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("x", SecretKeyM, x, x_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStruct key_marshal_info [] =
{
  { "elgamal-public-key", sizeof (PublicKeyM), FLUX_MARSHALLER_CLEAR, public_key_m_entries },
  { "elgamal-secret-key", sizeof (SecretKeyM), FLUX_MARSHALLER_CLEAR, secret_key_m_entries },
  { NULL }
};

/* Key structure allocators/initializers */

static FluxElgamalPublicKey *new_public (void);
static FluxElgamalSecretKey *new_secret (void);

/* Helpers for flux_elgamal_dup_key (). */

static FluxElgamalPublicKey *dup_public (FluxElgamalPublicKey *old);
static FluxElgamalSecretKey *dup_secret (FluxElgamalSecretKey *old);

/* Deeper static funcs. */

static guint     hex_to_int            (char hex);
static gchar     int_to_hex            (int n);

static void      mpz_to_raw            (mpz_t n, guchar *raw, guint len);
static void      raw_to_mpz            (mpz_t n, const guchar *raw, guint len);
static void      mpz_to_raw_alloc      (mpz_t n, guchar **out, guint *out_len);
static mpz_t    *raw_to_mpz_alloc      (const guchar *raw, guint len);

static void      m_out_of_n            (char *array, int m, int n);
static void      mulm                  (mpz_t w, mpz_t u, mpz_t v, mpz_t m);
static int       build_index           (mpz_t **exparray, int k, int i, int t);
static void      mulpowm               (mpz_t res, mpz_t **basearray, mpz_t **exparray, mpz_t m);

static mpz_t    *alloc_mpz             (void);
static void      free_mpz              (mpz_t *n);
static void      randomize_mpz         (mpz_t *n, unsigned int bits_num);

static mpz_t    *generate_random_prime (unsigned int bits_num, FluxRng *rng);
static mpz_t    *generate_elg_prime    (int mode, unsigned pbits, unsigned qbits, mpz_t *g,
                                        mpz_t **ret_factors, FluxRng *rng);

static gboolean  check_secret_key      (FluxElgamalSecretKey *sk);
static mpz_t    *gen_k                 (mpz_t *p, FluxRng *rng);

static void      elg_generate          (FluxElgamalSecretKey *sk, unsigned nbits, mpz_t **ret_factors);
static void      encrypt               (mpz_t *a, mpz_t *b, mpz_t *input, FluxElgamalPublicKey *pkey);
static void      decrypt               (mpz_t *output, mpz_t *a, mpz_t *b, FluxElgamalSecretKey *skey);
static void      sign                  (mpz_t *a, mpz_t *b, mpz_t *input, FluxElgamalSecretKey *skey);
static gboolean  verify                (mpz_t *a, mpz_t *b, mpz_t *input, FluxElgamalPublicKey *pkey);

static FluxRng *elg_rng = NULL;

static void
init_rng (void)
{
  if (!elg_rng)
    elg_rng = flux_rng_new ();
}

static void
init_marshaller (void)
{
  if (!key_marshaller)
  {
    key_marshaller = flux_marshaller_new ();
    flux_marshaller_register_many (key_marshaller, key_marshal_info);
  }
}

static void
m_out_of_n (char *array, int m, int n)
{
  int i = 0, i1 = 0, j = 0, jp = 0,  j1 = 0, k1 = 0, k2 = 0;

  if (!m || m >= n) return;

  if (m == 1)
  {
    /* Special case. */
    for (i = 0; i < n; i++)
    {
      if (array [i])
      {
        array [i++] = 0;
        if (i >= n) i = 0;
        array [i] = 1;
        return;
      }
    }
    g_error ("Cipher bug #2.");
  }

  for (j = 1; j < n; j++)
  {
    if (array [n - 1] == array [n - j - 1]) continue;
    j1 = j;
    break;
  }

  if (m & 1)
  {
    /* m is odd. */
    if (array [n-1])
    {
      if (j1 & 1)
      {
        k1 = n - j1;
        k2 = k1 + 2;
        if (k2 > n) k2 = n;
        goto leave;
      }
      goto scan;
    }
    
    k2 = n - j1 - 1;
    if (k2 == 0)
    {
      k1 = i;
      k2 = n - j1;
    }
    else if (array [k2] && array [k2 - 1]) k1 = n;
    else k1 = k2 + 1;
  }
  else
  {
    /* m is even. */
    if (!array [n - 1])
    {
      k1 = n - j1;
      k2 = k1 + 1;
      goto leave;
    }

    if (!(j1 & 1))
    {
      k1 = n - j1;
      k2 = k1 + 2;
      if (k2 > n) k2 = n;
      goto leave;
    }
    
    scan:
    
    jp = n - j1 - 1;
    for (i = 1; i <= jp; i++)
    {
      i1 = jp + 2 - i;
      if (array [i1 - 1])
      {
        if (array [i1 - 2])
        {
          k1 = i1 - 1;
          k2 = n - j1;
        }
        else
        {
          k1 = i1 - 1;
          k2 = n + 1 - j1;
        }
        goto leave;
      }
    }
    k1 = 1;
    k2 = n + 1 - m;
  }
  
  leave:
  
  array [k1 - 1] = !array [k1 - 1];
  array [k2 - 1] = !array [k2 - 1];
}

static void
mulm (mpz_t w, mpz_t u, mpz_t v, mpz_t m)
{
  mpz_mul (w, u, v);
  mpz_mod (w, w, m);
}

static int
build_index (mpz_t **exparray, int k, int i, int t)
{
  int j, bitno;
  int index = 0;

  bitno = t - i;
  for (j = k - 1; j >= 0; j--)
  {
    index <<= 1;
    if (mpz_tstbit (*exparray [j], bitno))
      index |= 1;
  }

  return index;
}

/* result = ((base [0] ^ exp [0]) * (base [1] ^ exp [1]) * ...) mod M */

static void
mulpowm (mpz_t res, mpz_t **basearray, mpz_t **exparray, mpz_t m)
{
  int k;          /* Number of elements. */
  int t;          /* Bit size of largest exponent. */
  int i, j, idx;
  mpz_t **G;      /* Table with precomputed values of size 2^k. */
  mpz_t tmp;
  mpz_t *mptr;

  for (k = 0; basearray [k]; k++) ;
  g_assert (k);

  for (t = 0, i = 0; (mptr = exparray [i]); i++)
  {
    j = bits_in_mpz (*mptr);
    if (j > t) t = j;
  }

  g_assert (i == k);
  g_assert (t);
  g_assert (k < 10);

  G = g_new0 (mpz_t *, (1 << k));

  /* Calculate. */

  mpz_init (tmp);
  mpz_set_ui (res, 1);

  for (i = 1; i <= t; i++)
  {
    mulm (tmp, res, res, m);
    idx = build_index (exparray, k, i, t);
    g_assert (idx >= 0 && idx < (1 << k));

    if (!G [idx])
    {
      if (!idx)
      {
        G [0] = alloc_mpz ();
        mpz_set_ui (*G [0], 1);
      }
      else
      {
        for (j = 0; j < k; j++)
        {
          if ((idx & (1 << j)))
          {
            if (!G [idx])
            {
              G [idx] = alloc_mpz ();
              mpz_set (*G [idx], *basearray [j]);
            }
            else
              mulm (*G [idx], *G [idx], *basearray [j], m);
          }
        }

        if (!G [idx])
        {
          G [idx] = alloc_mpz ();
        }
      }
    }

    mulm (res, tmp, *G [idx], m);
  }

  /* Cleanup. */

  mpz_clear (tmp);

  for (i = 0; i < (1 << k); i++)
  {
    if (G [i])
      free_mpz (G [i]);
  }

  g_free (G);
}

static mpz_t *
alloc_mpz (void)
{
  mpz_t *n;

  n = g_new (mpz_t, 1);
  mpz_init (*n);
  return n;
}

static void
free_mpz (mpz_t *n)
{
  mpz_clear (*n);
  g_free (n);
}

static void
randomize_mpz (mpz_t *n, guint bits_num)
{
  /* #warning FIXME: Do not use static. */
  static gmp_randstate_t rand_state;
  static gboolean got_rand_state = FALSE;

  if (!got_rand_state)
  {
    /* Initialize the random number generator. This scheme gives us "pure" random
     * chunks of max 128 bits. */

    gmp_randinit (rand_state, GMP_RAND_ALG_LC, 128);
    gmp_randseed_ui (rand_state, time (NULL));

    got_rand_state = TRUE;
  }

  /* Generate a random number between 0 and 2 ^ (bits_num - 1), inclusive. */

  mpz_urandomb (*n, rand_state, bits_num);
}

static void
randomize_mpz_strong (mpz_t *n, guint bits_num, FluxRng *rng)
{
  guchar *buf;

  buf = g_malloc (bits_to_bytes (bits_num));

  flux_rng_get_bytes (rng, buf, bits_to_bytes (bits_num));
  buf [0] &= 0xff >> (8 - (bits_num % 8));  /* Trim MSB */
  raw_to_mpz         (*n,  buf, bits_to_bytes (bits_num));

  g_free (buf);
}

static guint
hex_to_int (char hex)
{
  if (hex >= '0' && hex <= '9')
    return (hex - '0');

  if (hex >= 'a' && hex <= 'f')
    return (hex - 'a' + 0x0a);

  if (hex >= 'A' && hex <= 'F')
    return (hex - 'A' + 0x0a);

  g_error ("Got a digit not in base16.");
  return (-1);  /* Never executed */
}

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void 
mpz_to_raw (mpz_t n, guchar *raw, guint len)
{
  char *mpz_b16;
  guint mpz_b16_len, mpz_b256_len;
  guint i, j;

  if (mpz_sgn (n) < 0)
    g_error ("Got negative mpz.");

  /* Get base16 string for conversion to raw */

  mpz_b16 = mpz_get_str (NULL, 16, n);
  g_assert (mpz_b16);

  mpz_b16_len  = strlen (mpz_b16);
  mpz_b256_len = (mpz_b16_len + 1) / 2;

  if (len < mpz_b256_len)
    g_error ("Destination too small to fit mpz: %u vs %u.", len, mpz_b256_len);

  /* Convert base16 string to raw */

  memset (raw, 0, len);
  j = len - mpz_b256_len;
  i = 0;

  if (mpz_b16_len % 2)
    raw [j++] = hex_to_int (mpz_b16 [i++]);

  for ( ; mpz_b16 [i]; i += 2, j++)
  {
    raw [j] = (hex_to_int (mpz_b16 [i]) << 4) +
              (hex_to_int (mpz_b16 [i + 1]));
  }

  g_free (mpz_b16);
}

static void
mpz_to_raw_alloc (mpz_t n, guchar **out, guint *out_len)
{
  *out_len = bits_to_bytes (bits_in_mpz (n));
  *out = g_malloc (*out_len);

  mpz_to_raw (n, *out, *out_len);
}

static void
raw_to_mpz (mpz_t n, const guchar *raw, guint len)
{
  gchar *mpz_b16;
  guint i;

  /* Create a base16 string to import with mpz_set_str () */

  mpz_b16 = g_malloc ((len * 2) + 1);

  for (i = 0; i < len; i++)
  {
    mpz_b16 [i * 2]       = int_to_hex ((raw [i] & 0xf0) >> 4);
    mpz_b16 [(i * 2) + 1] = int_to_hex (raw [i] & 0x0f);
  }

  mpz_b16 [i * 2] = '\0';

  /* Import string */

  mpz_set_str (n, mpz_b16, 16);
  g_free (mpz_b16);
}

static mpz_t *
raw_to_mpz_alloc (const guchar *raw, guint len)
{
  mpz_t *n;

  n = alloc_mpz ();
  raw_to_mpz (*n, raw, len);
  return n;
}

static mpz_t *
generate_random_prime (guint bits_num, FluxRng *rng)
{
  mpz_t *prime;

  D (g_print ("In generate_random_prime.\n"));

  prime = alloc_mpz ();

  /* Loop until we find a probable prime. */

  for (;;)
  {
    /* Generate a random number between 0 and 2 ^ (bits_num - 1), inclusive. */

    D (g_print ("  Randomizing...\n"));

    randomize_mpz_strong (prime, bits_num, rng);

    /* Set high and low order bit. This guarantees size of bits_num and that
     * it's not divisible by 2. */

    D (g_print ("  Setting bits...\n"));

    mpz_setbit (*prime, bits_num - 1);
    mpz_setbit (*prime, 0);

    /* Do a Miller-Rabin probabilistic prime test with 7 reps out of [5-10]. */

    D (g_print ("  Checking primeness...\n"));

    if (mpz_probab_prime_p (*prime, 7) > 0)
      break;
  }

  D (g_print ("Finished generate_random_prime.\n"));

  return prime;
}

/*
 * We do not need to use the strongest RNG because we gain no extra
 * security from it - The prime number is public and we could also
 * offer the factors for those who are willing to check that it is
 * indeed a strong prime.
 *
 * mode 0: Standard
 *      1: Make sure that at least one factor is of size qbits.
 */

static mpz_t *
generate_elg_prime (int mode, unsigned pbits, unsigned qbits, mpz_t *g, mpz_t **ret_factors, FluxRng *rng)
{
  int n;               /* Number of factors. */
  int m;               /* Number of primes in pool. */
  unsigned fbits;      /* Length of prime factors. */
  mpz_t **factors;     /* Current factors. */
  mpz_t **pool;        /* Pool of primes. */
  mpz_t *q;            /* First prime factor (variable). */
  mpz_t *prime;        /* Prime test value. */
  mpz_t *q_factor;     /* Used for mode 1. */
  guchar *perms = NULL;
  int i, j;
  int count1, count2;
  unsigned nprime;
  unsigned req_qbits = qbits;  /* The requested q bits size */

  /* Find number of needed prime factors. */
  for (n = 1; (pbits - qbits - 1) / n >= qbits; n++) ;

  n--;
  if (!n || (mode == 1 && n < 2))
    g_error ("Can't generate prime.");

  if (mode == 1)
  {
    n--;
    fbits = (pbits - 2 * req_qbits - 1) / n;
    qbits =  pbits - req_qbits - n * fbits;
  }
  else
  {
    fbits = (pbits - req_qbits - 1) / n;
    qbits = pbits - n * fbits;
  }

  prime = alloc_mpz ();

  D (g_print ("Generating random prime (a)...\n"));

  q = generate_random_prime (qbits, rng);

  D (g_print ("Generating random prime (b)...\n"));

  q_factor = (mode == 1) ? generate_random_prime (req_qbits, rng) : NULL;

  D (g_print ("Random primes generated.\n"));

  /* Allocate an array to hold the factors + 2 for later usage. */
  factors = g_new0 (mpz_t *, n + 2);

  /* Make a pool of 3n+5 primes (this is an arbitrary value). */
  m = n * 3 + 5;
  if (mode == 1) m += 5; /* need some more for DSA */
  if (m < 25)    m = 25;

  pool = g_new0 (mpz_t *, m);

  /* Permutate over the pool of primes. */

  count1 = count2 = 0;

  D (g_print ("Running first big loop...\n"));

  do
  {
    next_try:

    if (!perms)
    {
      /* Allocate new primes. */
      for (i = 0; i < m; i++ )
      {
        if (pool [i])  /* Is this necessary? */
        {
          free_mpz (pool [i]);
          pool [i] = NULL;
        }
      }

      /* Init m_out_of_n(). */

      perms = g_new0 (guchar, m);

      for (i = 0; i < n; i++)
      {
        perms [i]   = 1;
        pool [i]    = generate_random_prime (fbits, rng);
        factors [i] = pool [i];
      }
    }
    else
    {
      m_out_of_n (perms, n, m);

      for (i = j = 0; i < m && j < n; i++ )
      {
        if (perms [i])
        {
          if (!pool [i])
            pool [i] = generate_random_prime (fbits, rng);

          factors [j++] = pool [i];
        }
      }

      if (i == n)
      {
        g_free (perms);
        perms = NULL;
        goto next_try;  /* Allocate new primes. */
      }
    }

    mpz_set    (*prime, *q);
    mpz_mul_ui (*prime, *prime, 2);
    
    if (mode == 1)
      mpz_mul (*prime, *prime, *q_factor);

    for (i = 0; i < n; i++)
      mpz_mul (*prime, *prime, *factors [i]);

    mpz_add_ui (*prime, *prime, 1);
    nprime = bits_in_mpz (*prime);

    if (nprime < pbits)
    {
      if (++count1 > 20)
      {
        count1 = 0;
        qbits++;
        free_mpz (q);
        q = generate_random_prime (qbits, rng);
        goto next_try;
      }
    }
    else count1 = 0;

    if (nprime > pbits)
    {
      if (++count2 > 20)
      {
        count2 = 0;
        qbits--;
        free_mpz (q);
        q = generate_random_prime (qbits, rng);
        goto next_try;
      }
    }
    else count2 = 0;
  }
  while (!(nprime == pbits && mpz_probab_prime_p (*prime, 7)));

#if 0
  if (ret_factors)
  {
    /* caller wants the factors */
    *ret_factors = malloc((n + 2) * sizeof **ret_factors);
    memset(*ret_factors, 0, (n + 2) * sizeof **ret_factors);

    if (mode == 1)
    {
      i = 0;
      (*ret_factors)[i++] = mpi_copy(q_factor);
      for( ; i <= n; i++) (*ret_factors)[i] = mpi_copy(factors[i]);
    }
    else
    {
      for( ; i < n; i++) (*ret_factors)[i] = mpi_copy(factors[i]);
    }
  }
#endif

  D (g_print ("Running second big loop...\n"));

  if (g)
  {
    mpz_t tmp;
    mpz_t b;
    mpz_t pmin1;

    /* Create a generator (start with 3). */

    mpz_init (tmp);
    mpz_init (b);
    mpz_init (pmin1);

    if (mode == 1)
      g_error ("Cipher bug #1."); /* not yet implemented */

    factors [n]     = q;
    factors [n + 1] = alloc_mpz ();
    mpz_set_ui (*factors [n + 1], 2);

    mpz_sub_ui (pmin1, *prime, 1);
    mpz_set_ui (*g, 2);

    do
    {
      mpz_add_ui (*g, *g, 1);
      
      for (i = 0; i < n + 2; i++)
      {
        mpz_fdiv_q (tmp, pmin1, *factors [i]);

        /* mpz_powm() is okay to use with mod prime. */
        mpz_powm (b, *g, tmp, *prime);
        if (!mpz_cmp_ui (b, 1)) break;
      }
    }
    while (i < n+2);

    free_mpz (factors [n + 1]);
    mpz_clear (tmp);
    mpz_clear (b);
    mpz_clear (pmin1);
  }

  /* Factors are shallow copies of entries in pool. */
  g_free (factors);

  for (i = 0; i < m; i++)
    if (pool [i])
      free_mpz (pool [i]);

  g_free (pool);
  g_free (perms);
  free_mpz (q);

  return prime;
}

/****************
 * generate a random secret exponent k from prime p, so
 * that k is relatively prime to p-1
 */
static mpz_t *
gen_k (mpz_t *p, FluxRng *rng)
{
  mpz_t *k, temp, p_1;
  unsigned int nbits  = bits_in_mpz (*p);

  k = alloc_mpz ();
  mpz_init      (temp);
  mpz_init_set  (p_1, *p);

  mpz_sub_ui (p_1, *p, 1);

  for (;;)
  {
    randomize_mpz_strong (k, nbits, rng);

    for (;;)
    {
      if (!(mpz_cmp (*k, p_1) < 0))   /* Is k < (p - 1)? */
        break;                        /* No.  */

      if (!(mpz_cmp_ui (*k, 0) > 0))  /* Is k > 0? */
        break;                        /* No. */

      /* Find greatest common divisor of k and p_1. If this is 1, we're done. */

      mpz_gcd (temp, *k, p_1);
      if (!mpz_cmp_ui (temp, 1))
        goto found;                   /* k is relatively prime to (p - 1) */

      mpz_add_ui (*k, *k, 1);
    }
  }

found:
  mpz_clear (p_1);
  mpz_clear (temp);

  return k;
}

/****************
 * Generate a key pair with a key of size NBITS
 * Returns: 2 structures filled with all needed values
 *      and an array with n-1 factors of (p-1)
 */

static void
elg_generate (FluxElgamalSecretKey *sk, unsigned nbits, mpz_t **ret_factors)
{
  mpz_t *p;                       /* the prime */
  mpz_t p_min1;
  mpz_t g;
  mpz_t x;                        /* the secret exponent */
  mpz_t y;
  mpz_t temp;
  unsigned qbits;
  guchar *rndbuf;

  mpz_init (p_min1);
  mpz_init (temp);

  if      (nbits < 512)   qbits = 120;
  else if (nbits <= 1024) qbits = 160;
  else if (nbits <= 2048) qbits = 200;
  else                    qbits = 240;

  mpz_init (g);

  D (g_print ("Generating elg prime...\n"));

  p = generate_elg_prime (0, nbits, qbits, &g, ret_factors, elg_rng);
  mpz_sub_ui (p_min1, *p, 1);

  /* Select a random number which has these properties: 0 < x < p-1.
   *
   * This must be a very good random number because this is the
   * secret part.  The prime is public and may be shared anyway,
   * so a random generator level of 1 is used for the prime.
   */

  /* #warning FIXME: Use secure memory. */
  mpz_init (x);

  rndbuf = NULL;

  D (g_print ("Randomizing mpz...\n"));

  do
  {
    randomize_mpz_strong (&x, nbits, elg_rng);
    mpz_clrbit (x, nbits + 1);
  }
  while (!(mpz_cmp_ui (x, 0) > 0 && mpz_cmp (x, p_min1) < 0));

  mpz_init (y);
  mpz_powm (y, g, x, *p);

  /* Copy the stuff to the key structures. */

  mpz_init_set (sk->key.p, *p);
  mpz_init_set (sk->key.g, g);
  mpz_init_set (sk->key.y, y);
  mpz_init_set (sk->x, x);

#if 0
  /* Test our keys (this should never fail). */

  test_keys (sk, nbits - 64);
#endif

  /* NEW FREES */
  free_mpz (p);
  mpz_clear (g);
  mpz_clear (y);
  mpz_clear (x);

  mpz_clear (p_min1);
  mpz_clear (temp);
}

/****************
 * Test whether the secret key is valid.
 * Returns: if this is a valid key.
 */

static gboolean
check_secret_key (FluxElgamalSecretKey *sk)
{
  int rc;
  mpz_t y;

  mpz_init (y);

  mpz_powm (y, sk->key.g, sk->x, sk->key.p);
  rc = mpz_cmp (y, sk->key.y);

  mpz_clear (y);

  return rc == 0;
}

static void
encrypt (mpz_t *a, mpz_t *b, mpz_t *input, FluxElgamalPublicKey *pkey)
{
  mpz_t *k;

  /* Note: maybe we should change the interface, so that it
   * is possible to check that input is < p and return an
   * error code.
   */

  k = gen_k (&pkey->key.p, elg_rng);
  mpz_powm (*a, pkey->key.g, *k, pkey->key.p);

  /* b = (y^k * input) mod p
   *   = ((y^k mod p) * (input mod p)) mod p
   *
   * and because input is < p
   *   = ((y^k mod p) * input) mod p
   */

  mpz_powm (*b, pkey->key.y, *k, pkey->key.p);
  mulm (*b, *b, *input, pkey->key.p);

  free_mpz (k);
}

static void
decrypt (mpz_t *output, mpz_t *a, mpz_t *b, FluxElgamalSecretKey *skey)
{
  mpz_t t1;

  /* #warning FIXME: Use secure memory. */
  mpz_init (t1);

  /* output = b/(a^x) mod p */

  mpz_powm (t1, *a, skey->x, skey->key.p);
  mpz_invert (t1, t1, skey->key.p);
  mulm (*output, *b, t1, skey->key.p);

  mpz_clear (t1);
}

/*
 * Make an Elgamal signature out of INPUT
 */
static void
sign (mpz_t *a, mpz_t *b, mpz_t *input, FluxElgamalSecretKey *skey)
{
  mpz_t *k, t, inv, p_1;

  mpz_init     (t);
  mpz_init     (inv);
  mpz_init_set (p_1, skey->key.p);

  /*
   * b = (t * inv) mod (p-1)
   * b = (t * inv(k,(p-1),(p-1)) mod (p-1)
   * b = (((M-x*a) mod (p-1)) * inv(k,(p-1),(p-1))) mod (p-1)
   *
   */

  mpz_sub_ui (p_1, p_1, 1);
  k = gen_k (&skey->key.p, elg_rng);
  mpz_powm (*a, skey->key.g, *k, skey->key.p);
  mpz_mul (t, skey->x, *a);

  mpz_sub (t, *input, t);
  mpz_mod (t, t, p_1);
  /* mpz_subm (t, *input, t, p_1); */

  while (mpz_sgn (t) < 0)
    mpz_add (t, t, p_1);

  mpz_invert (inv, *k, p_1);
  mulm (*b, t, inv, p_1);

  free_mpz  (k);
  mpz_clear (t);
  mpz_clear (inv);
  mpz_clear (p_1);
}

/*
 * Returns true if the signature composed of A and B is valid.
 */
static gboolean
verify (mpz_t *a, mpz_t *b, mpz_t *input, FluxElgamalPublicKey *pkey)
{
  int rc;
  mpz_t t1;
  mpz_t t2;
  mpz_t *base [4];
  mpz_t *exp [4];

  if (!(mpz_cmp_ui (*a, 0) > 0 && mpz_cmp (*a, pkey->key.p) < 0))
#ifdef D
    g_error ("Assertion 0 < a < p failed.");
#else
    return FALSE;                    /* assertion  0 < a < p  failed */
#endif

  mpz_init (t1);
  mpz_init (t2);

#if 0
  /* t1 = (y^a mod p) * (a^b mod p) mod p */
  mpi_powm(t1, pkey->key.y, a, pkey->key.p);
  mpi_powm(t2, a, b, pkey->key.p);
  mpi_mulm(t1, t1, t2, pkey->key.p);

  /* t2 = g ^ input mod p */
  mpi_powm(t2, pkey->key.g, input, pkey->key.p);

  rc = !mpi_cmp(t1, t2);
#elif 0
  /* t1 = (y^a mod p) * (a^b mod p) mod p */
  base[0] = pkey->key.y;
  exp[0] = a;
  base[1] = a;
  exp[1] = b;
  base[2] = NULL;
  exp[2] = NULL;
  mpi_mulpowm(t1, base, exp, pkey->key.p);

  /* t2 = g ^ input mod p */
  mpi_powm(t2, pkey->key.g, input, pkey->key.p);

  rc = !mpi_cmp(t1, t2);
#else

  /* t1 = g ^ - input * y ^ a * a ^ b  mod p */
  mpz_invert (t2, pkey->key.g, pkey->key.p);  /* I doubt this */
  /* mpz_invm (t2, pkey->key.g, pkey->key.p); */

  base [0] = &t2;
  exp  [0] = input;
  base [1] = &pkey->key.y;
  exp  [1] = a;
  base [2] = a;
  exp  [2] = b;
  base [3] = NULL;
  exp  [3] = NULL;

  mulpowm (t1, base, exp, pkey->key.p);
  rc = mpz_cmp_ui (t1, 1);

#endif

  mpz_clear (t1);
  mpz_clear (t2);
  return rc == 0;
}

/* --------- *
 * Interface *
 * --------- */

static FluxElgamalPublicKey *
new_public (void)
{
  FluxElgamalPublicKey *pkey;

  init_rng ();
  init_marshaller ();

  pkey = g_new0 (FluxElgamalPublicKey, 1);
  flux_object_init              (FLUX_OBJECT (pkey));
  flux_object_set_type          (FLUX_OBJECT (pkey), FLUX_ELGAMAL_PUBLIC_KEY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (pkey), (FluxFinalizeFunc *) flux_elgamal_finalize_public_key);

  return pkey;
}

static FluxElgamalSecretKey *
new_secret (void)
{
  FluxElgamalSecretKey *skey;

  init_rng ();
  init_marshaller ();

  skey = g_new0 (FluxElgamalSecretKey, 1);
  flux_object_init              (FLUX_OBJECT (skey));
  flux_object_set_type          (FLUX_OBJECT (skey), FLUX_ELGAMAL_SECRET_KEY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (skey), (FluxFinalizeFunc *) flux_elgamal_finalize_secret_key);

  return skey;
}

static FluxElgamalPublicKey *
dup_public (FluxElgamalPublicKey *old)
{
  FluxElgamalPublicKey *new;

  new = new_public ();
  mpz_init_set (new->key.p, old->key.p);
  mpz_init_set (new->key.g, old->key.g);
  mpz_init_set (new->key.y, old->key.y);

  return new;
}

static FluxElgamalSecretKey *
dup_secret (FluxElgamalSecretKey *old)
{
  FluxElgamalSecretKey *new;

  new = new_secret ();
  mpz_init_set (new->key.p, old->key.p);
  mpz_init_set (new->key.g, old->key.g);
  mpz_init_set (new->key.y, old->key.y);
  mpz_init_set (new->x,     old->x);

  return new;
}

static FluxElgamalPublicKey *
elg_fork_public_key (FluxElgamalSecretKey *skey)
{
  FluxElgamalPublicKey *pkey;

  flux_object_check (FLUX_OBJECT (skey), FLUX_ELGAMAL_SECRET_KEY_TYPE);

  pkey = new_public ();
  mpz_init_set (pkey->key.p, skey->key.p);
  mpz_init_set (pkey->key.g, skey->key.g);
  mpz_init_set (pkey->key.y, skey->key.y);

  return pkey;
}

gboolean
flux_elgamal_generate_keypair (guint key_length, FluxElgamalPublicKey **public_key,
                               FluxElgamalSecretKey **secret_key)
{
  g_return_val_if_fail (key_length >= 256, FALSE);

  flux_init_mp ();

  if (key_length & 1)
    key_length++;

  *secret_key = new_secret ();
  elg_generate (*secret_key, key_length, NULL);
  *public_key = elg_fork_public_key (*secret_key);

  return TRUE;
}

FluxElgamalKey *
flux_elgamal_dup_key (FluxElgamalKey *old)
{
  FluxElgamalKey *new;

  flux_object_check (FLUX_OBJECT (old), FLUX_ELGAMAL_KEY_TYPE);

  if (flux_object_get_type (FLUX_OBJECT (old)) == FLUX_ELGAMAL_PUBLIC_KEY_TYPE)
    new = FLUX_ELGAMAL_KEY (dup_public (FLUX_ELGAMAL_PUBLIC_KEY (old)));
  else
    new = FLUX_ELGAMAL_KEY (dup_secret (FLUX_ELGAMAL_SECRET_KEY (old)));

  return new;
}

static void
finalize_common (FluxElgamalKey *key)
{
  mpz_clear (key->p);
  mpz_clear (key->g);
  mpz_clear (key->y);
}

void
flux_elgamal_finalize_public_key (FluxElgamalPublicKey *pkey)
{
  finalize_common (FLUX_ELGAMAL_KEY (pkey));
}

void
flux_elgamal_finalize_secret_key (FluxElgamalSecretKey *skey)
{
  finalize_common (FLUX_ELGAMAL_KEY (skey));

  /* Try not to leave traces of the secret exponent. */
  randomize_mpz (&skey->x, bits_in_mpz (skey->x));
  mpz_clear (skey->x);
}

guint
flux_elgamal_get_key_length (FluxElgamalKey *key)
{
  flux_object_check (FLUX_OBJECT (key), FLUX_ELGAMAL_KEY_TYPE);
  return bits_in_mpz (key->p);
}

static guint
get_ciphertext_block_size (mpz_t p)
{
  return bits_to_bytes (bits_in_mpz (p)) * 2;
}

static guint
get_plaintext_block_size (mpz_t p)
{
  return (bits_in_mpz (p) - 2) >> 3;
}

guint
flux_elgamal_get_block_size (FluxElgamalKey *key)
{
  return get_plaintext_block_size (key->p);
}

typedef void (ProcessorFunc) (FluxElgamalKey *key,
                              guchar       **out, guint *out_len,
                              const guchar  *in,  guint  in_len);

static void
elgamal_process_multiple_blocks (FluxElgamalKey *key,
                                 guchar       **out, guint *out_len,
                                 const guchar  *in,  guint  in_len,
                                 guint in_increment,
                                 ProcessorFunc *func)
{
  guchar *buf;
  guint   buf_len;
  gint    in_pos = 0;

  *out     = NULL;
  *out_len = 0;

  for ( ; in_pos < in_len; in_pos += in_increment)
  {
    func (key,
          &buf,        &buf_len,
          in + in_pos, MIN (in_increment, in_len - in_pos));

    if (*out == NULL)
    {
      *out     = buf;
      *out_len = buf_len;
    }
    else if (buf)
    {
      *out = g_realloc (*out, *out_len + buf_len);
      memcpy (*out + *out_len, buf, buf_len);
      *out_len += buf_len;
      g_free (buf);
    }
  }
}

static void
elgamal_sign (FluxElgamalSecretKey *secret_key,
              guchar       **signature, guint *signature_len,
              const guchar  *hash,      guint  hash_len)
{
  guint block_size;
  mpz_t resarr [2];
  mpz_t data;

  block_size = get_ciphertext_block_size (secret_key->key.p);

  mpz_init (resarr [0]);
  mpz_init (resarr [1]);
  mpz_init (data);

  /* Encode hashtext -> mpz */

  raw_to_mpz (data, hash, hash_len);

  /* Sign it */

  sign (&resarr [0], &resarr [1], &data, secret_key);

  /* Encode mpz -> signaturetext */

  *signature     = g_malloc (block_size);
  *signature_len = block_size;

  mpz_to_raw (resarr [0], *signature,                    block_size / 2);
  mpz_to_raw (resarr [1], *signature + (block_size / 2), block_size / 2);

  /* Free scratch variables */
  
  mpz_clear (resarr[0]);
  mpz_clear (resarr[1]);
  mpz_clear (data);
  
  /* *signature = [ab] */
}

void
flux_elgamal_sign (FluxElgamalSecretKey *secret_key,
                   guchar       **signature, guint *signature_len,
                   const guchar  *hash,      guint  hash_len)
{
  flux_object_check (FLUX_OBJECT (secret_key), FLUX_ELGAMAL_SECRET_KEY_TYPE);

  elgamal_process_multiple_blocks (FLUX_ELGAMAL_KEY (secret_key),
                                   signature, signature_len,
                                   hash,      hash_len,
                                   get_plaintext_block_size (secret_key->key.p),
                                   (ProcessorFunc *) elgamal_sign);
}

gboolean
flux_elgamal_verify (FluxElgamalPublicKey *public_key,
                     const guchar *signature, guint signature_len,
                     const guchar *hash,      guint hash_len)
{
  gboolean result;
  mpz_t    hash_mpz;
  mpz_t    cip_a, cip_b;
  guint    signature_pos, hash_pos;
  guint    signature_increment, hash_increment;

  flux_object_check (FLUX_OBJECT (public_key), FLUX_ELGAMAL_PUBLIC_KEY_TYPE);

  signature_increment = get_ciphertext_block_size (public_key->key.p);
  hash_increment      = get_plaintext_block_size  (public_key->key.p);

  if (signature_len % signature_increment != 0)
    return FALSE;

  for (signature_pos = 0, hash_pos = 0, result = TRUE;

       signature_pos < signature_len && hash_pos < hash_len && result == TRUE;

       signature_pos += get_ciphertext_block_size (public_key->key.p),
       hash_pos      += get_plaintext_block_size  (public_key->key.p))
  {
    mpz_init (hash_mpz);
    mpz_init (cip_a);
    mpz_init (cip_b);

    /* Encode signaturetext / hash -> MPIs */

    raw_to_mpz (cip_a,    signature + signature_pos,                             signature_increment / 2);
    raw_to_mpz (cip_b,    signature + signature_pos + (signature_increment / 2), signature_increment / 2);
    raw_to_mpz (hash_mpz, hash + hash_pos, MIN (hash_increment, hash_len - hash_pos));

    /* Verify it */

    result = verify (&cip_a, &cip_b, &hash_mpz, public_key);

    /* Free scratch variables */

    mpz_clear (cip_a);
    mpz_clear (cip_b);
    mpz_clear (hash_mpz);
  }

  if (signature_pos < signature_len || hash_pos < hash_len)
    result = FALSE;

  return result;
}

static void
elgamal_encrypt (FluxElgamalPublicKey *public_key,
                 guchar       **ciphertext, guint *ciphertext_len,
                 const guchar  *plaintext,  guint  plaintext_len)
{
  guint  block_size;
  mpz_t  resarr [2];
  mpz_t *data;

  block_size = get_ciphertext_block_size (public_key->key.p);

  mpz_init (resarr [0]);
  mpz_init (resarr [1]);
  
  /* Encode plaintext -> mpz */

  data = raw_to_mpz_alloc (plaintext, plaintext_len);
  mpz_setbit (*data, bytes_to_bits (plaintext_len));

  /* Encrypt it */

  encrypt (&resarr [0], &resarr [1], data, public_key);

  /* Encode MPIs -> ciphertext */

  *ciphertext     = g_malloc (block_size);
  *ciphertext_len = block_size;

  mpz_to_raw (resarr [0], *ciphertext,                    block_size / 2);
  mpz_to_raw (resarr [1], *ciphertext + (block_size / 2), block_size / 2);

  /* Free scratch variables */

  free_mpz  (data);
  mpz_clear (resarr [0]);
  mpz_clear (resarr [1]);

  /* *ciphertext = [ab] */
}

void
flux_elgamal_encrypt (FluxElgamalPublicKey *public_key,
                      guchar       **ciphertext, guint *ciphertext_len,
                      const guchar  *plaintext,  guint  plaintext_len)
{
  flux_object_check (FLUX_OBJECT (public_key), FLUX_ELGAMAL_PUBLIC_KEY_TYPE);

  elgamal_process_multiple_blocks (FLUX_ELGAMAL_KEY (public_key),
                                   ciphertext, ciphertext_len,
                                   plaintext,  plaintext_len,
                                   get_plaintext_block_size (public_key->key.p),
                                   (ProcessorFunc *) elgamal_encrypt);
}

static void
elgamal_decrypt (FluxElgamalSecretKey *secret_key,
                 guchar       **plaintext,  guint *plaintext_len,
                 const guchar  *ciphertext, guint  ciphertext_len)
{
  guint  block_size;
  mpz_t  result;
  mpz_t *cip_a, *cip_b;

  block_size = get_ciphertext_block_size (secret_key->key.p);
  if (ciphertext_len != block_size)
  {
    *plaintext     = NULL;
    *plaintext_len = 0;
    return;
  }

  mpz_init (result);

  /* Encode ciphertext -> mpz */

  cip_a = raw_to_mpz_alloc (ciphertext,                    block_size / 2);
  cip_b = raw_to_mpz_alloc (ciphertext + (block_size / 2), block_size / 2);

  /* Decrypt it */

  decrypt (&result, cip_a, cip_b, secret_key);

  /* Encode mpz -> plaintext */

  *plaintext_len = bits_to_bytes (bits_in_mpz (result) - 1);
  mpz_clrbit (result, bits_in_mpz (result) - 1);

  *plaintext = g_malloc (*plaintext_len);
  mpz_to_raw (result, *plaintext, *plaintext_len);

  /* Free scratch variables */

  free_mpz  (cip_a);
  free_mpz  (cip_b);
  mpz_clear (result);
}

void
flux_elgamal_decrypt (FluxElgamalSecretKey *secret_key,
                      guchar       **plaintext,  guint *plaintext_len,
                      const guchar  *ciphertext, guint  ciphertext_len)
{
  flux_object_check (FLUX_OBJECT (secret_key), FLUX_ELGAMAL_SECRET_KEY_TYPE);

  elgamal_process_multiple_blocks (FLUX_ELGAMAL_KEY (secret_key),
                                   plaintext,  plaintext_len,
                                   ciphertext, ciphertext_len,
                                   get_ciphertext_block_size (secret_key->key.p),
                                   (ProcessorFunc *) elgamal_decrypt);
}

FluxElgamalPublicKey *
flux_elgamal_import_public_key (const guchar *in, guint in_len)
{
  FluxElgamalPublicKey *public_key;
  PublicKeyM           *public_key_m;
  mpz_t                 temp_mpz;

  g_return_val_if_fail (in != NULL, NULL);
  g_return_val_if_fail (in_len > 1, NULL);

  flux_init_mp ();

  public_key = new_public ();

  if (!flux_marshaller_decode_pack (key_marshaller, "elgamal-public-key",
                                    in, in_len, (gpointer) &public_key_m))
  {
    flux_object_unref (FLUX_OBJECT (public_key));
    return NULL;
  }

  if (public_key_m->p == NULL || public_key_m->g == NULL ||
      public_key_m->y == NULL)
  {
    if (public_key_m->p)
      g_free (public_key_m->p);
    if (public_key_m->g)
      g_free (public_key_m->g);
    if (public_key_m->y)
      g_free (public_key_m->y);

    flux_object_unref (FLUX_OBJECT (public_key));
    g_free (public_key_m);
    return NULL;
  }

  mpz_init (temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->p, public_key_m->p_len);
  mpz_init_set (public_key->key.p, temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->g, public_key_m->g_len);
  mpz_init_set (public_key->key.g, temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->y, public_key_m->y_len);
  mpz_init_set (public_key->key.y, temp_mpz);

  mpz_clear (temp_mpz);
  g_free (public_key_m->p);
  g_free (public_key_m->g);
  g_free (public_key_m->y);
  g_free (public_key_m);
  return public_key;
}

FluxElgamalSecretKey *
flux_elgamal_import_secret_key (const guchar *in, guint in_len)
{
  FluxElgamalSecretKey *secret_key;
  SecretKeyM           *secret_key_m;
  mpz_t                 temp_mpz;

  g_return_val_if_fail (in != NULL, NULL);
  g_return_val_if_fail (in_len > 1, NULL);

  flux_init_mp ();

  secret_key = new_secret ();

  if (!flux_marshaller_decode_pack (key_marshaller, "elgamal-secret-key",
                                    in, in_len, (gpointer) &secret_key_m))
  {
    flux_object_unref (FLUX_OBJECT (secret_key));
    return NULL;
  }

  if (secret_key_m->p == NULL || secret_key_m->g == NULL ||
      secret_key_m->y == NULL || secret_key_m->x == NULL)
  {
    if (secret_key_m->p)
      g_free (secret_key_m->p);
    if (secret_key_m->g)
      g_free (secret_key_m->g);
    if (secret_key_m->y)
      g_free (secret_key_m->y);
    if (secret_key_m->x)
      g_free (secret_key_m->x);

    flux_object_unref (FLUX_OBJECT (secret_key));
    g_free (secret_key_m);
    return NULL;
  }

  mpz_init (temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->p, secret_key_m->p_len);
  mpz_init_set (secret_key->key.p, temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->g, secret_key_m->g_len);
  mpz_init_set (secret_key->key.g, temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->y, secret_key_m->y_len);
  mpz_init_set (secret_key->key.y, temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->x, secret_key_m->x_len);
  mpz_init_set (secret_key->x, temp_mpz);

  mpz_clear (temp_mpz);
  g_free (secret_key_m->p);
  g_free (secret_key_m->g);
  g_free (secret_key_m->y);
  g_free (secret_key_m->x);
  g_free (secret_key_m);
  return secret_key;
}

void
flux_elgamal_export_public_key (FluxElgamalPublicKey *public_key, guchar **out, guint *out_len)
{
  PublicKeyM public_key_m;

  flux_object_check (FLUX_OBJECT (public_key), FLUX_ELGAMAL_PUBLIC_KEY_TYPE);
  g_return_if_fail (out != NULL);
  g_return_if_fail (out_len != NULL);

  mpz_to_raw_alloc (public_key->key.p, &public_key_m.p, &public_key_m.p_len);
  mpz_to_raw_alloc (public_key->key.g, &public_key_m.g, &public_key_m.g_len);
  mpz_to_raw_alloc (public_key->key.y, &public_key_m.y, &public_key_m.y_len);

  flux_marshaller_encode_pack (key_marshaller, "elgamal-public-key", &public_key_m, (gpointer *) out, out_len);

  g_free (public_key_m.p);
  g_free (public_key_m.g);
  g_free (public_key_m.y);
}

void
flux_elgamal_export_secret_key (FluxElgamalSecretKey *secret_key, guchar **out, guint *out_len)
{
  SecretKeyM secret_key_m;

  flux_object_check (FLUX_OBJECT (secret_key), FLUX_ELGAMAL_SECRET_KEY_TYPE);
  g_return_if_fail (out != NULL);
  g_return_if_fail (out_len != NULL);

  mpz_to_raw_alloc (secret_key->key.p, &secret_key_m.p, &secret_key_m.p_len);
  mpz_to_raw_alloc (secret_key->key.g, &secret_key_m.g, &secret_key_m.g_len);
  mpz_to_raw_alloc (secret_key->key.y, &secret_key_m.y, &secret_key_m.y_len);
  mpz_to_raw_alloc (secret_key->x, &secret_key_m.x, &secret_key_m.x_len);

  flux_marshaller_encode_pack (key_marshaller, "elgamal-secret-key", &secret_key_m, (gpointer *) out, out_len);

  g_free (secret_key_m.p);
  g_free (secret_key_m.g);
  g_free (secret_key_m.y);
  g_free (secret_key_m.x);
}
