/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-elgamal.h - The Elgamal asymmetric cipher.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: The GNUPG team
 *          Hans Petter Jansson <hpj@copyleft.no>
 *
 * Original code taken from GNUPG and adapted to link dynamically with
 * the GMP (GNU Multiple Precision) bignum library instead of including
 * an obsolete version of its sources, at the price of not having secure
 * memory (this can now be remedied by specifying your own allocator, but
 * we don't do that yet). Also adapted to conform to Flux semantics.
 */

#ifndef _FLUX_ELGAMAL_H
#define _FLUX_ELGAMAL_H

#ifdef __FLUX_BUILD
# include "flux-types.h"
# include "flux-object.h"
#else
# include <flux/flux-types.h>
# include <flux/flux-object.h>
#endif

#include <gmp.h>

#define FLUX_ELGAMAL_KEY(key)        ((FluxElgamalKey *) key)
#define FLUX_ELGAMAL_PUBLIC_KEY(key) ((FluxElgamalPublicKey *) key)
#define FLUX_ELGAMAL_SECRET_KEY(key) ((FluxElgamalSecretKey *) key)

struct _FluxElgamalKey
{
  FluxObject object;

  mpz_t p;
  mpz_t g;
  mpz_t y;
};

struct _FluxElgamalPublicKey
{
  FluxElgamalKey key;
};

struct _FluxElgamalSecretKey
{
  FluxElgamalKey key;

  mpz_t x;
};

/* --- */

gboolean              flux_elgamal_generate_keypair    (guint key_length, FluxElgamalPublicKey **public_key,
                                                        FluxElgamalSecretKey **secret_key);
FluxElgamalKey       *flux_elgamal_dup_key             (FluxElgamalKey *key);

void                  flux_elgamal_finalize_secret_key (FluxElgamalSecretKey *skey);
void                  flux_elgamal_finalize_public_key (FluxElgamalPublicKey *pkey);

FluxElgamalPublicKey *flux_elgamal_import_public_key   (const guchar *in, guint in_len);
FluxElgamalSecretKey *flux_elgamal_import_secret_key   (const guchar *in, guint in_len);

void                  flux_elgamal_export_public_key   (FluxElgamalPublicKey *public_key,
                                                        guchar **out, guint *out_len);
void                  flux_elgamal_export_secret_key   (FluxElgamalSecretKey *secret_key,
                                                        guchar **out, guint *out_len);

guint                 flux_elgamal_get_key_length      (FluxElgamalKey *key);
guint                 flux_elgamal_get_block_size      (FluxElgamalKey *key);

void                  flux_elgamal_encrypt             (FluxElgamalPublicKey *pkey,
                                                        guchar      **ciphertext, guint *ciphertext_len,
                                                        const guchar *plaintext,  guint  plaintext_len);
void                  flux_elgamal_decrypt             (FluxElgamalSecretKey *skey,
                                                        guchar      **plaintext,  guint *plaintext_len,
                                                        const guchar *ciphertext, guint  ciphertext_len);
void                  flux_elgamal_sign                (FluxElgamalSecretKey *skey,
                                                        guchar      **signature,  guint *signature_len,
                                                        const guchar *hash,       guint  hash_len);
gboolean              flux_elgamal_verify              (FluxElgamalPublicKey *pkey,
                                                        const guchar *signature,  guint  signature_len,
                                                        const guchar *hash,       guint  hash_len);

#endif  /* _FLUX_ELGAMAL_H */
