/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-envoy.c - N-ary tree-based invocation layer for FluxIO.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "flux-comm.h"
#include "flux-envoy.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_ENVOY_TYPE)

typedef struct _FluxEnvoyAssoc
{
  guint                 id;
  gchar                *token;
  FluxEnvoyDequeueFunc *func;
  gpointer              data;
}
FluxEnvoyAssoc;

static FluxEnvoyAssoc *
assoc_new (FluxEnvoy *envoy, const gchar *token, FluxEnvoyDequeueFunc *func, gpointer data)
{
  FluxEnvoyAssoc *assoc;

  assoc = g_new0 (FluxEnvoyAssoc, 1);

  assoc->token = g_strdup (token);
  assoc->func  = func;
  assoc->data  = data;
  assoc->id    = envoy->next_id++;

  return assoc;
}

static void
assoc_del (FluxEnvoyAssoc *assoc)
{
  g_free (assoc->token);
  g_free (assoc);
}

static void
dequeue (FluxComm *comm, FluxTree *tree, gpointer data)
{
  FluxEnvoy *envoy;
  GList *l;
  gboolean handled = FALSE;

  envoy = FLUX_ENVOY (comm);

  for (l = envoy->assoc_list; l; l = l->next)
  {
    FluxEnvoyAssoc *assoc = l->data;

    if (!flux_tree_strcmp (tree, assoc->token))
      handled = assoc->func (envoy, tree, assoc->data);

    if (handled)
      break;
  }

  if (!handled)
    flux_tree_destroy (tree);
}

void
flux_envoy_init (FluxEnvoy *envoy, FluxIO *io)
{
  flux_comm_init                (FLUX_COMM (envoy), io);
  flux_object_set_type          (FLUX_OBJECT (envoy), FLUX_ENVOY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (envoy), flux_envoy_finalize);

  flux_comm_set_dequeue_func (FLUX_COMM (envoy), dequeue, NULL);
}

FluxEnvoy *
flux_envoy_new (FluxIO *io)
{
  FluxEnvoy *envoy;

  envoy = g_new0 (FluxEnvoy, 1);
  flux_envoy_init (envoy, io);

  return envoy;
}

void
flux_envoy_finalize (FluxObject *obj)
{
  FluxEnvoy *envoy;
  GList *l;

  envoy = FLUX_ENVOY (obj);

  for (l = envoy->assoc_list; l; l = l->next)
    assoc_del ((FluxEnvoyAssoc *) l->data);

  g_list_free (envoy->assoc_list);

  flux_comm_finalize (obj);
}

guint
flux_envoy_add_dequeuer_before (FluxEnvoy *envoy, const gchar *token, FluxEnvoyDequeueFunc *func,
                                gpointer data)
{
  FluxEnvoyAssoc *assoc;

  assoc = assoc_new (envoy, token, func, data);
  envoy->assoc_list = g_list_prepend (envoy->assoc_list, assoc);

  return assoc->id;
}

guint
flux_envoy_add_dequeuer_after (FluxEnvoy *envoy, const gchar *token, FluxEnvoyDequeueFunc *func,
                               gpointer data)
{
  FluxEnvoyAssoc *assoc;

  assoc = assoc_new (envoy, token, func, data);
  envoy->assoc_list = g_list_append (envoy->assoc_list, assoc);

  return assoc->id;
}

void
flux_envoy_remove_dequeuer_by_id (FluxEnvoy *envoy, guint id)
{
  GList *l, *next;

  for (l = envoy->assoc_list; l; l = next)
  {
    FluxEnvoyAssoc *assoc = l->data;
    next = l->next;

    if (assoc->id == id)
    {
      envoy->assoc_list = g_list_remove_link (envoy->assoc_list, l);
      g_list_free_1 (l);
      assoc_del (assoc);
      return;
    }
  }

  g_warning ("Tried to remove nonexistent id %u from envoy.", id);
}

void
flux_envoy_remove_dequeuers_by_token (FluxEnvoy *envoy, const gchar *token)
{
  GList *l, *next;

  for (l = envoy->assoc_list; l; l = next)
  {
    FluxEnvoyAssoc *assoc = l->data;
    next = l->next;

    if (!strcmp (assoc->token, token))
    {
      envoy->assoc_list = g_list_remove_link (envoy->assoc_list, l);
      g_list_free_1 (l);
      assoc_del (assoc);
    }
  }
}

void
flux_envoy_remove_dequeuers_by_func (FluxEnvoy *envoy, FluxEnvoyDequeueFunc *func)
{
  GList *l, *next;

  for (l = envoy->assoc_list; l; l = next)
  {
    FluxEnvoyAssoc *assoc = l->data;
    next = l->next;

    if (assoc->func == func)
    {
      envoy->assoc_list = g_list_remove_link (envoy->assoc_list, l);
      g_list_free_1 (l);
      assoc_del (assoc);
    }
  }
}

void
flux_envoy_remove_dequeuers_by_data (FluxEnvoy *envoy, gpointer data)
{
  GList *l, *next;

  for (l = envoy->assoc_list; l; l = next)
  {
    FluxEnvoyAssoc *assoc = l->data;
    next = l->next;

    if (assoc->data == data)
    {
      envoy->assoc_list = g_list_remove_link (envoy->assoc_list, l);
      g_list_free_1 (l);
      assoc_del (assoc);
    }
  }
}

void
flux_envoy_remove_dequeuer_by_all (FluxEnvoy *envoy, const gchar *token,
                                   FluxEnvoyDequeueFunc *func, gpointer data)
{
  GList *l, *next;

  for (l = envoy->assoc_list; l; l = next)
  {
    FluxEnvoyAssoc *assoc = l->data;
    next = l->next;

    if (assoc->data == data && assoc->func == func && !strcmp (assoc->token, token))
    {
      envoy->assoc_list = g_list_remove_link (envoy->assoc_list, l);
      g_list_free_1 (l);
      assoc_del (assoc);
      break;
    }
  }
}

void
flux_envoy_remove_dequeuers_by_token_and_func (FluxEnvoy *envoy, const gchar *token,
                                               FluxEnvoyDequeueFunc *func)
{
  GList *l, *next;

  for (l = envoy->assoc_list; l; l = next)
  {
    FluxEnvoyAssoc *assoc = l->data;
    next = l->next;

    if (assoc->func == func && !strcmp (assoc->token, token))
    {
      envoy->assoc_list = g_list_remove_link (envoy->assoc_list, l);
      g_list_free_1 (l);
      assoc_del (assoc);
      break;
    }
  }
}

void
flux_envoy_add_dequeuer_list_before (FluxEnvoy *envoy, FluxEnvoyDequeuer *dequeuers,
                                     gpointer data)
{
  gint i;

  /* Need to add dequeuers backwards, so they appear before the old ones,
   * but in the right order internally. */

  for (i = 0; dequeuers [i].token; i++) ;

  for (i--; i >= 0; i--)
    flux_envoy_add_dequeuer_before (envoy, dequeuers [i].token, dequeuers [i].func, data);
}

void
flux_envoy_add_dequeuer_list_after (FluxEnvoy *envoy, FluxEnvoyDequeuer *dequeuers,
                                    gpointer data)
{
  gint i;

  for (i = 0; dequeuers [i].token; i++)
    flux_envoy_add_dequeuer_after (envoy, dequeuers [i].token, dequeuers [i].func, data);
}

void
flux_envoy_remove_dequeuer_list (FluxEnvoy *envoy, FluxEnvoyDequeuer *dequeuers)
{
  gint i;

  for (i = 0; dequeuers [i].token; i++)
    flux_envoy_remove_dequeuers_by_token_and_func (envoy, dequeuers [i].token,
                                                   dequeuers [i].func);
}
