/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-ext-hash-table.c - Hash table with multiple values per key.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <glib.h>

#include "flux-ext-hash-table.h"

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_EXT_HASH_TABLE_TYPE)

void
flux_ext_hash_table_init (FluxExtHashTable *ext_hash_table,
                          GHashFunc hash_func, GEqualFunc equal_func,
                          GDestroyNotify key_destroy_func,
                          GDestroyNotify value_destroy_func)
{
  flux_object_init              (FLUX_OBJECT (ext_hash_table));
  flux_object_set_type          (FLUX_OBJECT (ext_hash_table), FLUX_EXT_HASH_TABLE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (ext_hash_table),
                                 (FluxFinalizeFunc *) flux_ext_hash_table_finalize);

  ext_hash_table->hash_table = g_hash_table_new_full (hash_func, equal_func,
                                                      NULL, NULL);

  ext_hash_table->key_destroy_func   = key_destroy_func;
  ext_hash_table->value_destroy_func = value_destroy_func;
}

void
flux_ext_hash_table_finalize (FluxObject *object)
{
  FluxExtHashTable *ext_hash_table = FLUX_EXT_HASH_TABLE (object);

  g_hash_table_destroy (ext_hash_table->hash_table);
}

FluxExtHashTable *
flux_ext_hash_table_new (GHashFunc hash_func, GEqualFunc equal_func)
{
  return flux_ext_hash_table_new_full (hash_func, equal_func, NULL, NULL);
}

FluxExtHashTable *
flux_ext_hash_table_new_full (GHashFunc hash_func, GEqualFunc equal_func,
                              GDestroyNotify key_destroy_func,
                              GDestroyNotify value_destroy_func)
{
  FluxExtHashTable *ext_hash_table;

  ext_hash_table = g_new0 (FluxExtHashTable, 1);
  flux_ext_hash_table_init (ext_hash_table, hash_func, equal_func,
                            key_destroy_func, value_destroy_func);
  return ext_hash_table;
}

void
flux_ext_hash_table_insert (FluxExtHashTable *ext_hash_table, gpointer key, gpointer value)
{
  gboolean  have_old_key;
  gpointer  old_key;
  GList    *values;

  assert_types (ext_hash_table);

  have_old_key = g_hash_table_lookup_extended (ext_hash_table->hash_table, key,
                                               &old_key, (gpointer) &values);

  if (have_old_key)
  {
    if (ext_hash_table->key_destroy_func)
      ext_hash_table->key_destroy_func (key);

    key = old_key;
  }
  else
    values = NULL;

  values = g_list_prepend (values, value);
  g_hash_table_insert (ext_hash_table->hash_table, key, values);
}

guint
flux_ext_hash_table_size (FluxExtHashTable *ext_hash_table)
{
  assert_types (ext_hash_table);

  return g_hash_table_size (ext_hash_table->hash_table);
}

GList *
flux_ext_hash_table_lookup (FluxExtHashTable *ext_hash_table, gconstpointer key)
{
  GList *values;

  assert_types (ext_hash_table);

  values = g_hash_table_lookup (ext_hash_table->hash_table, key);
  return g_list_copy (values);
}

typedef struct
{
  FluxExtHashTable *ext_hash_table;
  GHFunc            func;
  gpointer          user_data;
}
ForEachContext;

static void
foreach_cb (gpointer key, GList *values, ForEachContext *context)
{
  GList *l;

  for (l = values; l; l = g_list_next (l))
    context->func (key, l->data, context->user_data);
}

void
flux_ext_hash_table_foreach (FluxExtHashTable *ext_hash_table, GHFunc func, gpointer user_data)
{
  ForEachContext context;

  assert_types (ext_hash_table);

  context.ext_hash_table = ext_hash_table;
  context.func           = func;
  context.user_data      = user_data;

  g_hash_table_foreach (ext_hash_table->hash_table, (GHFunc) foreach_cb, &context);
}

gboolean
flux_ext_hash_table_remove_key (FluxExtHashTable *ext_hash_table, gconstpointer key)
{
  gboolean  have_old_key;
  gpointer  old_key;
  GList    *values;

  assert_types (ext_hash_table);

  have_old_key = g_hash_table_lookup_extended (ext_hash_table->hash_table, key,
                                               &old_key, (gpointer) &values);

  if (!have_old_key)
    return FALSE;

  if (ext_hash_table->value_destroy_func)
  {
    GList *l;

    for (l = values; l; l = g_list_next (l))
      ext_hash_table->value_destroy_func (l->data);
  }

  g_list_free (values);
  g_hash_table_remove (ext_hash_table->hash_table, old_key);

  if (ext_hash_table->key_destroy_func)
    ext_hash_table->key_destroy_func (old_key);

  return TRUE;
}

gboolean
flux_ext_hash_table_remove_value (FluxExtHashTable *ext_hash_table, gconstpointer key, gconstpointer value)
{
  gboolean  have_old_key;
  gpointer  old_key;
  GList    *values;
  GList    *l;

  assert_types (ext_hash_table);

  have_old_key = g_hash_table_lookup_extended (ext_hash_table->hash_table, key,
                                               &old_key, (gpointer) &values);
  if (!have_old_key)
    return FALSE;

  l = g_list_find (values, value);
  if (!l)
    return FALSE;

  if (ext_hash_table->value_destroy_func)
    ext_hash_table->value_destroy_func (l->data);

  values = g_list_delete_link (values, l);

  if (values)
  {
    g_hash_table_insert (ext_hash_table->hash_table, old_key, values);
  }
  else
  {
    g_hash_table_remove (ext_hash_table->hash_table, old_key);

    if (ext_hash_table->key_destroy_func)
      ext_hash_table->key_destroy_func (old_key);
  }

  return TRUE;
}

gboolean
flux_ext_hash_table_steal_key (FluxExtHashTable *ext_hash_table, gconstpointer key)
{
  GList *values;

  assert_types (ext_hash_table);

  values = g_hash_table_lookup (ext_hash_table->hash_table, key);
  if (!values)
    return FALSE;

  g_list_free (values);
  g_hash_table_remove (ext_hash_table->hash_table, key);

  return TRUE;
}

/* Note: Destroy func may be called for key */
gboolean
flux_ext_hash_table_steal_value (FluxExtHashTable *ext_hash_table, gconstpointer key, gconstpointer value)
{
  gboolean  have_old_key;
  gpointer  old_key;
  GList    *values;
  GList    *l;

  assert_types (ext_hash_table);

  have_old_key = g_hash_table_lookup_extended (ext_hash_table->hash_table, key,
                                               &old_key, (gpointer) &values);
  if (!have_old_key)
    return FALSE;

  l = g_list_find (values, value);
  if (!l)
    return FALSE;

  values = g_list_delete_link (values, l);

  if (values)
  {
    g_hash_table_insert (ext_hash_table->hash_table, old_key, values);
  }
  else
  {
    g_hash_table_remove (ext_hash_table->hash_table, old_key);

    if (ext_hash_table->key_destroy_func)
      ext_hash_table->key_destroy_func (old_key);
  }

  return TRUE;
}

typedef struct
{
  FluxExtHashTable *ext_hash_table;
  GHashTable       *change_table;
  guint             steal : 1;
  guint             n_removals;

  GHRFunc           func;
  gpointer          user_data;
}
ForEachRemoveContext;

static gboolean
foreach_remove_cb (gpointer key, GList *values, ForEachRemoveContext *context)
{
  GList *new_values = values;
  GList *l, *next;

  for (l = values; l; l = next)
  {
    gboolean to_remove;

    next = g_list_next (l);
    to_remove = context->func (key, l->data, context->user_data);

    if (to_remove)
    {
      if (!context->steal && context->ext_hash_table->value_destroy_func)
        context->ext_hash_table->value_destroy_func (l->data);
      new_values = g_list_delete_link (new_values, l);

      context->n_removals++;
    }
  }

  if (!new_values)
  {
    if (context->ext_hash_table->key_destroy_func)
      context->ext_hash_table->key_destroy_func (key);

    g_hash_table_remove (context->change_table, key);
    return TRUE;
  }
  else if (new_values != values)
  {
    g_hash_table_insert (context->change_table, key, new_values);
  }

  return FALSE;
}

static void
foreach_change_cb (gpointer key, GList *values, ForEachRemoveContext *context)
{
  g_hash_table_insert (context->ext_hash_table->hash_table, key, values);
}

guint
flux_ext_hash_table_foreach_remove (FluxExtHashTable *ext_hash_table, GHRFunc func, gpointer user_data)
{
  ForEachRemoveContext context;

  assert_types (ext_hash_table);

  context.ext_hash_table = ext_hash_table;
  context.change_table   = g_hash_table_new (g_direct_hash, g_direct_equal);
  context.steal          = FALSE;
  context.n_removals     = 0;
  context.func           = func;
  context.user_data      = user_data;

  g_hash_table_foreach_remove (ext_hash_table->hash_table, (GHRFunc) foreach_remove_cb, &context);
  g_hash_table_foreach (context.change_table, (GHFunc) foreach_change_cb, &context);

  g_hash_table_destroy (context.change_table);
  return context.n_removals;
}

/* Note: Destroy func may be called for key */
guint
flux_ext_hash_table_foreach_steal (FluxExtHashTable *ext_hash_table, GHRFunc func, gpointer user_data)
{
  ForEachRemoveContext context;

  assert_types (ext_hash_table);

  context.ext_hash_table = ext_hash_table;
  context.change_table   = g_hash_table_new (g_direct_hash, g_direct_equal);
  context.steal          = TRUE;
  context.n_removals     = 0;
  context.func           = func;
  context.user_data      = user_data;

  g_hash_table_foreach_remove (ext_hash_table->hash_table, (GHRFunc) foreach_remove_cb, &context);
  g_hash_table_foreach (context.change_table, (GHFunc) foreach_change_cb, &context);

  g_hash_table_destroy (context.change_table);
  return context.n_removals;
}
