/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-ext-pat-tree.c - Extended on-disk indexing tree.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-ext-pat-tree.h"
#include <string.h>             /* memcpy */

#define assert_types(o)        flux_object_check (FLUX_OBJECT (o), FLUX_EXT_PAT_TREE_TYPE)
#define assert_cursor_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_EXT_PAT_TREE_CURSOR_TYPE)

static FluxExtPatTreeCursor *cursor_new (FluxExtPatTree *ext_pat_tree);

gboolean
flux_ext_pat_tree_insert (FluxExtPatTree *ext_pat_tree, gconstpointer key, guint key_len,
                          gconstpointer value, guint value_len)
{
  FluxItem *found_node;
  guint     found_node_len;
  gboolean  exists;

  assert_types (ext_pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (value != NULL, FALSE);

  exists = flux_pat_tree_find (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                               key, key_len, (gpointer) &found_node, &found_node_len);

  if (!exists || found_node_len != sizeof (FluxItem))
  {
    FluxItem new_node;

    new_node = flux_pat_tree_create_root (ext_pat_tree->value_trees);
    flux_pat_tree_insert (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                          key, key_len, &new_node, sizeof (FluxItem));
    flux_pat_tree_insert (ext_pat_tree->value_trees, new_node, value, value_len, NULL, 0);
  }
  else
  {
    flux_pat_tree_insert (ext_pat_tree->value_trees, *found_node, value, value_len, NULL, 0);
  }

  if (exists)
    g_free (found_node);
  return TRUE;
}

FluxExtPatTreeCursor *
flux_ext_pat_tree_find (FluxExtPatTree *ext_pat_tree, gconstpointer key, guint key_len)
{
  FluxExtPatTreeCursor *cursor;
  FluxItem             *node;
  guint                 node_len;

  assert_types (ext_pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);

  if (!flux_pat_tree_find (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                           key, key_len, (gpointer) &node, &node_len))
    return NULL;

  if (node_len != sizeof (FluxItem))
  {
    /* Corrupt tree */
    g_free (node);
    return NULL;
  }

  cursor = cursor_new (ext_pat_tree);
  cursor->value_cursor = flux_pat_tree_traverse (ext_pat_tree->value_trees, *node);
  g_free (node);

  if (!cursor->value_cursor)
  {
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  cursor->key = g_memdup (key, key_len);
  return cursor;
}

FluxExtPatTreeCursor *
flux_ext_pat_tree_find_closest (FluxExtPatTree *ext_pat_tree, gconstpointer key, guint key_len)
{
  FluxExtPatTreeCursor *cursor;
  FluxItem             *node;
  guint                 node_len;

  assert_types (ext_pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);

  cursor = cursor_new (ext_pat_tree);

  if (!flux_pat_tree_find_closest (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                                   key, key_len,
                                   &cursor->key, &cursor->key_len,
                                   (gpointer) &node, &node_len))
  {
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  if (node_len != sizeof (FluxItem))
  {
    /* Corrupt tree */
    g_free (node);
    return NULL;
  }

  cursor->value_cursor = flux_pat_tree_traverse (ext_pat_tree->value_trees, *node);
  g_free (node);

  if (!cursor->value_cursor)
  {
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  return cursor;
}

FluxExtPatTreeCursor *
flux_ext_pat_tree_find_with_mask (FluxExtPatTree *ext_pat_tree, gconstpointer key, gconstpointer key_mask,
                                  guint key_len)
{
  FluxExtPatTreeCursor *cursor;
  FluxPatTreeCursor    *key_cursor;
  FluxItem             *node;
  guint                 node_len = 0;

  assert_types (ext_pat_tree);

  key_cursor = flux_pat_tree_find_with_mask (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                                             key, key_mask, key_len);
  if (!key_cursor)
    return NULL;

  cursor = cursor_new (ext_pat_tree);
  cursor->traverse_values = TRUE;

  node     = NULL;
  node_len = 0;

  flux_pat_tree_cursor_read (key_cursor, &cursor->key, &cursor->key_len, (gpointer) &node, &node_len);

  if (node_len != sizeof (FluxItem))
  {
    /* Corrupt tree */
    flux_object_unref (FLUX_OBJECT (key_cursor));
    flux_object_unref (FLUX_OBJECT (cursor));
    g_free (node);
    return NULL;
  }

  cursor->key_cursor   = key_cursor;
  cursor->value_cursor = flux_pat_tree_traverse (ext_pat_tree->value_trees, *node);
  g_free (node);

  /* Skip to first key with values */

  while (!cursor->value_cursor && flux_cursor_next (FLUX_CURSOR (cursor->key_cursor)))
  {
    if (cursor->key)
    {
      g_free (cursor->key);
      cursor->key = NULL;
    }

    node     = NULL;
    node_len = 0;

    flux_pat_tree_cursor_read (key_cursor, &cursor->key, &cursor->key_len, (gpointer) &node, &node_len);

    if (node_len != sizeof (FluxItem))
    {
      /* Corrupt tree */
      flux_object_unref (FLUX_OBJECT (cursor));
      g_free (node);
      return NULL;
    }

    cursor->value_cursor = flux_pat_tree_traverse (ext_pat_tree->value_trees, *node);
    g_free (node);
  }

  if (!cursor->value_cursor)
  {
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  return cursor;
}

gboolean
flux_ext_pat_tree_remove_value (FluxExtPatTree *ext_pat_tree, gconstpointer key, guint key_len,
                                gconstpointer value, guint value_len)
{
  FluxItem *node;
  guint     node_len;
  gboolean  result;

  assert_types (ext_pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (value != NULL, FALSE);

  if (!flux_pat_tree_find (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                           key, key_len, (gpointer) &node, &node_len))
    return FALSE;

  if (node_len != sizeof (FluxItem))
  {
    g_free (node);
    return FALSE;
  }

  result = flux_pat_tree_remove (ext_pat_tree->value_trees, *node, value, value_len);

  g_free (node);
  return result;
}

gboolean
flux_ext_pat_tree_get_random (FluxExtPatTree *ext_pat_tree, gpointer *key, guint *key_len,
                              gpointer *value, guint *value_len)
{
  FluxItem *node;
  guint     node_len;
  gboolean  result;

  assert_types (ext_pat_tree);

  if (!flux_pat_tree_get_random (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                                 key, key_len, (gpointer) &node, &node_len))
    return FALSE;

  if (node_len != sizeof (FluxItem))
  {
    /* Corrupt tree */
    g_free (node);
    return FALSE;
  }

  if (value)
    result = flux_pat_tree_get_random (ext_pat_tree->value_trees, *node, value, value_len, NULL, NULL);
  else
    result = TRUE;

  g_free (node);
  return result;
}

gboolean
flux_ext_pat_tree_get_random_value (FluxExtPatTree *ext_pat_tree, gconstpointer key, guint key_len,
                                    gpointer *value, guint *value_len)
{
  FluxItem *node;
  guint     node_len;
  gboolean  result;

  assert_types (ext_pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (value != NULL, FALSE);

  if (!flux_pat_tree_find (ext_pat_tree->key_tree, FLUX_ITEM_NULL,
                           key, key_len, (gpointer) &node, &node_len))
    return FALSE;

  if (node_len != sizeof (FluxItem))
  {
    /* Corrupt tree */
    g_free (node);
    return FALSE;
  }

  result = flux_pat_tree_get_random (ext_pat_tree->value_trees, *node, value, value_len, NULL, NULL);

  g_free (node);
  return result;
}

FluxExtPatTreeCursor *
flux_ext_pat_tree_traverse (FluxExtPatTree *ext_pat_tree, gboolean traverse_values)
{
  FluxExtPatTreeCursor *cursor;
  FluxPatTreeCursor    *key_cursor;
  FluxItem             *node;
  guint                 node_len;

  assert_types (ext_pat_tree);

  key_cursor = flux_pat_tree_traverse (ext_pat_tree->key_tree, FLUX_ITEM_NULL);
  if (!key_cursor)
    return NULL;

  cursor = cursor_new (ext_pat_tree);
  cursor->traverse_values = traverse_values;
  cursor->key_cursor = key_cursor;

  if (!flux_pat_tree_cursor_read (key_cursor, &cursor->key, &cursor->key_len, (gpointer) &node, &node_len))
  {
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  if (node_len != sizeof (FluxItem))
  {
    /* Corrupt tree */
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  /* If user wants values, skip to first key with values */
  if (traverse_values)
  {
    cursor->value_cursor = flux_pat_tree_traverse (ext_pat_tree->value_trees, *node);
    g_free (node);

    while (!cursor->value_cursor && flux_cursor_next (FLUX_CURSOR (cursor->key_cursor)))
    {
      if (cursor->key)
      {
        g_free (cursor->key);
        cursor->key = NULL;
      }

      if (flux_pat_tree_cursor_read (key_cursor, &cursor->key, &cursor->key_len,
                                     (gpointer) &node, &node_len))
      {
        if (node_len == sizeof (FluxItem))
          cursor->value_cursor = flux_pat_tree_traverse (ext_pat_tree->value_trees, *node);
        g_free (node);
      }
    }

    if (!cursor->value_cursor)
    {
      flux_object_unref (FLUX_OBJECT (cursor));
      return NULL;
    }
  }
  else
    g_free (node);

  return cursor;
}

gboolean
flux_ext_pat_tree_init (FluxExtPatTree *ext_pat_tree, FluxStorage *key_storage, FluxStorage *value_storage,
                        const gchar *id, guint key_len, guint value_len)
{
  g_return_val_if_fail (key_len > 0, FALSE);
  g_return_val_if_fail (value_len > 0, FALSE);

  flux_object_init              (FLUX_OBJECT (ext_pat_tree));
  flux_object_set_type          (FLUX_OBJECT (ext_pat_tree), FLUX_EXT_PAT_TREE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (ext_pat_tree), flux_ext_pat_tree_finalize);

  ext_pat_tree->key_storage   = key_storage;
  flux_object_ref (FLUX_OBJECT (key_storage));
  ext_pat_tree->value_storage = value_storage;
  flux_object_ref (FLUX_OBJECT (value_storage));

  ext_pat_tree->key_len     = key_len;
  ext_pat_tree->value_len   = value_len;

  /* FIXME: Error checking */
  ext_pat_tree->key_tree    = flux_pat_tree_new (key_storage, id, key_len, sizeof (FluxItem));
  ext_pat_tree->value_trees = flux_pat_tree_new (value_storage, id, value_len, 0);

  if (!ext_pat_tree->key_tree || !ext_pat_tree->value_trees)
    g_error ("Could not open pat trees.");

  return TRUE;
}

FluxExtPatTree *
flux_ext_pat_tree_new (FluxStorage *key_storage, FluxStorage *value_storage, const gchar *id,
                       guint key_len, guint value_len)
{
  FluxExtPatTree *ext_pat_tree;

  ext_pat_tree = g_new (FluxExtPatTree, 1);
  if (!flux_ext_pat_tree_init (ext_pat_tree, key_storage, value_storage, id, key_len, value_len))
  {
    g_free (ext_pat_tree);
    return NULL;
  }

  return ext_pat_tree;
}

void
flux_ext_pat_tree_finalize (FluxObject *object)
{
  FluxExtPatTree *ext_pat_tree = FLUX_EXT_PAT_TREE (object);

  assert_types (object);

  flux_object_unref (FLUX_OBJECT (ext_pat_tree->key_tree));
  flux_object_unref (FLUX_OBJECT (ext_pat_tree->value_trees));
  flux_object_unref (FLUX_OBJECT (ext_pat_tree->key_storage));
  flux_object_unref (FLUX_OBJECT (ext_pat_tree->value_storage));
}

void
flux_ext_pat_tree_calc_node_sizes (guint key_len, guint value_len, guint *key_node_len, guint *value_node_len)
{
  *key_node_len   = flux_pat_tree_calc_node_size (key_len, sizeof (FluxItem));
  *value_node_len = flux_pat_tree_calc_node_size (value_len, 0);
}

static FluxExtPatTreeCursor *
cursor_new (FluxExtPatTree *ext_pat_tree)
{
  FluxExtPatTreeCursor *ext_pat_tree_cursor;

  ext_pat_tree_cursor = g_new0 (FluxExtPatTreeCursor, 1);
  flux_ext_pat_tree_cursor_init (ext_pat_tree_cursor, ext_pat_tree);

  return ext_pat_tree_cursor;
}

static gboolean
cursor_next (FluxCursor *base_cursor)
{
  FluxExtPatTreeCursor *cursor = FLUX_EXT_PAT_TREE_CURSOR (base_cursor);
  gboolean              result = FALSE;
  FluxItem             *node;
  guint                 node_len;

  assert_cursor_types (cursor);

  if (cursor->value_cursor)
    result = flux_cursor_next (FLUX_CURSOR (cursor->value_cursor));

  while (!result && cursor->key_cursor && flux_cursor_next (FLUX_CURSOR (cursor->key_cursor)))
  {
    if (cursor->key)
    {
      g_free (cursor->key);
      cursor->key = NULL;
    }

    result = flux_pat_tree_cursor_read (cursor->key_cursor, &cursor->key, &cursor->key_len,
                                        (gpointer) &node, &node_len);

    if (result)
    {
      if (cursor->traverse_values)
      {
        if (cursor->value_cursor)
          flux_object_unref (FLUX_OBJECT (cursor->value_cursor));

        cursor->value_cursor = flux_pat_tree_traverse (cursor->ext_pat_tree->value_trees, *node);
        if (cursor->random_order)
          flux_pat_tree_cursor_set_randomized (cursor->value_cursor, TRUE);
        if (!cursor->value_cursor)
          result = FALSE;
      }

      g_free (node);
    }
  }

  return result;
}

void
flux_ext_pat_tree_cursor_init (FluxExtPatTreeCursor *ext_pat_tree_cursor, FluxExtPatTree *ext_pat_tree)
{
  flux_cursor_init              (FLUX_CURSOR (ext_pat_tree_cursor));
  flux_object_set_type          (FLUX_OBJECT (ext_pat_tree_cursor), FLUX_EXT_PAT_TREE_CURSOR_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (ext_pat_tree_cursor), flux_ext_pat_tree_cursor_finalize);

  flux_object_ref (FLUX_OBJECT (ext_pat_tree));
  ext_pat_tree_cursor->ext_pat_tree = ext_pat_tree;

  FLUX_CURSOR (ext_pat_tree_cursor)->next = cursor_next;
}

void
flux_ext_pat_tree_cursor_finalize (FluxObject *cursor)
{
  FluxExtPatTreeCursor *ext_pat_tree_cursor = FLUX_EXT_PAT_TREE_CURSOR (cursor);

  assert_cursor_types (ext_pat_tree_cursor);

  if (ext_pat_tree_cursor->key)
    g_free (ext_pat_tree_cursor->key);
  if (ext_pat_tree_cursor->key_cursor)
    flux_object_unref (FLUX_OBJECT (ext_pat_tree_cursor->key_cursor));
  if (ext_pat_tree_cursor->value_cursor)
    flux_object_unref (FLUX_OBJECT (ext_pat_tree_cursor->value_cursor));

  flux_object_unref (FLUX_OBJECT (ext_pat_tree_cursor->ext_pat_tree));

  flux_cursor_finalize (cursor);
}

gboolean
flux_ext_pat_tree_cursor_read (FluxExtPatTreeCursor *cursor, gpointer *key, guint *key_len,
                               gpointer *value, guint *value_len)
{
  gboolean result = TRUE;

  assert_cursor_types (cursor);
  g_return_val_if_fail (cursor->traverse_values == TRUE || value == NULL, FALSE);

  if (cursor->value_cursor && value)
    result = flux_pat_tree_cursor_read (cursor->value_cursor, value, value_len, NULL, NULL);
  if (result && key)
  {
    if (cursor->key)
    {
      *key_len = cursor->key_len;
      *key = g_memdup (cursor->key, cursor->key_len);
    }
    else
    {
      if (key_len)
        *key_len = 0;
      *key = NULL;
    }
  }

  return result;
}

gboolean
flux_ext_pat_tree_cursor_get_randomized (FluxExtPatTreeCursor *cursor)
{
  assert_cursor_types (cursor);

  return cursor->random_order;
}

void
flux_ext_pat_tree_cursor_set_randomized (FluxExtPatTreeCursor *cursor, gboolean randomized)
{
  assert_cursor_types (cursor);

  cursor->random_order = randomized;

  if (cursor->key_cursor)
    flux_pat_tree_cursor_set_randomized (cursor->key_cursor, randomized);
  if (cursor->value_cursor)
    flux_pat_tree_cursor_set_randomized (cursor->value_cursor, randomized);
}
