/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-fifo.c - Byte FIFO with optional compression and ciphering.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

/* This is a ring buffer of variable size. It consists of nodes that hold
 * a specified number of bytes each. Node list is circular:
 *
 * .-> . <-> . <-> . <-> . <-> . <-> . <-.
 * |                                     |
 * |               node_out    node_in   |
 * |               |           |         |
 * `-> . <-> . <-> o <-> O <-> o <-> . <-'
 *       | |
 *       | next
 *       prev
 *
 * New nodes are always inserted right in front of the "in" node,
 * while node removal occurs only right behind the "out" node.
 *
 * ---
 *
 * Within nodes:
 *
 *     byte_out                       byte_in
 *     |                              |
 * [...ooooooooooooooo] [ooooooooooooo.....] 
 * node_out             node_in
 */

#include "config.h"
#include "flux-types.h"
#include "flux-fifo.h"
#include "flux-twofish.h"
#include "flux-cipher.h"
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#define DEBUG(x)

#define fit_to_nodes(fifo, len) (((len) / (fifo)->buf_size) + 2)
#define TEMP_BUFFER_SIZE 4096

static FluxFifoNode *add_node (FluxFifo *fifo);
static void          del_node (FluxFifo *fifo);

typedef struct
{
  gint cpu_level;
  gint dict_bits;
  gint mem_level;
}
CompressionParams;

static CompressionParams compression_params [10] =
{
  { 0,  -8, 0 },  /* 0 */
  { 1,  -9, 1 },  /* 1 */
  { 2, -10, 2 },  /* 2 */
  { 3, -11, 3 },  /* 3 */
  { 4, -12, 4 },  /* 4 */
  { 5, -13, 5 },  /* 5 */
  { 6, -13, 6 },  /* 6 */
  { 7, -14, 7 },  /* 7 */
  { 8, -14, 8 },  /* 8 */
  { 9, -15, 9 }   /* 9 */
};

DEBUG (

static gchar
int_to_hex (gint n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void
dump_hex (guchar *buf, guint len)
{
  gint i;

  for (i = 0; i < len; i++)
    g_print ("%c%c ", int_to_hex (buf [i] >> 4), int_to_hex (buf [i] & 0x0f));

  g_print ("\n");
}

)

static FluxFifoNode *
add_node (FluxFifo *fifo)
{
  FluxFifoNode *node_new;

  /* Allocate node structures. */

  node_new       = g_new (FluxFifoNode, 1);
  node_new->data = g_malloc (fifo->buf_size);

  if (!fifo->node_out)  /* Is this the first node allocated to buffer? */
  {
    node_new->next = node_new;
    node_new->prev = node_new;
    fifo->node_out = fifo->node_in = node_new;
  }
  else
  {
    node_new->next = fifo->node_in->next;  /* [IN]     [NEW]  -> [O] */
    node_new->prev = fifo->node_in;        /* [IN] <-  [NEW]     [O] */
    fifo->node_in->next->prev = node_new;  /* [IN]     [NEW] <-  [O] */
    fifo->node_in->next = node_new;        /* [IN]  -> [NEW]     [O] */
  }

  fifo->buffers_cur++;
  return node_new;
}

/* NOTE: Does not check node for data content. Happily deletes
   used nodes. */

static void
del_node (FluxFifo *fifo)
{
  FluxFifoNode *node_del;

  if (!fifo->buffers_cur)
    return;

  /* Unlink */

  node_del = fifo->node_out->prev;
  node_del->prev->next = fifo->node_out;  /* [O] -> (OLD)    [OUT] */
  fifo->node_out->prev = node_del->prev;  /* [O]    (OLD) <- [OUT] */

  /* Free */
  
  g_free (node_del->data);
  g_free (node_del);

  /* Logistics update */
  
  fifo->buffers_cur--;
}

static void
init_compression (FluxFifo *fifo)
{
  gint index;

  if (!fifo->compression_level || fifo->have_compression)
    return;

  index = fifo->compression_level;
  g_assert (index > 0 && index < 10);

  if (fifo->mode == FLUX_FIFO_MODE_ENCODE)
  {
    deflateInit2 (&fifo->zs,
                  compression_params [index].cpu_level,
                  Z_DEFLATED,
                  compression_params [index].dict_bits,
                  compression_params [index].mem_level,
                  Z_DEFAULT_STRATEGY);
  }
  else if (fifo->mode == FLUX_FIFO_MODE_DECODE)
  {
    inflateInit2 (&fifo->zs, compression_params [index].dict_bits);
  }

  fifo->have_compression = TRUE;
  DEBUG (g_print ("Compression on.\n"));
}

static void
end_compression (FluxFifo *fifo)
{
  if (!fifo->have_compression)
    return;

  /* FIXME: How about a flush? */

  if (fifo->mode == FLUX_FIFO_MODE_ENCODE)
  {
    deflateEnd (&fifo->zs);
  }
  else if (fifo->mode == FLUX_FIFO_MODE_DECODE)
  {
    inflateEnd (&fifo->zs);
  }

  fifo->have_compression = FALSE;
  DEBUG (g_print ("Compression off.\n"));
}

static void
change_compression (FluxFifo *fifo)
{
  end_compression  (fifo);
}

static void
reset_compression (FluxFifo *fifo)
{
  if (!fifo->have_compression)
    return;

  if (fifo->mode == FLUX_FIFO_MODE_ENCODE)
  {
    deflateReset (&fifo->zs);
  }
  else if (fifo->mode == FLUX_FIFO_MODE_DECODE)
  {
    inflateReset (&fifo->zs);
  }

  DEBUG (g_print ("Compression reset.\n"));
}

static void
flux_fifo_init (FluxFifo *fifo, guint buffers_min, guint buffers_max, guint buf_size)
{
  /* Invalid parameters? */

  if (buffers_min < 2)           buffers_min = 2;
  if (buffers_max < buffers_min) buffers_max = buffers_min;
  if (buf_size < 16)             buf_size = 16;

  /* Initialize */
  
  flux_object_init              (FLUX_OBJECT (fifo));
  flux_object_set_type          (FLUX_OBJECT (fifo), FLUX_FIFO_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (fifo), flux_fifo_finalize);

  fifo->mode = FLUX_FIFO_MODE_SIMPLE;
  fifo->buffers_min = buffers_min;
  fifo->buffers_max = buffers_max;
  fifo->buf_size    = buf_size;

  if (!g_thread_supported ())
    g_thread_init (NULL);

  fifo->mutex = g_mutex_new ();
}

FluxFifo *
flux_fifo_new (guint buffers_min, guint buffers_max, guint buf_size)
{
  FluxFifo *fifo;

  /* Initialize head */

  fifo = g_new0 (FluxFifo, 1);
  flux_fifo_init (fifo, buffers_min, buffers_max, buf_size);

  return fifo;
}

void
flux_fifo_finalize (FluxObject *object)
{
  FluxFifo *fifo = FLUX_FIFO (object);

  end_compression (fifo);

  /* Delete data nodes */

  while (fifo->buffers_cur)
    del_node (fifo);

  /* Unref cipher contexts */

  if (fifo->cipher)
    flux_object_unref (FLUX_OBJECT (fifo->cipher));

  g_mutex_free (fifo->mutex);
}

gint
flux_fifo_get_free_max (FluxFifo *fifo)
{
  gint free_max, enqueued;

  g_mutex_lock (fifo->mutex);

  free_max = fifo->buffers_max * fifo->buf_size;
  if (fifo->enqueued > free_max)
  {
    g_mutex_unlock (fifo->mutex);
    return 0;
  }

  enqueued = fifo->enqueued;

  g_mutex_unlock (fifo->mutex);

  return free_max - enqueued;
}

static void
enqueue_simple (FluxFifo *fifo, gconstpointer source, guint len)
{
  gint i, j;

  DEBUG (g_print ("FIFO: Queuing %d bytes.\n", len));

  /* Make sure there are enough nodes to hold the data */

  j = fit_to_nodes (fifo, fifo->enqueued + len) - fifo->buffers_cur;
  for (i = 0; i < j; i++)
    add_node (fifo);

  /* Enqueue data bytes */

  for (i = len; i; )
  {
    if (fifo->buf_size - fifo->byte_in)
    {
      /* Last buffer can hold more */

      gulong copied_len;
      
      if (fifo->buf_size - fifo->byte_in < i)
        copied_len = fifo->buf_size - fifo->byte_in;
      else
        copied_len = i;

      memcpy ((guchar *) fifo->node_in->data + fifo->byte_in, (guchar *) source + len - i, copied_len);

      i -= copied_len;
      fifo->byte_in += copied_len;
      fifo->enqueued += copied_len;
    }
    else
    {
      /* Goto next buffer */

      fifo->node_in = fifo->node_in->next;
      if (fifo->node_in == fifo->node_out)
        g_warning ("Enqueuing over dequeue zone.");
      fifo->byte_in = 0;
    }
  }
}

static void
enqueue_encode (FluxFifo *fifo, gconstpointer source, gint len, gboolean flush)
{
  guchar        buf [TEMP_BUFFER_SIZE];
  guint         buf_len;
  const guchar *source_p = source;

  DEBUG (g_print ("enqueue_encode(): %d bytes.\n", len));

  fifo->zs.next_in  = (guchar *) source_p;
  fifo->zs.avail_in = len;

  for ( ; len >= 0; )
  {
    if (fifo->compression_level)
    {
      gint read_len;

      DEBUG (g_print ("FIFO enqueue/encode: Deflating.\n"));

      fifo->zs.next_out  = buf;
      fifo->zs.avail_out = TEMP_BUFFER_SIZE;

      deflate (&fifo->zs, flush ? Z_SYNC_FLUSH : Z_NO_FLUSH);

      read_len  = len - fifo->zs.avail_in;
      source_p += read_len;
      len      -= read_len;
      buf_len   = TEMP_BUFFER_SIZE - fifo->zs.avail_out;
    }
    else
    {
      buf_len = MIN (len, TEMP_BUFFER_SIZE);
      if (buf_len)
        memcpy (buf, source_p, buf_len);
      len      -= buf_len;
      source_p += buf_len;
    }

    if (!buf_len)
    {
      if (!len)
        break;
      else
        continue;
    }

    if (fifo->cipher && flux_cipher_is_ready (fifo->cipher))
    {
      DEBUG (g_print ("FIFO enqueue/encode: Encrypting %d bytes.\n", buf_len));
      flux_cipher_encrypt_buffer (fifo->cipher, buf, buf_len);
    }

    DEBUG (dump_hex (buf, buf_len));

    enqueue_simple (fifo, buf, buf_len);
  }
}

static gboolean
enqueue_decode (FluxFifo *fifo, gconstpointer source, guint len)
{
  gboolean  success = TRUE;
  guchar    buf [TEMP_BUFFER_SIZE];
  guint     buf_len;
  guchar   *source_start, *source_p;

  DEBUG (g_print ("enqueue_decode(): %d bytes.\n", len));

  if (!len)
    return TRUE;

  source_start = source_p = g_malloc (len);
  memcpy (source_start, source, len);

  DEBUG (dump_hex (source_start, len));

  if (fifo->cipher && flux_cipher_is_ready (fifo->cipher))
  {
    DEBUG (g_print ("FIFO enqueue/decode: Decrypting %d bytes.\n", len));
    flux_cipher_decrypt_buffer (fifo->cipher, source_p, len);
  }

  fifo->zs.next_in = source_p;
  fifo->zs.avail_in = len;

  for ( ; len > 0; )
  {
    gint result = Z_OK;

    DEBUG (g_print ("len = %u\n", len));

    if (fifo->compression_level)
    {
      gint read_len;

      DEBUG (g_print ("FIFO enqueue/decode: Inflating.\n"));

      fifo->zs.next_out  = buf;
      fifo->zs.avail_out = TEMP_BUFFER_SIZE;

      result = inflate (&fifo->zs, Z_SYNC_FLUSH);

      read_len  = len - fifo->zs.avail_in;
      source_p += read_len;
      len      -= read_len;
      buf_len   = TEMP_BUFFER_SIZE - fifo->zs.avail_out;

      /* If last operation was an unsuccessful sync, inflate will return Z_OK
       * even if it didn't do any processing. Raise the flag here, triggering
       * another sync. */
      if (!read_len && !buf_len)
        result = Z_DATA_ERROR;
    }
    else
    {
      buf_len = MIN (len, TEMP_BUFFER_SIZE);
      memcpy (buf, source_p, buf_len);
      len      -= buf_len;
      source_p += buf_len;
    }

    if (buf_len)
      enqueue_simple (fifo, buf, buf_len);

    if (result == Z_DATA_ERROR || result == Z_NEED_DICT)
    {
      gint skipped_len;

      result = inflateSync (&fifo->zs);

      skipped_len  = len - fifo->zs.avail_in;
      source_p    += skipped_len;
      len         -= skipped_len;

      success = FALSE;
    }
  }

  g_free (source_start);
  return success;
}

gboolean
flux_fifo_enqueue (FluxFifo *fifo, gconstpointer source, guint len)
{
  gboolean success = TRUE;
  gint     i;

  if (!len)
    return TRUE;

  g_mutex_lock (fifo->mutex);

  /* Ensure minimum number of data nodes */
  for (i = fifo->buffers_cur; i < fifo->buffers_min; i++)
    add_node (fifo);

  if (fifo->compression_level && !fifo->have_compression)
    init_compression (fifo);

  if (fifo->mode == FLUX_FIFO_MODE_SIMPLE)
    enqueue_simple (fifo, source, len);
  else if (fifo->mode == FLUX_FIFO_MODE_ENCODE)
    enqueue_encode (fifo, source, len, FALSE);
  else if (fifo->mode == FLUX_FIFO_MODE_DECODE)
    success = enqueue_decode (fifo, source, len);
  else
    g_error ("Invalid FIFO mode set.");

  g_mutex_unlock (fifo->mutex);
  return success;
}

void
flux_fifo_flush (FluxFifo *fifo)
{
  guchar buf [1];
  gint   i;

  g_mutex_lock (fifo->mutex);

  if (!fifo->have_compression || !fifo->compression_level)
  {
    g_mutex_unlock (fifo->mutex);
    return;
  }

  /* Ensure minimum number of data nodes */
  for (i = fifo->buffers_cur; i < fifo->buffers_min; i++)
    add_node (fifo);

  DEBUG (g_print ("FIFO: Flushing compressed data.\n"));

  if (fifo->mode == FLUX_FIFO_MODE_ENCODE)
    enqueue_encode (fifo, buf, 0, TRUE);

  g_mutex_unlock (fifo->mutex);
}

gint
flux_fifo_dequeue (FluxFifo *fifo, gpointer dest, guint len)
{
  guint i;

  g_mutex_lock (fifo->mutex);

  if (len > fifo->enqueued)
    len = fifo->enqueued;
  
  DEBUG (g_print ("FIFO: Dequeuing %d bytes.\n", len));

  for (i = len; i; )
  {
    if (fifo->buf_size - fifo->byte_out)
    {
      gulong copied_len;

      if (fifo->buf_size - fifo->byte_out < i)
        copied_len = fifo->buf_size - fifo->byte_out;
      else
        copied_len = i;

      memcpy ((guchar *) dest + len - i, (guchar *) fifo->node_out->data + fifo->byte_out, copied_len);

      i -= copied_len;
      fifo->byte_out += copied_len;
      fifo->enqueued -= copied_len;
    }
    else
    {
      fifo->node_out = fifo->node_out->next;
      fifo->byte_out = 0;
    }
  }

  while (((fifo->buffers_cur - 2) * fifo->buf_size) > fifo->enqueued)
  {
    del_node (fifo);
  }

  g_mutex_unlock (fifo->mutex);

  return len;
}

void *
flux_fifo_dequeue_dup (FluxFifo *fifo, guint len)
{
  void *data;

  g_mutex_lock (fifo->mutex);

  if (len > fifo->enqueued)
  {
    g_mutex_unlock (fifo->mutex);
    return NULL;
  }
  
  data = g_malloc (len + 1);
  *((unsigned char *) data + len) = 0;

  flux_fifo_dequeue (fifo, data, len);

  g_mutex_unlock (fifo->mutex);

  return data;
}

gint
flux_fifo_peek (FluxFifo *fifo, gpointer dest, guint len)
{
  guint i, byte_out;
  FluxFifoNode *node_out;

  g_mutex_lock (fifo->mutex);

  if (len > fifo->enqueued)
    len = fifo->enqueued;

  for (i = len, byte_out = fifo->byte_out, node_out = fifo->node_out; i; )
  {
    if (fifo->buf_size - byte_out)
    {
      gulong copied_len;
        
      if (fifo->buf_size - byte_out < i)
        copied_len = fifo->buf_size - byte_out;
      else
        copied_len = i;
        
      memcpy ((guchar *) dest + len - i, (guchar *) node_out->data + byte_out,
              copied_len);
        
      i -= copied_len;
      byte_out += copied_len;
    }
    else
    {
      node_out = node_out->next;
      byte_out = 0;
    }
  }

  g_mutex_unlock (fifo->mutex);

  return len;
}

static gint
do_drop (FluxFifo *fifo, guint len)
{
  if (!len)
    return fifo->enqueued;

  if (len > fifo->enqueued)
    len = fifo->enqueued;

  fifo->enqueued -= len;
  
  while (len > fifo->buf_size - fifo->byte_out)
  {
    fifo->node_out = fifo->node_out->next;
    len -= fifo->buf_size - fifo->byte_out;
    fifo->byte_out = 0;
  }

  fifo->byte_out += len;

  return fifo->enqueued;
}

gint
flux_fifo_drop (FluxFifo *fifo, guint len)
{
  gint enqueued;

  g_mutex_lock (fifo->mutex);
  enqueued = do_drop (fifo, len);
  g_mutex_unlock (fifo->mutex);

  return enqueued;
}

gint
flux_fifo_drop_all (FluxFifo *fifo)
{
  gint enqueued;

  g_mutex_lock (fifo->mutex);
  enqueued = do_drop (fifo, fifo->enqueued);
  g_mutex_unlock (fifo->mutex);

  return enqueued;
}

void
flux_fifo_reset (FluxFifo *fifo)
{
  g_mutex_lock (fifo->mutex);

  do_drop (fifo, fifo->enqueued);
  reset_compression (fifo);

  g_mutex_unlock (fifo->mutex);
}

gint
flux_fifo_for_data (FluxFifo *fifo, guint len, FluxByteIteratorFunc *func, gpointer user_data)
{
  FluxFifoNode *node_out;
  guint         i, byte_out;
  gint          r;

  g_mutex_lock (fifo->mutex);
  
  if (len < fifo->enqueued)
    len = fifo->enqueued;

  for (i = len, byte_out = fifo->byte_out, node_out = fifo->node_out; i; )
  {
    if (fifo->buf_size - byte_out)
    {
      guint copied_len;
        
      if (fifo->buf_size - byte_out < i)
        copied_len = fifo->buf_size - byte_out;
      else
        copied_len = i;

      r = func (node_out->data + byte_out, copied_len, user_data);
      if (r < 0)
        break;
      
      i -= copied_len;
      byte_out += copied_len;
    }
    else
    {
      node_out = node_out->next;
      byte_out = 0;
    }
  }

  g_mutex_unlock (fifo->mutex);
  return i;
}

void
flux_fifo_set_cipher (FluxFifo *fifo, FluxCipher *cipher)
{
  g_mutex_lock (fifo->mutex);

  if (cipher)
    flux_object_ref (FLUX_OBJECT (cipher));

  if (fifo->cipher)
    flux_object_unref (FLUX_OBJECT (fifo->cipher));

  fifo->cipher = cipher;

  g_mutex_unlock (fifo->mutex);
}

void
flux_fifo_set_compression (FluxFifo *fifo, guint compression)
{
  g_mutex_lock (fifo->mutex);

  if (fifo->compression_level == compression)
  {
    g_mutex_unlock (fifo->mutex);
    return;
  }

  fifo->compression_level = compression;
  change_compression (fifo);

  g_mutex_unlock (fifo->mutex);
}

void
flux_fifo_set_mode (FluxFifo *fifo, FluxFifoMode mode)
{
  g_mutex_lock (fifo->mutex);

  if (mode == fifo->mode)
  {
    g_mutex_unlock (fifo->mutex);
    return;
  }

  fifo->mode = mode;
  change_compression (fifo);

  g_mutex_unlock (fifo->mutex);
}

gfloat
flux_fifo_get_compression_ratio (FluxFifo *fifo)
{
  gfloat ratio;

  g_mutex_lock (fifo->mutex);

  if (fifo->compression_level == 0 || !fifo->have_compression)
  {
    g_mutex_unlock (fifo->mutex);
    return 1.0;
  }

  ratio = (gfloat) fifo->zs.total_out / (gfloat) fifo->zs.total_in;

  g_mutex_unlock (fifo->mutex);

  return ratio;
}
