/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-fifo.h - Byte FIFO with optional compression and ciphering.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef _FLUX_FIFO_H
#define _FLUX_FIFO_H

#ifdef __FLUX_BUILD
# include "flux-types.h"
# include "flux-object.h"
# include "flux-cipher.h"
#else
# include <flux/flux-types.h>
# include <flux/flux-object.h>
# include <flux/flux-cipher.h>
#endif

#include <zlib.h>

#define FLUX_FIFO(fifo) ((FluxFifo *) (fifo))

typedef struct _FluxFifoNode FluxFifoNode;

typedef enum
{
  FLUX_FIFO_MODE_SIMPLE,
  FLUX_FIFO_MODE_ENCODE,
  FLUX_FIFO_MODE_DECODE
}
FluxFifoMode;

struct _FluxFifoNode
{
  FluxFifoNode *prev, *next;
  gchar *data;
};

struct _FluxFifo
{
  FluxObject object;

  GMutex *mutex;

  FluxFifoMode mode;

  FluxFifoNode *node_out, *node_in;
  FluxFifoNode *node_cur;
  gint byte_out, byte_in;
  gint enqueued;
  gint buf_size;
  gint i;
  gint buffers_min, buffers_max, buffers_cur;

  FluxCipher *cipher;

  guint    compression_level;
  guint    have_compression : 1;
  z_stream zs;
};

/* --- */

FluxFifo *flux_fifo_new                   (guint buffers_min, guint buffers_max, guint buf_size);
void      flux_fifo_finalize              (FluxObject *fifo);

gint      flux_fifo_get_free_max          (FluxFifo *fifo);

gboolean  flux_fifo_enqueue               (FluxFifo *fifo, gconstpointer source, guint len);
gint      flux_fifo_dequeue               (FluxFifo *fifo, gpointer dest, guint len);
gpointer  flux_fifo_dequeue_dup           (FluxFifo *fifo, guint len);

gint      flux_fifo_peek                  (FluxFifo *fifo, gpointer dest, guint len);
gint      flux_fifo_drop                  (FluxFifo *fifo, guint len);
gint      flux_fifo_drop_all              (FluxFifo *fib);
void      flux_fifo_reset                 (FluxFifo *fib);
void      flux_fifo_flush                 (FluxFifo *fifo);

gint      flux_fifo_for_data              (FluxFifo *fifo, guint len, FluxByteIteratorFunc *func,
                                           void *user_data);

void      flux_fifo_set_mode              (FluxFifo *fifo, FluxFifoMode mode);
void      flux_fifo_set_cipher            (FluxFifo *fifo, FluxCipher *cipher);
void      flux_fifo_set_compression       (FluxFifo *fifo, guint compression);

gfloat    flux_fifo_get_compression_ratio (FluxFifo *fifo);

#define   flux_fifo_get_used(fib)         ((fib)->enqueued)
#define   flux_fifo_get_free_current(fib) flux_fifo_get_free_max (fib)
#define   flux_fifo_get_free_min(fib)     flux_fifo_get_free_max (fib)

#define   FLUX_FIFO_FOR_DATA(fib, lim, p, len)               \
  for(fib->i = lim < fib->enqueued ? lim : fib->enqueued,    \
      fib->node_cur = fib->node_out,                         \
      p = fib->node_out->data + fib->byte_out,               \
      len = (fib->buf_size - fib->byte_out) < fib->i ?       \
            (fib->buf_size - fib->byte_out) : fib->i;        \
                                                             \
      fib->i;                                                \
                                                             \
      fib->i -= len,                                         \
      fib->node_cur = fib->node_cur->next,                   \
      p = fib->node_cur->data,                               \
      len = fib->buf_size < fib->i ? fib->buf_size : fib->i)

#endif  /* _FLUX_FIFO_H */
