/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-file-io.c - File I/O channel.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

/* Large file support */
#undef  _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64
#undef  __REDIRECT
#define __REDIRECT 1

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/time.h>
#else
# include <io.h>
# include <winsock2.h>
# include <time.h>
# include "win32/win32-compat.h"
#endif

#include "flux-fifo.h"
#include "flux-io.h"
#include "flux-file-io.h"

#define DEBUG(x)

#define MAX_READ           4096
#define MAX_WRITE          4096
#define MAX_IDLE_BUFFER_IN 8192
#define THREAD_STACK_SIZE  (MAX_READ + MAX_WRITE + 16384)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_FILE_IO_TYPE);

static void
cancel_read_buffer (FluxFileIO *file_io)
{
  FluxFifo *fifo;
  guint     offset;

  fifo   = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (file_io));
  offset = flux_fifo_get_used (fifo);

  if (offset)
  {
    file_io->real_position -= offset;
    lseek (file_io->fd, file_io->real_position, SEEK_SET);
    flux_fifo_reset (fifo);
  }
}

static void
cancel_write_buffer (FluxFileIO *file_io)
{
  FluxFifo *fifo;
  guint     offset;

  fifo   = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (file_io));
  offset = flux_fifo_get_used (fifo);

  if (offset)
  {
    file_io->real_position += offset;
    lseek (file_io->fd, file_io->real_position, SEEK_SET);
    flux_fifo_reset (fifo);
  }
}

static void
goto_read_state (FluxFileIO *file_io)
{
  if (file_io->state == FLUX_FILE_IO_READING)
    return;

  cancel_write_buffer (file_io);
  file_io->state = FLUX_FILE_IO_READING;
}

static void
goto_write_state (FluxFileIO *file_io)
{
  if (file_io->state == FLUX_FILE_IO_WRITING)
    return;

  cancel_read_buffer (file_io);
  file_io->state = FLUX_FILE_IO_WRITING;
}

static void
user_write_cb (gpointer data)
{
  FluxFileIO *file_io = data;

  assert_types (file_io);

  g_mutex_lock (file_io->mutex);
  goto_write_state (file_io);
  g_cond_signal (file_io->wakeup_cond);
  g_mutex_unlock (file_io->mutex);
}

static void
user_read_cb (gpointer data)
{
  FluxFileIO *file_io = data;

  assert_types (file_io);

  g_mutex_lock (file_io->mutex);
  goto_read_state (file_io);
  g_cond_signal (file_io->wakeup_cond);
  g_mutex_unlock (file_io->mutex);
}

static gboolean
eof_cb (FluxFileIO *file_io)
{
  flux_object_ref (FLUX_OBJECT (file_io));
  g_mutex_lock (file_io->mutex);

  if (file_io->eof_func)
  {
    FluxFifo *fifo;
    guint     bytes_left;

    fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (file_io));
    bytes_left = flux_fifo_get_used (fifo);

    g_mutex_unlock (file_io->mutex);
    file_io->eof_func (file_io, bytes_left, file_io->eof_func_data);
    g_mutex_lock (file_io->mutex);
  }

  file_io->user_cb_id = 0;

  DEBUG (g_print ("Main EOF on %p done.\n", file_io));

  g_cond_signal (file_io->wakeup_cond);
  g_mutex_unlock (file_io->mutex);
  flux_object_unref (FLUX_OBJECT (file_io));

  return FALSE;
}

static gboolean
error_cb (FluxFileIO *file_io)
{
  flux_object_ref (FLUX_OBJECT (file_io));
  g_mutex_lock (file_io->mutex);

  if (file_io->error_func)
  {
    g_mutex_unlock (file_io->mutex);
    file_io->error_func (FLUX_OBJECT (file_io), file_io->error, file_io->error_func_data);
    g_mutex_lock (file_io->mutex);
  }

  file_io->user_cb_id = 0;
  file_io->error      = NULL;

  DEBUG (g_print ("Main error on %p done.\n", file_io));

  g_cond_signal (file_io->wakeup_cond);
  g_mutex_unlock (file_io->mutex);
  flux_object_unref (FLUX_OBJECT (file_io));

  return FALSE;
}

static gboolean
read_cb (FluxFileIO *file_io)
{
  FluxFifo *fifo;

  flux_object_ref (FLUX_OBJECT (file_io));
  g_mutex_lock (file_io->mutex);

  fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (file_io));

  while (FLUX_IO (file_io)->read_threshold &&
         flux_fifo_get_used (fifo) >= FLUX_IO (file_io)->read_threshold &&
         FLUX_IO (file_io)->read_func &&
         flux_object_get_refs (FLUX_OBJECT (file_io)) > 1)
  {
    g_mutex_unlock (file_io->mutex);
    FLUX_IO (file_io)->read_func (FLUX_IO (file_io), FLUX_IO (file_io)->read_threshold,
                                  FLUX_IO (file_io)->read_func_data);
    g_mutex_lock (file_io->mutex);
  }

  file_io->user_cb_id = 0;

  DEBUG (g_print ("Main read on %p done.\n", file_io));

  g_cond_signal (file_io->wakeup_cond);
  g_mutex_unlock (file_io->mutex);
  flux_object_unref (FLUX_OBJECT (file_io));

  return FALSE;
}

static gboolean
write_cb (FluxFileIO *file_io)
{
  FluxFifo *fifo;
  guint     threshold;

  flux_object_ref (FLUX_OBJECT (file_io));
  g_mutex_lock (file_io->mutex);

  fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (file_io));
  threshold = FLUX_IO (file_io)->write_threshold;

  if (threshold && flux_fifo_get_free_current (fifo) >= threshold && FLUX_IO (file_io)->write_func)
  {
    g_mutex_unlock (file_io->mutex);
    FLUX_IO (file_io)->write_func (FLUX_IO (file_io), threshold, FLUX_IO (file_io)->write_func_data);
    g_mutex_lock (file_io->mutex);
  }

  if (file_io->eow_func && flux_fifo_get_used (fifo) == 0)
  {
    g_mutex_unlock (file_io->mutex);
    file_io->eow_func (file_io, file_io->eow_func_data);
    g_mutex_lock (file_io->mutex);
  }

  file_io->user_cb_id = 0;

  DEBUG (g_print ("Main write on %p done.\n", file_io));

  g_cond_signal (file_io->wakeup_cond);
  g_mutex_unlock (file_io->mutex);
  flux_object_unref (FLUX_OBJECT (file_io));

  return FALSE;
}

static GError *
new_error_from_errno (gint err_no)
{
  GError          *error = NULL;
  FluxFileIOError  file_io_error;

  switch (err_no)
  {
    case EBADF:
      file_io_error = FLUX_FILE_IO_ERROR_INVALID_FD;
      break;

    case EIO:
      file_io_error = FLUX_FILE_IO_ERROR_DEVICE_IO;
      break;

    case ENOSPC:
      file_io_error = FLUX_FILE_IO_ERROR_NO_SPACE;
      break;

    default:
      file_io_error = FLUX_FILE_IO_ERROR_OTHER;
      break;
  }

  g_set_error (&error, FLUX_FILE_IO_ERROR, file_io_error, "%s", g_strerror (err_no));
  return error;
}

static void
worker_read (FluxFileIO *file_io)
{
  while (file_io->running && file_io->state == FLUX_FILE_IO_READING)
  {
    FluxFifo *fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (file_io));
    gchar     buf [MAX_READ];
    gint      result;

    if (flux_fifo_get_used (fifo) >= FLUX_IO (file_io)->read_threshold)
    {
      /* If we can satisfy the user's read threshold, ensure read callback */
      if (!file_io->user_cb_id)
        file_io->user_cb_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) read_cb, file_io, NULL);

      /* If buffers are full, wait for user action */
      if (flux_fifo_get_used (fifo) >= MAX_IDLE_BUFFER_IN)
      {
        DEBUG (g_print ("Buffers full...\n"));
        g_cond_wait (file_io->wakeup_cond, file_io->mutex);
        DEBUG (g_print ("Woke up.\n"));
        continue;
      }
    }

    g_mutex_unlock (file_io->mutex);
    result = read (file_io->fd, buf, MAX_READ);
    g_mutex_lock (file_io->mutex);

    if (result >= 1)
    {
      /* Got data */

      flux_fifo_enqueue (fifo, buf, result);
      file_io->real_position += result;
    }
    else if (result == 0)
    {
      /* Got EOF */

      if (file_io->user_cb_id)
      {
        /* Wait for last read callback to be handled by user */
        DEBUG (g_print ("EOF, but read callback pending... real_position = %lld, ftell = %d\n", file_io->real_position, lseek (file_io->fd, 0, SEEK_CUR)));
        g_cond_wait (file_io->wakeup_cond, file_io->mutex);
        DEBUG (g_print ("Woke up.\n"));
      }
      else if (file_io->running)
      {
        /* Set up EOF callback and wait for it to be handled */
        file_io->user_cb_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) eof_cb, file_io, NULL);
        DEBUG (g_print ("EOF... real_position = %lld, ftell = %d\n", file_io->real_position, lseek (file_io->fd, 0, SEEK_CUR)));
        g_cond_wait (file_io->wakeup_cond, file_io->mutex);
        DEBUG (g_print ("Woke up.\n"));
      }
    }
    else  /* result < 0 */
    {
      /* Got error */

      switch (errno)
      {
        case EINTR:
          DEBUG (g_print ("Interrupted by signal (fd %d).\n", file_io->fd));
          break;

        case EAGAIN:
          DEBUG (g_print ("Got EAGAIN (fd %d).\n", file_io->fd));
          break;

        default:
          DEBUG (g_print ("Bad, bad read error.\n"));
          if (file_io->user_cb_id)
          {
            /* Wait for last read callback to be handled by user */
            DEBUG (g_print ("Error, but read callback pending...\n"));
            g_cond_wait (file_io->wakeup_cond, file_io->mutex);
            DEBUG (g_print ("Woke up.\n"));
          }
          else if (file_io->running)
          {
            /* Set up error callback and wait for it to be handled */
            file_io->error = new_error_from_errno (errno);
            file_io->user_cb_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) error_cb, file_io, NULL);
            DEBUG (g_print ("Error...\n"));
            g_cond_wait (file_io->wakeup_cond, file_io->mutex);
            DEBUG (g_print ("Woke up.\n"));
          }
          break;
      }
    }
  }
}

static void
worker_write (FluxFileIO *file_io)
{
  while (file_io->running && file_io->state == FLUX_FILE_IO_WRITING)
  {
    FluxFifo *fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (file_io));
    gchar     buf [MAX_WRITE];
    gint      result;
    gint      len;

    if (flux_fifo_get_free_current (fifo) >= FLUX_IO (file_io)->write_threshold)
    {
      /* If we can satisfy the user's write threshold, ensure write callback */
      if (!file_io->user_cb_id)
        file_io->user_cb_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) write_cb, file_io, NULL);

      /* If nothing to write, wait for user action */
      if (flux_fifo_get_used (fifo) == 0)
      {
        g_cond_wait (file_io->wakeup_cond, file_io->mutex);
        continue;
      }
    }

    len = flux_fifo_get_used (fifo);
    if (len > MAX_WRITE)
      len = MAX_WRITE;

    g_assert (len > 0);

    flux_fifo_peek (fifo, buf, len);

    g_mutex_unlock (file_io->mutex);
    result = write (file_io->fd, buf, len);
    g_mutex_lock (file_io->mutex);

    if (result >= 1)
    {
      /* Got data */

      flux_fifo_drop (fifo, result);
      file_io->real_position += result;
    }
    else
    {
      /* Got error */

      switch (errno)
      {
        case EINTR:
          DEBUG (g_print ("Interrupted by signal (fd %d).\n", file_io->fd));
          break;

        case EAGAIN:
          DEBUG (g_print ("Got EAGAIN (fd %d).\n", file_io->fd));
          break;

        default:
          if (file_io->user_cb_id)
          {
            /* Wait for last write callback to be handled by user */
            g_cond_wait (file_io->wakeup_cond, file_io->mutex);
          }
          else if (file_io->running)
          {
            /* Set up error callback and wait for it to be handled */
            file_io->error = new_error_from_errno (errno);
            file_io->user_cb_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) error_cb, file_io, NULL);
            g_cond_wait (file_io->wakeup_cond, file_io->mutex);
          }
          break;
      }
    }
  }
}

static gboolean
unref_on_idle (FluxFileIO *file_io)
{
  DEBUG (g_print ("Idle unref of %p.\n", file_io));
  flux_object_unref (FLUX_OBJECT (file_io));
  return FALSE;
}

static gpointer
worker_main (FluxFileIO *file_io)
{
  g_mutex_lock (file_io->mutex);

  while (file_io->running)
  {
    DEBUG (g_print ("FluxFileIO iterating.\n"));

    switch (file_io->state)
    {
      case FLUX_FILE_IO_READING:
        worker_read (file_io);
        break;

      case FLUX_FILE_IO_WRITING:
        worker_write (file_io);
        break;

      case FLUX_FILE_IO_IDLE:
      default:
        g_cond_wait (file_io->wakeup_cond, file_io->mutex);
        break;
    }
  }

  DEBUG (g_print ("Worker %p exiting.\n", file_io));

  g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) unref_on_idle, file_io, NULL);
  g_mutex_unlock (file_io->mutex);
  return NULL;
}

void
flux_file_io_init (FluxFileIO *file_io, gint fd)
{
  DEBUG (g_print ("Initializing FileIO (fd=%d).\n", fd));

  flux_io_init                  (FLUX_IO (file_io));
  flux_object_set_type          (FLUX_OBJECT (file_io), FLUX_FILE_IO_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (file_io), flux_file_io_finalize);

  if (!g_thread_supported ())
    g_thread_init (NULL);

  flux_dual_fifo_set_write_hook (FLUX_DUAL_FIFO (file_io), user_write_cb, file_io);
  flux_dual_fifo_set_read_hook  (FLUX_DUAL_FIFO (file_io), user_read_cb, file_io);

  file_io->state         = FLUX_FILE_IO_IDLE;
  file_io->fd            = fd;
  file_io->running       = TRUE;

  file_io->mutex         = g_mutex_new ();
  file_io->wakeup_cond   = g_cond_new ();
  file_io->worker_thread = g_thread_create_full ((GThreadFunc) worker_main, file_io,
                                                 THREAD_STACK_SIZE, FALSE, FALSE,
                                                 G_THREAD_PRIORITY_NORMAL, NULL);

  DEBUG (g_print ("Initialized FileIO.\n"));
}

FluxFileIO *
flux_file_io_new (gint fd)
{
  FluxFileIO *file_io;

  file_io = g_new0 (FluxFileIO, 1);
  flux_file_io_init (file_io, fd);

  return file_io;
}

void
flux_file_io_finalize (FluxObject *object)
{
  FluxFileIO *file_io = FLUX_FILE_IO (object);

  assert_types (file_io);

  DEBUG (g_print ("Finalizing FileIO.\n"));

  g_mutex_lock (file_io->mutex);

  if (file_io->running)
  {
    DEBUG (g_print ("First unref of %p; worker still running.\n", file_io));

    file_io->running = FALSE;
    file_io->eow_func = NULL;
    file_io->eof_func = NULL;
    FLUX_IO (file_io)->read_func  = NULL;
    FLUX_IO (file_io)->write_func = NULL;

    if (file_io->user_cb_id)
    {
      g_source_remove (file_io->user_cb_id);
      file_io->user_cb_id = 0;
    }

    flux_object_ref (FLUX_OBJECT (file_io));

    g_cond_signal (file_io->wakeup_cond);
    g_mutex_unlock (file_io->mutex);
  }
  else
  {
    DEBUG (g_print ("Final unref of %p.\n", file_io));

    close (FLUX_FILE_IO (file_io)->fd);
    g_mutex_unlock (file_io->mutex);

    g_cond_free (file_io->wakeup_cond);
    g_mutex_free (file_io->mutex);

    flux_io_finalize (FLUX_OBJECT (file_io));
    DEBUG (g_print ("Finalized FileIO.\n"));
  }
}

guint64
flux_file_io_get_size (FluxFileIO *file_io)
{
  struct stat sbuf;

  assert_types (file_io);

  fstat (file_io->fd, &sbuf);
  return sbuf.st_size;
}

static guint64 
get_position (FluxFileIO *file_io)
{
  guint64 pos;

  /* Idle state means we could still be doing read-ahead. When doing write-behind,
   * the state is always FLUX_FILE_IO_WRITING. */

  if (file_io->state == FLUX_FILE_IO_READING || file_io->state == FLUX_FILE_IO_IDLE)
  {
    FluxFifo *fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (file_io));
    pos = file_io->real_position - flux_fifo_get_used (fifo);
  }
  else if (file_io->state == FLUX_FILE_IO_WRITING)
  {
    FluxFifo *fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (file_io));
    pos = file_io->real_position + flux_fifo_get_used (fifo);
  }
  else /* Never happens */
  {
    pos = file_io->real_position;
  }

  return pos;
}

static void
seek_to (FluxFileIO *file_io, guint64 position)
{
  guint64 virtual_position;

  /* FIXME: Seeking probably doesn't work at all */

  virtual_position = get_position (file_io);
  if (virtual_position != position)
  {
    lseek (file_io->fd, position, SEEK_SET);
    cancel_read_buffer (file_io);
    cancel_write_buffer (file_io);
    /* goto_idle_state (file_io); */
    file_io->real_position = position;
  }
}

/**
 * flux_file_io_get_position: Get file position.
 *
 * Returns: The file position as it would be if all buffers were flushed. This
 * value changes immediately when the user reads or writes data, even if it
 * hasn't been read from/committed to disk.
 */
guint64
flux_file_io_get_position (FluxFileIO *file_io)
{
  guint64 pos;

  assert_types (file_io);

  g_mutex_lock (file_io->mutex);
  pos = get_position (file_io);
  g_mutex_unlock (file_io->mutex);

  return pos;
}

void
flux_file_io_seek_to (FluxFileIO *file_io, guint64 position)
{
  assert_types (file_io);

  g_mutex_lock (file_io->mutex);
  seek_to (file_io, position);
  g_mutex_unlock (file_io->mutex);
}

void
flux_file_io_seek_relative (FluxFileIO *file_io, gint64 offset)
{
  guint64 virtual_position;

  assert_types (file_io);

  g_mutex_lock (file_io->mutex);

  virtual_position = get_position (file_io);
  if (offset < 0 && abs (offset) > virtual_position)
    virtual_position = 0;
  else
    virtual_position += offset;

  seek_to (file_io, virtual_position);

  g_mutex_unlock (file_io->mutex);
}

void
flux_file_io_set_eof_func (FluxFileIO *file_io, FluxEofFunc *func, gpointer data)
{
  assert_types (file_io);

  g_mutex_lock (file_io->mutex);

  file_io->eof_func      = func;
  file_io->eof_func_data = data;

  g_mutex_unlock (file_io->mutex);
}

void
flux_file_io_set_eow_func (FluxFileIO *file_io, FluxEowFunc *func, gpointer data)
{
  assert_types (file_io);

  g_mutex_lock (file_io->mutex);

  file_io->eow_func      = func;
  file_io->eow_func_data = data;

  g_mutex_unlock (file_io->mutex);
}

void
flux_file_io_set_error_func (FluxFileIO *file_io, FluxErrorFunc *func, gpointer data)
{
  assert_types (file_io);

  g_mutex_lock (file_io->mutex);

  file_io->error_func      = func;
  file_io->error_func_data = data;

  g_mutex_unlock (file_io->mutex);
}

GQuark
flux_file_io_error_quark (void)
{
  static GQuark q = 0;

  if (q == 0)
    q = g_quark_from_static_string ("flux-file-io-error-quark");

  return q;
}
