/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-interval.c - Maps input values to a continuous curve of output values.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "flux-interval.h"

typedef struct
{
  gdouble        in;
  gdouble        out;
  FluxTransition transition;
}
Point;

FluxInterval *
flux_interval_new (gdouble default_value_out, gboolean repeat_defined)
{
  FluxInterval *interval;

  interval = g_new0 (FluxInterval, 1);
  flux_interval_init (interval, default_value_out, repeat_defined);

  return interval;
}

void
flux_interval_init (FluxInterval *interval, gdouble default_value_out, gboolean repeat_defined)
{
  flux_object_init              (FLUX_OBJECT (interval));
  flux_object_set_type          (FLUX_OBJECT (interval), FLUX_INTERVAL_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (interval), flux_interval_finalize);

  interval->points = g_array_new (FALSE, FALSE, sizeof (Point));
  interval->default_value_out = default_value_out;
  interval->repeat_defined = repeat_defined;
}

void
flux_interval_finalize (FluxObject *object)
{
  FluxInterval *interval = FLUX_INTERVAL (object);

  g_array_free (interval->points, TRUE);
}

static gint
find_closest_index (GArray *points, gdouble value_in)
{
  gint len;
  gint left, middle = 0, right;

  len = points->len;

  if (len < 1)
    return -1;

  left  = 0;
  right = len - 1;

  while (left <= right)
  {
    gint middle    = (left + right) >> 1;
    gint middle_in = g_array_index (points, Point, middle).in;

    if (middle_in > value_in)
      right = middle - 1;
    else if (middle_in < value_in)
      left  = middle + 1;
    else
      break;
  }

  return middle;
}

void
flux_interval_add_point (FluxInterval *interval, gdouble value_in, gdouble value_out,
                         FluxTransition transition)
{
  Point p;
  gint  len;

  g_return_if_fail (transition <= FLUX_TRANSITION_LAST);

  p.in         = value_in;
  p.out        = value_out;
  p.transition = transition;

  len = interval->points->len;

  if (len == 0 || value_in >= g_array_index (interval->points, Point, len - 1).in)
  {
    /* If user is adding points in sorted order, we can skip find_closest_index () */
    g_array_append_val (interval->points, p);
  }
  else
  {
    gint n = find_closest_index (interval->points, value_in);

    if (g_array_index (interval->points, Point, n).in < value_in)
      n++;

    g_array_insert_val (interval->points, n, p);
  }
}

static gdouble
transform_constant (gdouble f)
{
  return 0.0;
}

static gdouble
transform_linear (gdouble f)
{
  return f;
}

static gdouble
transform_exponential (gdouble f)
{
  return (pow (2.0, f + 1.0) / 2.0) - 1.0;
}

static gdouble
transform_reverse_exponential (gdouble f)
{
  return (pow (2.0, 2.0 - f) / 2.0) - 1.0;
}

static gdouble
transform_half_sinusoid (gdouble f)
{
  return (cos (M_PI * f) + 1.0) / 2.0;
}

static gdouble
transform_quarter_sinusoid (gdouble f)
{
  return (cos (M_PI * f * 0.5) + 1.0) / 2.0;
}

static gdouble
transform_reverse_quarter_sinusoid (gdouble f)
{
  return (cos (M_PI * 0.5 * (1.0 + f)) + 1.0) / 2.0;
}

struct
{
  gdouble (*func) (gdouble);
}
transform_table [FLUX_TRANSITION_LAST + 1] =
{
  { transform_constant                 },
  { transform_linear                   },
  { transform_exponential              },
  { transform_reverse_exponential      },
  { transform_half_sinusoid            },
  { transform_quarter_sinusoid         },
  { transform_reverse_quarter_sinusoid }
};

gdouble
flux_interval_transform (FluxInterval *interval, gdouble value_in)
{
  gdouble n_val;
  gdouble next_val;
  gint    len;
  gint    n;

  len = interval->points->len;

  if (len == 0)
    return interval->default_value_out;

  if (interval->repeat_defined)
  {
    gdouble first = g_array_index (interval->points, Point, 0).in;
    gdouble last  = g_array_index (interval->points, Point, len - 1).in;
    gdouble range = last - first;

    if (len == 1 || range == 0.0)
      return g_array_index (interval->points, Point, 0).out;

    value_in = fmod (value_in - first, range);

    if (value_in < 0)
      value_in += range;

    value_in += first;
  }

  n = find_closest_index (interval->points, value_in);

  if (g_array_index (interval->points, Point, n).in > value_in)
    n--;

  if (n < 0 || n == len - 1)
    return interval->default_value_out;

  n_val    = g_array_index (interval->points, Point, n).out;
  next_val = g_array_index (interval->points, Point, n + 1).out;

  return n_val + transform_table [g_array_index (interval->points, Point, n).transition].func (
    (value_in - n_val) / (next_val - n_val)) * (next_val - n_val);
}
