/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-interval.h - Maps input values to a continuous curve of output values.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef _FLUX_INTERVAL_H
#define _FLUX_INTERVAL_H

#ifdef __FLUX_BUILD
# include "flux-types.h"
# include "flux-object.h"
#else
# include <flux/flux-types.h>
# include <flux/flux-object.h>
#endif

/* Constant:
 *
 *         b  a_______
 *         |          |
 * a_______|          |
 *                    b
 *
 * Linear:
 *
 *   b  a
 *  /    \
 * a      b
 *
 * Exponential:
 *
 *       b  a.__
 *      |       '.
 * a_..'         |
 *                b
 *
 * Reverse exponential:
 *
 *    __.b  a
 *  .'       |
 *  |         '.._b
 * a
 *
 * Half sinusoid:
 *
 *     ,-b  a-.
 * a_.'        '._b
 *
 * Quarter sinusoid:
 *
 *    b  a-.
 * a_.'     'b
 *
 * Reverse quarter sinusoid:
 *
 *   .-b  a
 * a'      '._b
 *
 */

typedef enum
{
  FLUX_TRANSITION_CONSTANT,
  FLUX_TRANSITION_LINEAR,
  FLUX_TRANSITION_EXPONENTIAL,
  FLUX_TRANSITION_REVERSE_EXPONENTIAL,
  FLUX_TRANSITION_HALF_SINUSOID,
  FLUX_TRANSITION_QUARTER_SINUSOID,
  FLUX_TRANSITION_REVERSE_QUARTER_SINUSOID,

  FLUX_TRANSITION_LAST = FLUX_TRANSITION_REVERSE_QUARTER_SINUSOID
}
FluxTransition;

#define FLUX_INTERVAL(o) ((FluxInterval *) (o))

struct _FluxInterval
{
  FluxObject  object;

  guint       repeat_defined : 1;
  gdouble     default_value_out;

  GArray     *points;
};

FluxInterval *flux_interval_new       (gdouble default_value_out, gboolean repeat_defined);
void          flux_interval_init      (FluxInterval *interval, gdouble default_value_out,
                                       gboolean repeat_defined);
void          flux_interval_finalize  (FluxObject *object);

void          flux_interval_add_point (FluxInterval *interval, gdouble value_in, gdouble value_out,
                                       FluxTransition transition);

gdouble       flux_interval_transform (FluxInterval *interval, gdouble value_in);

#endif /* _FLUX_INTERVAL_H */
