/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-io-group.c - Resource sharing for FluxIOs.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <glib.h>

#include "flux-io-group.h"

#define PRI_RANGE_PAD   1.0
#define CHANGE_FRACTION 0.05

typedef struct
{
  guint   is_group : 1;

  gdouble priority_in;
  gdouble priority_out;
}
Member;

typedef struct
{
  FluxIOGroup *io_group;

  /* Highest priority seen among members */
  gdouble      highest_priority_in;
  gdouble      highest_priority_out;

  /* Lowest priority seen among members */
  gdouble      lowest_priority_in;
  gdouble      lowest_priority_out;

  /* Sum of member priorities */
  gdouble      priority_sum_in;
  gdouble      priority_sum_out;

  /* Number of members where used CPS > min CPS */
  gdouble      over_min_priority_sum_in;
  gdouble      over_min_priority_sum_out;

  /* Number of members where used CPS ~= set CPS || used CPS > set_CPS */
  gdouble      capped_priority_sum_in;
  gdouble      capped_priority_sum_out;

  /* Capped CPS use relative to goal; negative is overlimit, positive is underlimit */
  gint         spare_cps_in;
  gint         spare_cps_out;
}
Stats;

typedef struct
{
  Stats    *stats;
  gboolean  in;
}
Change;

#define capped_cps(cps, cap) MIN (cps, cap)

static guint   group_slice_timeout_id = 0;
static GSList *group_slice_list       = NULL;

static gboolean
is_close (guint a, guint b)
{
  if (a + 100 >= b && b + 100 >= a)
    return TRUE;

  if (a * 1.02 >= b && b * 1.02 >= a)
    return TRUE;

  return FALSE;
}

static gboolean
is_capped (gint cps, gint cap)
{
  if (is_close (cps, cap) || cps > cap)
    return TRUE;

  return FALSE;
}

static guint
calc_min_cps (guint goal_cps, gdouble sum_priorities, gdouble lowest_priority, gdouble member_priority)
{
  gdouble member_priority_relative;
  gdouble a;

  member_priority_relative = member_priority - lowest_priority + PRI_RANGE_PAD;
  a = goal_cps * member_priority_relative;

  return (guint) (a / sum_priorities);
}

static void
set_cap (gpointer member, Member *member_record, gint cap, gboolean in)
{
  if (in)
  {
    if (member_record->is_group)
    {
      flux_io_group_set_cps_goal (member, cap, -1);
    }
    else
    {
      flux_io_set_read_cps_goal (member, cap);
    }
  }
  else
  {
    if (member_record->is_group)
    {
      flux_io_group_set_cps_goal (member, -1, cap);
    }
    else
    {
      flux_io_set_write_cps_goal (member, cap);
    }
  }
}

static void
gather_stats_1_cb (gpointer member, Member *member_record, Stats *stats)
{
  guint cps_in, cps_out;
  guint cps_cap_in, cps_cap_out;

  if (member_record->priority_in > stats->highest_priority_in)
    stats->highest_priority_in = member_record->priority_in;
  if (member_record->priority_out > stats->highest_priority_out)
    stats->highest_priority_out = member_record->priority_out;

  if (member_record->priority_in < stats->lowest_priority_in)
    stats->lowest_priority_in = member_record->priority_in;
  if (member_record->priority_out < stats->lowest_priority_out)
    stats->lowest_priority_out = member_record->priority_out;

  if (member_record->is_group)
  {
    flux_io_group_get_cps (member, &cps_in, &cps_out);
    flux_io_group_get_cps_goal (member, &cps_cap_in, &cps_cap_out);
  }
  else
  {
    cps_in  = flux_io_get_read_cps  (member);
    cps_out = flux_io_get_write_cps (member);

    cps_cap_in  = flux_io_get_read_cps_goal (member);
    cps_cap_out = flux_io_get_write_cps_goal (member);
  }

  if (is_capped (cps_in, cps_cap_in))
    stats->capped_priority_sum_in += member_record->priority_in + PRI_RANGE_PAD;
  if (is_capped (cps_out, cps_cap_out))
    stats->capped_priority_sum_out += member_record->priority_out + PRI_RANGE_PAD;

  stats->io_group->used_cps_in  += cps_in;
  stats->io_group->used_cps_out += cps_out;

  stats->priority_sum_in  += member_record->priority_in + PRI_RANGE_PAD;
  stats->priority_sum_out += member_record->priority_out + PRI_RANGE_PAD;
}

static void
gather_stats_2_cb (gpointer member, Member *member_record, Stats *stats)
{
  guint cps_in, cps_out;
  guint cps_cap_in, cps_cap_out;
  guint cps_min_in, cps_min_out;

  if (member_record->is_group)
  {
    flux_io_group_get_cps (member, &cps_in, &cps_out);
    flux_io_group_get_cps_goal (member, &cps_cap_in, &cps_cap_out);
  }
  else
  {
    cps_in  = flux_io_get_read_cps  (member);
    cps_out = flux_io_get_write_cps (member);

    cps_cap_in  = flux_io_get_read_cps_goal (member);
    cps_cap_out = flux_io_get_write_cps_goal (member);
  }

  cps_min_in  = calc_min_cps (stats->io_group->goal_cps_in, stats->priority_sum_in,
                              stats->lowest_priority_in, member_record->priority_in);

  cps_min_out = calc_min_cps (stats->io_group->goal_cps_out, stats->priority_sum_out,
                              stats->lowest_priority_out, member_record->priority_out);

  stats->spare_cps_in  += cps_min_in  - capped_cps (cps_in,  cps_cap_in);
  stats->spare_cps_out += cps_min_out - capped_cps (cps_out, cps_cap_out);

  if (capped_cps (cps_in, cps_cap_in) > cps_min_in)
    stats->over_min_priority_sum_in += member_record->priority_in + PRI_RANGE_PAD;

  if (capped_cps (cps_out, cps_cap_out) > cps_min_out)
    stats->over_min_priority_sum_out += member_record->priority_out + PRI_RANGE_PAD;
}

static void
gather_stats (FluxIOGroup *io_group, Stats *stats)
{
  io_group->used_cps_in  = 0;
  io_group->used_cps_out = 0;

  stats->io_group                  = io_group;
  stats->highest_priority_in       = G_MINDOUBLE;
  stats->highest_priority_out      = G_MINDOUBLE;
  stats->lowest_priority_in        = G_MAXDOUBLE;
  stats->lowest_priority_out       = G_MAXDOUBLE;
  stats->priority_sum_in           = 0;
  stats->priority_sum_out          = 0;
  stats->over_min_priority_sum_in  = 0;
  stats->over_min_priority_sum_out = 0;
  stats->capped_priority_sum_in    = 0;
  stats->capped_priority_sum_out   = 0;
  stats->spare_cps_in              = 0;
  stats->spare_cps_out             = 0;

  g_hash_table_foreach (io_group->member_table, (GHFunc) gather_stats_1_cb, stats);

  stats->priority_sum_in  -= g_hash_table_size (io_group->member_table) * stats->lowest_priority_in;
  stats->priority_sum_out -= g_hash_table_size (io_group->member_table) * stats->lowest_priority_out;

  g_hash_table_foreach (io_group->member_table, (GHFunc) gather_stats_2_cb, stats);
}

static void
tighten_caps_cb (gpointer member, Member *member_record, Change *change)
{
  guint   cps;
  guint   cps_cap;
  guint   cps_min;
  gdouble priority_inverse;
  gint    spare_cps;
  gint    excess_cps;
  gdouble over_min_priority_sum;

  /* Get stats */

  if (change->in)
  {
    if (member_record->is_group)
    {
      flux_io_group_get_cps (member, &cps, NULL);
      flux_io_group_get_cps_goal (member, &cps_cap, NULL);
    }
    else
    {
      cps     = flux_io_get_read_cps (member);
      cps_cap = flux_io_get_read_cps_goal (member);
    }

    cps_min = calc_min_cps (change->stats->io_group->goal_cps_in, change->stats->priority_sum_in,
                            change->stats->lowest_priority_in, member_record->priority_in);

    priority_inverse      = change->stats->highest_priority_in - member_record->priority_in + PRI_RANGE_PAD;
    spare_cps             = change->stats->spare_cps_in;
    over_min_priority_sum = change->stats->over_min_priority_sum_in;
  }
  else
  {
    if (member_record->is_group)
    {
      flux_io_group_get_cps (member, NULL, &cps);
      flux_io_group_get_cps_goal (member, NULL, &cps_cap);
    }
    else
    {
      cps     = flux_io_get_write_cps (member);
      cps_cap = flux_io_get_write_cps_goal (member);
    }

    cps_min = calc_min_cps (change->stats->io_group->goal_cps_out, change->stats->priority_sum_out,
                            change->stats->lowest_priority_out, member_record->priority_out);

    priority_inverse      = change->stats->highest_priority_out - member_record->priority_out + PRI_RANGE_PAD;
    spare_cps             = change->stats->spare_cps_out;
    over_min_priority_sum = change->stats->over_min_priority_sum_out;
  }

  if (over_min_priority_sum == 0)
    return;

  /* Calculate */

  excess_cps = capped_cps (cps, cps_cap) - cps_min;
  if (excess_cps <= 0)
    return;

  cps_cap =
    capped_cps (cps, cps_cap) - MIN (excess_cps, ((-spare_cps) * priority_inverse) / over_min_priority_sum);

  /* Set new cap */

  set_cap (member, member_record, cps_cap, change->in);
}

static void
loosen_caps_cb (gpointer member, Member *member_record, Change *change)
{
  guint   cps;
  guint   cps_cap;
  guint   cps_min;
  gdouble priority;
  gint    goal_cps;
  gint    spare_cps;
  gint    avail_cps;
  gdouble capped_priority_sum;

  /* Get stats */

  if (change->in)
  {
    if (member_record->is_group)
    {
      flux_io_group_get_cps (member, &cps, NULL);
      flux_io_group_get_cps_goal (member, &cps_cap, NULL);
    }
    else
    {
      cps     = flux_io_get_read_cps (member);
      cps_cap = flux_io_get_read_cps_goal (member);
    }

    cps_min = calc_min_cps (change->stats->io_group->goal_cps_in, change->stats->priority_sum_in,
                            change->stats->lowest_priority_in, member_record->priority_in);

    priority            = member_record->priority_in - change->stats->lowest_priority_in + PRI_RANGE_PAD;
    goal_cps            = change->stats->io_group->goal_cps_in;
    spare_cps           = change->stats->spare_cps_in;
    capped_priority_sum = change->stats->capped_priority_sum_in;
  }
  else
  {
    if (member_record->is_group)
    {
      flux_io_group_get_cps (member, NULL, &cps);
      flux_io_group_get_cps_goal (member, NULL, &cps_cap);
    }
    else
    {
      cps     = flux_io_get_write_cps (member);
      cps_cap = flux_io_get_write_cps_goal (member);
    }

    cps_min = calc_min_cps (change->stats->io_group->goal_cps_out, change->stats->priority_sum_out,
                            change->stats->lowest_priority_out, member_record->priority_out);

    priority            = member_record->priority_out - change->stats->lowest_priority_out + PRI_RANGE_PAD;
    goal_cps            = change->stats->io_group->goal_cps_out;
    spare_cps           = change->stats->spare_cps_out;
    capped_priority_sum = change->stats->capped_priority_sum_out;
  }

  if (capped_priority_sum == 0)
    return;

  if (!is_capped (cps, cps_cap))
    return;

  /* Calculate */

  avail_cps = goal_cps - cps_cap;
  cps_cap += MIN (avail_cps, (spare_cps * priority) / capped_priority_sum);

  /* Set new cap */

  set_cap (member, member_record, cps_cap, change->in);
}

static void
tighten_caps (FluxIOGroup *io_group, Stats *stats, gboolean in)
{
  Change change;

  change.stats = stats;
  change.in    = in;

  g_hash_table_foreach (io_group->member_table, (GHFunc) tighten_caps_cb, &change);
}

static void
loosen_caps (FluxIOGroup *io_group, Stats *stats, gboolean in)
{
  Change change;

  change.stats = stats;
  change.in    = in;

  g_hash_table_foreach (io_group->member_table, (GHFunc) loosen_caps_cb, &change);
}

static void iterate_group (FluxIOGroup *io_group);

static void
iterate_child_cb (gpointer member, Member *member_record)
{
  if (member_record->is_group)
    iterate_group (member);
}

static void
iterate_group (FluxIOGroup *io_group)
{
  Stats stats;

  gather_stats (io_group, &stats);

  if (stats.spare_cps_in < 0)
    tighten_caps (io_group, &stats, TRUE);
  else if (stats.spare_cps_in > 0)
    loosen_caps (io_group, &stats, TRUE);

  if (stats.spare_cps_out < 0)
    tighten_caps (io_group, &stats, FALSE);
  else if (stats.spare_cps_out > 0)
    loosen_caps (io_group, &stats, FALSE);

  g_hash_table_foreach (io_group->member_table, (GHFunc) iterate_child_cb, NULL);
}

static gboolean
iterate_groups (void)
{
  GSList *l;

  for (l = group_slice_list; l; l = g_slist_next (l))
  {
    FluxIOGroup *group = l->data;

    iterate_group (group);
  }

  return TRUE;
}

void
flux_io_group_init (FluxIOGroup *io_group)
{
  flux_object_init              (FLUX_OBJECT (io_group));
  flux_object_set_type          (FLUX_OBJECT (io_group), FLUX_IO_GROUP_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (io_group), (FluxFinalizeFunc *) flux_io_group_finalize);

  io_group->member_table = g_hash_table_new_full (g_direct_hash,
                                                  g_direct_equal,
                                                  (GDestroyNotify) flux_object_unref,
                                                  (GDestroyNotify) g_free);

  group_slice_list = g_slist_prepend (group_slice_list, io_group);

  if (!group_slice_timeout_id)
    group_slice_timeout_id = g_timeout_add (1000, (GSourceFunc) iterate_groups, NULL);
}

FluxIOGroup *
flux_io_group_new (void)
{
  FluxIOGroup *io_group;

  io_group = g_new0 (FluxIOGroup, 1);
  flux_io_group_init (io_group);

  return io_group;
}

void
flux_io_group_finalize (FluxObject *object)
{
  FluxIOGroup *io_group = FLUX_IO_GROUP (object);

  g_hash_table_destroy (io_group->member_table);
  group_slice_list = g_slist_remove (group_slice_list, io_group);

  if (!group_slice_list && group_slice_timeout_id)
  {
    g_source_remove (group_slice_timeout_id);
    group_slice_timeout_id = 0;
  }
}

void
flux_io_group_add_group (FluxIOGroup *io_group, FluxIOGroup *child_io_group)
{
  Member *member_record;

  if (!g_slist_find (group_slice_list, child_io_group))
  {
    g_warning ("Child FluxIOGroup does not exist or is not a toplevel group.");
    return;
  }

  group_slice_list = g_slist_remove (group_slice_list, child_io_group);

  member_record = g_new (Member, 1);
  member_record->is_group     = TRUE;
  member_record->priority_in  = 0;
  member_record->priority_out = 0;

  flux_object_ref (FLUX_OBJECT (child_io_group));
  g_hash_table_insert (io_group->member_table, child_io_group, member_record);
}

void
flux_io_group_add_io (FluxIOGroup *io_group, FluxIO *io)
{
  Member *member_record;

  member_record = g_new (Member, 1);
  member_record->is_group     = FALSE;
  member_record->priority_in  = 0;
  member_record->priority_out = 0;

  flux_object_ref (FLUX_OBJECT (io));
  g_hash_table_insert (io_group->member_table, io, member_record);
}

gboolean
flux_io_group_remove_member (FluxIOGroup *io_group, gpointer member)
{
  Member *member_record;

  member_record = g_hash_table_lookup (io_group->member_table, member);
  if (!member_record)
    return FALSE;

  /* Child group becomes toplevel */
  if (member_record->is_group)
    group_slice_list = g_slist_prepend (group_slice_list, member);

  return g_hash_table_remove (io_group->member_table, member);
}

void
flux_io_group_get_cps_goal (FluxIOGroup *io_group, gint *cps_in, gint *cps_out)
{
  if (cps_in)
    *cps_in  = io_group->goal_cps_in;
  if (cps_out)
    *cps_out = io_group->goal_cps_out;
}

void
flux_io_group_set_cps_goal (FluxIOGroup *io_group, gint cps_in, gint cps_out)
{
  if (cps_in > -1)
    io_group->goal_cps_in  = cps_in;
  if (cps_out > -1)
    io_group->goal_cps_out = cps_out;
}

void
flux_io_group_get_cps (FluxIOGroup *io_group, gint *cps_in, gint *cps_out)
{
  if (cps_in)
    *cps_in  = io_group->used_cps_in;
  if (cps_out)
    *cps_out = io_group->used_cps_out;
}

gboolean
flux_io_group_get_member_priority (FluxIOGroup *io_group, gconstpointer member,
                                   gdouble *priority_in, gdouble *priority_out)
{
  Member *member_record;

  member_record = g_hash_table_lookup (io_group->member_table, member);
  if (!member_record)
    return FALSE;

  *priority_in  = member_record->priority_in;
  *priority_out = member_record->priority_out;
  return TRUE;
}

gboolean
flux_io_group_set_member_priority (FluxIOGroup *io_group, gpointer member,
                                   gdouble priority_in, gdouble priority_out)
{
  Member *member_record;

  member_record = g_hash_table_lookup (io_group->member_table, member);
  if (!member_record)
    return FALSE;

  member_record->priority_in  = priority_in;
  member_record->priority_out = priority_out;
  return TRUE;
}

