/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-io.c - Transport-independent full-duplex stream class.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/time.h>
#else
# include <io.h>
# include <winsock2.h>
# include <time.h>
#endif

#include "flux-io.h"
#include "flux-tcp-io.h"

#define CPS_WEIGHT 0.1

static GStaticMutex  cap_slice_mutex      = G_STATIC_MUTEX_INIT;
static guint         cap_slice_timeout_id = 0;
static GSList       *cap_slice_io_list    = NULL;

void
_flux_io_lock (void)
{
  g_static_mutex_lock (&cap_slice_mutex);
}

void
_flux_io_unlock (void)
{
  g_static_mutex_unlock (&cap_slice_mutex);
}

static gboolean
cap_slice_timeout (gpointer data)
{
  GSList *l;

  _flux_io_lock ();

  for (l = cap_slice_io_list; l; l = g_slist_next (l))
  {
    FluxIO *io = l->data;

    flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

    io->read_cps  = io->read_cps  * (1.0 - CPS_WEIGHT) + io->read_count  * CPS_WEIGHT;
    io->write_cps = io->write_cps * (1.0 - CPS_WEIGHT) + io->write_count * CPS_WEIGHT;

    io->read_count  = 0;
    io->write_count = 0;

#if 0
    io->read_slice_cap  = MAX (0, io->read_slice_cap  - io->read_count)  / 2 + io->read_cap;
    io->write_slice_cap = MAX (0, io->write_slice_cap - io->write_count) / 2 + io->write_cap;
#else
    io->read_slice_cap  = io->read_cap;
    io->write_slice_cap = io->write_cap;
#endif

    /* NOTE: Callback can't alter cap_slice_io_list */
    if (io->read_cap > 0 && FLUX_DUAL_FIFO (io)->read_hook)
      FLUX_DUAL_FIFO (io)->read_hook (FLUX_DUAL_FIFO (io)->read_hook_data);

    /* NOTE: Callback can't alter cap_slice_io_list */
    if (io->write_cap > 0 && FLUX_DUAL_FIFO (io)->write_hook)
      FLUX_DUAL_FIFO (io)->write_hook (FLUX_DUAL_FIFO (io)->write_hook_data);
  }

  _flux_io_unlock ();

  return TRUE;
}

void
flux_io_init (FluxIO *io)
{
  flux_dual_fifo_init           (FLUX_DUAL_FIFO (io), 2, 8, 1024);
  flux_object_set_type          (FLUX_OBJECT (io), FLUX_IO_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (io), flux_io_finalize);

  io->read_compression_level = 0;
  io->write_compression_level = 0;

  io->read_threshold  = 0;
  io->write_threshold = 0;

  io->read_cap  = -1;
  io->write_cap = -1;

  io->read_count  = 0;
  io->write_count = 0;

  io->read_cps  = 0;
  io->write_cps = 0;

  _flux_io_lock ();

  if (!cap_slice_timeout_id)
    cap_slice_timeout_id = g_timeout_add (1000, cap_slice_timeout, NULL);

  cap_slice_io_list = g_slist_prepend (cap_slice_io_list, io);

  _flux_io_unlock ();
}

void
flux_io_finalize (FluxObject *object)
{
  FluxIO *io = FLUX_IO (object);

  _flux_io_lock ();

  cap_slice_io_list = g_slist_remove (cap_slice_io_list, io);

  if (!cap_slice_io_list)
  {
    g_source_remove (cap_slice_timeout_id);
    cap_slice_timeout_id = 0;
  }

  _flux_io_unlock ();

  flux_dual_fifo_finalize (object);
}

gboolean
flux_io_is_connected (FluxIO *io)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  if (flux_object_is_type (FLUX_OBJECT (io), FLUX_TCP_IO_TYPE) &&
      flux_tcp_io_get_state (FLUX_TCP_IO (io)) == FLUX_TCP_IO_STATE_CONNECTED)
    return TRUE;

  return FALSE;
}

void
flux_io_set_cipher (FluxIO *io, FluxCipher *cipher)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  if (io->cipher)
    flux_object_unref (FLUX_OBJECT (io->cipher));

  flux_object_ref (FLUX_OBJECT (cipher));
  io->cipher = cipher;
}

void
flux_io_set_read_func (FluxIO *io, FluxIOReadFunc *func, gpointer data)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->read_func      = func;
  io->read_func_data = data;

  if (io->read_func && io->read_threshold > 0 && FLUX_DUAL_FIFO (io)->read_hook)
    FLUX_DUAL_FIFO (io)->read_hook (FLUX_DUAL_FIFO (io)->read_hook_data);

  _flux_io_unlock ();
}

void
flux_io_set_write_func (FluxIO *io, FluxIOWriteFunc *func, gpointer data)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->write_func      = func;
  io->write_func_data = data;

  if (io->write_func && io->write_threshold > 0 && FLUX_DUAL_FIFO (io)->write_hook)
    FLUX_DUAL_FIFO (io)->write_hook (FLUX_DUAL_FIFO (io)->write_hook_data);

  _flux_io_unlock ();
}

void
flux_io_set_read_threshold (FluxIO *io, guint threshold)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->read_threshold = threshold;

  if (threshold > 0 && FLUX_DUAL_FIFO (io)->read_hook)
    FLUX_DUAL_FIFO (io)->read_hook (FLUX_DUAL_FIFO (io)->read_hook_data);

  _flux_io_unlock ();
}

void
flux_io_set_write_threshold (FluxIO *io, guint threshold)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->write_threshold = threshold;

  if (threshold > 0 && FLUX_DUAL_FIFO (io)->write_hook)
    FLUX_DUAL_FIFO (io)->write_hook (FLUX_DUAL_FIFO (io)->write_hook_data);

  _flux_io_unlock ();
}

gint
flux_io_get_read_cps_goal (FluxIO *io)
{
  gint read_cap;

  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();
  read_cap = io->read_cap;
  _flux_io_unlock ();

  return read_cap;
}

void
flux_io_set_read_cps_goal (FluxIO *io, gint cps)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->read_cap       = cps;
  io->read_slice_cap = MAX (io->read_cap, io->read_slice_cap);

  if ((io->read_cap == -1 || io->read_count < io->read_slice_cap) &&
      FLUX_DUAL_FIFO (io)->read_hook)
    FLUX_DUAL_FIFO (io)->read_hook (FLUX_DUAL_FIFO (io)->read_hook_data);

  _flux_io_unlock ();
}

gint
flux_io_get_write_cps_goal (FluxIO *io)
{
  gint write_cap;

  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();
  write_cap = io->write_cap;
  _flux_io_unlock ();

  return write_cap;
}

void
flux_io_set_write_cps_goal (FluxIO *io, gint cps)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->write_cap       = cps;
  io->write_slice_cap = MAX (io->write_cap, io->write_slice_cap);

  if ((io->write_cap == -1 || io->write_count < io->write_slice_cap) &&
      FLUX_DUAL_FIFO (io)->write_hook)
    FLUX_DUAL_FIFO (io)->write_hook (FLUX_DUAL_FIFO (io)->write_hook_data);

  _flux_io_unlock ();
}

gint
flux_io_get_read_cps (FluxIO *io)
{
  gint read_cps;

  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();
  read_cps = io->read_cps;
  _flux_io_unlock ();

  return read_cps;
}

gint
flux_io_get_write_cps (FluxIO *io)
{
  gint write_cps;

  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();
  write_cps = io->write_cps;
  _flux_io_unlock ();

  return write_cps;
}

void
flux_io_set_compression (FluxIO *io, gint read_level, gint write_level)
{
  flux_object_check (FLUX_OBJECT (io), FLUX_IO_TYPE);

  _flux_io_lock ();

  io->read_compression_level  = read_level;
  io->write_compression_level = write_level;

  if (flux_io_is_connected (io))
    flux_dual_fifo_set_compression (FLUX_DUAL_FIFO (io), read_level, write_level);

  _flux_io_unlock ();
}
