/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-ip-addr.c - Internet address.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/time.h>
# include <sys/socket.h>
# include <sys/ioctl.h>
# include <net/if.h>
# include <ifaddrs.h>
#else
# include <io.h>
# include <winsock2.h>
# include <ws2tcpip.h>
# include <time.h>
#endif

#include "flux-init.h"
#include "flux-util.h"
#include "flux-socket-util.h"
#include "flux-ip-addr.h"

void
flux_ip_addr_init (FluxIPAddr *ip_addr)
{
  flux_object_init              (FLUX_OBJECT (ip_addr));
  flux_object_set_type          (FLUX_OBJECT (ip_addr), FLUX_IP_ADDR_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (ip_addr), flux_ip_addr_finalize);

  ip_addr->type = FLUX_IP_ADDR_INVALID;
  memset (ip_addr->addr, 0, sizeof (ip_addr->addr));
}

void
flux_ip_addr_finalize (FluxObject *object)
{
}

FluxIPAddr *
flux_ip_addr_new (void)
{
  FluxIPAddr *ip_addr;

  ip_addr = g_new (FluxIPAddr, 1);
  flux_ip_addr_init (ip_addr);

  return ip_addr;
}

gboolean
flux_ip_addr_is_valid (FluxIPAddr *ip_addr)
{
  return ip_addr->type == FLUX_IP_ADDR_INVALID ? FALSE : TRUE;
}

FluxIPAddrType
flux_ip_addr_get_type (FluxIPAddr *ip_addr)
{
  return ip_addr->type;
}

gchar *
flux_ip_addr_get_string (FluxIPAddr *ip_addr)
{
  switch (ip_addr->type)
  {
    case FLUX_IP_ADDR_IPV4:
      return g_strdup_printf ("%"  G_GUINT16_FORMAT ".%" G_GUINT16_FORMAT
                              ".%" G_GUINT16_FORMAT ".%" G_GUINT16_FORMAT,
                              ip_addr->addr [0], ip_addr->addr [1],
                              ip_addr->addr [2], ip_addr->addr [3]);

    case FLUX_IP_ADDR_IPV6:
      /* TODO: Output compressed form if possible */
      return g_strdup_printf ("%"   /* G_GINT16_MODIFIER */ "X:%" /* G_GINT16_MODIFIER */
                              "X:%" /* G_GINT16_MODIFIER */ "X:%" /* G_GINT16_MODIFIER */
                              "X:%" /* G_GINT16_MODIFIER */ "X:%" /* G_GINT16_MODIFIER */
                              "X:%" /* G_GINT16_MODIFIER */ "X:%" /* G_GINT16_MODIFIER */ "X",
                              (ip_addr->addr  [0] << 8) | ip_addr->addr  [1],
                              (ip_addr->addr  [2] << 8) | ip_addr->addr  [3],
                              (ip_addr->addr  [4] << 8) | ip_addr->addr  [5],
                              (ip_addr->addr  [6] << 8) | ip_addr->addr  [7],
                              (ip_addr->addr  [8] << 8) | ip_addr->addr  [9],
                              (ip_addr->addr [10] << 8) | ip_addr->addr [11],
                              (ip_addr->addr [12] << 8) | ip_addr->addr [13],
                              (ip_addr->addr [14] << 8) | ip_addr->addr [15]);
    default:
      break;
  }

  /* Invalid */
  return NULL;
}

static gint
parse_ipv4_field (const gchar *field, const gchar *punct)
{
  guint val = 0;

  for ( ; field < punct; field++)
  {
    gint d;

    val *= 10;
    d = *field - '0';

    if (d < 0 || d > 9)
      return -1;

    val += d;
  }

  if (val > 0xff)
    return -1;

  return val;
}

static gint
parse_ipv6_field (const gchar *field, const gchar *punct)
{
  guint val = 0;

  for ( ; field < punct; field++)
  {
    gint d;

    val *= 16;
    d = *field - '0';

    if (d < 0 || d > 9)
    {
      d = tolower (*field) - 'a' + 10;
      if (d < 10 || d > 15)
        return -1;
    }

    val += d;
  }

  if (val > 0xffff)
    return -1;

  return val;
}

gboolean
flux_ip_addr_set_string (FluxIPAddr *ip_addr, const gchar *string)
{
  static gchar   *valid_chars     = "0123456789abcdefABCDEF.:";
  static gchar   *separator_chars = ".:";
  gboolean        well_formed     = TRUE;
  gint            compress_start  = -1;
  FluxIPAddrType  last_field_type = FLUX_IP_ADDR_INVALID;
  FluxIPAddrType  addr_type       = FLUX_IP_ADDR_INVALID;
  gint            ipv4_fields     = 0;
  gint            string_len;
  const gchar    *field;
  const gchar    *punct;
  guint8          addr [16];
  gint            t, i;

  g_return_val_if_fail (string != NULL, FALSE);

  string_len = strlen (string);

  /* Bail if contains invalid characters */

  if (strspn (string, valid_chars) != string_len)
    return FALSE;

  /* Parse */

  for (i = 0, field = string; ; )
  {
    punct = field + strcspn (field, separator_chars);

    if (punct == field)
    {
      /* Two puncts in a row, or punct at beginning - must be "::". If first char is a '\0', or
       * if we have a single punctuation at the end of the string, we'll also end up here. */

      if (*punct != ':' || (i == 0 && *(punct + 1) != ':') || (i > 0 && *(punct - 1) != ':'))
      {
        well_formed = FALSE;
        break;
      }

      /* Can't have more than one :: in an address */
      if (compress_start >= 0)
      {
        well_formed = FALSE;
        break;
      }

      addr_type = FLUX_IP_ADDR_IPV6;
      last_field_type = FLUX_IP_ADDR_IPV6;
      compress_start = i;

      if (i == 0)
        field += 2;
      else
        field += 1;

      if (!*field)
        break;

      continue;
    }

    if (*punct == '.' || (*punct == '\0' && last_field_type == FLUX_IP_ADDR_IPV4))
    {
      /* IPv4 style - decimal */

      if (addr_type == FLUX_IP_ADDR_INVALID)
        addr_type = FLUX_IP_ADDR_IPV4;
      last_field_type = FLUX_IP_ADDR_IPV4;
      t = parse_ipv4_field (field, punct);

      if (t < 0 || i > 15 || ++ipv4_fields > 4)
      {
        /* Bad field contents, too long address, or too long IPv4 portion of IPv6 address */
        well_formed = FALSE;
        break;
      }

      addr [i++] = t;
    }
    else if (*punct == ':' || last_field_type == FLUX_IP_ADDR_IPV6)
    {
      /* IPv6 style - hexadecimal */

      /* Can have IPv4 syntax only at the end of an IPv6 string */
      if (last_field_type == FLUX_IP_ADDR_IPV4)
      {
        well_formed = FALSE;
        break;
      }

      addr_type = FLUX_IP_ADDR_IPV6;
      last_field_type = FLUX_IP_ADDR_IPV6;
      t = parse_ipv6_field (field, punct);

      if (t < 0 || i > 14)
      {
        /* Bad field contents, or too long address */
        well_formed = FALSE;
        break;
      }

      addr [i++] = t >> 8;
      addr [i++] = t & 0xff;
    }

    if (!*punct)
      break;

    field = punct + 1;
  }

  /* Post-process */

  if (well_formed)
  {
    if (compress_start >= 0)
    {
      /* Maybe decompress IPv6 shorthand */
      memmove (&addr [16 - (i - compress_start)], &addr [compress_start], i - compress_start);
      memset (&addr [compress_start], 0, 16 - i);
    }
    else if (addr_type == FLUX_IP_ADDR_IPV6)
    {
      /* IPv6 addresses with no compression in them must be specific */
      if (i != 16)
        well_formed = FALSE;
    }
    else if (addr_type == FLUX_IP_ADDR_IPV4)
    {
      /* IPv4 addresses must always be specific. If it is, clear the rest of the
       * field for cleanliness. */
      if (i != 4)
        well_formed = FALSE;
      else
        memset (&addr [4], 0, 16 - 4);
    }
    else
    {
      /* addr_type == FLUX_IP_ADDR_INVALID */
      well_formed = FALSE;
    }
  }

  /* If it's still well formed, set our address */

  if (well_formed)
  {
    ip_addr->type = addr_type;
    memcpy (ip_addr->addr, addr, 16);
  }

  return well_formed;
}

gboolean
flux_ip_addr_get_raw (FluxIPAddr *ip_addr, guint8 *dest, gint *len)
{
  g_return_val_if_fail (dest != NULL, FALSE);

  switch (ip_addr->type)
  {
    case FLUX_IP_ADDR_IPV4:
      memcpy (dest, ip_addr->addr, 4);
      *len = 4;
      break;

    case FLUX_IP_ADDR_IPV6:
      memcpy (dest, ip_addr->addr, 16);
      *len = 16;
      break;

    default:
      return FALSE;
  }

  return TRUE;
}

gboolean
flux_ip_addr_set_raw (FluxIPAddr *ip_addr, const guint8 *src, gint len)
{
  g_return_val_if_fail (src != NULL, FALSE);
  g_return_val_if_fail (len > 0, FALSE);

  switch (len)
  {
    case 4:
      ip_addr->type = FLUX_IP_ADDR_IPV4;
      memset (ip_addr->addr, 0, 16);
      memcpy (ip_addr->addr, src, 4);
      break;

    case 16:
      ip_addr->type = FLUX_IP_ADDR_IPV6;
      memcpy (ip_addr->addr, src, 16);
      break;

    default:
      return FALSE;
  }

  return TRUE;
}

struct sockaddr *
flux_ip_addr_get_sockaddr (FluxIPAddr *ip_addr, gint port)
{
  struct sockaddr     *sa;
  gchar               *addr_str;

  addr_str = flux_ip_addr_get_string (ip_addr);
  if (!addr_str)
    return NULL;

  sa = flux_sockaddr_new (ip_addr->type, addr_str, port);
  g_free (addr_str);

  return sa;
}

gboolean
flux_ip_addr_set_sockaddr (FluxIPAddr *ip_addr, struct sockaddr *sa)
{
  gchar    *addr_str;
  gboolean  result;

  g_return_val_if_fail (sa != NULL, FALSE);

  addr_str = flux_sockaddr_to_string (sa);
  if (!addr_str)
    return FALSE;

  result = flux_ip_addr_set_string (ip_addr, addr_str);
  g_free (addr_str);

  return result;
}

gboolean
flux_ip_addr_is_loopback (FluxIPAddr *ip_addr)
{
  if (ip_addr->type == FLUX_IP_ADDR_IPV4)
  {
    guint32 addr;

    addr = GUINT32_FROM_BE (*((guint32 *) ip_addr->addr));

    if ((addr & 0xFF000000) == 127 << 24)
      return TRUE;
  }

  /* TODO: IPv6 */

  return FALSE;
}

gboolean
flux_ip_addr_is_multicast (FluxIPAddr *ip_addr)
{
  if (ip_addr->type == FLUX_IP_ADDR_IPV4)
  {
    guint32 addr;

    addr = GUINT32_FROM_BE (*((guint32 *) ip_addr->addr));

    if ((addr & 0xF0000000) == 0xE0000000)
      return TRUE;
  }

  /* TODO: IPv6 */

  return FALSE;
}

gboolean
flux_ip_addr_is_broadcast (FluxIPAddr *ip_addr)
{
  if (ip_addr->type == FLUX_IP_ADDR_IPV4)
  {
    guint32 addr;

    addr = GUINT32_FROM_BE (*((guint32 *) ip_addr->addr));

    if (addr == 0xFFFFFFFF)
      return TRUE;
  }

  /* TODO: IPv6 */

  return FALSE;
}

gboolean
flux_ip_addr_is_reserved (FluxIPAddr *ip_addr)
{
  if (ip_addr->type == FLUX_IP_ADDR_IPV4)
  {
    guint32 addr;

    addr = GUINT32_FROM_BE (*((guint32 *) ip_addr->addr));

    if ((addr & 0xFFFF0000) == 0)
      return TRUE;

    if ((addr & 0xF8000000) == 0xF0000000)
      return TRUE;
  }

  /* TODO: IPv6 */

  return FALSE;
}

gboolean
flux_ip_addr_is_private (FluxIPAddr *ip_addr)
{
  if (ip_addr->type == FLUX_IP_ADDR_IPV4)
  {
    guint32 addr;

    addr = GUINT32_FROM_BE (*((guint32 *) ip_addr->addr));

    if ((addr & 0xFF000000) == (10 << 24))
      return TRUE;

    if ((addr & 0xFFF00000) == 0xAC100000)
      return TRUE;

    if ((addr & 0xFFFF0000) == 0xC0A80000)
      return TRUE;
  }

  /* TODO: IPv6 */

  return FALSE;
}

gboolean
flux_ip_addr_is_internet (FluxIPAddr *ip_addr)
{
  if (!flux_ip_addr_is_private   (ip_addr) &&
      !flux_ip_addr_is_reserved  (ip_addr) &&
      !flux_ip_addr_is_loopback  (ip_addr) &&
      !flux_ip_addr_is_multicast (ip_addr) &&
      !flux_ip_addr_is_broadcast (ip_addr))
  {
    return TRUE;
  }

  return FALSE;
}

FluxIPAddr *
flux_ip_addr_get_interface_to (FluxIPAddr *ip_addr)
{
  gint        fd = -1;
  gpointer    remote_sa = NULL;
  gchar       interface_sa [512];
  gint        interface_sa_len = sizeof (interface_sa);  /* socklen_t */
  FluxIPAddr *interface_addr;

  flux_init_sockets ();

#ifdef HAVE_IPV6
  fd = socket (ip_addr->type == FLUX_IP_ADDR_IPV4 ? AF_INET : AF_INET6,
               SOCK_DGRAM, 0);
#else
  fd = socket (AF_INET, SOCK_DGRAM, 0);
#endif

  if (fd < 0)
    goto error;

  remote_sa = flux_ip_addr_get_sockaddr (ip_addr, 0);

  if (connect (fd, (struct sockaddr *) remote_sa, flux_sockaddr_get_len ((struct sockaddr *) remote_sa)) < 0)
    goto error;

  if (getsockname (fd, (struct sockaddr *) interface_sa, &interface_sa_len) != 0)
    goto error;

  interface_addr = flux_ip_addr_new ();
  flux_ip_addr_set_sockaddr (interface_addr, (struct sockaddr *) interface_sa);

  flux_socket_close (fd);
  g_free (remote_sa);
  return interface_addr;

error:
  if (fd >= 0)
    flux_socket_close (fd);

  g_free (remote_sa);
  return NULL;
}

FluxIPAddr *
flux_ip_addr_find_internet_interface (void)
{
  FluxIPAddr *ip_addr;
  FluxIPAddr *if_addr;
  GList      *if_list;
  GList      *l;

  /* Try fix.no's address */
  ip_addr = flux_ip_addr_new ();
  flux_ip_addr_set_string (ip_addr, "212.71.72.21");

  if_addr = flux_ip_addr_get_interface_to (ip_addr);
  flux_object_unref (FLUX_OBJECT (ip_addr));

  if (if_addr && flux_ip_addr_is_internet (if_addr))
    return if_addr;

  if (if_addr)
  {
    flux_object_unref (FLUX_OBJECT (if_addr));
    if_addr = NULL;
  }

  if_list = flux_ip_addr_get_interfaces ();
  for (l = if_list; l; l = g_list_next (l))
  {
    ip_addr = l->data;

    if (flux_ip_addr_is_internet (ip_addr))
    {
      if_addr = ip_addr;
      break;
    }
  }

  if (if_addr)
    flux_object_ref (FLUX_OBJECT (if_addr));

  for (l = if_list; l; l = g_list_next (l))
    flux_object_unref (FLUX_OBJECT (l->data));

  g_list_free (if_list);
  return if_addr;
}

#ifndef G_PLATFORM_WIN32  /* Unix-specific code follows */

GList *
flux_ip_addr_get_interfaces (void)
{
  GList         *interface_list = NULL;
  gint           len, lastlen;
  gchar         *buf, *p0;
  gint           fd;
  struct ifconf  ifc;
  struct ifreq  *ifr;

  flux_init_sockets ();

  /* TODO: IPv6 method */

  fd = socket (AF_INET, SOCK_DGRAM, 0);
  if (fd < 0)
    return NULL;

  len     = 8 * sizeof (struct ifreq);
  lastlen = 0;

  for (;;)
  {
    buf = g_new0 (gchar, len);

    ifc.ifc_len = len;
    ifc.ifc_buf = buf;

    if (ioctl (fd, SIOCGIFCONF, &ifc) < 0)
    {
      /* Might have failed because our buffer was too small */
      if (errno != EINVAL || lastlen != 0)
      {
        g_free (buf);
        return NULL;
      }
    }
    else
    {
      /* Break if we got all the interfaces */
      if (ifc.ifc_len == lastlen)
        break;

      lastlen = ifc.ifc_len;
    }

    /* Did not allocate big enough buffer - try again */
    len += 8 * sizeof (struct ifreq);
    g_free (buf);
  }

  for (p0 = buf; p0 < (buf + ifc.ifc_len); p0 += sizeof (struct ifreq))
  {
    struct sockaddr_storage  addr;
    int                      rv;
    FluxIPAddr              *ip_addr;

    ifr = (struct ifreq*) p0;

    /* Ignore non-AF_INET */
    if (ifr->ifr_addr.sa_family != AF_INET &&
        ifr->ifr_addr.sa_family != AF_INET6)
      continue;

    /* Save the address - the next call will clobber it */
    memcpy (&addr, &ifr->ifr_addr, sizeof(addr));

    /* Get the flags */
    rv = ioctl (fd, SIOCGIFFLAGS, ifr);
    if (rv == -1)
      continue;

    /* Ignore loopback and entries that aren't up */
    if (!(ifr->ifr_flags & IFF_UP) ||
        (ifr->ifr_flags & IFF_LOOPBACK))
      continue;

    /* Create a FluxIPAddr for this one and add it to our list */
    ip_addr = flux_ip_addr_new ();
    flux_ip_addr_set_sockaddr (ip_addr, (struct sockaddr *) &addr);
    interface_list = g_list_prepend (interface_list, ip_addr);
  }

  g_free (buf);

  interface_list = g_list_reverse (interface_list);
  return interface_list;
}

#else  /* Win32-specific code follows */

GList *
flux_ip_addr_get_interfaces (void)
{
  GList                  *interface_list = NULL;
  SOCKET                  s;
  DWORD                   bytesReturned;
  SOCKADDR_IN            *pAddrInet;
  u_long                  SetFlags;
  INTERFACE_INFO          localAddr [10];  /* No more than 10 IP interfaces */
  struct sockaddr_in      addr;
  gint                    numLocalAddr; 
  gint                    wsError;
  gint                    i;

  flux_init_sockets ();

  if ((s = WSASocket (AF_INET, SOCK_DGRAM, IPPROTO_UDP, NULL, 0, 0)) == INVALID_SOCKET)
    return NULL;

  /* Enumerate all IP interfaces */

  wsError = WSAIoctl (s, SIO_GET_INTERFACE_LIST, NULL, 0, &localAddr,
                      sizeof (localAddr), &bytesReturned, NULL, NULL);
  if (wsError == SOCKET_ERROR)
  {
    flux_socket_close (s);
    return NULL;
  }

  flux_socket_close (s);

  /* Display interface information */

  numLocalAddr = (bytesReturned / sizeof (INTERFACE_INFO));
  for (i = 0; i < numLocalAddr; i++)
  {
    pAddrInet = (SOCKADDR_IN *) &localAddr [i].iiAddress;
    memcpy (&addr, pAddrInet, sizeof (addr));

    SetFlags = localAddr [i].iiFlags;
    if ((SetFlags & IFF_UP) || (SetFlags & IFF_LOOPBACK))
    {
      FluxIPAddr *ip_addr;

      /* Create a FluxIPAddr for this one and add it to our list */
      ip_addr = flux_ip_addr_new ();
      flux_ip_addr_set_sockaddr (ip_addr, (struct sockaddr *) &addr);
      interface_list = g_list_prepend (interface_list, ip_addr);
    }
  }

  return interface_list;
}

#endif
