/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-ip-resolver.c - DNS resolver.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/socket.h>
# include <netdb.h>
# include <netinet/in.h>
# include <arpa/inet.h>
#else
# include <io.h>
# include <winsock2.h>
#endif

#include "flux-init.h"
#include "flux-socket-util.h"
#include "flux-ip-addr.h"
#include "flux-ip-resolver.h"

#define DEBUG(x)

#define ID_BITS 28

typedef struct
{
  FluxIPLookupFunc *func;                  /* User callback */
  gpointer          data;                  /* User data */

  FluxIPResolver   *resolver;              /* Resolver we belong to */

  guint             id         : ID_BITS;  /* ID of this request */
  guint             is_reverse : 1;        /* TRUE if we're translating IP -> name */
  guint             is_running : 1;        /* TRUE if thread picked it up */
  guint             is_wanted  : 1;        /* TRUE unless cancelled */

  gpointer          arg;                   /* (gchar *), or (FluxIPAddr *) on reverse lookup */
  GList            *results;
}
Lookup;

static FluxIPResolver *default_resolver = NULL;

static void
free_object_list (GList *list)
{
  GList *l;

  for (l = list; l; l = g_list_next (l))
    flux_object_unref (FLUX_OBJECT (l->data));

  g_list_free (list);
}

static void
free_deep_list (GList *list)
{
  GList *l;

  for (l = list; l; l = g_list_next (l))
    g_free (l->data);

  g_list_free (list);
}

#ifdef HAVE_GETADDRINFO

static GList *
lookup_getaddrinfo (const gchar *name)
{
  struct addrinfo  hints;
  struct addrinfo *res       = NULL;
  GList           *addr_list = NULL;
  gint             rv;

  memset (&hints, 0, sizeof (struct addrinfo));
  hints.ai_socktype = SOCK_STREAM;

  rv = getaddrinfo (name, NULL, &hints, &res);
  if (rv == 0)
  {
    struct addrinfo *i;

    for (i = res; i != NULL; i = i->ai_next)
    {
      FluxIPAddr *ip_addr;

      if (!flux_addr_family_is_ip (i->ai_family))
        continue;

      ip_addr = flux_ip_addr_new ();

      if (!flux_ip_addr_set_sockaddr (ip_addr, i->ai_addr))
      {
        flux_object_unref (FLUX_OBJECT (ip_addr));
        continue;
      }

      addr_list = g_list_prepend (addr_list, ip_addr);
    }
  }

  if (res)
    freeaddrinfo (res);

  return addr_list;
}

static GList *
lookup_getnameinfo (FluxIPAddr *ip_addr)
{
  struct sockaddr *sa;
  gchar            name [NI_MAXHOST];
  gint             rv;

  sa = flux_ip_addr_get_sockaddr (ip_addr, 0);
  if (!sa)
    return NULL;

  do
  {
    rv = getnameinfo (sa, flux_sockaddr_get_len (sa),
                      name, sizeof (name),
                      NULL, 0, NI_NAMEREQD);
  }
  while (rv != 0 && errno == EAI_AGAIN);

  g_free (sa);

  if (rv != 0)
    return NULL;

  return g_list_prepend (NULL, g_strdup (name));
}

#else

static GList *
lookup_getaddrinfo (const gchar *name)
{
  static GStaticMutex  mutex = G_STATIC_MUTEX_INIT;
  GList               *addr_list = NULL;
  struct hostent      *he;
  gint                 i;

  g_static_mutex_lock (&mutex);

  DEBUG (g_print ("Looking up '%s'\n", name));

  he = gethostbyname (name);

  DEBUG (g_print ("Looked up '%s'\n", name));

  if (!he || !flux_addr_family_is_ipv4 (he->h_addrtype))
  {
    g_static_mutex_unlock (&mutex);
    DEBUG (g_print ("No IPv4 hostent!\n"));
    return NULL;
  }

  for (i = 0; he->h_addr_list [i]; i++)
  {
    FluxIPAddr         *ip_addr;
    struct sockaddr_in  sa;

    DEBUG (g_print ("Checking hostent [%d].\n", i));

    ((struct sockaddr *) &sa)->sa_family = he->h_addrtype;
    memcpy (flux_sockaddr_get_addr ((struct sockaddr *) &sa), he->h_addr_list [i], he->h_length);

    ip_addr = flux_ip_addr_new ();
    if (!flux_ip_addr_set_sockaddr (ip_addr, (struct sockaddr *) &sa))
    {
      DEBUG (g_print ("Could not set sockaddr!\n"));
      flux_object_unref (FLUX_OBJECT (ip_addr));
      continue;
    }

    DEBUG (g_print ("Got address: '%s'\n", flux_ip_addr_get_string (ip_addr)));

    addr_list = g_list_prepend (addr_list, ip_addr);
  }

  g_static_mutex_unlock (&mutex);

  return addr_list;
}

static GList *
lookup_getnameinfo (FluxIPAddr *ip_addr)
{
  static GStaticMutex  mutex     = G_STATIC_MUTEX_INIT;
  GList               *name_list = NULL;
  struct hostent      *he;
  struct sockaddr_in  *sa;

  if (flux_ip_addr_get_type (ip_addr) != FLUX_IP_ADDR_IPV4)
  {
    DEBUG (g_print ("Supporting IPv4 only.\n"));
    return NULL;
  }

  sa = (struct sockaddr_in *) flux_ip_addr_get_sockaddr (ip_addr, 0);
  if (!sa)
  {
    DEBUG (g_print ("No SA.\n"));
    return NULL;
  }

  g_static_mutex_lock (&mutex);

  he = gethostbyaddr (flux_sockaddr_get_addr ((struct sockaddr *) sa),
                      flux_sockaddr_get_addr_len ((struct sockaddr *) sa),
                      flux_sockaddr_get_family ((struct sockaddr *) sa));

  g_free (sa);

  if (he && he->h_name)
    name_list = g_list_prepend (name_list, g_strdup (he->h_name));
  else
    DEBUG (g_print ("No he.\n"));

  g_static_mutex_unlock (&mutex);

  return name_list;
}

#endif

static void
destroy_lookup (Lookup *lookup)
{
  /* Remove from lookup table */

  g_hash_table_remove (lookup->resolver->lookup_table, GUINT_TO_POINTER (lookup->id));

  /* Free memory */

  if (lookup->is_reverse)
  {
    flux_object_unref (FLUX_OBJECT (lookup->arg));
    free_deep_list (lookup->results);
  }
  else
  {
    g_free (lookup->arg);
    free_object_list (lookup->results);
  }

  g_free (lookup);
}

static gboolean
dispatch_lookup (Lookup *lookup)
{
  FluxIPResolver *resolver = lookup->resolver;

  g_mutex_lock (resolver->mutex);

  if (lookup->is_wanted)
  {
    gboolean  is_reverse;
    GList    *arg_list = NULL;
    GList    *result_list;

    is_reverse = lookup->is_reverse;
    
    arg_list    = g_list_append (arg_list, lookup->arg);
    result_list = lookup->results;

    g_mutex_unlock (resolver->mutex);

    if (is_reverse)
      lookup->func (arg_list, result_list, lookup->data);
    else
      lookup->func (result_list, arg_list, lookup->data);

    g_mutex_lock (resolver->mutex);

    g_list_free (arg_list);
  }

  destroy_lookup (lookup);
  g_mutex_unlock (resolver->mutex);

  return FALSE;
}

static void
do_lookup (Lookup *lookup, FluxIPResolver *resolver)
{
  g_mutex_lock (lookup->resolver->mutex);

  if (lookup->is_wanted)
  {
    gboolean is_reverse;
    gpointer arg;

    lookup->is_running = TRUE;

    is_reverse = lookup->is_reverse;
    arg        = lookup->arg;

    g_mutex_unlock (lookup->resolver->mutex);

    if (is_reverse)
    {
      /* Address in, names out */
      lookup->results = lookup_getnameinfo (lookup->arg);
    }
    else
    {
      /* Name in, addresses out */
      lookup->results = lookup_getaddrinfo (lookup->arg);
    }

    g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) dispatch_lookup, lookup, NULL);
  }
  else
  {
    FluxIPResolver *resolver = lookup->resolver;

    destroy_lookup (lookup);
    g_mutex_unlock (resolver->mutex);
  }
}

static guint
create_lookup (FluxIPResolver *resolver, gboolean is_reverse, gpointer arg,
               FluxIPLookupFunc *func, gpointer data)
{
  Lookup *lookup;

  lookup = g_new0 (Lookup, 1);

  lookup->func       = func;
  lookup->data       = data;
  lookup->resolver   = resolver;
  lookup->is_reverse = is_reverse;
  lookup->is_wanted  = TRUE;
  lookup->arg        = arg;

  /* Assign an ID that is not 0 and not in use */

  g_mutex_lock (resolver->mutex);

  for ( ; ; resolver->next_lookup_id++)
  {
    resolver->next_lookup_id %= (1 << ID_BITS);
    if (!resolver->next_lookup_id)
      continue;

    if (!g_hash_table_lookup (resolver->lookup_table, GUINT_TO_POINTER (resolver->next_lookup_id)))
      break;
  }

  lookup->id = resolver->next_lookup_id;

  /* Insert in lookup table */

  g_hash_table_insert (resolver->lookup_table, GUINT_TO_POINTER (lookup->id), lookup);

  g_mutex_unlock (resolver->mutex);

  /* Queue request */

  g_thread_pool_push (resolver->thread_pool, lookup, NULL);

  return lookup->id;
}

gboolean
flux_ip_resolver_init (FluxIPResolver *resolver)
{
  flux_object_init              (FLUX_OBJECT (resolver));
  flux_object_set_type          (FLUX_OBJECT (resolver), FLUX_IP_RESOLVER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (resolver), flux_ip_resolver_finalize);

  flux_init_sockets ();

  if (!g_thread_supported ())
    g_thread_init (NULL);

  resolver->thread_pool = g_thread_pool_new ((GFunc) do_lookup, resolver, 10, FALSE, NULL);
  if (!resolver->thread_pool)
  {
    g_warning ("Could not create thread pool for FluxIPResolver.");
    return FALSE;
  }

  g_thread_pool_set_max_unused_threads (1);

  resolver->mutex = g_mutex_new ();
  resolver->lookup_table = g_hash_table_new (g_direct_hash, g_direct_equal);

  return TRUE;
}

void
flux_ip_resolver_finalize (FluxObject *object)
{
  FluxIPResolver *resolver = FLUX_IP_RESOLVER (object);

}

FluxIPResolver *
flux_ip_resolver_new (void)
{
  FluxIPResolver *resolver;

  resolver = g_new0 (FluxIPResolver, 1);
  if (!flux_ip_resolver_init (resolver))
  {
    g_free (resolver);
    return NULL;
  }

  return resolver;
}

FluxIPResolver *
flux_ip_resolver_get_default (void)
{
  if (!default_resolver)
    default_resolver = flux_ip_resolver_new ();

  return default_resolver;
}

guint
flux_ip_resolver_lookup (FluxIPResolver *resolver, const gchar *name,
                         FluxIPLookupFunc *func, gpointer data)
{
  g_return_val_if_fail (name != NULL, FALSE);
  g_return_val_if_fail (func != NULL, FALSE);

  return create_lookup (resolver, FALSE, g_strdup (name), func, data);
}

guint
flux_ip_resolver_lookup_reverse (FluxIPResolver *resolver, FluxIPAddr *addr,
                                 FluxIPLookupFunc *func, gpointer data)
{
  g_return_val_if_fail (addr != NULL, FALSE);
  g_return_val_if_fail (func != NULL, FALSE);

  flux_object_ref (FLUX_OBJECT (addr));

  return create_lookup (resolver, TRUE, addr, func, data);
}

void
flux_ip_resolver_cancel_lookup (FluxIPResolver *resolver, guint lookup_id)
{
  Lookup *lookup;

  g_mutex_lock (resolver->mutex);

  lookup = g_hash_table_lookup (resolver->lookup_table, GUINT_TO_POINTER (lookup_id));
  if (lookup)
    lookup->is_wanted = FALSE;

  g_mutex_unlock (resolver->mutex);
}

