/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-marshaller.c - Platform-independent en/decoder for C structs.
 *                     Can also generate/parse human-readable data.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-util.h"
#include "flux-marshaller.h"
#include "flux-tree.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define D(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_MARSHALLER_TYPE)

/* Byte-ordering aliases */

/* We use little-endian since X86 is the most common platform and this saves some
 * processing on those */

#define PACK_S16(val)   GINT16_TO_LE    (val)
#define UNPACK_S16(val) GINT16_FROM_LE  (val)
#define PACK_U16(val)   GUINT16_TO_LE   (val)
#define UNPACK_U16(val) GUINT16_FROM_LE (val)
#define PACK_S32(val)   GINT32_TO_LE    (val)
#define UNPACK_S32(val) GINT32_FROM_LE  (val)
#define PACK_U32(val)   GUINT32_TO_LE   (val)
#define UNPACK_U32(val) GUINT32_FROM_LE (val)
#define PACK_S64(val)   GINT64_TO_LE    (val)
#define UNPACK_S64(val) GINT64_FROM_LE  (val)
#define PACK_U64(val)   GUINT64_TO_LE   (val)
#define UNPACK_U64(val) GUINT64_FROM_LE (val)

/* Read values from a struct. Used by encode_pack (). */

#define GET_STRUCT_INT(entry, struct_ptr)                                                                  \
  ((entry)->size == sizeof (gint8)  ? G_STRUCT_MEMBER (gint8,  struct_ptr, (entry)->offset) :              \
   (entry)->size == sizeof (gint16) ? G_STRUCT_MEMBER (gint16, struct_ptr, (entry)->offset) :              \
   (entry)->size == sizeof (gint32) ? G_STRUCT_MEMBER (gint32, struct_ptr, (entry)->offset) :              \
   G_STRUCT_MEMBER (gint64, struct_ptr, (entry)->offset))

#define GET_STRUCT_UINT(entry, struct_ptr)                                                                 \
  ((entry)->size == sizeof (guint8)  ? G_STRUCT_MEMBER (guint8,  struct_ptr, (entry)->offset) :            \
   (entry)->size == sizeof (guint16) ? G_STRUCT_MEMBER (guint16, struct_ptr, (entry)->offset) :            \
   (entry)->size == sizeof (guint32) ? G_STRUCT_MEMBER (guint32, struct_ptr, (entry)->offset) :            \
   G_STRUCT_MEMBER (guint64, struct_ptr, (entry)->offset))

#define GET_STRUCT_FLOAT(entry, struct_ptr)                                                                \
  ((entry)->size == sizeof (gfloat) ? G_STRUCT_MEMBER (gfloat, struct_ptr, (entry)->offset) :              \
   G_STRUCT_MEMBER (gdouble, struct_ptr, (entry)->offset))

#define GET_STRUCT_BOOLEAN(entry, struct_ptr)                                                              \
  (GET_STRUCT_INT (entry, struct_ptr) ? TRUE : FALSE)

#define GET_STRUCT_META_UINT(entry, struct_ptr)                                                            \
  ((entry)->meta_size == sizeof (guint8)  ? G_STRUCT_MEMBER (guint8,  struct_ptr, (entry)->meta_offset) :  \
   (entry)->meta_size == sizeof (guint16) ? G_STRUCT_MEMBER (guint16, struct_ptr, (entry)->meta_offset) :  \
   (entry)->meta_size == sizeof (guint32) ? G_STRUCT_MEMBER (guint32, struct_ptr, (entry)->meta_offset) :  \
   G_STRUCT_MEMBER (guint64, struct_ptr, (entry)->meta_offset))

/* Write values to a struct. Used by decode_pack (). */

#define SET_STRUCT_INT(entry, struct_ptr, value)                                                           \
  G_STMT_START{                                                                                            \
    if      ((entry)->size == sizeof (gint8))                                                              \
      *((gint8 *)  G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                               \
    else if ((entry)->size == sizeof (gint16))                                                             \
      *((gint16 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                               \
    else if ((entry)->size == sizeof (gint32))                                                             \
      *((gint32 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                               \
    else                                                                                                   \
      *((gint64 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                               \
  }G_STMT_END

#define SET_STRUCT_UINT(entry, struct_ptr, value)                                                          \
  G_STMT_START{                                                                                            \
    if      ((entry)->size == sizeof (guint8))                                                             \
      *((guint8 *)  G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                              \
    else if ((entry)->size == sizeof (guint16))                                                            \
      *((guint16 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                              \
    else if ((entry)->size == sizeof (guint32))                                                            \
      *((guint32 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                              \
    else                                                                                                   \
      *((guint64 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                              \
  }G_STMT_END

#define SET_STRUCT_FLOAT(entry, struct_ptr, value)                                                         \
  G_STMT_START{                                                                                            \
    if ((entry)->size == sizeof (gfloat))                                                                  \
      *((gfloat *)  G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                              \
    else                                                                                                   \
      *((gdouble *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value;                              \
  }G_STMT_END

#define SET_STRUCT_BOOLEAN(entry, struct_ptr, value)                                                       \
  G_STMT_START{                                                                                            \
    if      ((entry)->size == sizeof (gint8))                                                              \
      *((gint8 *)  G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value ? TRUE : FALSE;                \
    else if ((entry)->size == sizeof (gint16))                                                             \
      *((gint16 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value ? TRUE : FALSE;                \
    else if ((entry)->size == sizeof (gint32))                                                             \
      *((gint32 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value ? TRUE : FALSE;                \
    else                                                                                                   \
      *((gint64 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->offset)) = value ? TRUE : FALSE;                \
  }G_STMT_END

#define SET_STRUCT_META_UINT(entry, struct_ptr, value)                                                     \
  G_STMT_START{                                                                                            \
    if      ((entry)->meta_size == sizeof (guint8))                                                        \
      *((guint8 *)  G_STRUCT_MEMBER_P (struct_ptr, (entry)->meta_offset)) = value;                         \
    else if ((entry)->meta_size == sizeof (guint16))                                                       \
      *((guint16 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->meta_offset)) = value;                         \
    else if ((entry)->meta_size == sizeof (guint32))                                                       \
      *((guint32 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->meta_offset)) = value;                         \
    else                                                                                                   \
      *((guint64 *) G_STRUCT_MEMBER_P (struct_ptr, (entry)->meta_offset)) = value;                         \
  }G_STMT_END

/* Used by decode_pack() for bounds checking of input */

#define ADVANCE_INPUT_BY_TYPE(ptr, decoded_len, type)           \
  (ptr.type##_p)++;                                             \
  decoded_len += sizeof (type);

#define ADVANCE_INPUT_BY_LENGTH(ptr, decoded_len, length)       \
  (ptr.guint8_p) += length;                                     \
  (decoded_len) += (length);

#define CHECK_BOUNDS_FOR_TYPE(pack_len, decoded_len, type)      \
  G_STMT_START{                                                 \
    if (sizeof (type) > (pack_len) - (decoded_len))             \
      return NULL;                                              \
  }G_STMT_END

#define CHECK_BOUNDS_FOR_LENGTH(pack_len, decoded_len, length)  \
  G_STMT_START{                                                 \
    if ((length) > (pack_len) - (decoded_len))                  \
      return NULL;                                              \
  }G_STMT_END

#define GET_REMAINING_LENGTH(pack_len, decoded_len)             \
  ((pack_len) - (decoded_len))

/* Internal metadata for a marshalled struct */

typedef struct
{
  gint                       size;
  guint                      flags;
  FluxMarshalledStructEntry *entries;
  gpointer                   user_data;
}
StructMeta;

/* Forward declarations */

static FluxMarshalledStructEntry *encode_tree          (FluxMarshaller *marshaller, FluxTree *parent_tree,
                                                        FluxMarshalledStructEntry *entry,
                                                        gconstpointer struct_ptr);

static FluxMarshalledStructEntry *decode_tree          (FluxMarshaller *marshaller, FluxTree *parent_tree,
                                                        FluxMarshalledStructEntry *entry,
                                                        gpointer struct_ptr, gboolean must_fill);

static guint                      determine_pack_len   (FluxMarshaller *marshaller,
                                                        FluxMarshalledStructEntry *entry,
                                                        gconstpointer struct_ptr);

static gpointer                   encode_pack          (FluxMarshaller *marshaller,
                                                        gpointer pack_ptr,
                                                        FluxMarshalledStructEntry *entry,
                                                        gconstpointer struct_ptr);

static gconstpointer              decode_pack          (FluxMarshaller *marshaller,
                                                        gconstpointer pack_ptr, guint pack_len,
                                                        FluxMarshalledStructEntry *entry,
                                                        gpointer struct_ptr);

static gboolean                   free_struct_contents (FluxMarshaller *marshaller,
                                                        FluxMarshalledStructEntry *entry, gpointer struct_ptr);

FluxMarshaller *
flux_marshaller_new (void)
{
  FluxMarshaller *marshaller;
  
  marshaller = g_new0 (FluxMarshaller, 1);
  flux_marshaller_init (marshaller);

  return marshaller;
}

void
flux_marshaller_init (FluxMarshaller *marshaller)
{
  flux_object_init              (FLUX_OBJECT (marshaller));
  flux_object_set_type          (FLUX_OBJECT (marshaller), FLUX_MARSHALLER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (marshaller), flux_marshaller_finalize);

  marshaller->types = g_hash_table_new (g_str_hash, g_str_equal);
}

void
flux_marshaller_finalize (FluxObject *object)
{
  flux_marshaller_unregister_all ((FluxMarshaller *) object);
  g_hash_table_destroy (((FluxMarshaller *) object)->types);
}

/* --- */

gboolean
flux_marshaller_register (FluxMarshaller *marshaller, gchar *struct_name, gint struct_size,
                          guint flags, FluxMarshalledStructEntry *struct_entries)
{
  StructMeta *meta;

  assert_types (marshaller);

  meta            = g_new (StructMeta, 1);
  meta->size      = struct_size;
  meta->flags     = flags;
  meta->entries   = struct_entries;

  if (g_hash_table_lookup (marshaller->types, struct_name))
    return FALSE;

  g_hash_table_insert (marshaller->types, struct_name, meta);
  return TRUE;
}

gint
flux_marshaller_register_many (FluxMarshaller *marshaller, FluxMarshalledStruct *structs)
{
  gint i;
  gint registered_num = 0;

  assert_types (marshaller);

  for (i = 0; structs [i].name; i++)
  {
    if (flux_marshaller_register (marshaller, structs [i].name, structs [i].size,
                                  structs [i].flags, structs [i].entries))
    {
      registered_num++;
    }
  }

  return registered_num;
}

static void
unregister_meta (StructMeta *meta)
{
  g_free (meta);
}

gboolean
flux_marshaller_unregister (FluxMarshaller *marshaller, const gchar *struct_name)
{
  StructMeta *meta;

  assert_types (marshaller);

  meta = g_hash_table_lookup (marshaller->types, struct_name);
  if (!meta)
    return FALSE;

  unregister_meta (meta);
  g_hash_table_remove (marshaller->types, struct_name);
  return TRUE;
}

gint
flux_marshaller_unregister_many (FluxMarshaller *marshaller, FluxMarshalledStruct *structs)
{
  gint unregistered_num = 0;
  gint i;

  assert_types (marshaller);

  for (i = 0; structs [i].name; i++)
  {
    if (flux_marshaller_unregister (marshaller, structs [i].name))
    {
      unregistered_num++;
    }
  }

  return unregistered_num;
}

static gboolean
remove_always (gpointer key, gpointer value, gpointer user_data)
{
  unregister_meta (value);
  (*((gint *) user_data))++;
  return TRUE;
}

gint
flux_marshaller_unregister_all (FluxMarshaller *marshaller)
{
  gint unregistered_num = 0;

  assert_types (marshaller);
  g_hash_table_foreach_remove (marshaller->types, remove_always, &unregistered_num);
  return unregistered_num;
}

static FluxTree *
encode_tree_boolean (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  gchar    *true_str = "true", *false_str = "false";
  gchar    *str;
  gpointer  member;

  str = false_str;
  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);

  switch (entry->size)
  {
    case sizeof (guint8):
      if (*((guint8 *) member)) str = true_str;
      break;
    case sizeof (guint16):
      if (*((guint16 *) member)) str = true_str;
      break;
    case sizeof (guint32):
      if (*((guint32 *) member)) str = true_str;
      break;
    case sizeof (guint64):
      if (*((guint64 *) member)) str = true_str;
      break;
    default:
      g_error ("Unrecognized pre-marshalling boolean size %d.", entry->size);
  }

  tree = flux_tree_new_with_data (str, strlen (str));
  return tree;
}

static FluxTree *
encode_tree_int (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  gchar     str [128];

  switch (entry->size)
  {
    case sizeof (gint8):
      g_snprintf (str, 128, "%d",
                  *((gint8 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (gint16):
      g_snprintf (str, 128, "%" G_GINT16_FORMAT,
                  *((gint16 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (gint32):
      g_snprintf (str, 128, "%" G_GINT32_FORMAT,
                  *((gint32 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (gint64):
      g_snprintf (str, 128, "%" G_GINT64_FORMAT,
                  *((gint64 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    default:
      g_error ("Unrecognized pre-marshalling integer size %d.", entry->size);
      str [0] = '\0';
  }

  tree = flux_tree_new_with_data (str, strlen (str));
  return tree;
}

static FluxTree *
encode_tree_uint (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  gchar     str [128];

  switch (entry->size)
  {
    case sizeof (guint8):
      g_snprintf (str, 128, "%u",
                  *((guint8 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (guint16):
      g_snprintf (str, 128, "%" G_GUINT16_FORMAT,
                  *((guint16 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (guint32):
      g_snprintf (str, 128, "%" G_GUINT32_FORMAT,
                  *((guint32 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (guint64):
      g_snprintf (str, 128, "%" G_GUINT64_FORMAT,
                  *((guint64 *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    default:
      g_error ("Unrecognized pre-marshalling integer size %d.", entry->size);
      str [0] = '\0';
  }

  tree = flux_tree_new_with_data (str, strlen (str));
  return tree;
}

static FluxTree *
encode_tree_float (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  gchar     str [128];

  switch (entry->size)
  {
    case sizeof (gfloat):
      g_snprintf (str, 128, "%f", *((gfloat *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    case sizeof (gdouble):
      g_snprintf (str, 128, "%f", *((gdouble *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset)));
      break;

    default:
      g_error ("Unrecognized pre-marshalling float size %d.", entry->size);
      str [0] = '\0';
  }

  tree = flux_tree_new_with_data (str, strlen (str));
  return tree;
}

static FluxTree *
encode_tree_raw_dynamic (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  guchar   *raw;
  guint     len;

  raw = G_STRUCT_MEMBER (guchar *, struct_ptr, entry->offset);
  len = GET_STRUCT_META_UINT (entry, struct_ptr);

  if (!raw || !len)
    return NULL;

  return flux_tree_new_with_data (raw, len);
}

static FluxTree *
encode_tree_raw_fixed (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  guchar   *raw;
  guint     len;

  raw = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  len = entry->size;

  tree = flux_tree_new_with_data (raw, len);
  return tree;
}

static FluxTree *
encode_tree_string_dynamic (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  gchar    *str;

  str = G_STRUCT_MEMBER (gchar *, struct_ptr, entry->offset);
  if (!str)
    return NULL;

  tree = flux_tree_new_with_data (str, strlen (str));
  return tree;
}

static FluxTree *
encode_tree_string_fixed (FluxMarshalledStructEntry *entry, gconstpointer struct_ptr)
{
  FluxTree *tree;
  gchar    *str, *last;
  gint      len;

  str = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  last = memchr (str, '\0', entry->size);

  if (last)
    len = last - str;
  else
    len = entry->size;

  tree = flux_tree_new_with_data (str, len);
  return tree;
}

static FluxTree *
encode_tree_struct_dynamic (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry,
                            gconstpointer struct_ptr)
{
  FluxTree *tree;
  gpointer  sub_struct_ptr;
  gchar    *sub_struct_name;

  sub_struct_ptr = G_STRUCT_MEMBER (gpointer, struct_ptr, entry->offset);
  sub_struct_name = entry->default_value;

  if (sub_struct_ptr)
  {
    tree = flux_marshaller_encode_tree (marshaller, sub_struct_name, sub_struct_ptr);
  }
  else
  {
    tree = NULL;
  }

  return tree;
}

static FluxTree *
encode_tree_struct_fixed (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry,
                          gconstpointer struct_ptr)
{
  FluxTree *tree;
  gpointer  sub_struct_ptr;
  gchar    *sub_struct_name;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  sub_struct_name = entry->default_value;

  if (sub_struct_ptr)
  {
    tree = flux_marshaller_encode_tree (marshaller, sub_struct_name, sub_struct_ptr);
  }
  else
  {
    tree = NULL;
  }

  return tree;
}

static FluxMarshalledStructEntry *
encode_tree (FluxMarshaller *marshaller, FluxTree *parent_tree, FluxMarshalledStructEntry *entry,
             gconstpointer struct_ptr)
{
  for (;;)
  {
    FluxTree *key_tree, *value_tree;

    switch (entry->type)
    {
      case FLUX_MARSHAL_BOOLEAN_TYPE:
        key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                       strlen (entry->name));
        flux_tree_add_as_last_child (key_tree, encode_tree_boolean (entry, struct_ptr));
        entry++;
        break;

      case FLUX_MARSHAL_INT8_TYPE:
      case FLUX_MARSHAL_INT16_TYPE:
      case FLUX_MARSHAL_INT32_TYPE:
      case FLUX_MARSHAL_INT64_TYPE:
        key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                       strlen (entry->name));
        flux_tree_add_as_last_child (key_tree, encode_tree_int (entry, struct_ptr));
        entry++;
        break;

      case FLUX_MARSHAL_UINT8_TYPE:
      case FLUX_MARSHAL_UINT16_TYPE:
      case FLUX_MARSHAL_UINT32_TYPE:
      case FLUX_MARSHAL_UINT64_TYPE:
        key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                       strlen (entry->name));
        flux_tree_add_as_last_child (key_tree, encode_tree_uint (entry, struct_ptr));
        entry++;
        break;

      case FLUX_MARSHAL_FLOAT32_TYPE:
      case FLUX_MARSHAL_FLOAT64_TYPE:
        key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                       strlen (entry->name));
        flux_tree_add_as_last_child (key_tree, encode_tree_float (entry, struct_ptr));
        entry++;
        break;

      case FLUX_MARSHAL_RAW_DYNAMIC_TYPE:
        value_tree = encode_tree_raw_dynamic (entry, struct_ptr);
        if (value_tree)
        {
          key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                         strlen (entry->name));
          flux_tree_add_as_last_child (key_tree, value_tree);
        }
        entry++;
        break;

      case FLUX_MARSHAL_RAW_FIXED_TYPE:
        value_tree = encode_tree_raw_fixed (entry, struct_ptr);
        if (value_tree)
        {
          key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                         strlen (entry->name));
          flux_tree_add_as_last_child (key_tree, value_tree);
        }
        entry++;
        break;

      case FLUX_MARSHAL_STRING_DYNAMIC_TYPE:
        value_tree = encode_tree_string_dynamic (entry, struct_ptr);
        if (value_tree)
        {
          key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                         strlen (entry->name));
          flux_tree_add_as_last_child (key_tree, value_tree);
        }
        entry++;
        break;

      case FLUX_MARSHAL_STRING_FIXED_TYPE:
        value_tree = encode_tree_string_fixed (entry, struct_ptr);
        if (value_tree)
        {
          key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name,
                                                         strlen (entry->name));
          flux_tree_add_as_last_child (key_tree, value_tree);
        }
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_DYNAMIC_TYPE:
        value_tree = encode_tree_struct_dynamic (marshaller, entry, struct_ptr);
        if (value_tree)
        {
          flux_tree_data_set_str (value_tree, entry->name);
          flux_tree_add_as_last_child (parent_tree, value_tree);
        }
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_FIXED_TYPE:
        value_tree = encode_tree_struct_fixed (marshaller, entry, struct_ptr);
        if (value_tree)
        {
          flux_tree_data_set_str (value_tree, entry->name);
          flux_tree_add_as_last_child (parent_tree, value_tree);
        }
        entry++;
        break;

      case FLUX_MARSHAL_BRANCH_BEGIN_TYPE:
        key_tree = flux_tree_new_with_parent_and_data (parent_tree, entry->name, strlen (entry->name));
        entry = encode_tree (marshaller, key_tree, entry + 1, struct_ptr);
        break;

      case FLUX_MARSHAL_BRANCH_END_TYPE:
        return entry + 1;

      case FLUX_MARSHAL_LAST_TYPE:
        return entry;

      default:
        g_error ("Error in marshalling metadata - invalid entry type.");
    }
  }
}

FluxTree *
flux_marshaller_encode_tree (FluxMarshaller *marshaller, const gchar *struct_name, gconstpointer struct_ptr)
{
  StructMeta *meta;
  FluxTree   *tree;

  assert_types (marshaller);
  g_return_val_if_fail (struct_name != NULL, NULL);
  g_return_val_if_fail (struct_ptr != NULL, NULL);

  meta = g_hash_table_lookup (marshaller->types, struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when encoding.", struct_name);
    return NULL;
  }

  tree = flux_tree_new_with_data (struct_name, strlen (struct_name));
  encode_tree (marshaller, tree, meta->entries, struct_ptr);
  return tree;
}

static void
decode_tree_boolean (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                     gboolean must_fill)
{
  gchar    *true_str = "true", *yes_str = "yes", *on_str = "on", *one_str = "1";
  gpointer  member;
  gboolean  value = FALSE;

  /* Determine the value */

  if (value_tree)
  {
    if (!flux_tree_strcasecmp (value_tree, true_str) || !flux_tree_strcasecmp (value_tree, yes_str) ||
        !flux_tree_strcasecmp (value_tree, on_str)   || !flux_tree_strcasecmp (value_tree, one_str))
      value = TRUE;
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = GPOINTER_TO_UINT (entry->default_value);
  }

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);

  /* Store it */

  switch (entry->size)
  {
    case sizeof (guint8):
      *((guint8 *) member) = value;
      break;
    case sizeof (guint16):
      *((guint16 *) member) = value;
      break;
    case sizeof (guint32):
      *((guint32 *) member) = value;
      break;
    case sizeof (guint64):
      *((guint64 *) member) = value;
      break;
    default:
      g_error ("Unrecognized post-marshalling boolean size %d.", entry->size);
  }
}

static void
decode_tree_int (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                 gboolean must_fill)
{
  gpointer  member;
  gint64    value;

  /* Determine the value */

  if (value_tree)
  {
    gchar *str;

    str = flux_tree_data_get_str (value_tree);
    sscanf (str, "%" G_GINT64_FORMAT, &value);
    g_free (str);
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = GPOINTER_TO_UINT (entry->default_value);
  }

  /* Store it */

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);

  switch (entry->size)
  {
    case sizeof (gint8):
      *((gint8 *) member) = value;
      break;

    case sizeof (gint16):
      *((gint16 *) member) = value;
      break;

    case sizeof (gint32):
      *((gint32 *) member) = value;
      break;

    case sizeof (gint64):
      *((gint64 *) member) = value;
      break;

    default:
      g_error ("Unrecognized post-marshalling integer size %d.", entry->size);
  }
}

static void
decode_tree_uint (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                  gboolean must_fill)
{
  gpointer member;
  guint64  value;

  /* Determine the value */

  if (value_tree)
  {
    gchar *str;

    str = flux_tree_data_get_str (value_tree);
    sscanf (str, "%" G_GUINT64_FORMAT, &value);
    g_free (str);
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = GPOINTER_TO_UINT (entry->default_value);
  }

  /* Store it */

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);

  switch (entry->size)
  {
    case sizeof (guint8):
      *((guint8 *) member) = value;
      break;

    case sizeof (guint16):
      *((guint16 *) member) = value;
      break;

    case sizeof (guint32):
      *((guint32 *) member) = value;
      break;

    case sizeof (guint64):
      *((guint64 *) member) = value;
      break;

    default:
      g_error ("Unrecognized post-marshalling integer size %d.", entry->size);
  }
}

static void
decode_tree_float (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                   gboolean must_fill)
{
  gpointer  member;
  gdouble   value;

  /* Determine the value */

  if (value_tree)
  {
    gchar *str;

    str = flux_tree_data_get_str (value_tree);
    value = atof (str);
    g_free (str);
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = FLUX_POINTER_TO_FLOAT (entry->default_value);
  }

  /* Store it */

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);

  switch (entry->size)
  {
    case sizeof (gfloat):
      *((gfloat *) member) = value;
      break;

    case sizeof (gdouble):
      *((gdouble *) member) = value;
      break;

    default:
      g_error ("Unrecognized post-marshalling float size %d.", entry->size);
  }
}

static void
decode_tree_raw_dynamic (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                         gboolean must_fill)
{
  gpointer  member;
  gchar    *value;
  guint     meta_value;

  /* Determine the value */

  if (value_tree)
  {
    gint len;

    len = flux_tree_data_get_size (value_tree);
    if (len)
    {
      meta_value = len;
      value      = g_malloc (len);
      flux_tree_data_get_bytes (value_tree, value, 0, len);
    }
    else
    {
      meta_value = 0;
      value      = NULL;
    }
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    meta_value = 0;
    value      = NULL;
  }

  /* Store it */

  SET_STRUCT_META_UINT (entry, struct_ptr, meta_value);
  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  *((gchar **) member) = value;
}

static void
decode_tree_raw_fixed (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                       gboolean must_fill)
{
  gchar    *member;
  gpointer  value     = NULL;
  guint     value_len = 0;

  /* Determine the value */

  if (value_tree)
  {
    value_len = flux_tree_data_get_size (value_tree);
    value = g_malloc (value_len);
    flux_tree_data_get_bytes (value_tree, value, 0, value_len);
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = (gchar *) entry->default_value;
    if (value)
    {
      value = g_strdup (value);
      value_len = strlen (value);
    }
  }

  /* Store it */

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  if (value)
  {
    memcpy (member, value, MIN (entry->size, value_len));
    g_free (value);
  }
  else
  {
    memset (member, 0, entry->size);
  }
}

static void
decode_tree_string_dynamic (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                            gboolean must_fill)
{
  gpointer  member;
  gchar    *value;

  /* Determine the value */

  if (value_tree)
  {
    value = flux_tree_data_get_str (value_tree);
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = (gchar *) entry->default_value;
    if (value)
      value = g_strdup (value);
  }

  /* Store it */

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  *((gchar **) member) = value;
}

static void
decode_tree_string_fixed (FluxTree *value_tree, FluxMarshalledStructEntry *entry, gpointer struct_ptr,
                          gboolean must_fill)
{
  gchar *member;
  gchar *value;

  /* Determine the value */

  if (value_tree)
  {
    value = flux_tree_data_get_str (value_tree);
  }
  else
  {
    if (!entry->substitute_default && !must_fill)
      return;

    value = (gchar *) entry->default_value;
    if (value)
      value = g_strdup (value);
  }

  /* Store it */

  member = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  if (value)
  {
    strncpy (member, value, entry->size - 1);
    *(member + entry->size - 1) = '\0';
    g_free (value);
  }
  else
  {
    *member = '\0';
  }
}

static void
decode_tree_struct_fixed (FluxMarshaller *marshaller, FluxTree *value_tree,
                          FluxMarshalledStructEntry *entry, gpointer struct_ptr, gboolean must_fill)
{
  StructMeta *meta;
  gchar      *struct_name;
  gpointer   sub_struct_ptr;

  if (!value_tree)
  {
    if (!entry->substitute_default && !must_fill)
      return;
  }

  struct_name = (gchar *) entry->default_value;
  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);

  meta = g_hash_table_lookup (marshaller->types, struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when decoding.", struct_name);
    return;
  }

  /* Since the struct is allocated by us, we must fill in all values */
  decode_tree (marshaller, value_tree, meta->entries, sub_struct_ptr, TRUE);
}

static void
decode_tree_struct_dynamic (FluxMarshaller *marshaller, FluxTree *value_tree,
                            FluxMarshalledStructEntry *entry, gpointer struct_ptr, gboolean must_fill)
{
  gboolean  ret;
  gchar    *struct_name;
  gchar    *old_name;
  gpointer *sub_struct_ptr;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  old_name = flux_tree_data_get_str (value_tree);
  flux_tree_data_set_str (value_tree, (gchar *) entry->default_value);

  *sub_struct_ptr = NULL;
  ret = flux_marshaller_decode_tree (marshaller, value_tree, &struct_name, sub_struct_ptr);
  g_free (struct_name);

  flux_tree_data_set_str (value_tree, old_name);
  g_free (old_name);
}

static FluxMarshalledStructEntry *
decode_tree (FluxMarshaller *marshaller, FluxTree *parent_tree, FluxMarshalledStructEntry *entry,
             gpointer struct_ptr, gboolean must_fill)
{
  for (;;)
  {
    switch (entry->type)
    {
      FluxTree *key_tree;

      case FLUX_MARSHAL_BOOLEAN_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_boolean (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                             struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_INT8_TYPE:
      case FLUX_MARSHAL_INT16_TYPE:
      case FLUX_MARSHAL_INT32_TYPE:
      case FLUX_MARSHAL_INT64_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_int (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                         struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_UINT8_TYPE:
      case FLUX_MARSHAL_UINT16_TYPE:
      case FLUX_MARSHAL_UINT32_TYPE:
      case FLUX_MARSHAL_UINT64_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_uint (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                          struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_FLOAT32_TYPE:
      case FLUX_MARSHAL_FLOAT64_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_float (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                           struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_RAW_DYNAMIC_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_raw_dynamic (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                                 struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_RAW_FIXED_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_raw_fixed (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                               struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_STRING_DYNAMIC_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_string_dynamic (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                                    struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_STRING_FIXED_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_string_fixed (key_tree ? flux_tree_get_first_child (key_tree) : NULL, entry,
                                  struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_DYNAMIC_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_struct_dynamic (marshaller, key_tree, entry, struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_FIXED_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        decode_tree_struct_fixed (marshaller, key_tree, entry, struct_ptr, must_fill);
        entry++;
        break;

      case FLUX_MARSHAL_BRANCH_BEGIN_TYPE:
        key_tree = parent_tree ?
          flux_tree_find_first_child (parent_tree, entry->name, strlen (entry->name)) : NULL;

        entry = decode_tree (marshaller, key_tree, entry + 1, struct_ptr, must_fill);
        break;

      case FLUX_MARSHAL_BRANCH_END_TYPE:
        return entry + 1;

      case FLUX_MARSHAL_LAST_TYPE:
        return entry;

      default:
        g_error ("Error in marshalling metadata - invalid entry type.");
    }
  }
}

gboolean
flux_marshaller_decode_tree (FluxMarshaller *marshaller, FluxTree *tree, gchar **struct_name,
                             gpointer *struct_ptr)
{
  StructMeta *meta;
  gchar      *local_struct_name;

  assert_types (marshaller);
  g_return_val_if_fail (tree != NULL, FALSE);
  g_return_val_if_fail (struct_ptr != NULL, FALSE);

  local_struct_name = flux_tree_data_get_str (tree);
  if (!local_struct_name)
  {
    g_warning ("Tree has no name for decoding.");
    return FALSE;
  }

  if (struct_name)
    *struct_name = local_struct_name;

  meta = g_hash_table_lookup (marshaller->types, local_struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when decoding.", *struct_name);
    g_free (local_struct_name);

    if (struct_name)
      *struct_name = NULL;

    return FALSE;
  }

  /* The struct is allocated by us */
  *struct_ptr = g_malloc (meta->size);

  /* Since we're allocating, we must clear it so we can determine what to
   * free later. */
  memset (*struct_ptr, 0, meta->size);

  /* And fill in specified defaults where values can't be found */
  decode_tree (marshaller, tree, meta->entries, *struct_ptr, TRUE);

  if (!struct_name)
    g_free (local_struct_name);

  return TRUE;
}

static guint
determine_dynamic_struct_len (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry,
                              gconstpointer struct_ptr)
{
  gpointer sub_struct_ptr;

  sub_struct_ptr = G_STRUCT_MEMBER (gpointer, struct_ptr, entry->offset);

  if (sub_struct_ptr)
  {
    gchar      *sub_struct_name;
    StructMeta *meta;

    sub_struct_name = entry->default_value;
    meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
    if (!meta)
    {
      g_warning ("No marshalling info for struct \"%s\" when encoding.", sub_struct_name);
      return 0;
    }

    return determine_pack_len (marshaller, meta->entries, sub_struct_ptr);
  }

  return 0;
}

static guint
determine_fixed_struct_len (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry,
                            gconstpointer struct_ptr)
{
  gpointer    sub_struct_ptr;
  gchar      *sub_struct_name;
  StructMeta *meta;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  sub_struct_name = entry->default_value;

  meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when encoding.", sub_struct_name);
    return 0;
  }

  return determine_pack_len (marshaller, meta->entries, sub_struct_ptr);
}

static guint
determine_pack_len (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry,
                    gconstpointer struct_ptr)
{
  guint  pack_len = 0;
  gchar *str, *last;
  guint  n;

  for (;;)
  {
    switch (entry->type)
    {
      case FLUX_MARSHAL_BOOLEAN_TYPE:
      case FLUX_MARSHAL_INT8_TYPE:
      case FLUX_MARSHAL_UINT8_TYPE:
        pack_len += sizeof (guint8);
        entry++;
        break;

      case FLUX_MARSHAL_INT16_TYPE:
      case FLUX_MARSHAL_UINT16_TYPE:
        pack_len += sizeof (guint16);
        entry++;
        break;

      case FLUX_MARSHAL_INT32_TYPE:
      case FLUX_MARSHAL_UINT32_TYPE:
      case FLUX_MARSHAL_FLOAT32_TYPE:
        pack_len += sizeof (guint32);
        entry++;
        break;

      case FLUX_MARSHAL_INT64_TYPE:
      case FLUX_MARSHAL_UINT64_TYPE:
      case FLUX_MARSHAL_FLOAT64_TYPE:
        pack_len += sizeof (guint64);
        entry++;
        break;

      case FLUX_MARSHAL_RAW_DYNAMIC_TYPE:
        n = GET_STRUCT_META_UINT (entry, struct_ptr);
        /* Length length */
        pack_len += flux_compress_uint (n, NULL);
        /* Data length */
        pack_len += n;
        entry++;
        break;

      case FLUX_MARSHAL_RAW_FIXED_TYPE:
        pack_len += entry->size;
        entry++;
        break;

      case FLUX_MARSHAL_STRING_DYNAMIC_TYPE:
        str = G_STRUCT_MEMBER (gchar *, struct_ptr, entry->offset);
        n = str ? strlen (str) : 0;
        /* String length */
        pack_len += flux_compress_uint (n, NULL);
        /* String */
        pack_len += n;
        entry++;
        break;

      case FLUX_MARSHAL_STRING_FIXED_TYPE:
        /* String length */
        if (entry->size < 256)
          pack_len += sizeof (guint8);
        else if (entry->size < 65536)
          pack_len += sizeof (guint16);
        else
          pack_len += sizeof (guint32);
        /* String */
        str = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
        last = memchr (str, '\0', entry->size);

        if (last)
          pack_len += last - str;
        else
          pack_len += entry->size;

        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_DYNAMIC_TYPE:
        /* Struct is present; boolean */
        pack_len += sizeof (guint8);
        /* Struct itself, if present */
        pack_len += determine_dynamic_struct_len (marshaller, entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_FIXED_TYPE:
        pack_len += determine_fixed_struct_len (marshaller, entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_BRANCH_BEGIN_TYPE:
      case FLUX_MARSHAL_BRANCH_END_TYPE:
        entry++;
        break;

      case FLUX_MARSHAL_LAST_TYPE:
        return pack_len;

      default:
        g_error ("Error in marshalling metadata - invalid entry type.");
    }
  }
}

static gpointer
encode_dynamic_struct (FluxMarshaller *marshaller, gpointer dest,
                       FluxMarshalledStructEntry *entry,
                       gconstpointer struct_ptr)
{
  gconstpointer  sub_struct_ptr;
  guint8        *dest_guint8 = dest;

  sub_struct_ptr = G_STRUCT_MEMBER (gpointer, struct_ptr, entry->offset);

  if (sub_struct_ptr)
  {
    gchar      *sub_struct_name;
    StructMeta *meta;

    sub_struct_name = entry->default_value;
    meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
    if (!meta)
    {
      g_warning ("No marshalling info for struct \"%s\" when encoding.", sub_struct_name);
      *(dest_guint8++) = FALSE;
      return dest;
    }

    *(dest_guint8++) = TRUE;
    dest_guint8 = encode_pack (marshaller, dest_guint8, meta->entries, sub_struct_ptr);
  }
  else
    *(dest_guint8++) = FALSE;

  return dest_guint8;
}

static gpointer
encode_fixed_struct (FluxMarshaller *marshaller, gpointer dest,
                     FluxMarshalledStructEntry *entry,
                     gconstpointer struct_ptr)
{
  gconstpointer  sub_struct_ptr;
  gchar         *sub_struct_name;
  StructMeta    *meta;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  sub_struct_name = entry->default_value;

  meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when encoding.", sub_struct_name);
    return dest;
  }

  return encode_pack (marshaller, dest, meta->entries, sub_struct_ptr);
}

static gpointer
encode_pack (FluxMarshaller *marshaller, gpointer pack_ptr, FluxMarshalledStructEntry *entry,
             gconstpointer struct_ptr)
{
  guchar   *data;
  gchar    *str;
  gchar    *last;
  guint     len;
  union  /* Accumulator for temporary storage */
  {
    gchar   gchar_a;
    gint8   gint8_a;
    gint16  gint16_a;
    gint32  gint32_a;
    gint64  gint64_a;
    gfloat  gfloat_a;
    gdouble gdouble_a;

    guchar  guchar_a;
    guint8  guint8_a;
    guint16 guint16_a;
    guint32 guint32_a;
    guint64 guint64_a;
  }
  a;
  union
  {
    gpointer  p;

    guchar   *guchar_p;
    guint8   *guint8_p;
    guint16  *guint16_p;
    guint32  *guint32_p;
    guint64  *guint64_p;
    gfloat   *gfloat_p;
    gdouble  *gdouble_p;

    gchar    *gchar_p;
    gint8    *gint8_p;
    gint16   *gint16_p;
    gint32   *gint32_p;
    gint64   *gint64_p;
  }
  dest;

  dest.p = pack_ptr;

  for (;;)
  {
    switch (entry->type)
    {
      case FLUX_MARSHAL_BOOLEAN_TYPE:
        *(dest.gint8_p++) = GET_STRUCT_BOOLEAN (entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_INT8_TYPE:
        *(dest.gint8_p++) = GET_STRUCT_INT (entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_UINT8_TYPE:
        *(dest.guint8_p++) = GET_STRUCT_UINT (entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_INT16_TYPE:
        a.gint16_a = GET_STRUCT_INT (entry, struct_ptr);
        *(dest.gint16_p++) = PACK_S16 (a.gint16_a);
        entry++;
        break;

      case FLUX_MARSHAL_UINT16_TYPE:
        a.guint16_a = GET_STRUCT_UINT (entry, struct_ptr);
        *(dest.guint16_p++) = PACK_U16 (a.guint16_a);
        entry++;
        break;

      case FLUX_MARSHAL_INT32_TYPE:
        a.gint32_a = GET_STRUCT_INT (entry, struct_ptr);
        *(dest.gint32_p++) = PACK_S32 (a.gint32_a);
        entry++;
        break;

      case FLUX_MARSHAL_UINT32_TYPE:
        a.guint32_a = GET_STRUCT_UINT (entry, struct_ptr);
        *(dest.guint32_p++) = PACK_U32 (a.guint32_a);
        entry++;
        break;

      case FLUX_MARSHAL_FLOAT32_TYPE:
        *(dest.gfloat_p++) = GET_STRUCT_FLOAT (entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_INT64_TYPE:
        a.gint64_a = GET_STRUCT_INT (entry, struct_ptr);
        *(dest.gint64_p++) = PACK_S64 (a.gint64_a);
        entry++;
        break;

      case FLUX_MARSHAL_UINT64_TYPE:
        a.guint64_a = GET_STRUCT_UINT (entry, struct_ptr);
        *(dest.guint64_p++) = PACK_U64 (a.guint64_a);
        entry++;
        break;

      case FLUX_MARSHAL_FLOAT64_TYPE:
        *(dest.gdouble_p++) = GET_STRUCT_FLOAT (entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_RAW_DYNAMIC_TYPE:
        len  = GET_STRUCT_META_UINT (entry, struct_ptr);
        data = G_STRUCT_MEMBER (guchar *, struct_ptr, entry->offset);

        /* Length */
        if (data)
          dest.guint8_p += flux_compress_uint (len, dest.p);
        else
          dest.guint8_p += flux_compress_uint (0, dest.p);

        /* Data */
        if (data && len)
        {
          memcpy (dest.p, data, len);
          dest.guint8_p += len;
        }

        entry++;
        break;

      case FLUX_MARSHAL_RAW_FIXED_TYPE:
        memcpy (dest.p, G_STRUCT_MEMBER_P (struct_ptr, entry->offset), entry->size);
        dest.guint8_p += entry->size;
        entry++;
        break;

      case FLUX_MARSHAL_STRING_DYNAMIC_TYPE:
        str = G_STRUCT_MEMBER (gchar *, struct_ptr, entry->offset);
        if (str)
        {
          len = strlen (str);

          dest.guint8_p += flux_compress_uint (len, dest.p);

          if (len)
          {
            memcpy (dest.p, str, len);
            dest.guint8_p += len;
          }
        }
        else
        {
          dest.guint8_p += flux_compress_uint (0, dest.p);
        }
        entry++;
        break;

      case FLUX_MARSHAL_STRING_FIXED_TYPE:
        str = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
        last = memchr (str, '\0', entry->size);

        if (last)
          len = last - str;
        else
          len = entry->size;

        /* String length */
        if (entry->size < 256)
          *(dest.guint8_p++) = len;
        else if (entry->size < 65536)
          *(dest.guint16_p++) = PACK_U16 (len);
        else
          *(dest.guint32_p++) = PACK_U32 (len);

        /* String */
        if (len)
        {
          memcpy (dest.p, str, len);
          dest.guint8_p += len;
        }

        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_DYNAMIC_TYPE:
        dest.p = encode_dynamic_struct (marshaller, dest.p, entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_FIXED_TYPE:
        dest.p = encode_fixed_struct (marshaller, dest.p, entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_BRANCH_BEGIN_TYPE:
      case FLUX_MARSHAL_BRANCH_END_TYPE:
        entry++;
        break;

      case FLUX_MARSHAL_LAST_TYPE:
        return dest.p;

      default:
        g_error ("Error in marshalling metadata - invalid entry type.");
    }
  }
}

gboolean
flux_marshaller_encode_pack (FluxMarshaller *marshaller, const gchar *struct_name, gconstpointer struct_ptr,
                             gpointer *pack_ptr, guint *pack_len)
{
  StructMeta *meta;

  assert_types (marshaller);
  g_return_val_if_fail (struct_name != NULL, FALSE);
  g_return_val_if_fail (struct_ptr != NULL, FALSE);

  meta = g_hash_table_lookup (marshaller->types, struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when encoding.", struct_name);
    return FALSE;
  }

  *pack_len = determine_pack_len (marshaller, meta->entries, struct_ptr);
  if (!*pack_len)
    return FALSE;

  *pack_ptr = g_malloc (*pack_len);
  encode_pack (marshaller, *pack_ptr, meta->entries, struct_ptr);
  return TRUE;
}

static gconstpointer
decode_dynamic_struct (FluxMarshaller *marshaller, gconstpointer src, guint remaining_len,
                       FluxMarshalledStructEntry *entry,
                       gpointer struct_ptr)
{
  gpointer     *sub_struct_ptr;
  const guint8 *src_guint8 = src;

  if (remaining_len < sizeof (guint8))
    return NULL;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  remaining_len -= sizeof (guint8);

  if (*(src_guint8++))
  {
    gchar      *sub_struct_name;
    StructMeta *meta;

    sub_struct_name = entry->default_value;
    meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
    if (!meta)
    {
      g_warning ("No marshalling info for struct \"%s\" when decoding.", sub_struct_name);
      return src_guint8;
    }

    *sub_struct_ptr = g_malloc (meta->size);
    memset (*sub_struct_ptr, 0, meta->size);

    src_guint8 = decode_pack (marshaller, src_guint8, remaining_len, meta->entries, *sub_struct_ptr);
    if (src_guint8 == NULL)
    {
      g_free (*sub_struct_ptr);
      *sub_struct_ptr = NULL;
    }
  }
  else
    *sub_struct_ptr = NULL;

  return src_guint8;
}

static gconstpointer
decode_fixed_struct (FluxMarshaller *marshaller, gconstpointer src, guint remaining_len,
                     FluxMarshalledStructEntry *entry,
                     gpointer struct_ptr)
{
  gpointer    sub_struct_ptr;
  gchar      *sub_struct_name;
  StructMeta *meta;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  sub_struct_name = entry->default_value;

  meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when decoding.", sub_struct_name);
    return src;
  }

  return decode_pack (marshaller, src, remaining_len, meta->entries, sub_struct_ptr);
}

static gconstpointer
decode_pack (FluxMarshaller *marshaller, gconstpointer pack_ptr, guint pack_len,
             FluxMarshalledStructEntry *entry, gpointer struct_ptr)
{
  gint              decoded_len = 0;
  guint             len;
  gint              n;
  union
  {
    gconstpointer   p;

    const guchar   *guchar_p;
    const guint8   *guint8_p;
    const guint16  *guint16_p;
    const guint32  *guint32_p;
    const guint64  *guint64_p;
    const gfloat   *gfloat_p;
    const gdouble  *gdouble_p;

    const gchar    *gchar_p;
    const gint8    *gint8_p;
    const gint16   *gint16_p;
    const gint32   *gint32_p;
    const gint64   *gint64_p;
  }
  src;

  src.p = pack_ptr;

  for (;;)
  {
    switch (entry->type)
    {
      case FLUX_MARSHAL_BOOLEAN_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gint8);
        SET_STRUCT_BOOLEAN (entry, struct_ptr, *(src.gint8_p));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gint8);
        entry++;
        break;

      case FLUX_MARSHAL_INT8_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gint8);
        SET_STRUCT_INT (entry, struct_ptr, *(src.gint8_p));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gint8);
        entry++;
        break;

      case FLUX_MARSHAL_UINT8_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint8);
        SET_STRUCT_UINT (entry, struct_ptr, *(src.guint8_p));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint8);
        entry++;
        break;

      case FLUX_MARSHAL_INT16_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gint16);
        SET_STRUCT_INT (entry, struct_ptr, UNPACK_S16 (*(src.gint16_p)));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gint16);
        entry++;
        break;

      case FLUX_MARSHAL_UINT16_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint16);
        SET_STRUCT_UINT (entry, struct_ptr, UNPACK_U16 (*(src.guint16_p)));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint16);
        entry++;
        break;

      case FLUX_MARSHAL_INT32_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gint32);
        SET_STRUCT_INT (entry, struct_ptr, UNPACK_S32 (*(src.gint32_p)));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gint32);
        entry++;
        break;

      case FLUX_MARSHAL_UINT32_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint32);
        SET_STRUCT_UINT (entry, struct_ptr, UNPACK_U32 (*(src.guint32_p)));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint32);
        entry++;
        break;

      case FLUX_MARSHAL_FLOAT32_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gfloat);
        SET_STRUCT_FLOAT (entry, struct_ptr, *(src.gfloat_p));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gfloat);
        entry++;
        break;

      case FLUX_MARSHAL_INT64_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gint64);
        SET_STRUCT_INT (entry, struct_ptr, UNPACK_S64 (*(src.gint64_p)));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gint64);
        entry++;
        break;

      case FLUX_MARSHAL_UINT64_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint64);
        SET_STRUCT_UINT (entry, struct_ptr, UNPACK_U64 (*(src.guint64_p)));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint64);
        entry++;
        break;

      case FLUX_MARSHAL_FLOAT64_TYPE:
        CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, gdouble);
        SET_STRUCT_FLOAT (entry, struct_ptr, *(src.gdouble_p));
        ADVANCE_INPUT_BY_TYPE (src, decoded_len, gdouble);
        entry++;
        break;

      case FLUX_MARSHAL_RAW_DYNAMIC_TYPE:
        n = flux_decompress_uint (&len, src.p, GET_REMAINING_LENGTH (pack_len, decoded_len));
        if (n < 1) return NULL;
        SET_STRUCT_META_UINT (entry, struct_ptr, len);
        ADVANCE_INPUT_BY_LENGTH (src, decoded_len, n);

        if (len)
        {
          guchar  *data;
          guchar **dest;

          CHECK_BOUNDS_FOR_LENGTH (pack_len, decoded_len, len);
          data = g_malloc (len);
          memcpy (data, src.p, len);
          ADVANCE_INPUT_BY_LENGTH (src, decoded_len, len);

          dest = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
          *dest = data;
        }
        else
        {
          guchar **dest;

          dest = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
          *dest = NULL;
        }

        entry++;
        break;

      case FLUX_MARSHAL_RAW_FIXED_TYPE:
        CHECK_BOUNDS_FOR_LENGTH (pack_len, decoded_len, entry->size);
        memcpy (G_STRUCT_MEMBER_P (struct_ptr, entry->offset), src.p, entry->size);
        ADVANCE_INPUT_BY_LENGTH (src, decoded_len, entry->size);
        entry++;
        break;

      case FLUX_MARSHAL_STRING_DYNAMIC_TYPE:
        n = flux_decompress_uint (&len, src.p, GET_REMAINING_LENGTH (pack_len, decoded_len));
        if (n < 1) return NULL;
        ADVANCE_INPUT_BY_LENGTH (src, decoded_len, n);

        if (len)
        {
          gchar  *str;
          gchar **dest;

          CHECK_BOUNDS_FOR_LENGTH (pack_len, decoded_len, len);
          str = g_malloc (len + 1);
          memcpy (str, src.p, len);
          *(str + len) = '\0';
          ADVANCE_INPUT_BY_LENGTH (src, decoded_len, len);

          dest = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
          *dest = str;
        }
        else
        {
          gchar **dest;

          dest = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
          *dest = NULL;
        }
        entry++;
        break;

      case FLUX_MARSHAL_STRING_FIXED_TYPE:
        /* String length */
        if (entry->size < 256)
        {
          CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint8);
          len = *(src.guint8_p);
          ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint8);
        }
        else if (entry->size < 65536)
        {
          CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint16);
          len = UNPACK_U16 (*(src.guint16_p));
          ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint16);
        }
        else
        {
          CHECK_BOUNDS_FOR_TYPE (pack_len, decoded_len, guint32);
          len = UNPACK_U32 (*(src.guint32_p));
          ADVANCE_INPUT_BY_TYPE (src, decoded_len, guint32);
        }

        /* String */
        if (len <= entry->size)
        {
          CHECK_BOUNDS_FOR_LENGTH (pack_len, decoded_len, len);
          memcpy ((gchar *) G_STRUCT_MEMBER_P (struct_ptr, entry->offset), src.p, len);
          ADVANCE_INPUT_BY_LENGTH (src, decoded_len, len);

          if (len < entry->size)
          {
            gchar *dest;

            dest = G_STRUCT_MEMBER_P (struct_ptr, entry->offset) + len;
            *dest = '\0';
          }
        }
        else
          return NULL;

        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_DYNAMIC_TYPE:
        {
          const gchar *old_src = src.gchar_p;

          src.p = decode_dynamic_struct (marshaller, src.p, pack_len - decoded_len,
                                         entry, struct_ptr);
          if (src.p == NULL)
            return NULL;

          decoded_len += (gchar *) src.gchar_p - (gchar *) old_src;
          entry++;
        }
        break;

      case FLUX_MARSHAL_STRUCT_FIXED_TYPE:
        {
          const gchar *old_src = src.gchar_p;

          src.p = decode_fixed_struct (marshaller, src.p, pack_len - decoded_len,
                                       entry, struct_ptr);
          if (src.p == NULL)
            return NULL;

          decoded_len += (gchar *) src.gchar_p - (gchar *) old_src;
          entry++;
        }
        break;

      case FLUX_MARSHAL_BRANCH_BEGIN_TYPE:
      case FLUX_MARSHAL_BRANCH_END_TYPE:
        entry++;
        break;

      case FLUX_MARSHAL_LAST_TYPE:
        return src.p;

      default:
        g_error ("Error in marshalling metadata - invalid entry type.");
    }
  }

  g_assert_not_reached ();
}

gboolean
flux_marshaller_decode_pack (FluxMarshaller *marshaller, const gchar *struct_name,
                             gconstpointer pack_ptr, guint pack_len, gpointer *struct_ptr)
{
  StructMeta *meta;

  assert_types (marshaller);
  g_return_val_if_fail (struct_name != NULL, FALSE);
  g_return_val_if_fail (struct_ptr != NULL, FALSE);
  g_return_val_if_fail (pack_ptr != NULL, FALSE);

  meta = g_hash_table_lookup (marshaller->types, struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when decoding.", struct_name);
    return FALSE;
  }

  /* The struct is always allocated by us */
  *struct_ptr = g_malloc (meta->size);

  /* We always clear it first, so we (and the user) know what to free */
  memset (*struct_ptr, 0, meta->size);

  /* With packs, we always have complete information */
  if (decode_pack (marshaller, pack_ptr, pack_len, meta->entries, *struct_ptr) == NULL)
  {
    free_struct_contents (marshaller, meta->entries, *struct_ptr);
    g_free (*struct_ptr);
    *struct_ptr = NULL;
    return FALSE;
  }

  return TRUE;
}

static gboolean
free_dynamic_struct (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry, gpointer struct_ptr)
{
  gpointer sub_struct_ptr;
  gboolean result = TRUE;

  sub_struct_ptr = G_STRUCT_MEMBER (gpointer, struct_ptr, entry->offset);

  if (sub_struct_ptr)
  {
    gchar      *sub_struct_name;
    StructMeta *meta;

    sub_struct_name = entry->default_value;
    meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
    if (!meta)
    {
      g_warning ("No marshalling info for struct \"%s\" when encoding.", sub_struct_name);
      return FALSE;
    }

    result = free_struct_contents (marshaller, meta->entries, sub_struct_ptr);
    g_free (sub_struct_ptr);
  }

  return result;
}

static gboolean
free_fixed_struct (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry, gpointer struct_ptr)
{
  gpointer       sub_struct_ptr;
  gchar         *sub_struct_name;
  StructMeta    *meta;

  sub_struct_ptr = G_STRUCT_MEMBER_P (struct_ptr, entry->offset);
  sub_struct_name = entry->default_value;

  meta = g_hash_table_lookup (marshaller->types, sub_struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when encoding.", sub_struct_name);
    return FALSE;
  }

  return free_struct_contents (marshaller, meta->entries, sub_struct_ptr);
}

static gboolean
free_struct_contents (FluxMarshaller *marshaller, FluxMarshalledStructEntry *entry, gpointer struct_ptr)
{
  guchar   *data;

  for (;;)
  {
    switch (entry->type)
    {
      case FLUX_MARSHAL_BOOLEAN_TYPE:
      case FLUX_MARSHAL_INT8_TYPE:
      case FLUX_MARSHAL_UINT8_TYPE:
      case FLUX_MARSHAL_INT16_TYPE:
      case FLUX_MARSHAL_UINT16_TYPE:
      case FLUX_MARSHAL_INT32_TYPE:
      case FLUX_MARSHAL_UINT32_TYPE:
      case FLUX_MARSHAL_FLOAT32_TYPE:
      case FLUX_MARSHAL_INT64_TYPE:
      case FLUX_MARSHAL_UINT64_TYPE:
      case FLUX_MARSHAL_FLOAT64_TYPE:
      case FLUX_MARSHAL_RAW_FIXED_TYPE:
      case FLUX_MARSHAL_STRING_FIXED_TYPE:
      case FLUX_MARSHAL_BRANCH_BEGIN_TYPE:
      case FLUX_MARSHAL_BRANCH_END_TYPE:
        entry++;
        break;

      case FLUX_MARSHAL_RAW_DYNAMIC_TYPE:
        data = G_STRUCT_MEMBER (guchar *, struct_ptr, entry->offset);
        if (data)
          g_free (data);
        entry++;
        break;

      case FLUX_MARSHAL_STRING_DYNAMIC_TYPE:
        data = G_STRUCT_MEMBER (guchar *, struct_ptr, entry->offset);
        if (data)
          g_free (data);
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_DYNAMIC_TYPE:
        free_dynamic_struct (marshaller, entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_STRUCT_FIXED_TYPE:
        free_fixed_struct (marshaller, entry, struct_ptr);
        entry++;
        break;

      case FLUX_MARSHAL_LAST_TYPE:
        return TRUE;

      default:
        g_error ("Error in marshalling metadata - invalid entry type.");
    }
  }

  return TRUE;
}

gboolean
flux_marshaller_free_struct (FluxMarshaller *marshaller, const gchar *struct_name, gpointer struct_ptr)
{
  StructMeta *meta;

  assert_types (marshaller);
  g_return_val_if_fail (struct_name != NULL, FALSE);
  g_return_val_if_fail (struct_ptr != NULL, FALSE);

  meta = g_hash_table_lookup (marshaller->types, struct_name);
  if (!meta)
  {
    g_warning ("No marshalling info for struct \"%s\" when freeing.", struct_name);
    return FALSE;
  }

  free_struct_contents (marshaller, meta->entries, struct_ptr);
  g_free (struct_ptr);

  return TRUE;
}
