/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-mem-chunk.c - Low-overhead same-size memory block allocator.
 *
 * Copyright (C) 2000 Helix Code Inc.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Michael Zucchi      <notzed@helixcode.com>
 *          Hans Petter Jansson <hpj@copyleft.no>
 *
 * Originally part of Evolution, Ximian's groupware client.
 */

#include "config.h"
#include "flux-mem-chunk.h"
#include <stdlib.h>          /* alloca */
#include <string.h>          /* memset */
#include <glib.h>

#define STRUCT_ALIGN (4)

struct _MemChunkFreeNode
{
  struct _MemChunkFreeNode *next;
  guint                     atoms;
};

/**
 * flux_mem_chunk_alloc:
 * @m: 
 * 
 * Allocate a new atom size block of memory from a memchunk.
 **/
gpointer
flux_mem_chunk_alloc (FluxMemChunk *mem_chunk)
{
  gchar            *b;
  MemChunkFreeNode *f;
  gpointer          mem;

  f = mem_chunk->free;
  if (f)
  {
    f->atoms--;
    if (f->atoms > 0)
    {
      mem = ((char *) f) + (f->atoms * mem_chunk->atomsize);
    }
    else
    {
      mem = f;
      mem_chunk->free = mem_chunk->free->next;
    }

    return mem;
  }
  else
  {
    b = g_malloc (mem_chunk->blocksize * mem_chunk->atomsize);
    g_ptr_array_add (mem_chunk->blocks, b);
    f = (MemChunkFreeNode *) &b[mem_chunk->atomsize];
    f->atoms = mem_chunk->blocksize - 1;
    f->next = NULL;
    mem_chunk->free = f;

    return b;
  }
}

gpointer
flux_mem_chunk_alloc0 (FluxMemChunk *mem_chunk)
{
  gpointer mem;

  mem = flux_mem_chunk_alloc (mem_chunk);
  memset (mem, 0, mem_chunk->atomsize);

  return mem;
}

/**
 * flux_mem_chunk_free:
 * @m: 
 * @mem: Address of atom to free.
 * 
 * Free a single atom back to the free pool of atoms in the given
 * memchunk.
 **/
void
flux_mem_chunk_free (FluxMemChunk *mem_chunk, gpointer mem)
{
  MemChunkFreeNode *f;

  /* Put the location back in the free list.  If we knew if the preceeding or following
     cells were free, we could merge the free nodes, but it doesn't really add much. */

  f = mem;
  f->next = mem_chunk->free;
  mem_chunk->free = f;
  f->atoms = 1;

  /* We could store the free list sorted - we could then do the above, and also
   * probably improve the locality of reference properties for the allocator.
   * And, it would simplify some other algorithms at that, but slow this one down
   * significantly */
}

/**
 * flux_mem_chunk_reset:
 * @m: 
 * 
 * Clean out the memchunk buffers.  Marks all allocated memory as free blocks,
 * but does not give it back to the system.  Can be used if the memchunk
 * is to be used repeatedly.
 **/
void
flux_mem_chunk_reset (FluxMemChunk *mem_chunk)
{
  gint              i;
  MemChunkFreeNode *f, *h = NULL;

  for (i = 0; i < mem_chunk->blocks->len; i++)
  {
    f = (MemChunkFreeNode *) mem_chunk->blocks->pdata[i];
    f->atoms = mem_chunk->blocksize;
    f->next = h;
    h = f;
  }

  mem_chunk->free = h;
}

struct _cleaninfo
{
  struct _cleaninfo *next;
  gchar             *base;
  gint               count;
  gint               size;  /* Just so tree_search has it */
};

static gint
tree_compare (struct _cleaninfo *a, struct _cleaninfo *b)
{
  if (a->base < b->base)
    return -1;
  else if (a->base > b->base)
    return 1;

  return 0;
}

static gint
tree_search (struct _cleaninfo *a, gchar *mem)
{
  if (a->base <= mem)
  {
    if (mem < &a->base [a->size])
      return 0;

    return 1;
  }

  return -1;
}

/**
 * flux_mem_chunk_clean:
 * @m: 
 * 
 * Scan all empty blocks and check for blocks which can be free'd 
 * back to the system.
 *
 * This routine may take a while to run if there are many allocated
 * memory blocks (if the total number of allocations is many times
 * greater than atomcount).
 **/
void
flux_mem_chunk_clean (FluxMemChunk *mem_chunk)
{
  GTree             *tree;
  gint               i;
  MemChunkFreeNode  *f;
  struct _cleaninfo *ci, *hi = NULL;

  f = mem_chunk->free;
  if (mem_chunk->blocks->len == 0 || f == NULL)
    return;

  /* First, setup the tree/list so we can map free block addresses to block addresses. */

  tree = g_tree_new ((GCompareFunc) tree_compare);

  for (i = 0; i < mem_chunk->blocks->len; i++)
  {
    ci = alloca (sizeof (*ci));
    ci->count = 0;
    ci->base = mem_chunk->blocks->pdata [i];
    ci->size = mem_chunk->blocksize * mem_chunk->atomsize;
    g_tree_insert (tree, ci, ci);
    ci->next = hi;
    hi = ci;
  }

  /* Now, scan all free nodes, and count them in their tree node. */

  while (f)
  {
    ci = g_tree_search (tree, (GCompareFunc) tree_search, f);

    if (ci)
      ci->count += f->atoms;
    else
      g_warning ("Can't find free node in memory block");

    f = f->next;
  }

  /* If any nodes are all free, free & unlink them. */

  for (ci = hi; ci; ci = ci->next)
  {
    if (ci->count == mem_chunk->blocksize)
    {
      MemChunkFreeNode *prev = NULL;

      for (f = mem_chunk->free; f; f = f->next)
      {
        if (tree_search (ci, (void *) f) == 0)
        {
          if (prev)
            prev->next = f->next;
          else
            mem_chunk->free = f->next;
        }
        else
        {
          prev = f;
        }
      }

      g_ptr_array_remove_fast (mem_chunk->blocks, ci->base);
      g_free (ci->base);
    }
  }

  g_tree_destroy (tree);
}

static guint
mc_get_arena_size (FluxStorage *storage)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);

  return mem_chunk->blocks->len * mem_chunk->blocksize * mem_chunk->atomsize;
}

static guint
mc_get_used_size (FluxStorage *storage)
{
  return FLUX_MEM_CHUNK (storage)->size;
}

static FluxItem
mc_get_root (FluxStorage *storage, const gchar *tag)
{
  FluxItem item;

  item = FLUX_POINTER_TO_ITEM (g_hash_table_lookup (FLUX_MEM_CHUNK (storage)->tag_table, tag));
  if (!item)
    return FLUX_ITEM_NULL;

  return item;
}

static gboolean
mc_set_root (FluxStorage *storage, const gchar *tag, FluxItem item)
{
  g_hash_table_replace (FLUX_MEM_CHUNK (storage)->tag_table, g_strdup (tag), FLUX_ITEM_TO_POINTER (item));
  return TRUE;
}

static FluxItem
mc_alloc (FluxStorage *storage, guint len)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);
  gpointer      data;

  g_return_val_if_fail (len != mem_chunk->atomsize, FLUX_ITEM_NULL);

  data = flux_mem_chunk_alloc (mem_chunk);
  mem_chunk->size += len;

  return FLUX_POINTER_TO_ITEM (data);
}

static FluxItem
mc_alloc0 (FluxStorage *storage, guint len)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);
  gpointer      data;

  g_return_val_if_fail (len != mem_chunk->atomsize, FLUX_ITEM_NULL);

  data = flux_mem_chunk_alloc (mem_chunk);
  mem_chunk->size += len;

  memset (data, 0, len);

  return FLUX_POINTER_TO_ITEM (data);
}

static FluxItem
mc_realloc (FluxStorage *storage, FluxItem item, guint len)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);

  g_return_val_if_fail (len != mem_chunk->atomsize, FLUX_ITEM_NULL);
  return item;
}

static gboolean
mc_free (FluxStorage *storage, FluxItem item)
{
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  flux_mem_chunk_free (FLUX_MEM_CHUNK (storage), FLUX_ITEM_TO_POINTER (item));
  FLUX_MEM_CHUNK (storage)->size -= FLUX_MEM_CHUNK (storage)->atomsize;
  return TRUE;
}

static gconstpointer
mc_peek (FluxStorage *storage, FluxItem item, guint offset, guint len)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);

  g_return_val_if_fail (item != FLUX_ITEM_NULL, NULL);
  g_return_val_if_fail (offset + len <= mem_chunk->atomsize, NULL);

  return (guchar *) FLUX_ITEM_TO_POINTER (item) + offset;
}

static gboolean
mc_read (FluxStorage *storage, FluxItem item, gpointer dest)
{
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (dest != NULL, FALSE);

  memcpy (dest, FLUX_ITEM_TO_POINTER (item), FLUX_MEM_CHUNK (storage)->atomsize);
  return TRUE;
}

static gboolean
mc_read_full (FluxStorage *storage, FluxItem item, gpointer dest,
              guint offset, guint len)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);

  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (offset + len <= mem_chunk->atomsize, FALSE);

  if (dest == NULL)
    return TRUE;

  memcpy (dest, (guchar *) FLUX_ITEM_TO_POINTER (item) + offset, len);
  return TRUE;
}

static gboolean
mc_write (FluxStorage *storage, FluxItem item, gconstpointer src)
{
  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);

  memcpy (FLUX_ITEM_TO_POINTER (item), src, FLUX_MEM_CHUNK (storage)->atomsize);
  return TRUE;
}

static gboolean
mc_write_full (FluxStorage *storage, FluxItem item,
               gconstpointer src, guint offset, guint len)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (storage);

  g_return_val_if_fail (item != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (offset + len <= mem_chunk->atomsize, FALSE);

  memcpy ((guchar *) FLUX_ITEM_TO_POINTER (item) + offset, src, len);
  return TRUE;
}

static guint
mc_get_item_len (FluxStorage *storage, FluxItem item)
{
  return FLUX_MEM_CHUNK (storage)->atomsize;
}

void
flux_mem_chunk_init (FluxMemChunk *mem_chunk, guint atomcount, guint atomsize)
{
  FluxStorage *storage;

  flux_storage_init             (FLUX_STORAGE (mem_chunk));
  flux_object_set_type          (FLUX_OBJECT (mem_chunk), FLUX_MEM_CHUNK_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (mem_chunk), flux_mem_chunk_finalize);

  mem_chunk->blocksize = atomcount;
  mem_chunk->atomsize  = MAX (atomsize, sizeof (MemChunkFreeNode));
  mem_chunk->blocks    = g_ptr_array_new ();
  mem_chunk->free      = NULL;

  /* Init storage */

  storage = FLUX_STORAGE (mem_chunk);

  storage->capabilities   = 0;

  storage->get_arena_size = mc_get_arena_size;
  storage->get_used_size  = mc_get_used_size;

  storage->get_root       = mc_get_root;
  storage->set_root       = mc_set_root;

  storage->alloc          = mc_alloc;
  storage->alloc0         = mc_alloc0;
  storage->realloc        = mc_realloc;
  storage->free           = mc_free;

  storage->peek           = mc_peek;
  storage->do_read        = mc_read;
  storage->do_read_full   = mc_read_full;
  storage->do_write       = mc_write;
  storage->do_write_full  = mc_write_full;

  storage->get_item_len   = mc_get_item_len;

  mem_chunk->tag_table    = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
}

/**
 * flux_mem_chunk_new:
 * @atomcount: The number of atoms stored in a single malloc'd block of memory.
 * @atomsize: The size of each allocation.
 * 
 * Create a new memchunk header.  Memchunks are an efficient way to allocate
 * and deallocate identical sized blocks of memory quickly, and space efficiently.
 * 
 * flux_mem_chunks are effectively the same as gmemchunks, only faster (much), and
 * they use less memory overhead for housekeeping.
 *
 * Return value: The new header.
 **/
FluxMemChunk *
flux_mem_chunk_new (guint atomcount, guint atomsize)
{
  FluxMemChunk *mem_chunk;

  mem_chunk = g_new (FluxMemChunk, 1);
  flux_mem_chunk_init (mem_chunk, atomcount, atomsize);

  return mem_chunk;
}

void
flux_mem_chunk_finalize (FluxObject *object)
{
  FluxMemChunk *mem_chunk = FLUX_MEM_CHUNK (object);
  gint          i;

  for (i = 0; i < mem_chunk->blocks->len; i++)
    g_free (mem_chunk->blocks->pdata [i]);

  g_ptr_array_free (mem_chunk->blocks, TRUE);
  g_hash_table_destroy (mem_chunk->tag_table);

  flux_storage_finalize (object);
}
