/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-mem-heap.c - Storage backend: In-memory heap.
 *
 * Copyright (C) 2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-mem-heap.h"
#include <stdlib.h>          /* alloca */
#include <string.h>          /* memset */
#include <glib.h>

static guint
mh_get_arena_size (FluxStorage *storage)
{
  return FLUX_MEM_HEAP (storage)->size;
}

static guint
mh_get_used_size (FluxStorage *storage)
{
  return FLUX_MEM_HEAP (storage)->size;
}

static FluxItem
mh_get_root (FluxStorage *storage, const gchar *tag)
{
  FluxItem item;

  item = FLUX_POINTER_TO_ITEM (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->tag_table, tag));
  if (!item)
    return FLUX_ITEM_NULL;

  return item;
}

static gboolean
mh_set_root (FluxStorage *storage, const gchar *tag, FluxItem item)
{
  g_hash_table_replace (FLUX_MEM_HEAP (storage)->tag_table, g_strdup (tag), FLUX_ITEM_TO_POINTER (item));
  return TRUE;
}

static FluxItem
mh_alloc (FluxStorage *storage, guint len)
{
  gpointer data;

  if (!len)
    return FLUX_ITEM_NULL;

  data = g_malloc (len);
  FLUX_MEM_HEAP (storage)->size += len;

  g_hash_table_insert (FLUX_MEM_HEAP (storage)->alloc_table, data, GUINT_TO_POINTER (len));
  return FLUX_ITEM (FLUX_POINTER_TO_ITEM (data));
}

static FluxItem
mh_alloc0 (FluxStorage *storage, guint len)
{
  FluxItem item;

  item = mh_alloc (storage, len);
  if (item != FLUX_ITEM_NULL)
    memset ((gpointer) item, 0, len);

  return item;
}

static FluxItem
mh_realloc (FluxStorage *storage, FluxItem item, guint len)
{
  gpointer data;
  guint    old_len;

  old_len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                                   FLUX_ITEM_TO_POINTER (item)));
  g_hash_table_remove (FLUX_MEM_HEAP (storage)->alloc_table, FLUX_ITEM_TO_POINTER (item));

  FLUX_MEM_HEAP (storage)->size -= old_len;
  data = g_realloc (FLUX_ITEM_TO_POINTER (item), len);
  FLUX_MEM_HEAP (storage)->size += len;

  if (!data)
    return FLUX_ITEM_NULL;

  g_hash_table_insert (FLUX_MEM_HEAP (storage)->alloc_table, data, GUINT_TO_POINTER (len));
  return FLUX_ITEM (FLUX_POINTER_TO_ITEM (data));
}

static gboolean
mh_free (FluxStorage *storage, FluxItem item)
{
  guint old_len;

  old_len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                                   FLUX_ITEM_TO_POINTER (item)));

  if (!g_hash_table_remove (FLUX_MEM_HEAP (storage)->alloc_table, FLUX_ITEM_TO_POINTER (item)))
  {
    g_warning ("Invalid item passed to free.");
    return FALSE;
  }

  FLUX_MEM_HEAP (storage)->size -= old_len;
  g_free (FLUX_ITEM_TO_POINTER (item));
  return TRUE;
}

static gconstpointer
mh_peek (FluxStorage *storage, FluxItem item, guint offset, guint len)
{
  guint item_len;

  item_len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                                    FLUX_ITEM_TO_POINTER (item)));
  if (!item_len)
  {
    g_warning ("Invalid item passed to peek.");
    return NULL;
  }

  if (offset + len > item_len)
  {
    g_warning ("Out-of-bounds peek.");
    return NULL;
  }

  return (gconstpointer) ((guchar *) FLUX_ITEM_TO_POINTER (item) + offset);
}

static gboolean
mh_read (FluxStorage *storage, FluxItem item, gpointer dest)
{
  guint len;

  len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                               FLUX_ITEM_TO_POINTER (item)));
  if (!len)
  {
    g_warning ("Invalid item passed to read.");
    return FALSE;
  }

  memcpy (dest, FLUX_ITEM_TO_POINTER (item), len);
  return TRUE;
}

static gboolean
mh_read_full (FluxStorage *storage, FluxItem item, gpointer dest,
              guint offset, guint len)
{
  guint item_len;

  item_len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                                    FLUX_ITEM_TO_POINTER (item)));
  if (!item_len)
  {
    g_warning ("Invalid item passed to read_full.");
    return FALSE;
  }

  if (offset + len > item_len)
  {
    g_warning ("Out-of-bounds read_full.");
    return FALSE;
  }

  memcpy (dest, ((guchar *) FLUX_ITEM_TO_POINTER (item)) + offset, len);
  return TRUE;
}

static gboolean
mh_write (FluxStorage *storage, FluxItem item, gconstpointer src)
{
  guint len;

  len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                               FLUX_ITEM_TO_POINTER (item)));
  if (!len)
  {
    g_warning ("Invalid item passed to write.");
    return FALSE;
  }

  memcpy (FLUX_ITEM_TO_POINTER (item), src, len);
  return TRUE;
}

static gboolean
mh_write_full (FluxStorage *storage, FluxItem item,
               gconstpointer src, guint offset, guint len)
{
  guint item_len;

  item_len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                                    FLUX_ITEM_TO_POINTER (item)));
  if (!item_len)
  {
    g_warning ("Invalid item passed to write_full.");
    return FALSE;
  }

  if (offset + len > item_len)
  {
    g_warning ("Out-of-bounds write_full!");
    return FALSE;
  }

  memcpy (((guchar *) FLUX_ITEM_TO_POINTER (item)) + offset, src, len);
  return TRUE;
}

static guint
mh_get_item_len (FluxStorage *storage, FluxItem item)
{
  guint len;

  len = GPOINTER_TO_UINT (g_hash_table_lookup (FLUX_MEM_HEAP (storage)->alloc_table,
                                               FLUX_ITEM_TO_POINTER (item)));
  if (!len)
  {
    g_warning ("Invalid item passed to get_item_len.");
    return FALSE;
  }

  return len;
}

void
flux_mem_heap_init (FluxMemHeap *mem_heap)
{
  FluxStorage *storage;

  flux_storage_init             (FLUX_STORAGE (mem_heap));
  flux_object_set_type          (FLUX_OBJECT (mem_heap), FLUX_MEM_HEAP_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (mem_heap), flux_mem_heap_finalize);

  /* Init storage */

  storage = FLUX_STORAGE (mem_heap);

  storage->capabilities = FLUX_STORAGE_VARIABLE_LENGTH;

  storage->get_arena_size = mh_get_arena_size;
  storage->get_used_size  = mh_get_used_size;

  storage->get_root       = mh_get_root;
  storage->set_root       = mh_set_root;

  storage->alloc          = mh_alloc;
  storage->alloc0         = mh_alloc0;
  storage->realloc        = mh_realloc;
  storage->free           = mh_free;

  storage->peek           = mh_peek;
  storage->do_read        = mh_read;
  storage->do_read_full   = mh_read_full;
  storage->do_write       = mh_write;
  storage->do_write_full  = mh_write_full;

  storage->get_item_len   = mh_get_item_len;

  mem_heap->tag_table   = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  mem_heap->alloc_table = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, NULL);
}

FluxMemHeap *
flux_mem_heap_new (void)
{
  FluxMemHeap *mem_heap;

  mem_heap = g_new0 (FluxMemHeap, 1);
  flux_mem_heap_init (mem_heap);

  return mem_heap;
}

void
flux_mem_heap_finalize (FluxObject *object)
{
  FluxMemHeap *mem_heap = FLUX_MEM_HEAP (object);

  g_hash_table_foreach (mem_heap->alloc_table, (GHFunc) g_free, NULL);

  g_hash_table_destroy (mem_heap->tag_table);
  g_hash_table_destroy (mem_heap->alloc_table);

  flux_storage_finalize (object);
}
