/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-object.c - Base class for Flux' naive but lightweight objects.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-object.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define D(x)

typedef struct
{
  guint  id;
  gchar *name;

  guint  parent_type;
}
ClassInfo;

typedef struct
{
  gchar *name;

  guint  parent_type;
}
InternalClassInfo;

static GStaticMutex object_mutex             = G_STATIC_MUTEX_INIT;

static guint        class_last_id            = FLUX_LAST_TYPE;  /* Last valid id + 1 */

static ClassInfo  **class_by_id_array        = NULL;
static GHashTable  *class_by_name_hash_table = NULL;

/* Static information on internal types, indexed by FluxObjectType */
static InternalClassInfo internal_class_info [] =
{
  /* Class name             Parent class ID */
  { NULL,                   FLUX_INVALID_TYPE     },  /* Class IDs start at 1 */
  { "FluxObject",           FLUX_INVALID_TYPE     },
  { "FluxBoolArray",        FLUX_OBJECT_TYPE      },
  { "FluxBinPath",          FLUX_BOOL_ARRAY_TYPE  },
  { "FluxExtHashTable",     FLUX_OBJECT_TYPE      },
  { "FluxDataCache",        FLUX_OBJECT_TYPE      },
  { "FluxCachedFile",       FLUX_OBJECT_TYPE      },
  { "FluxRmd160",           FLUX_OBJECT_TYPE      },
  { "FluxRng",              FLUX_OBJECT_TYPE      },
  { "FluxTwofishKey",       FLUX_OBJECT_TYPE      },
  { "FluxElgamalKey",       FLUX_OBJECT_TYPE      },
  { "FluxElgamalPublicKey", FLUX_ELGAMAL_KEY_TYPE },
  { "FluxElgamalSecretKey", FLUX_ELGAMAL_KEY_TYPE },
  { "FluxRSAKey",           FLUX_OBJECT_TYPE      },
  { "FluxRSAPublicKey",     FLUX_RSA_KEY_TYPE     },
  { "FluxRSASecretKey",     FLUX_RSA_KEY_TYPE     },
  { "FluxECKey",            FLUX_OBJECT_TYPE      },
  { "FluxECPublicKey",      FLUX_EC_KEY_TYPE      },
  { "FluxECSecretKey",      FLUX_EC_KEY_TYPE      },
  { "FluxCipher",           FLUX_OBJECT_TYPE      },
  { "FluxIPAddr",           FLUX_OBJECT_TYPE      },
  { "FluxIPResolver",       FLUX_OBJECT_TYPE      },
  { "FluxFifo",             FLUX_OBJECT_TYPE      },
  { "FluxDualFifo",         FLUX_OBJECT_TYPE      },
  { "FluxIO",               FLUX_DUAL_FIFO_TYPE   },
  { "FluxFileIO",           FLUX_IO_TYPE          },
  { "FluxTcpIO",            FLUX_IO_TYPE          },
  { "FluxTcpListener",      FLUX_OBJECT_TYPE      },
  { "FluxIOGroup",          FLUX_OBJECT_TYPE      },
  { "FluxCursor",           FLUX_OBJECT_TYPE      },
  { "FluxComm",             FLUX_OBJECT_TYPE      },
  { "FluxEnvoy",            FLUX_COMM_TYPE        },
  { "FluxMarshaller",       FLUX_OBJECT_TYPE      },
  { "FluxStorage",          FLUX_OBJECT_TYPE      },
  { "FluxMemChunk",         FLUX_STORAGE_TYPE     },
  { "FluxMemHeap",          FLUX_STORAGE_TYPE     },
  { "FluxDiskChunk",        FLUX_STORAGE_TYPE     },
  { "FluxDiskHeap",         FLUX_STORAGE_TYPE     },
  { "FluxDiskHeapCursor",   FLUX_CURSOR_TYPE      },
  { "FluxPackedBinTree",    FLUX_OBJECT_TYPE      },
  { "FluxPatTree",          FLUX_OBJECT_TYPE      },
  { "FluxPatTreeCursor",    FLUX_CURSOR_TYPE      },
  { "FluxExtPatTree",       FLUX_OBJECT_TYPE      },
  { "FluxExtPatTreeCursor", FLUX_CURSOR_TYPE      },
  { "FluxInterval",         FLUX_OBJECT_TYPE      }
};

static ClassInfo *
find_class_by_name (const gchar *name)
{
  ClassInfo *class_info;

  class_info = g_hash_table_lookup (class_by_name_hash_table, name);
  return class_info;
}

static ClassInfo *
find_class_by_type (guint type)
{
  ClassInfo *class_info;

  class_info = class_by_id_array [type];
  return class_info;
}

static gboolean
is_valid_ancestry (guint child, guint ancestor)
{
  while (child != ancestor && child != FLUX_INVALID_TYPE)
  {
    child = flux_get_class_parent (child);
  }

  return child != FLUX_INVALID_TYPE;
}

guint
flux_register_class (const gchar *name, guint parent_type)
{
  ClassInfo *class_info;

  if (!class_by_name_hash_table)
    class_by_name_hash_table = g_hash_table_new (g_str_hash, g_str_equal);
  if (!class_by_id_array)
    class_by_id_array = g_new (ClassInfo *, class_last_id);

  class_info = find_class_by_name (name);
  if (class_info)
  {
    g_warning ("Register: User class '%s' already registered.", name);
    return class_info->id;
  }

  class_info = g_new (ClassInfo, 1);

  class_info->id          = class_last_id++;
  class_info->name        = g_strdup (name);
  class_info->parent_type = parent_type;

  g_hash_table_insert (class_by_name_hash_table, (gpointer) name, class_info);
  class_by_id_array = g_renew (ClassInfo *, class_by_id_array, class_last_id);
  class_by_id_array [class_info->id] = class_info;

  return class_info->id;
}

static void
unregister_class (ClassInfo *class_info)
{
  g_hash_table_remove (class_by_name_hash_table, class_info->name);
  class_by_id_array [class_info->id] = NULL;

  g_free (class_info->name);
  g_free (class_info);
}

void
flux_unregister_class_by_name (const gchar *name)
{
  ClassInfo *class_info;

  class_info = find_class_by_name (name);
  if (!class_info)
    g_error ("Unregister: User class '%s' is not registered.", name);

  unregister_class (class_info);
}

void
flux_unregister_class_by_type (guint type)
{
  ClassInfo *class_info;

  if (type >= class_last_id)
    g_error ("Unregister: User class %d is not registered (out of range).", type);

  class_info = find_class_by_type (type);
  if (!class_info)
    g_error ("Unregister: User class %d is already unregistered.", type);

  unregister_class (class_info);
}

const gchar *
flux_get_class_name (guint type)
{
  ClassInfo *class_info;

  if (type < FLUX_LAST_TYPE)
    return internal_class_info [type].name;

  if (type >= class_last_id)
    g_error ("User class %d is not registered (out of range).", type);

  class_info = find_class_by_type (type);
  if (!class_info)
    g_error ("User class %d was previously unregistered.", type);

  return class_info->name;
}

guint
flux_get_class_type (const gchar *name)
{
  ClassInfo *class_info;

  if (strlen (name) > 4 && name [0] == 'F' && name [1] == 'l' && name [2] == 'u' && name [3] == 'x')
  {
    const gchar *s0 = name + 4;
    gint         i;

    for (i = 1; i < FLUX_LAST_TYPE; i++)
    {
      if (!strcmp (s0, internal_class_info [i].name + 4))
        return i;
    }
  }

  class_info = find_class_by_name (name);
  if (!class_info)
    g_error ("Class %s is not registered.", name);

  return class_info->id;
}

guint
flux_get_class_parent (guint type)
{
  ClassInfo *class_info;

  if (type < FLUX_LAST_TYPE)
    return internal_class_info [type].parent_type;

  if (type >= class_last_id)
    g_error ("User class %d is not registered (out of range).", type);

  class_info = find_class_by_type (type);
  if (!class_info)
    g_error ("User class %d was previously unregistered.", type);

  return class_info->parent_type;
}

FluxObject *
flux_object_new (void)
{
  FluxObject *obj;
  
  obj = g_new0 (FluxObject, 1);
  flux_object_init (obj);
  
  return obj;
}

void
flux_object_init (FluxObject *obj)
{
  obj->type = FLUX_OBJECT_TYPE;
  obj->refs = 1;
}

void
flux_object_set_type (FluxObject *obj, guint type)
{
  g_static_mutex_lock (&object_mutex);
  obj->type = type;
  g_static_mutex_unlock (&object_mutex);
}

guint
flux_object_get_type (FluxObject *obj)
{
  guint type;

  g_static_mutex_lock (&object_mutex);
  type = obj->type;
  g_static_mutex_unlock (&object_mutex);

  return type;
}

gboolean
flux_object_is_type (FluxObject *obj, guint type)
{
  gboolean valid;

  flux_object_assert (obj);

  valid = is_valid_ancestry (flux_object_get_type (obj), type);

  return valid;
}

void
flux_object_set_finalize_func (FluxObject *obj, FluxFinalizeFunc *func)
{
  obj->finalize_func = func;
}

FluxFinalizeFunc *
flux_object_get_finalize_func (FluxObject *obj)
{
  return obj->finalize_func;
}

gint
flux_object_get_refs (FluxObject *obj)
{
  gint refs;

  flux_object_assert (obj);

  g_static_mutex_lock (&object_mutex);
  refs = obj->refs;
  g_static_mutex_unlock (&object_mutex);

  return refs;
}

void
flux_object_ref (FluxObject *obj)
{
  flux_object_assert (obj);

  g_static_mutex_lock (&object_mutex);
  obj->refs++;
  g_static_mutex_unlock (&object_mutex);

  D (g_print ("Added object reference: %d.\n", flux_object_get_refs (obj)));
}

void
flux_object_unref (FluxObject *obj)
{
  flux_object_assert (obj);

  D (g_print ("Removing object reference: %d.\n", flux_object_get_refs (obj)));

  g_static_mutex_lock (&object_mutex);
  obj->refs--;

  if (!obj->refs)
  {
    g_static_mutex_unlock (&object_mutex);
    flux_object_destroy (obj);
  }
  else
    g_static_mutex_unlock (&object_mutex);
}

void
flux_object_destroy (FluxObject *obj)
{
  flux_object_assert (obj);

  if (obj->finalize_func)
    obj->finalize_func (obj);

  /* The object may have been referenced during finalization - maybe one of
   * the finalizers is doing cleanup, flushing buffers, etc. In that case, don't
   * destroy the object, but wait until we're invoked again. See FluxTcpIO for
   * an example. */

  if (!obj->refs)
  {
    flux_object_set_type (obj, FLUX_FREE_TYPE);
    g_free (obj);
  }
}

void
flux_object_assert (FluxObject *obj)
{
  guint type;

  g_assert (obj);
  type = flux_object_get_type (obj);

  if (type == FLUX_FREE_TYPE)
    g_error ("Invalid object; might be previously freed.");

  if (type == 0 || type >= class_last_id)
    g_error ("Invalid object.");
}

void
flux_object_check_real (FluxObject *obj, guint type)
{
  guint obj_type;

  if (obj == NULL)
    g_error ("NULL object passed as %s.", flux_get_class_name (type));

  obj_type = flux_object_get_type (obj);

  if (obj_type == FLUX_FREE_TYPE)
    g_error ("Freed object passed as %s.", flux_get_class_name (type));

  if (obj_type == 0 || obj_type >= class_last_id)
    g_error ("Invalid object passed as %s.", flux_get_class_name (type));

  if (!is_valid_ancestry (obj_type, type))
    g_error ("Parameter mismatch: %s is not an ancestor of %s.", flux_get_class_name (type),
             flux_get_class_name (obj_type));
}
