/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-packed-bin-tree.c - Packed, static binary tree.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

/* Examples
 * ========
 *
 * Tree with 3 leaves
 * ------------------
 *
 * X = internal, O = external (leaf)
 *
 * OXOXO    n_elems = 5
 * 01234    root    = 3
 *
 *
 * Tree with 4 leaves
 * ------------------
 *
 * OXOXOXO  n_elems = 7
 * 0123456  root    = 3 */

#include "config.h"
#include "flux-packed-bin-tree.h"
#include "flux-mem-heap.h"
#include <string.h>         /* memcpy, memset */

#define DEBUG(x)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_PACKED_BIN_TREE_TYPE)

#define ITEMS_FOR_LEAVES(n_leaves) ((n_leaves) * 2 - 1)
#define LEAVES_FOR_ITEMS(n_items)  (((n_items) + 1) / 2)

FluxPackedBinTree *
flux_packed_bin_tree_new (FluxStorage *storage, guint elem_len)
{
  FluxPackedBinTree *bin_tree;

  g_return_val_if_fail (elem_len > 0, NULL);

  bin_tree = g_new0 (FluxPackedBinTree, 1);
  if (!flux_packed_bin_tree_init (bin_tree, storage, elem_len))
  {
    g_free (bin_tree);
    return NULL;
  }

  return bin_tree;
}

gboolean
flux_packed_bin_tree_init (FluxPackedBinTree *bin_tree, FluxStorage *storage, guint elem_len)
{
  flux_object_init              (FLUX_OBJECT (bin_tree));
  flux_object_set_type          (FLUX_OBJECT (bin_tree), FLUX_PACKED_BIN_TREE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (bin_tree), flux_packed_bin_tree_finalize);

  if (storage)
  {
    if (!(flux_storage_get_capabilities (storage) & FLUX_STORAGE_VARIABLE_LENGTH))
    {
      g_warning ("FluxPackedBinTree needs a variable length storage!");
      return FALSE;
    }

    flux_object_ref (FLUX_OBJECT (storage));
  }
  else
  {
    storage = FLUX_STORAGE (flux_mem_heap_new ());
  }

  bin_tree->storage  = storage;
  bin_tree->elem_len = elem_len;
  return TRUE;
}

void
flux_packed_bin_tree_finalize (FluxObject *object)
{
  FluxPackedBinTree *bin_tree = FLUX_PACKED_BIN_TREE (object);

  flux_object_unref (FLUX_OBJECT (bin_tree->storage));
}

guint
flux_packed_bin_tree_get_elem_len (FluxPackedBinTree *bin_tree)
{
  assert_types (bin_tree);

  return bin_tree->elem_len;
}

static guint
get_n_elems (FluxPackedBinTree *bin_tree, FluxItem tree)
{
  guint len;

  len = flux_storage_get_item_len (bin_tree->storage, tree);
  if (!len)
    return 0;

  /* Must be a multiple of elem_len */
  if (len % bin_tree->elem_len)
    return 0;

  len /= bin_tree->elem_len;

  /* Must have an odd number of elements */
  if (!(len & 1))
    return 0;

  return len;
}

gint
flux_packed_bin_tree_get_n_leaves (FluxPackedBinTree *bin_tree, FluxItem tree)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, -1);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems)
    return -1;

  return LEAVES_FOR_ITEMS (n_elems);
}

FluxItem
flux_packed_bin_tree_create_tree (FluxPackedBinTree *bin_tree, guint n_leaves)
{
  FluxItem item;
  guint    size;

  assert_types (bin_tree);
  g_return_val_if_fail (n_leaves > 0, FLUX_ITEM_NULL);

  size = ITEMS_FOR_LEAVES (n_leaves) * bin_tree->elem_len;
  item = flux_storage_alloc0 (bin_tree->storage, size);

  return item;
}

void
flux_packed_bin_tree_delete_tree (FluxPackedBinTree *bin_tree, FluxItem tree)
{
  assert_types (bin_tree);
  g_return_if_fail (tree != FLUX_ITEM_NULL);

  flux_storage_free (bin_tree->storage, tree);
}

static gint
path_to_index (guint n_elems, FluxBinPath *path)
{
  guint path_depth;
  gint  index;
  guint bound_start;
  guint bound_len;
  guint i;

  path_depth = flux_bin_path_get_depth (path);

  if (n_elems == 1)
    index = 0;
  else
    index = (n_elems >> 1) | 1;

  bound_start = 0;
  bound_len   = n_elems;

  for (i = 0; i < path_depth; i++)
  {
    gboolean go_right;
    gint     step;

    /* If we're at an even index, we've reached a leaf. We can't go on, so path is invalid. */
    if (!(index & 1))
      return -1;

    go_right = flux_bin_path_get_value (path, i);

    if (go_right)
    {
      bound_len -= (index - bound_start + 1);  /* Discount left side + index we're at */
      bound_start = index + 1;
    }
    else
    {
      bound_len -= (bound_start + bound_len - index);
    }

    step = (bound_len + 1) >> 1;

    /* Step > 1 means we're going to another internal node. Offset is even. */
    if (step > 1)
      step &= 0xfffffffe;

#if 0
    /* Step == 0 means we're going to a leaf. Offset is 1 (odd). */
    else if (step == 0)
      step = 1;
#endif

    index += go_right ? step : -step;
  }

  return index;
}

static FluxBinPath *
index_to_path (guint n_elems, gint target_index)
{
  FluxBinPath *path;
  gint         index;
  guint        bound_start;
  guint        bound_len;

  if (n_elems == 1)
    index = 0;
  else
    index = (n_elems >> 1) | 1;

  bound_start = 0;
  bound_len   = n_elems;

  path = flux_bin_path_new ();

  while (index != target_index)
  {
    gboolean go_right;
    gint     step;

#if 0
    /* If we're at an even index, we've reached a leaf. We can't go on. This should never happen. */
    if (!(index & 1))
      return NULL;
#endif

    go_right = target_index > index ? TRUE : FALSE;

    if (go_right)
    {
      bound_len -= (index - bound_start + 1);  /* Discount left side + index we're at */
      bound_start = index + 1;

      flux_bin_path_right (path);
    }
    else
    {
      bound_len -= (bound_start + bound_len - index);

      flux_bin_path_left (path);
    }

    step = (bound_len + 1) >> 1;

    /* Step > 1 means we're going to another internal node. Offset is even. */
    if (step > 1)
      step &= 0xfffffffe;

#if 0
    /* Step == 0 means we're going to a leaf. Offset is 1 (odd). */
    else if (step == 0)
      step = 1;
#endif

    index += go_right ? step : -step;
  }

  return path;
}

gint
flux_packed_bin_tree_path_to_index (FluxPackedBinTree *bin_tree, FluxItem tree, FluxBinPath *path)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, -1);
  g_return_val_if_fail (path != NULL, -1);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems)
    return -1;

  return path_to_index (n_elems, path);
}

gint
flux_packed_bin_tree_path_to_leaf (FluxPackedBinTree *bin_tree, FluxItem tree, FluxBinPath *path)
{
  guint n_elems;
  gint  index;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, -1);
  g_return_val_if_fail (path != NULL, -1);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems)
    return -1;

  index = path_to_index (n_elems, path);
  if (index < 0)
    return -1;

  /* No internal nodes allowed */
  if (index & 1)
    return -1;

  return index >> 1;
}

FluxBinPath *
flux_packed_bin_tree_index_to_path (FluxPackedBinTree *bin_tree, FluxItem tree, gint index)
{
  guint        n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, NULL);
  g_return_val_if_fail (index >= 0, NULL);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || index >= n_elems)
    return NULL;

  return index_to_path (n_elems, index);
}

gint
flux_packed_bin_tree_index_to_leaf (FluxPackedBinTree *bin_tree, FluxItem tree, gint index)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, -1);
  g_return_val_if_fail (index >= 0, -1);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || index >= n_elems)
    return -1;

  /* No internal nodes allowed */
  if (index & 1)
    return -1;

  return index >> 1;
}

FluxBinPath *
flux_packed_bin_tree_leaf_to_path (FluxPackedBinTree *bin_tree, FluxItem tree, gint leaf)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, NULL);
  g_return_val_if_fail (leaf >= 0, NULL);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || (leaf << 1) >= n_elems)
    return NULL;

  return index_to_path (n_elems, leaf << 1);
}

gint
flux_packed_bin_tree_leaf_to_index (FluxPackedBinTree *bin_tree, FluxItem tree, gint leaf)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, -1);
  g_return_val_if_fail (leaf >= 0, -1);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || (leaf << 1) >= n_elems)
    return -1;

  return leaf << 1;
}

gconstpointer
flux_packed_bin_tree_peek (FluxPackedBinTree *bin_tree, FluxItem tree, gint index)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, NULL);
  g_return_val_if_fail (index >= 0, NULL);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || index >= n_elems)
    return NULL;

  return flux_storage_peek (bin_tree->storage, tree,
                            index * bin_tree->elem_len, bin_tree->elem_len);
}

gboolean
flux_packed_bin_tree_read (FluxPackedBinTree *bin_tree, FluxItem tree, gint index, gpointer dest)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (index >= 0, FALSE);
  g_return_val_if_fail (dest != NULL, FALSE);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || index >= n_elems)
    return FALSE;

  return flux_storage_read_full (bin_tree->storage, tree, dest,
                                 index * bin_tree->elem_len, bin_tree->elem_len);
}

gboolean
flux_packed_bin_tree_write (FluxPackedBinTree *bin_tree, FluxItem tree, gint index, gconstpointer src)
{
  guint n_elems;

  assert_types (bin_tree);
  g_return_val_if_fail (tree != FLUX_ITEM_NULL, FALSE);
  g_return_val_if_fail (index >= 0, FALSE);
  g_return_val_if_fail (src != NULL, FALSE);

  n_elems = get_n_elems (bin_tree, tree);
  if (!n_elems || index >= n_elems)
    return FALSE;

  return flux_storage_write_full (bin_tree->storage, tree, src,
                                  index * bin_tree->elem_len, bin_tree->elem_len);
}
