/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-pat-tree.c - On-disk indexing tree.
 *
 * Copyright (C) 2002-2003 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-mem-heap.h"
#include "flux-pat-tree.h"
#include <string.h>         /* memcpy, memset */

#define DEBUG(x)

#define assert_types(o)        flux_object_check (FLUX_OBJECT (o), FLUX_PAT_TREE_TYPE)
#define assert_cursor_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_PAT_TREE_CURSOR_TYPE)

#define DB_NAME    "FluxPatTree"
#define DB_VERSION 1

#define HIGHEST_BIT_INDEX 0x7fffffff

#define GET_INDEXED_BIT(data, data_len, bit_index) \
  ((data_len) > (bit_index >> 3) ? \
  *((guint8 *) (data) + (bit_index >> 3)) & (0x80 >> (bit_index & 0x7)) : \
  ((data_len << 3) == (bit_index)) ? TRUE : FALSE)

#define BIT_INDEX_GT(a_bit, b_bit)         ((a_bit) > (b_bit))
#define NODE_BIT_INDEX_GT(a, b)            BIT_INDEX_GT ((a)->bit_index, (b)->bit_index)
#define CALC_ATOM_SIZE(key_len, value_len) (sizeof (PTNodeFixedLen) + (key_len) + (value_len))
#define ATOM_SIZE(pat_tree)                CALC_ATOM_SIZE ((pat_tree)->key_len, (pat_tree)->value_len)

#define ROOT_NODE_IS_ACTIVE(root_node)     ((root_node)->active)
#define ROOT_NODE_SET_ACTIVE(root_node, a) ((root_node)->active = (a))

#define STORAGE_IS_VARLEN(pat_tree)        (flux_storage_get_capabilities ((pat_tree)->storage) & \
                                            FLUX_STORAGE_VARIABLE_LENGTH)

typedef struct
{
  FluxItem item;
  guint    went_left  : 1;
  guint    went_right : 1;
}
StackItem;

typedef struct
{
  FluxItem    left_child;
  FluxItem    right_child;
  guint32     bit_index : 31;
  guint32     active : 1;

  /* Pointers to key/value. Necessary, since we can't realloc the node head when
   * key/value length change - there are pointers to it that are difficult to update. */
  FluxItem    key;
  FluxItem    value;
}
PTNode;

typedef struct
{
  FluxItem    left_child;
  FluxItem    right_child;
  guint32     bit_index : 31;
  guint32     active : 1;

  /* Key data */
  /* Value data */
}
PTNodeFixedLen;

static FluxPatTreeCursor *cursor_new (FluxPatTree *pat_tree, FluxItem root_index, guint key_len);

#if 0

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void
dump_hex (guchar *buf, guint len)
{
  gint i;

  for (i = 0; i < len; i++)
    g_print ("%c%c ", int_to_hex (buf [i] >> 4), int_to_hex (buf [i] & 0x0f));

  g_print ("\n");
}

#endif

static StackItem *
stack_item_find (GArray *array, FluxItem item)
{
  gint i;

  for (i = 0; i < array->len; i++)
  {
    StackItem *stack_item;

    stack_item = &g_array_index (array, StackItem, i);
    if (stack_item->item == item)
      return stack_item;
  }

  return NULL;
}

static StackItem *
stack_item_last (GArray *array)
{
  if (array->len == 0)
    return NULL;

  return (StackItem *) &g_array_index (array, StackItem, array->len - 1);
}

static void
stack_item_append (GArray *array, FluxItem item, gboolean went_left, gboolean went_right)
{
  StackItem stack_item;

  stack_item.item       = item;
  stack_item.went_left  = went_left;
  stack_item.went_right = went_right;

  g_array_append_val (array, stack_item);
}

static gboolean
stack_item_remove (GArray *array, FluxItem item)
{
  gint i;

  for (i = 0; i < array->len; i++)
  {
    StackItem *stack_item;

    stack_item = &g_array_index (array, StackItem, i);
    if (stack_item->item == item)
    {
      g_array_remove_index (array, i);
      return TRUE;
    }
  }

  return FALSE;
}

static void
stack_item_remove_last (GArray *array)
{
  if (array->len == 0)
    return;

  g_array_remove_index (array, array->len - 1);
}

static PTNode *
alloc_node_mem (FluxPatTree *pat_tree)
{
  return flux_mem_chunk_alloc (pat_tree->mem_chunk);
}

static void
free_node_mem (FluxPatTree *pat_tree, PTNode *node)
{
  return flux_mem_chunk_free (pat_tree->mem_chunk, node);
}

static FluxItem
alloc_node_disk (FluxPatTree *pat_tree)
{
  if (STORAGE_IS_VARLEN (pat_tree))
    return flux_storage_alloc (pat_tree->storage, sizeof (PTNode));
  else
    return flux_storage_alloc (pat_tree->storage, ATOM_SIZE (pat_tree));
}

static gboolean
free_node_disk (FluxPatTree *pat_tree, FluxItem node_index)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    const PTNode *node;
    FluxItem      key, value;

    node = flux_storage_peek (pat_tree->storage, node_index, 0, sizeof (PTNode));
    if (!node)
      return FALSE;

    key   = node->key;
    value = node->value;

    if (key != FLUX_ITEM_NULL)
      flux_storage_free (pat_tree->storage, key);
    if (value != FLUX_ITEM_NULL)
      flux_storage_free (pat_tree->storage, value);

    return flux_storage_free (pat_tree->storage, node_index);
  }
  else
  {
    return flux_storage_free (pat_tree->storage, node_index);
  }
}

static gboolean
read_node (FluxPatTree *pat_tree, FluxItem node_index, PTNode *dest)
{
  return flux_storage_read_full (pat_tree->storage, node_index, dest, 0,
                                 STORAGE_IS_VARLEN (pat_tree) ? sizeof (PTNode) : sizeof (PTNodeFixedLen));
}

static const PTNode *
peek_node (FluxPatTree *pat_tree, FluxItem node_index)
{
  return flux_storage_peek (pat_tree->storage, node_index, 0,
                            STORAGE_IS_VARLEN (pat_tree) ? sizeof (PTNode) : sizeof (PTNodeFixedLen));
}

static gboolean
write_node (FluxPatTree *pat_tree, FluxItem node_index, PTNode *src)
{
  return flux_storage_write_full (pat_tree->storage, node_index, src, 0,
                                  STORAGE_IS_VARLEN (pat_tree) ? sizeof (PTNode) : sizeof (PTNodeFixedLen));
}

static gboolean
read_node_key (FluxPatTree *pat_tree, FluxItem node_index,
               gpointer *key, guint *key_len)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    const PTNode *node;
    FluxItem      key_item;

    /* Get pointers */

    node = flux_storage_peek (pat_tree->storage, node_index, 0, sizeof (PTNode));
    if (!node)
      return FALSE;

    key_item = node->key;

    /* Get key */

    if (key_item == FLUX_ITEM_NULL)
    {
      /* Empty root node or corrupt tree */
      return FALSE;
    }

    *key_len = flux_storage_get_item_len (pat_tree->storage, key_item);
    if (*key_len == 0)
    {
      /* Corrupt tree or I/O error! */
      return FALSE;
    }

    *key = g_malloc (*key_len);
    if (!flux_storage_read (pat_tree->storage, key_item, *key))
    {
      /* I/O error */
      g_free (*key);
      return FALSE;
    }
  }
  else
  {
    *key_len = pat_tree->key_len;
    *key     = g_malloc (*key_len);

    if (!flux_storage_read_full (pat_tree->storage, node_index, *key,
                                 sizeof (PTNodeFixedLen),       *key_len))
    {
      /* I/O error! */
      g_free (*key);
      return FALSE;
    }
  }

  return TRUE;
}

static gboolean
write_node_key (FluxPatTree *pat_tree, FluxItem node_index,
                gconstpointer key, guint key_len)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    const PTNode *node;
    FluxItem      key_item, old_key_item;
    gboolean      can_reuse = FALSE;

    /* Get pointer */

    node = flux_storage_peek (pat_tree->storage, node_index, 0, sizeof (PTNode));
    if (!node)
      return FALSE;

    old_key_item = node->key;

    /* Replace old key */

    if (old_key_item != FLUX_ITEM_NULL)
    {
      /* See if we can re-use old key memory */
      if (flux_storage_get_item_len (pat_tree->storage, old_key_item) >= key_len)
        can_reuse = TRUE;
    }

    if (can_reuse)
    {
      key_item = flux_storage_realloc (pat_tree->storage, old_key_item, key_len);
      if (key_item == FLUX_ITEM_NULL)
        return FALSE;  /* I/O error */
    }
    else
    {
      key_item = flux_storage_alloc (pat_tree->storage, key_len);
      if (key_item == FLUX_ITEM_NULL)
        return FALSE;  /* I/O error */
    }

    if (!flux_storage_write (pat_tree->storage, key_item, key))
    {
      /* I/O error */
      if (!can_reuse)
        flux_storage_free (pat_tree->storage, key_item);
      return FALSE;
    }

    /* If key item location changed, update node head's pointer */

    if (key_item != old_key_item)
    {
      PTNode changed_node;

      if (!read_node (pat_tree, node_index, &changed_node))
        return FALSE;  /* I/O error */

      changed_node.key = key_item;
      if (!write_node (pat_tree, node_index, &changed_node))
        return FALSE;  /* I/O error */
    }

    /* Free old key */

    if (old_key_item != FLUX_ITEM_NULL && !can_reuse)
    {
      flux_storage_free (pat_tree->storage, old_key_item);
    }
  }
  else
  {
    if (!flux_storage_write_full (pat_tree->storage, node_index, key,
                                  sizeof (PTNodeFixedLen), pat_tree->key_len))
      return FALSE;  /* I/O error */
  }

  return TRUE;
}

static gboolean
read_node_value (FluxPatTree *pat_tree, FluxItem node_index,
                 gpointer *value, guint *value_len)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    const PTNode *node;
    FluxItem      value_item;

    /* Get pointers */

    node = flux_storage_peek (pat_tree->storage, node_index, 0, sizeof (PTNode));
    if (!node)
      return FALSE;

    value_item = node->value;

    /* Get key */

    if (value_item == FLUX_ITEM_NULL)
    {
      /* NULL value */
      return FALSE;
    }

    *value_len = flux_storage_get_item_len (pat_tree->storage, value_item);
    if (*value_len == 0)
    {
      /* Corrupt tree or I/O error! */
      return FALSE;
    }

    *value = g_malloc (*value_len);
    if (!flux_storage_read (pat_tree->storage, value_item, *value))
    {
      /* I/O error */
      g_free (*value);
      return FALSE;
    }
  }
  else
  {
    *value_len = pat_tree->value_len;
    *value     = g_malloc (pat_tree->value_len);

    if (!flux_storage_read_full (pat_tree->storage, node_index, *value,
                                 sizeof (PTNodeFixedLen) + pat_tree->key_len, pat_tree->value_len))
    {
      /* I/O error! */
      g_free (*value);
      return FALSE;
    }
  }

  return TRUE;
}

static gboolean
write_node_value (FluxPatTree *pat_tree, FluxItem node_index,
                  gconstpointer value, guint value_len)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    const PTNode *node;
    FluxItem      value_item, old_value_item;
    gboolean      can_reuse = FALSE;

    /* Get pointer */

    node = flux_storage_peek (pat_tree->storage, node_index, 0, sizeof (PTNode));
    if (!node)
      return FALSE;

    old_value_item = node->value;

    /* Replace old value */

    if (value)
    {
      if (old_value_item != FLUX_ITEM_NULL)
      {
        /* See if we can re-use old value memory */
        if (flux_storage_get_item_len (pat_tree->storage, old_value_item) >= value_len)
          can_reuse = TRUE;
      }

      if (can_reuse)
      {
        value_item = flux_storage_realloc (pat_tree->storage, old_value_item, value_len);
        if (value_item == FLUX_ITEM_NULL)
          return FALSE;  /* I/O error */
      }
      else
      {
        value_item = flux_storage_alloc (pat_tree->storage, value_len);
        if (value_item == FLUX_ITEM_NULL)
          return FALSE;  /* I/O error */
      }

      if (!flux_storage_write (pat_tree->storage, value_item, value))
      {
        /* I/O error */
        if (!can_reuse)
          flux_storage_free (pat_tree->storage, value_item);
        return FALSE;
      }
    }
    else  /* value == NULL */
      value_item = FLUX_ITEM_NULL;

    /* If value item location changed, update node head's pointer */

    if (value_item != old_value_item)
    {
      PTNode changed_node;

      if (!read_node (pat_tree, node_index, &changed_node))
        return FALSE;  /* I/O error */

      changed_node.value = value_item;
      if (!write_node (pat_tree, node_index, &changed_node))
        return FALSE;  /* I/O error */
    }

    /* Free old value */

    if (old_value_item != FLUX_ITEM_NULL && !can_reuse)
    {
      flux_storage_free (pat_tree->storage, old_value_item);
    }
  }
  else
  {
    if (!flux_storage_write_full (pat_tree->storage, node_index, value,
                                  sizeof (PTNodeFixedLen) + pat_tree->key_len, pat_tree->value_len))
      return FALSE;  /* I/O error */
  }

  return TRUE;
}

static gboolean
read_node_data (FluxPatTree *pat_tree, FluxItem node_index,
                gpointer *key, guint *key_len,
                gpointer *value, guint *value_len)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    const PTNode *node;
    FluxItem      key_item, value_item;

    /* Get pointers */

    node = flux_storage_peek (pat_tree->storage, node_index, 0, sizeof (PTNode));
    if (!node)
      return FALSE;

    key_item   = node->key;
    value_item = node->value;

    /* Get key */

    if (key_item == FLUX_ITEM_NULL)
    {
      /* Corrupt tree! */
      return FALSE;
    }

    *key_len = flux_storage_get_item_len (pat_tree->storage, key_item);
    if (*key_len == 0)
    {
      /* Corrupt tree or I/O error! */
      return FALSE;
    }

    *key = g_malloc (*key_len);
    if (!flux_storage_read (pat_tree->storage, key_item, *key))
    {
      /* I/O error */
      g_free (*key);
      return FALSE;
    }

    /* Get value, if set */

    if (value_item != FLUX_ITEM_NULL)
    {
      *value_len = flux_storage_get_item_len (pat_tree->storage, value_item);
      if (*value_len == 0)
      {
        /* I/O error! */
        *value = NULL;
        return TRUE;
      }

      *value = g_malloc (*value_len);
      if (!flux_storage_read (pat_tree->storage, value_item, *value))
      {
        /* I/O error! */
        g_free (*value);
        *value_len = 0;
        *value = NULL;
      }
    }
    else
    {
      *value_len = 0;
      *value = NULL;
    }
  }
  else
  {
    *key_len   = pat_tree->key_len;
    *value_len = pat_tree->value_len;

    *key   = g_malloc (*key_len);
    *value = g_malloc (*value_len);

    if (!flux_storage_read_full (pat_tree->storage, node_index, *key,
                                 sizeof (PTNodeFixedLen),            *key_len) ||
        !flux_storage_read_full (pat_tree->storage, node_index, *value,
                                 sizeof (PTNodeFixedLen) + *key_len, *value_len))
    {
      /* I/O error! */
      g_free (*key);
      g_free (*value);
      return FALSE;
    }
  }

  return TRUE;
}

#if 0

static void
dump_index (FluxPatTree *pat_tree, FluxItem index)
{
  PTNode *node;
  PTNode *left;
  PTNode *right;

  node = alloc_node_mem (pat_tree);
  left = alloc_node_mem (pat_tree);
  right = alloc_node_mem (pat_tree);

  read_node (pat_tree, index, node);
  read_node (pat_tree, node->left_child, left);
  read_node (pat_tree, node->right_child, right);

  g_print ("(%d) left=(%d) right=(%d) ibit=%d\n",
           index,
           node->left_child,
           node->right_child,
           node->bit_index);

  free_node_mem (pat_tree, node);
  free_node_mem (pat_tree, left);
  free_node_mem (pat_tree, right);
}

#endif

static void
get_differentiating_key_bit (FluxPatTree *pat_tree,
                             gconstpointer data_0, guint data_0_len,
                             gconstpointer data_1, guint data_1_len,
                             guint32 *diff_bit_index)
{
  guint byte_index;

  if (data_0_len != data_1_len)
  {
    /* We differentiate keys with different lengths with a virtual stop bit
     * at the end of the record */
    *diff_bit_index = MAX (data_0_len, data_1_len) << 3;
    return;
  }

  for (byte_index = MAX (data_0_len, data_1_len) - 1; ; byte_index--)
  {
    guchar d0, d1;

    d0 = byte_index < data_0_len ? ((const guchar *) data_0) [byte_index] : 0;
    d1 = byte_index < data_1_len ? ((const guchar *) data_1) [byte_index] : 0;

    if (d0 != d1)
    {
      guint bit_index = 7;

      /* Scan bits in this char */
      for ( ; (d0 & 0x01) == (d1 & 0x01); d0 >>= 1, d1 >>= 1)
      {
        bit_index--;
      }

      *diff_bit_index = (byte_index << 3) + bit_index;
      return;
    }

    if (!byte_index)
      break;
  }

  g_assert_not_reached ();
}

/* Gets the index where the key will be if we have it; if we don't have it, it can
 * be used as the basis for insertion. */
static FluxItem
get_closest_index (FluxPatTree *pat_tree, FluxItem root_index, gconstpointer key, guint key_len)
{
  FluxItem      p_index, t_index;
  PTNode       *p_node;
  const PTNode *t_node;

  p_node = alloc_node_mem (pat_tree);

  /* Set p to root node */
  p_index = root_index;
  if (!read_node (pat_tree, root_index, p_node))
  {
    /* Corrupt tree or I/O error! */
    t_index = FLUX_ITEM_NULL;
    goto out;
  }

  /* t starts out as root node's left child */
  t_index = (p_node->left_child == FLUX_ITEM_NULL ? root_index : p_node->left_child);
  t_node = peek_node (pat_tree, t_index);
  if (!t_node)
  {
    /* Corrupt tree or I/O error! */
    t_index = FLUX_ITEM_NULL;
    goto out;
  }

  /* Go deeper until we find a node whose bit index is equal to or
   * greater than its parent's bit index. */

  while (NODE_BIT_INDEX_GT (p_node, t_node))
  {
    p_index = t_index;
    memcpy (p_node, t_node, sizeof (PTNode));  /* Node head only, not key or value */

    /* If the key's bit as referenced by this node is TRUE, go right.
     * Else, go left. */

    if (GET_INDEXED_BIT (key, key_len, p_node->bit_index))
      t_index = p_node->right_child;
    else
      t_index = p_node->left_child;

    t_node = peek_node (pat_tree, t_index);
    if (!t_node)
    {
      /* Corrupt tree or I/O error !*/
      t_index = FLUX_ITEM_NULL;
      goto out;
    }
  }

out:
  free_node_mem (pat_tree, p_node);
  return t_index;
}

static FluxItem
find_parent_of (FluxPatTree *pat_tree, FluxItem root_index, FluxItem child_index,
                gconstpointer child_key, guint child_key_len)
{
  FluxItem      p_index;
  const PTNode *p_node;

  p_index = root_index;
  p_node = peek_node (pat_tree, p_index);
  if (!p_node)
  {
    /* Corrupt tree or I/O error! */
    p_index = FLUX_ITEM_NULL;
    goto out;
  }

  while (p_node->left_child != child_index &&
         p_node->right_child != child_index)
  {
    if (GET_INDEXED_BIT (child_key, child_key_len, p_node->bit_index))
      p_index = p_node->right_child;
    else
      p_index = p_node->left_child;

    p_node = peek_node (pat_tree, p_index);
    if (!p_node)
    {
      /* Corrupt tree or I/O error! */
      p_index = FLUX_ITEM_NULL;
      goto out;
    }
  }

out:
  return p_index;
}

static gboolean
get_insertion_data (FluxPatTree *pat_tree, FluxItem root_index,
                    FluxItem closest_index, gconstpointer key, guint key_len,
                    FluxItem *parent_index, FluxItem *child_index,
                    guint32 *diff_bit_index)
{
  FluxItem      p_index, x_index;
  PTNode       *p_node;
  const PTNode *x_node;
  const PTNode *closest_node;
  guint32       d_bit_index;
  gboolean      success = FALSE;
  gpointer      closest_key;
  guint         closest_key_len;

  p_node = alloc_node_mem (pat_tree);

  /* Get the key data for the closest match node */
  closest_node = peek_node (pat_tree, closest_index);
  if (!closest_node)
    goto out;  /* Corrupt tree or I/O error! */

  /* Calculate byte/bit indexes differentiating key from closest_key */
  if (closest_index == root_index && !ROOT_NODE_IS_ACTIVE (closest_node))
  {
    closest_key     = NULL;
    closest_key_len = 0;
  }
  else if (!read_node_key (pat_tree, closest_index, &closest_key, &closest_key_len))
    goto out;  /* Empty root node or corrupt tree */

  get_differentiating_key_bit (pat_tree, key, key_len, closest_key, closest_key_len, &d_bit_index);

  if (closest_key)
    g_free (closest_key);

  /* Set p to root node */
  p_index = root_index;
  if (!read_node (pat_tree, p_index, p_node))
    goto out;  /* Corrupt tree or I/O error! */

  /* x starts out as root node's left child */
  x_index = (p_node->left_child == FLUX_ITEM_NULL ? root_index : p_node->left_child);
  x_node = peek_node (pat_tree, x_index);
  if (!x_node)
    goto out;  /* Corrupt tree or I/O error! */

  while (NODE_BIT_INDEX_GT (p_node, x_node) &&
         BIT_INDEX_GT (x_node->bit_index, d_bit_index))
  {
    p_index = x_index;
    memcpy (p_node, x_node, sizeof (PTNodeFixedLen));  /* Node head only, not key or value */

    if (GET_INDEXED_BIT (key, key_len, p_node->bit_index))
      x_index = p_node->right_child;
    else
      x_index = p_node->left_child;

    x_node = peek_node (pat_tree, x_index);
    if (!x_node)
      goto out;  /* Corrupt tree or I/O error! */
  }

  *parent_index    = p_index;
  *child_index     = x_index;
  *diff_bit_index  = d_bit_index;

  success = TRUE;

out:
  free_node_mem (pat_tree, p_node);
  return success;
}

static gboolean
insert_node (FluxPatTree *pat_tree,
             gconstpointer key, guint key_len,
             gconstpointer value, guint value_len,
             FluxItem new_index,
             FluxItem parent_index, FluxItem child_index,
             guint32 diff_bit_index)
{
  PTNode   *new_node, *parent_node;
  gboolean  success = FALSE;

  new_node    = alloc_node_mem (pat_tree);
  parent_node = alloc_node_mem (pat_tree);

  /* Read and prepare */

  if (!read_node (pat_tree, parent_index, parent_node))
    goto out;  /* Corrupt tree or I/O error! */

  memset (new_node, 0, sizeof (PTNode));
  new_node->bit_index = diff_bit_index;
  new_node->key   = FLUX_ITEM_NULL;
  new_node->value = FLUX_ITEM_NULL;

#if 0
  memcpy (PT_NODE_KEY (pat_tree, new_node), key, pat_tree->key_len);
  if (value)
    memcpy (PT_NODE_VALUE (pat_tree, new_node), value, pat_tree->value_len);
#endif

  if (GET_INDEXED_BIT (key, key_len, diff_bit_index))
  {
    new_node->left_child  = child_index;
    new_node->right_child = new_index;
  }
  else
  {
    new_node->left_child  = new_index;
    new_node->right_child = child_index;
  }

  if (GET_INDEXED_BIT (key, key_len, parent_node->bit_index))
  {
    parent_node->right_child = new_index;
  }
  else
  {
    parent_node->left_child  = new_index;
  }

  /* Commit changes */

  if (!write_node (pat_tree, new_index, new_node))
    goto out;  /* I/O error! */
  if (!write_node_key (pat_tree, new_index, key, key_len))
    goto out;  /* I/O error! */
  if (value && !write_node_value (pat_tree, new_index, value, value_len))
    goto out;  /* I/O error! */
  if (!write_node (pat_tree, parent_index, parent_node))
    goto out;  /* I/O error! The new node will exist physically, but won't be linked in */

  success = TRUE;

out:
  free_node_mem (pat_tree, parent_node);
  free_node_mem (pat_tree, new_node);
  return success;
}

static gboolean
compare_key_with_node_key (FluxPatTree *pat_tree, FluxItem node_index,
                           gconstpointer key, guint key_len)
{
  gpointer node_key;
  guint    node_key_len;
  gboolean result;

  if (!read_node_key (pat_tree, node_index, &node_key, &node_key_len))
    return FALSE;

  if (key_len != node_key_len)
    result = FALSE;
  else if (memcmp (key, node_key, key_len))
    result = FALSE;
  else
    result = TRUE;

  g_free (node_key);
  return result;
}

static gboolean
compare_key_with_node_key_masked (FluxPatTree *pat_tree, FluxItem node_index,
                                  gconstpointer key, gconstpointer mask, guint key_len)
{
  gpointer node_key;
  guint    node_key_len;
  gboolean result = TRUE;
  guint    i;

  if (!read_node_key (pat_tree, node_index, &node_key, &node_key_len))
    return FALSE;

  for (i = 0; i < key_len; i++)
  {
    if ((((guchar *) key)      [i] & ((guchar *) mask) [i]) !=
        (((guchar *) node_key) [i] & ((guchar *) mask) [i]))
    {
      result = FALSE;
      break;
    }
  }

  g_free (node_key);
  return result;  
}

/* Updates stack and returns key index */
static FluxItem
traverse_down (FluxPatTreeCursor *cursor, FluxItem start_index)
{
  FluxItem      p_index, t_index;
  PTNode       *p_node;
  const PTNode *t_node;

  DEBUG (g_print ("left\n"));

  p_node = alloc_node_mem (cursor->pat_tree);

  /* XXXXX: Now that we're checking bits, we might have to check the first bit too */

  /* Set t to start node */
  if (start_index == FLUX_ITEM_NULL)
    t_index = p_index = stack_item_last (cursor->index_stack)->item;
  else
    t_index = p_index = start_index;

  t_node = peek_node (cursor->pat_tree, t_index);
  if (!t_node)
  {
    /* Corrupt tree! Initial node can't be read; return NULL */
    free_node_mem (cursor->pat_tree, p_node);
    return FLUX_ITEM_NULL;
  }

  do
  {
    p_index = t_index;
    memcpy (p_node, t_node, sizeof (PTNode));  /* Node head only, not key or value */

    if ((cursor->find_mask && GET_INDEXED_BIT (cursor->find_mask, cursor->key_len, p_node->bit_index)) ||
        (p_node->bit_index / 8 > cursor->key_len && FALSE /* mask_extend_bit == TRUE */))
    {
      /* Key bit matters; do one of the subtrees depending on key bit. Since we're
       * only doing one subtree, don't push on stack */

      if (GET_INDEXED_BIT (cursor->find_key, cursor->key_len, p_node->bit_index))
      {
        DEBUG (g_print ("key right\n"));
        t_index = p_node->right_child;
      }
      else
      {
        DEBUG (g_print ("key left\n"));
        t_index = p_node->left_child;
      }

      t_node = peek_node (cursor->pat_tree, t_index);
      if (!t_node)
      {
        /* Corrupt tree! We can't follow the branch we want; return NULL */
        t_index = FLUX_ITEM_NULL;
        break;
      }
    }
    else
    {
      gboolean go_left;

      /* Key bit doesn't matter; do both subtrees (save one for later) */

      if (cursor->randomize)
        go_left = g_random_boolean ();
      else
        go_left = TRUE;

      if (go_left)
      {
        /* Go left */
        DEBUG (g_print ("any left\n"));
        t_index = p_node->left_child;
        t_node  = peek_node (cursor->pat_tree, t_index);
        if (!t_node)
        {
          /* Corrupt tree! Skip the left branch and do the right one instead */
          t_index = p_node->right_child;
          t_node  = peek_node (cursor->pat_tree, t_index);
          if (!t_node)
          {
            /* Corrupt tree! Neither branch can be followed; return NULL */
            t_index = FLUX_ITEM_NULL;
            break;
          }
        }
        else
        {
          /* Put index on stack, so we can come back to this branch point and do the right subtree too */
          stack_item_append (cursor->index_stack, p_index, TRUE, FALSE);
        }
      }
      else
      {
        /* Go right */
        DEBUG (g_print ("any right\n"));
        t_index = p_node->right_child;
        t_node  = peek_node (cursor->pat_tree, t_index);
        if (!t_node)
        {
          /* Corrupt tree! Skip the right branch and do the left one instead */
          t_index = p_node->left_child;
          t_node  = peek_node (cursor->pat_tree, t_index);
          if (!t_node)
          {
            /* Corrupt tree! Neither branch can be followed; return NULL */
            t_index = FLUX_ITEM_NULL;
            break;
          }
        }
        else
        {
          /* Put index on stack, so we can come back to this branch point and do the left subtree too */
          stack_item_append (cursor->index_stack, p_index, FALSE, TRUE);
        }
      }
    }
  }
  while (NODE_BIT_INDEX_GT (p_node, t_node));

  free_node_mem (cursor->pat_tree, p_node);
  return t_index;
}

/* Returns index of right node, removing its parent from the stack. Returns TRUE if
 * new_index is through a downwards link, or FALSE if we went up, and new_index is a key.
 * If the tree is corrupt, new_index will be set to FLUX_ITEM_NULL, in which case
 * the return value doesn't matter. */
static gboolean
traverse_up (FluxPatTreeCursor *cursor, FluxItem *new_index)
{
  FluxItem      t_index;
  PTNode       *p_node;
  const PTNode *t_node;
  gboolean      result = FALSE;
  gboolean      got_node;

  DEBUG (g_print ("right\n"));

  p_node = alloc_node_mem (cursor->pat_tree);

  /* Repeat until we find a stack entry that points to a non-corrupt node */
  do
  {
    gboolean went_left;

    /* Get parent from stack, and pop it, since after going down from it for the 2nd time,
     * we don't need it anymore */
    t_index = stack_item_last (cursor->index_stack)->item;
    went_left = stack_item_last (cursor->index_stack)->went_left;
    stack_item_remove_last (cursor->index_stack);

    /* Read parent */
    got_node = read_node (cursor->pat_tree, t_index, p_node);
    if (got_node)
    {
      if (went_left)
        t_index = p_node->right_child;
      else
        t_index = p_node->left_child;

      t_node = peek_node (cursor->pat_tree, t_index);

      /* See if we went up or down in the tree */
      if (t_node)
      {
        if (NODE_BIT_INDEX_GT (p_node, t_node))
          result = TRUE;
        else
          result = FALSE;

        DEBUG (g_print ("%d vs %d\n", p_node->bit_index, t_node->bit_index));
        *new_index = t_index;
      }
      else
      {
        /* Corrupt tree! Try next stack entry... */
        got_node = FALSE;
      }
    }
  }
  while (!got_node && cursor->index_stack->len > 0);

  if (!got_node)
    *new_index = FLUX_ITEM_NULL;

  free_node_mem (cursor->pat_tree, p_node);
  return result;
}

static gboolean
traverse_next (FluxPatTreeCursor *cursor)
{
  gboolean result = FALSE;
  FluxItem index;

  DEBUG (g_print ("next\n"));

  while (result == FALSE)
  {
    /* If the last direction was none, go to leftmost node (this only happens
     * the first time, i.e. at the root node) */
    if (cursor->started == FALSE)
    {
      FluxItem      root_index;
      const PTNode *root_node;

      root_index = g_array_index (cursor->index_stack, StackItem, 0).item;

      cursor->key_index = traverse_down (cursor, FLUX_ITEM_NULL);
      cursor->started = TRUE;

      if (cursor->key_index == FLUX_ITEM_NULL)
      {
        /* Corrupt tree! */
        result = FALSE;
        break;
      }

      if (cursor->key_index == root_index)
      {
        /* Special case root node */
        root_node = peek_node (cursor->pat_tree, root_index);
        if (!root_node)
        {
          /* Corrupt tree! */
          result = FALSE;
          break;
        }

        if (ROOT_NODE_IS_ACTIVE (root_node))
        {
          DEBUG (g_print ("active root node\n"));
          result = TRUE;
        }
        else
        {
          DEBUG (g_print ("inactive root node\n"));
          result = FALSE;  /* For clarity */
        }
      }
      else
      {
        DEBUG (g_print ("ok initial down\n"));
        result = TRUE;
      }
    }

    /* If the last direction was right, go up to the last left branch, then
     * go right from that, then go all the way left */
    if (result == FALSE && cursor->index_stack->len > 1)
    {
      gboolean went_down;

      went_down = traverse_up (cursor, &index);
      if (index == FLUX_ITEM_NULL)
      {
        /* Corrupt tree! */
        result = FALSE;
        break;
      }

      if (went_down)
      {
        FluxItem new_index = traverse_down (cursor, index);

        if (new_index == FLUX_ITEM_NULL)
        {
          /* Corrupt tree! Try other items on stack */
          result = FALSE;
          continue;
        }

        cursor->key_index = new_index;
        DEBUG (g_print ("went down\n"));
      }
      else
      {
        cursor->key_index = index;
        DEBUG (g_print ("went up\n"));
      }

      result = TRUE;
    }
    else
    {
      /* A) We're sitting at the root node and the traversal is done, or
       * B) The initial down succeeded. */
      DEBUG (g_print ("traversal done\n"));
    }

    if (result == TRUE)
    {
      /* Make sure key really matches the wildcard */
#if 0
      key_node = peek_node (cursor->pat_tree, cursor->key_index);
      if (!key_node)
      {
        /* Corrupt tree! Try other keys */
        result = FALSE;
        continue;
      }

      for (i = 0; i < cursor->pat_tree->key_len; i++)
      {
        if ((((guchar *) cursor->find_key) [i]                         & ((guchar *) cursor->find_mask) [i]) !=
            (((guchar *) PT_NODE_KEY (cursor->pat_tree, key_node)) [i] & ((guchar *) cursor->find_mask) [i]))
        {
          result = FALSE;
          break;  /* Break out of inner loop and proceed to next key */
        }
      }
#endif

      result = compare_key_with_node_key_masked (cursor->pat_tree, cursor->key_index,
                                                 cursor->find_key, cursor->find_mask, cursor->key_len);
    }
    else
      break;
  }

  if (result == TRUE)
    cursor->index_invalidated = FALSE;

  return result;
}

static FluxItem
find_upwards_link (FluxPatTree *pat_tree, FluxItem start_index,
                   gconstpointer start_key, guint start_key_len,
                   FluxItem *parent_index, PTNode *result_node)
{
  FluxItem      search_index;
  gboolean      found;
  PTNode       *start_node;
  const PTNode *search_node;

  start_node = alloc_node_mem (pat_tree);

  if (!read_node (pat_tree, start_index, start_node))
  {
    /* Corrupt tree or I/O error! */
    search_index = FLUX_ITEM_NULL;
    goto out;
  }

  search_node = start_node;

  /* Find node with upwards pointer to us */
  for (search_index = start_index, found = FALSE; !found;
       search_node = peek_node (pat_tree, search_index))
  {
    if (!search_node)
    {
      /* Corrupt tree or I/O error! */
      search_index = FLUX_ITEM_NULL;
      goto out;
    }

    if (search_node->left_child  == start_index ||
        search_node->right_child == start_index)
    {
      found = TRUE;
    }
    else if (GET_INDEXED_BIT (start_key, start_key_len, search_node->bit_index))
    {
      if (parent_index)
        *parent_index = search_index;
      search_index = search_node->right_child;
    }
    else
    {
      if (parent_index)
        *parent_index = search_index;
      search_index = search_node->left_child;
    }

    if (search_index == FLUX_ITEM_NULL)
      goto out;  /* Corrupt tree! */
  }

  if (result_node)
    memcpy (result_node, search_node, sizeof (PTNode));

out:
  free_node_mem (pat_tree, start_node);
  return search_index;
}

static gboolean
replace_upwards_link (FluxPatTree *pat_tree, FluxItem start_index,
                      gconstpointer start_key, guint start_key_len, FluxItem replacement_index)
{
  FluxItem  decision_index;
  PTNode   *decision_node;
  gboolean  success = FALSE;

  decision_node = alloc_node_mem (pat_tree);

  /* Read and prepare */

  decision_index = find_upwards_link (pat_tree, start_index, start_key, start_key_len,
                                      NULL, decision_node);
  if (decision_index == FLUX_ITEM_NULL)
    goto out;  /* Corrupt tree or I/O error! */

  if (decision_node->left_child == start_index)
    decision_node->left_child = replacement_index;
  else
    decision_node->right_child = replacement_index;

  /* Commit changes */

  if (!write_node (pat_tree, decision_index, decision_node))
    goto out;  /* I/O error! No changes made. */

  success = TRUE;

out:
  free_node_mem (pat_tree, decision_node);
  return success;
}

static gboolean
unlink_node_from_parent (FluxPatTree *pat_tree, FluxItem root_index,
                         FluxItem parent_index, FluxItem unlink_index,
                         gconstpointer unlink_key, guint unlink_key_len,
                         FluxItem lose_index)
{
  FluxItem  child_index;
  PTNode   *parent_node, *unlink_node;
  gboolean  success = FALSE;

  parent_node = alloc_node_mem (pat_tree);
  unlink_node = alloc_node_mem (pat_tree);

  /* Read and prepare */

  if (!read_node (pat_tree, unlink_index, unlink_node))
    goto out;  /* Corrupt tree or I/O error! */

  if (parent_index == FLUX_ITEM_NULL)
  {
    parent_index = find_parent_of (pat_tree, root_index, unlink_index, unlink_key, unlink_key_len);
  }

  if (parent_index == FLUX_ITEM_NULL)
    goto out;  /* Corrupt tree or I/O error! */

  if (!read_node (pat_tree, parent_index, parent_node))
    goto out;  /* Corrupt tree or I/O error! */

  child_index = (unlink_node->left_child == lose_index ?
                 unlink_node->right_child : unlink_node->left_child);

  if (parent_node->left_child == unlink_index)
    parent_node->left_child = child_index;
  else
    parent_node->right_child = child_index;

  /* Commit changes */

  if (!write_node (pat_tree, parent_index, parent_node))
    goto out;  /* I/O error! No changes made. */

  success = TRUE;

out:
  free_node_mem (pat_tree, unlink_node);
  free_node_mem (pat_tree, parent_node);
  return success;
}

static gboolean
reassign_data (FluxPatTree *pat_tree, FluxItem src_index, FluxItem dest_index)
{
  if (STORAGE_IS_VARLEN (pat_tree))
  {
    PTNode   src_node, dest_node;
    FluxItem old_key, old_value;

    /* Varlen storage; copy the data pointers */

    /* Read headers */
    if (!read_node (pat_tree, src_index, &src_node))
      return FALSE;  /* I/O error */
    if (!read_node (pat_tree, dest_index, &dest_node))
      return FALSE;  /* I/O error */

    /* Move src's data to dest */
    old_key   = dest_node.key;
    old_value = dest_node.value;

    dest_node.key   = src_node.key;
    dest_node.value = src_node.value;

    src_node.key   = FLUX_ITEM_NULL;
    src_node.value = FLUX_ITEM_NULL;

    /* Write headers */
    if (!write_node (pat_tree, src_index, &src_node))
      return FALSE;  /* I/O error */
    if (!write_node (pat_tree, dest_index, &dest_node))
      return FALSE;  /* I/O error */

    /* Free dest's data */
    if (old_key != FLUX_ITEM_NULL)
      flux_storage_free (pat_tree->storage, old_key);
    if (old_value != FLUX_ITEM_NULL)
      flux_storage_free (pat_tree->storage, old_value);
  }
  else
  {
    gpointer data;
    guint    data_len;

    /* Fixed-len storage; copy the data itself */

    if (!read_node_key (pat_tree, src_index, &data, &data_len))
      return FALSE;
    if (!write_node_key (pat_tree, dest_index, data, data_len))
      return FALSE;

    g_free (data);

    if (!read_node_value (pat_tree, src_index, &data, &data_len))
    {
      data = NULL;
      data_len = 0;
    }

    write_node_value (pat_tree, dest_index, data, data_len);
    if (data)
      g_free (data);
  }

  return TRUE;
}

static void
cursors_update_removal (FluxPatTree *pat_tree, FluxItem delete_index, FluxItem replace_index)
{
  GList     *l;
  StackItem *stack_item;

  for (l = pat_tree->cursors; l; l = g_list_next (l))
  {
    FluxPatTreeCursor *cursor = l->data;

    if (cursor->key_index == delete_index)
    {
      if (replace_index != FLUX_ITEM_NULL)
      {
        /* Our key replaced a deleted key */
        cursor->key_index = replace_index;
      }
      else
      {
        /* Our key was deleted */
        cursor->index_invalidated = TRUE;
      }
    }
    else if (cursor->key_index == replace_index)
    {
      /* Our key was deleted (replaced by another key) */
      cursor->index_invalidated = TRUE;
    }

    stack_item = stack_item_find (cursor->index_stack, delete_index);
    if (stack_item)
    {
      PTNode   delete_node;
      gboolean went_left;
      FluxItem child_index;

      went_left = stack_item->went_left;
      stack_item_remove (cursor->index_stack, stack_item->item);

#if 0
      g_print ("Deletion:\n");

      g_print ("  delete_index  ==  %d\n", delete_index);
      if (delete_index != FLUX_ITEM_NULL)
        { g_print ("                   "); dump_index (pat_tree, delete_index); }
      g_print ("  replace_index ==  %d\n", replace_index);
      if (replace_index != FLUX_ITEM_NULL)
        { g_print ("                   "); dump_index (pat_tree, replace_index); }
      g_print ("  key_index     ==  %d\n", cursor->key_index);
      if (cursor->key_index != FLUX_ITEM_NULL)
        { g_print ("                   "); dump_index (pat_tree, cursor->key_index); }

      g_print ("  stacked_index ==  %d\n", delete_index);
#endif

      if (!read_node (pat_tree, delete_index, &delete_node))
        return;  /* Corrupt tree or I/O error */

      if (went_left)
        child_index = delete_node.right_child;
      else
        child_index = delete_node.left_child;

      if (child_index != delete_index && child_index != FLUX_ITEM_NULL)
      {
        const PTNode *child_node;

        child_node = peek_node (pat_tree, child_index);
        if (child_node && NODE_BIT_INDEX_GT (&delete_node, child_node))
          traverse_down (cursor, child_index);
#if 0
        g_print ("  delete->r     ==  %d\n", delete_node->right_child);
        g_print ("                   "); dump_index (pat_tree, delete_node->right_child);
#endif
      }

#if 0
      if (array_has_duplicates (cursor->index_stack))
        g_error ("Cursor stack has duplicate references!");
#endif
    }
  }
}

static gboolean
remove_key_at_index (FluxPatTree *pat_tree, FluxItem root_index,
                     FluxItem remove_index, FluxItem parent_index,
                     gconstpointer remove_key, guint remove_key_len)
{
  FluxItem  search_index, search_parent_index = FLUX_ITEM_NULL;
  PTNode   *remove_node, *node;
  gboolean  success = FALSE;

  remove_node = alloc_node_mem (pat_tree);
  node        = alloc_node_mem (pat_tree);

  if (!read_node (pat_tree, remove_index, remove_node))
    goto out;  /* Corrupt tree! */

  search_index = find_upwards_link (pat_tree, remove_index, remove_key, remove_key_len,
                                    &search_parent_index, NULL);
  if (!read_node (pat_tree, search_index, node))
    goto out;  /* Corrupt tree! */

  if (search_index == remove_index)
  {
    /* We hold the upwards pointer to ourselves. The other pointer
     * should be assigned to our parent, replacing the pointer to us,
     * and we can free ourselves. */

    /* Fix up running cursors */
    cursors_update_removal (pat_tree, remove_index, FLUX_ITEM_NULL);

    /* Unlink from tree */
    if (!unlink_node_from_parent (pat_tree, root_index, search_parent_index, remove_index,
                                  remove_key, remove_key_len, remove_index))
      goto out;  /* Corrupt tree or I/O error! */

    free_node_disk (pat_tree, remove_index);
    success = TRUE;
  }
  else
  {
    gpointer search_key;
    guint    search_key_len;

    /* Another node holds the upwards pointer to us. Let that node's data
     * replace ours, and do a recursive run to see what should replace
     * that node. */

    read_node_key (pat_tree, search_index, &search_key, &search_key_len);

    /* Fix up running cursors */
    cursors_update_removal (pat_tree, search_index, remove_index);

    /* Replace data to be deleted with data from decision node */
    reassign_data (pat_tree, search_index, remove_index);

    /* The original data is gone. There is still some cleanup to do, however. */
    success = TRUE;

    /* Change upwards link pointing to decision node to point to data node */
    if (!replace_upwards_link (pat_tree, search_index, search_key, search_key_len, remove_index))
      goto out;  /* Corrupt tree or I/O error! Deletion succeeded, but we can't go through with cleanup */

    /* Unlink decision node from parent (link past it) */
    if (!unlink_node_from_parent (pat_tree, root_index, search_parent_index, search_index,
                                  search_key, search_key_len, remove_index))
      goto out;  /* Corrupt tree or I/O error! Deletion succeeded, but we can't free */

    free_node_disk (pat_tree, search_index);

    g_free (search_key);
  }

out:
  free_node_mem (pat_tree, node);
  free_node_mem (pat_tree, remove_node);
  return success;
}

FluxItem
flux_pat_tree_create_root (FluxPatTree *pat_tree)
{
  FluxItem  node_index;
  PTNode   *node;

  node_index = alloc_node_disk (pat_tree);
  if (node_index == FLUX_ITEM_NULL)
    return FLUX_ITEM_NULL;

  node = alloc_node_mem (pat_tree);

  node->bit_index   = HIGHEST_BIT_INDEX;
  node->active      = FALSE;
  node->left_child  = node_index;
  node->right_child = node_index;

  /* Stored for varlen only */
  node->key   = FLUX_ITEM_NULL;
  node->value = FLUX_ITEM_NULL;

  if (!write_node (pat_tree, node_index, node))
    node_index = FLUX_ITEM_NULL;  /* I/O error! TODO: Free node. */
  else if (!STORAGE_IS_VARLEN (pat_tree))
  {
    /* NOTE: I don't think we need to clear key/value */
  }

  free_node_mem (pat_tree, node);
  return node_index;
}

gboolean
flux_pat_tree_insert (FluxPatTree *pat_tree, FluxItem root_index,
                      gconstpointer key, guint key_len,
                      gconstpointer value, guint value_len)
{
  FluxItem     closest_index;
  FluxItem     parent_index;
  FluxItem     child_index;
  FluxItem     new_index;
  PTNode      *closest_node;
  guint32      diff_bit_index;
  gboolean     result = FALSE;

  assert_types (pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (key_len > 0, FALSE);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  closest_node  = alloc_node_mem (pat_tree);
  closest_index = get_closest_index (pat_tree, root_index, key, key_len);
  if (closest_index == FLUX_ITEM_NULL)
    goto out;  /* Corrupt tree or I/O error! */

  if (!read_node (pat_tree, closest_index, closest_node))
    goto out;  /* Corrupt tree or I/O error! */

  if (!compare_key_with_node_key (pat_tree, closest_index, key, key_len))
  {
    /* Key did not exist previously, so create a new node */

    if (!get_insertion_data (pat_tree, root_index, closest_index, key, key_len,
                             &parent_index, &child_index, &diff_bit_index))
      goto out;  /* Corrupt tree or I/O error! */

    DEBUG (g_print ("insert: root=%d closest=%d parent=%d child=%d diff_bit=%d\n",
                    root_index, closest_index, parent_index, child_index, diff_bit_index));

    new_index = alloc_node_disk (pat_tree);
    if (new_index == FLUX_ITEM_NULL)
      goto out;  /* I/O error! */

    result = insert_node (pat_tree, key, key_len, value, value_len,
                          new_index, parent_index, child_index, diff_bit_index);
    if (!result)
    {
      /* Corrupt tree or I/O error! We won't need the node we allocated */
      free_node_disk (pat_tree, new_index);
    }
  }
  else
  {
    /* Overwrite old value */

    if (closest_index == root_index)
    {
      ROOT_NODE_SET_ACTIVE (closest_node, TRUE);
      if (!write_node (pat_tree, closest_index, closest_node))
        goto out;
    }

    if (write_node_key (pat_tree, closest_index, key, key_len))
      result = write_node_value (pat_tree, closest_index, value, value_len);
    else
      goto out;
  }

out:
  free_node_mem (pat_tree, closest_node);
  return result;
}

gboolean
flux_pat_tree_replace_closest (FluxPatTree *pat_tree, FluxItem root_index,
                               gconstpointer key, guint key_len,
                               gconstpointer value, guint value_len)
{
  FluxItem  closest_index;
  PTNode   *closest_node;
  gboolean  result = FALSE;

  assert_types (pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  closest_node  = alloc_node_mem (pat_tree);
  closest_index = get_closest_index (pat_tree, root_index, key, key_len);
  if (closest_index == FLUX_ITEM_NULL)
    goto out;  /* Corrupt tree or I/O error! */

  if (!read_node (pat_tree, closest_index, closest_node))
    goto out;  /* Corrupt tree or I/O error! */

  /* Overwrite old key and value */

  if (closest_index == root_index)
  {
    ROOT_NODE_SET_ACTIVE (closest_node, TRUE);
    if (!write_node (pat_tree, closest_index, closest_node))
      goto out;  /* I/O error! No changes made. */
  }

  if (!write_node_key (pat_tree, closest_index, key, key_len))
    goto out;
  if (!write_node_value (pat_tree, closest_index, value, value_len))
    goto out;

  result = TRUE;

out:
  free_node_mem (pat_tree, closest_node);
  return result;
}

gboolean
flux_pat_tree_find (FluxPatTree *pat_tree, FluxItem root_index,
                    gconstpointer key, guint key_len,
                    gpointer *value, guint *value_len)
{
  FluxItem  closest_index;
  PTNode   *closest_node;
  gboolean  result = FALSE;

  assert_types (pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  closest_node  = alloc_node_mem (pat_tree);
  closest_index = get_closest_index (pat_tree, root_index, key, key_len);
  if (closest_index == FLUX_ITEM_NULL)
    goto out;  /* Corrupt tree or I/O error! */

  if (!read_node (pat_tree, closest_index, closest_node))
    goto out;  /* Corrupt tree or I/O error! */

  if (compare_key_with_node_key (pat_tree, closest_index, key, key_len) &&
      (closest_index != root_index || ROOT_NODE_IS_ACTIVE (closest_node)))
  {
    result = read_node_value (pat_tree, closest_index, value, value_len);
  }

out:
  free_node_mem (pat_tree, closest_node);
  return result;
}

gboolean
flux_pat_tree_find_closest (FluxPatTree *pat_tree, FluxItem root_index,
                            gconstpointer key, guint key_len,
                            gpointer *closest_key, guint *closest_key_len,
                            gpointer *closest_value, guint *closest_value_len)
{
  FluxItem  closest_index;
  PTNode   *closest_node;
  gboolean  result = FALSE;

  assert_types (pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  closest_node  = alloc_node_mem (pat_tree);
  closest_index = get_closest_index (pat_tree, root_index, key, key_len);
  if (closest_index == FLUX_ITEM_NULL)
    goto out;  /* Corrupt tree or I/O error! */

  if (!read_node (pat_tree, closest_index, closest_node))
    goto out;  /* Corrupt tree or I/O error! */

  if (closest_index != root_index || ROOT_NODE_IS_ACTIVE (closest_node))
  {
    if (closest_key && !read_node_key (pat_tree, closest_index, closest_key, closest_key_len))
      goto out;

    if (closest_value)
      read_node_value (pat_tree, closest_index, closest_value, closest_value_len);

    result = TRUE;
  }
  else
  {
    /* FIXME: We'll return FALSE if the user searches for the zero key, and it
     * doesn't exist. With some special casing, we could return the closest key
     * if we have at least one key in the tree */
    result = FALSE;
  }

out:
  free_node_mem (pat_tree, closest_node);
  return result;
}

gboolean
flux_pat_tree_remove (FluxPatTree *pat_tree, FluxItem root_index, gconstpointer key, guint key_len)
{
  FluxItem  closest_index;
  PTNode   *closest_node;
  gboolean  result;

  assert_types (pat_tree);
  g_return_val_if_fail (key != NULL, FALSE);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  closest_node  = alloc_node_mem (pat_tree);
  closest_index = get_closest_index (pat_tree, root_index, key, key_len);

  if (closest_index != FLUX_ITEM_NULL /* If fail, tree is corrupt */ &&
      read_node (pat_tree, closest_index, closest_node) /* If fail, tree is corrupt */ &&
      compare_key_with_node_key (pat_tree, closest_index, key, key_len))
  {
    if (closest_index == root_index)
    {
      ROOT_NODE_SET_ACTIVE (closest_node, FALSE);
      result = write_node (pat_tree, closest_index, closest_node);
    }
    else
    {
      result = remove_key_at_index (pat_tree, root_index, closest_index, FLUX_ITEM_NULL, key, key_len);
    }

    /* If result == FALSE here, tree is corrupt or a disk operation failed */
  }
  else
    result = FALSE;

  free_node_mem (pat_tree, closest_node);
  return result;
}

gboolean
flux_pat_tree_get_random (FluxPatTree *pat_tree, FluxItem root_index,
                          gpointer *key, guint *key_len, gpointer *value, guint *value_len)
{
  FluxItem      p_index, t_index;
  PTNode       *p_node;
  const PTNode *t_node;
  gboolean      root_is_active;

  assert_types (pat_tree);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  p_node = alloc_node_mem (pat_tree);

  /* Set p to root node */
  p_index = root_index;
  if (!read_node (pat_tree, root_index, p_node))
  {
    /* Corrupt tree! */
    free_node_mem (pat_tree, p_node);
    return FALSE;
  }

  root_is_active = ROOT_NODE_IS_ACTIVE (p_node);

  /* t starts out as root node's left child */
  t_index = (p_node->left_child == FLUX_ITEM_NULL ? root_index : p_node->left_child);
  if (t_index == p_index)
  {
    /* We have root node only */
    if (root_is_active)
    {
      t_node = p_node;
    }
    else
    {
      free_node_mem (pat_tree, p_node);
      return FALSE;
    }
  }
  else
  {
    t_node = peek_node (pat_tree, t_index);
    if (!t_node)
    {
      /* Corrupt tree! */
      free_node_mem (pat_tree, p_node);
      return FALSE;
    }

    /* Go deeper until we find a node whose bit index is equal to or
     * greater than its parent's bit index. */

    while (NODE_BIT_INDEX_GT (p_node, t_node))
    {
      p_index = t_index;
      memcpy (p_node, t_node, sizeof (PTNode));  /* Node head only, not key or value */

      /* Follow a random path */

      if (p_node->left_child == root_index && !root_is_active)
        t_index = p_node->right_child;
      else if (p_node->right_child == root_index && !root_is_active)
        t_index = p_node->left_child;
      else if (g_random_boolean ())
        t_index = p_node->right_child;
      else
        t_index = p_node->left_child;

      t_node = peek_node (pat_tree, t_index);
      if (!t_node)
      {
        /* Corrupt tree! */
        free_node_mem (pat_tree, p_node);
        return FALSE;
      }
    }
  }

  if (key)
  {
    if (!read_node_key (pat_tree, t_index, key, key_len))
      return FALSE;
  }
  if (value)
  {
    if (!read_node_key (pat_tree, t_index, value, value_len))
      return FALSE;
  }

  free_node_mem (pat_tree, p_node);
  return TRUE;
}

FluxPatTreeCursor *
flux_pat_tree_find_with_mask (FluxPatTree *pat_tree, FluxItem root_index,
                              gconstpointer key, gconstpointer key_mask, guint key_len)
{
  FluxPatTreeCursor *cursor;

  assert_types (pat_tree);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  cursor = cursor_new (pat_tree, root_index, key_len);

  if (key)
  {
    cursor->find_key = g_memdup (key, key_len);
  }
  else if (key_len)
  {
    /* FIXME: What if key_len == 0? */
    cursor->find_key = g_malloc (key_len);
    memset (cursor->find_key, 0, key_len);
  }

  if (key_mask)
  {
    cursor->find_mask = g_memdup (key_mask, key_len);
  }
  else if (key_len)
  {
    cursor->find_mask = g_malloc (key_len);
    memset (cursor->find_mask, 0, key_len);
  }

  if (!traverse_next (cursor))
  {
    flux_object_unref (FLUX_OBJECT (cursor));
    return NULL;
  }

  return cursor;
}

FluxPatTreeCursor *
flux_pat_tree_traverse (FluxPatTree *pat_tree, FluxItem root_index)
{
  assert_types (pat_tree);

  if (root_index == FLUX_ITEM_NULL)
    root_index = pat_tree->default_root;

  return flux_pat_tree_find_with_mask (pat_tree, root_index, NULL, NULL, 0);
}

#if 0

guint
flux_pat_tree_get_n_records (FluxPatTree *pat_tree)
{
  gint n;

  assert_types (pat_tree);

  return pat_tree->n_records;
}

typedef struct
{
  FluxItem default_root;
  guint    n_records;
}
Config;

static gboolean
read_config (FluxPatTree *pat_tree)
{
  FluxItem config_item;
  Config   config;

  config_item = flux_storage_get_root (pat_tree->storage, pat_tree->id);
  if (config_item == FLUX_ITEM_NULL)
    return FALSE;

  if (!flux_storage_read_full (pat_tree->storage, config_item, &config, 0, sizeof (Config)))
    return FALSE;

  pat_tree->default_root = config.default_root;
  pat_tree->n_records    = config.n_records;

  return TRUE;
}

static gboolean
write_config (FluxPatTree *pat_tree)
{
  FluxItem config_item;
  Config   config;

  config_item = flux_storage_get_root (pat_tree->storage, pat_tree->id);
  if (config_item == FLUX_ITEM_NULL)
  {
    config_item = flux_storage_alloc (pat_tree->storage, 

    if (!flux_storage_set_root (pat_tree->storage, pat_tree->id, config_item))
      return FALSE;
  }

  if (!flux_storage_read_full (pat_tree->storage, config_item, &config, 0, sizeof (Config)))
    return FALSE;

  pat_tree->default_root = config.default_root;
  pat_tree->n_records    = config.n_records;

  return TRUE;
}

#endif

gboolean
flux_pat_tree_init (FluxPatTree *pat_tree, FluxStorage *storage, const gchar *id,
                    guint key_len, guint value_len)
{
  g_return_val_if_fail (pat_tree != NULL, FALSE);
  g_return_val_if_fail (id != NULL, FALSE);

  flux_object_init              (FLUX_OBJECT (pat_tree));
  flux_object_set_type          (FLUX_OBJECT (pat_tree), FLUX_PAT_TREE_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (pat_tree), flux_pat_tree_finalize);

  if (storage)
    flux_object_ref (FLUX_OBJECT (storage));
  else
    storage = FLUX_STORAGE (flux_mem_heap_new ());

  pat_tree->storage = storage;

  pat_tree->key_len   = key_len;
  pat_tree->value_len = value_len;
  pat_tree->fixed_len = ATOM_SIZE (pat_tree);

  pat_tree->mem_chunk = flux_mem_chunk_new (16, sizeof (PTNode));

  pat_tree->default_root = flux_storage_get_root (pat_tree->storage, id);
  if (pat_tree->default_root == FLUX_ITEM_NULL)
  {
    pat_tree->default_root = flux_pat_tree_create_root (pat_tree);
    if (pat_tree->default_root == FLUX_ITEM_NULL)
    {
      /* I/O error */
      flux_object_unref (FLUX_OBJECT (pat_tree->storage));
      flux_object_unref (FLUX_OBJECT (pat_tree->mem_chunk));
      return FALSE;
    }

    flux_storage_set_root (pat_tree->storage, id, pat_tree->default_root);
  }

  pat_tree->id = g_strdup (id);
  return TRUE;
}

FluxPatTree *
flux_pat_tree_new (FluxStorage *storage, const gchar *id, guint key_len, guint value_len)
{
  FluxPatTree *pat_tree;

  g_return_val_if_fail (id != NULL, NULL);

  pat_tree = g_new0 (FluxPatTree, 1);
  if (!flux_pat_tree_init (pat_tree, storage, id, key_len, value_len))
  {
    g_free (pat_tree);
    return NULL;
  }

  return pat_tree;
}

void
flux_pat_tree_finalize (FluxObject *object)
{
  FluxPatTree *pat_tree = FLUX_PAT_TREE (object);

  g_free (pat_tree->id);
  flux_object_unref (FLUX_OBJECT (pat_tree->storage));
  flux_object_unref (FLUX_OBJECT (pat_tree->mem_chunk));
}

guint
flux_pat_tree_calc_node_size (guint key_len, guint value_len)
{
  return sizeof (PTNode) + key_len + value_len;
}

static FluxPatTreeCursor *
cursor_new (FluxPatTree *pat_tree, FluxItem root_index, guint key_len)
{
  FluxPatTreeCursor *cursor;

  cursor = g_new0 (FluxPatTreeCursor, 1);
  flux_pat_tree_cursor_init (cursor, pat_tree, root_index, key_len);

  return cursor;
}

static gboolean
cursor_next (FluxCursor *cursor)
{
  FluxPatTreeCursor *pat_tree_cursor = FLUX_PAT_TREE_CURSOR (cursor);

  assert_cursor_types (pat_tree_cursor);

  if (pat_tree_cursor->key_index == FLUX_ITEM_NULL)
    return FALSE;

  return traverse_next (pat_tree_cursor);
}

void
flux_pat_tree_cursor_init (FluxPatTreeCursor *pat_tree_cursor, FluxPatTree *pat_tree,
                           FluxItem root_index, guint key_len)
{
  flux_cursor_init              (FLUX_CURSOR (pat_tree_cursor));
  flux_object_set_type          (FLUX_OBJECT (pat_tree_cursor), FLUX_PAT_TREE_CURSOR_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (pat_tree_cursor), flux_pat_tree_cursor_finalize);

  flux_object_ref (FLUX_OBJECT (pat_tree));

  pat_tree_cursor->pat_tree = pat_tree;
  pat_tree_cursor->key_len = key_len;

  pat_tree_cursor->index_stack = g_array_new (FALSE, FALSE, sizeof (StackItem));
  stack_item_append (pat_tree_cursor->index_stack, root_index, FALSE, FALSE);
  pat_tree_cursor->started = FALSE;
  pat_tree_cursor->randomize = FALSE;

  FLUX_CURSOR (pat_tree_cursor)->next = cursor_next;

  pat_tree->cursors = g_list_prepend (pat_tree->cursors, pat_tree_cursor);
}

void
flux_pat_tree_cursor_finalize (FluxObject *cursor)
{
  FluxPatTreeCursor *pat_tree_cursor = FLUX_PAT_TREE_CURSOR (cursor);

  assert_cursor_types (pat_tree_cursor);

  pat_tree_cursor->pat_tree->cursors = g_list_remove (pat_tree_cursor->pat_tree->cursors, cursor);

  g_array_free (pat_tree_cursor->index_stack, TRUE);
  if (pat_tree_cursor->find_key)
    g_free (pat_tree_cursor->find_key);
  if (pat_tree_cursor->find_mask)
    g_free (pat_tree_cursor->find_mask);

  flux_object_unref (FLUX_OBJECT (pat_tree_cursor->pat_tree));

  flux_cursor_finalize (cursor);
}

gboolean
flux_pat_tree_cursor_read (FluxPatTreeCursor *cursor, gpointer *key, guint *key_len,
                           gpointer *value, guint *value_len)
{
  FluxItem root_index;

  assert_cursor_types (cursor);

  if (cursor->key_index == FLUX_ITEM_NULL ||
      cursor->index_invalidated)
    return FALSE;

  root_index = g_array_index (cursor->index_stack, StackItem, 0).item;

  if (cursor->key_index == root_index)
  {
    const PTNode *root_node;

    /* Special case root node */

    root_node = peek_node (cursor->pat_tree, root_index);
    if (!ROOT_NODE_IS_ACTIVE (root_node))
      return FALSE;
  }

  if (key && !read_node_key (cursor->pat_tree, cursor->key_index, key, key_len))
    return FALSE;
  if (value && !read_node_value (cursor->pat_tree, cursor->key_index, value, value_len))
    return FALSE;

  return TRUE;
}

gboolean
flux_pat_tree_cursor_get_randomized (FluxPatTreeCursor *cursor)
{
  assert_cursor_types (cursor);

  return cursor->randomize;
}

void
flux_pat_tree_cursor_set_randomized (FluxPatTreeCursor *cursor, gboolean randomized)
{
  assert_cursor_types (cursor);

  cursor->randomize = randomized;
}
