/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-rng.c - Entropy-gathering random number generator.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/time.h>
#else
# include <io.h>
# include <time.h>
# include "win32/win32-compat.h"
#endif

#include "flux-rmd160.h"
#include "flux-rng.h"

/* Set to size of an RMD160 hash for now. */

#define DEFAULT_POOL_SIZE 20

typedef struct
{
  gchar *path;
  gint   bytes_num;
}
FluxRngFileSource;

static FluxRngFileSource dynamic_files [] =
{
  { "/proc/net/dev",    450 },
  { "/proc/interrupts", 480 },
  { NULL }
};

static void
read_to_buffer (gchar *file, guint max_len, gboolean nonblock, gpointer *dest, guint *len)
{
  guchar *buf;
  gint    fd;
  gint    result;

  *len = 0;

  /* FIXME: Always blocks on win32 */
#ifndef G_PLATFORM_WIN32
  fd = open (file, O_RDONLY | (nonblock ? O_NONBLOCK : 0));
#else
  fd = open (file, O_RDONLY);
#endif
  if (fd < 0)
  {
    *dest = NULL;
    return;
  }

  buf = g_malloc (max_len);
  result = read (fd, buf, max_len);
  close (fd);

  if (result >= 0)
    *len = result;

  if (!*len)
  {
    g_free (buf);
    buf = NULL;
  }

  *dest = buf;
}

static void
mix (FluxRng *rng)
{
  flux_rmd160_hash_single (rng->pool, rng->pool, rng->len);
}

static void
add (FluxRng *rng, guchar *data, guint len)
{
  guint i;

  for (i = 0; i < len; i++)
  {
    rng->pool [rng->insert_pos++] ^= data [i];

    /* Could use modulo operation, but it requires a division. Not
     * sure which is faster, and not sure it matters. */

    if (rng->insert_pos == rng->len)
      rng->insert_pos = 0;
  }
}

/* Adding a timeval does not produce a lot of entropy. If it's used at
 * all, it should only be added at init, since more low-order bits count
 * for something then. */

static void
add_timeval (FluxRng *rng)
{
#ifndef G_PLATFORM_WIN32
  struct timeval tv;

  gettimeofday (&tv, NULL);
  add (rng, (guchar *) &tv, sizeof (tv));
#else
  guint t;

  t = time (NULL);
  add (rng, (guchar *) &t, sizeof (t));
#endif
}

static void
add_random (FluxRng *rng, guint bytes_num)
{
#ifdef HAVE_DEV_RANDOM
  gpointer buf;
  guint    len;

  read_to_buffer (NAME_OF_DEV_RANDOM, bytes_num, TRUE, &buf, &len);
  if (!buf)
    return;

  add (rng, buf, len);

  g_free (buf);
#endif
}

static void
add_files (FluxRng *rng, FluxRngFileSource *files)
{
  gint     i;
  gpointer buf;
  guint    len;

  for (i = 0; files [i].path; i++)
  {
    read_to_buffer (files [i].path, files [i].bytes_num, FALSE, &buf, &len);
    if (!buf)
      continue;

    add (rng, buf, len);
    g_free (buf);
  }
}

static void
init (FluxRng *rng)
{
  add_timeval (rng);
  mix (rng);
  add_random (rng, 20);
  mix (rng);
  add_files (rng, dynamic_files);
  mix (rng);
}

static void
stir (FluxRng *rng)
{
  add_random (rng, 4);
  add_files (rng, dynamic_files);
  mix (rng);
}

static void
get_bytes (FluxRng *rng, guchar *buf, guint len)
{
  guint i;

  for (i = 0; i < len; i++)
  {
    /* Should memcpy() segments here. */

    buf [i] = rng->pool [rng->get_pos++];

    if (rng->get_pos == rng->len)
    {
      rng->get_pos = 0;
      stir (rng);
    }
  }
}

/* --- */

void
flux_rng_init (FluxRng *rng)
{
  flux_object_init              (FLUX_OBJECT (rng));
  flux_object_set_type          (FLUX_OBJECT (rng), FLUX_RNG_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (rng), (FluxFinalizeFunc *) flux_rng_finalize);

  rng->len        = DEFAULT_POOL_SIZE;
  rng->insert_pos = 0;
  rng->get_pos    = 0;
  rng->pool       = g_malloc0 (rng->len);
  init (rng);
}

FluxRng *
flux_rng_new (void)
{
  FluxRng *rng;

  rng = g_new0 (FluxRng, 1);
  flux_rng_init (rng);

  return rng;
}

void
flux_rng_finalize (FluxRng *rng)
{
  g_free (rng->pool);
}

void
flux_rng_get_bytes (FluxRng *rng, gpointer out, guint len)
{
  flux_object_check (FLUX_OBJECT (rng), FLUX_RNG_TYPE);
  get_bytes (rng, out, len);
}

guint
flux_rng_get_guint (FluxRng *rng)
{
  guint ret;

  flux_object_check (FLUX_OBJECT (rng), FLUX_RNG_TYPE);
  get_bytes (rng, (guchar *) &ret, sizeof (ret));
  return ret;
}
