/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-rsa.c - The RSA asymmetric cipher.
 *
 * Copyright (C) 2003 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>          /* time */
#include <glib.h>
#include <gmp.h>
#include "flux-init.h"
#include "flux-marshaller.h"
#include "flux-rng.h"
#include "flux-rsa.h"

#define D(x)

#define bits_in_mpz(n)   (mpz_sizeinbase (n, 2))
#define bits_to_bytes(b) (((b) + 7) >> 3)
#define bytes_to_bits(b) ((b) << 3)

typedef struct
{
  guchar *n;
  guint   n_len;

  guchar *e;
  guint   e_len;
}
PublicKeyM;

typedef struct
{
  guchar *n;
  guint   n_len;

  guchar *d;
  guint   d_len;
}
SecretKeyM;

static FluxMarshaller *key_marshaller = NULL;

static FluxMarshalledStructEntry public_key_m_entries [] =
{
  FLUX_MARSHAL_RAW_DYNAMIC ("n", PublicKeyM, n, n_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("e", PublicKeyM, e, e_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStructEntry secret_key_m_entries [] =
{
  FLUX_MARSHAL_RAW_DYNAMIC ("n", SecretKeyM, n, n_len, TRUE, NULL),
  FLUX_MARSHAL_RAW_DYNAMIC ("d", SecretKeyM, d, d_len, TRUE, NULL),
  FLUX_MARSHAL_END         ()
};

static FluxMarshalledStruct key_marshal_info [] =
{
  { "rsa-public-key", sizeof (PublicKeyM), FLUX_MARSHALLER_CLEAR, public_key_m_entries },
  { "rsa-secret-key", sizeof (SecretKeyM), FLUX_MARSHALLER_CLEAR, secret_key_m_entries },
  { NULL }
};

static guint
hex_to_int (char hex)
{
  if (hex >= '0' && hex <= '9')
    return (hex - '0');

  if (hex >= 'a' && hex <= 'f')
    return (hex - 'a' + 0x0a);

  if (hex >= 'A' && hex <= 'F')
    return (hex - 'A' + 0x0a);

  g_error ("Got a digit not in base16.");
  return (-1);  /* Never executed */
}

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void 
mpz_to_raw (mpz_t n, guchar *raw, guint len)
{
  char  *mpz_b16;
  guint  mpz_b16_len, mpz_b256_len;
  guint  i, j;

  if (mpz_sgn (n) < 0)
    g_error ("Got negative mpz.");

  /* Get base16 string for conversion to raw */

  mpz_b16 = mpz_get_str (NULL, 16, n);
  g_assert (mpz_b16);

  mpz_b16_len  = strlen (mpz_b16);
  mpz_b256_len = (mpz_b16_len + 1) / 2;

  if (len < mpz_b256_len)
    g_error ("Destination too small to fit mpz: %u vs %u.", len, mpz_b256_len);

  /* Convert base16 string to raw */

  memset (raw, 0, len);
  j = len - mpz_b256_len;
  i = 0;

  if (mpz_b16_len % 2)
    raw [j++] = hex_to_int (mpz_b16 [i++]);

  for ( ; mpz_b16 [i]; i += 2, j++)
  {
    raw [j] = (hex_to_int (mpz_b16 [i]) << 4) +
              (hex_to_int (mpz_b16 [i + 1]));
  }

  /* NOTE: free() crashes on win32. No idea why, maybe I need to see if I can attach
   * glib memory management functions. */
  g_free (mpz_b16);
}

static void
mpz_to_raw_alloc (mpz_t n, guchar **out, guint *out_len)
{
  *out_len = bits_to_bytes (bits_in_mpz (n));
  *out = g_malloc (*out_len);

  mpz_to_raw (n, *out, *out_len);
}

static void
raw_to_mpz (mpz_t n, const guchar *raw, guint len)
{
  gchar *mpz_b16;
  guint  i;

  /* Create a base16 string to import with mpz_set_str () */

  mpz_b16 = g_malloc ((len * 2) + 1);

  for (i = 0; i < len; i++)
  {
    mpz_b16 [i * 2]       = int_to_hex ((raw [i] & 0xf0) >> 4);
    mpz_b16 [(i * 2) + 1] = int_to_hex (raw [i] & 0x0f);
  }

  mpz_b16 [i * 2] = '\0';

  /* Import string */

  mpz_set_str (n, mpz_b16, 16);
  g_free (mpz_b16);
}

static void
init_marshaller (void)
{
  if (!key_marshaller)
  {
    key_marshaller = flux_marshaller_new ();
    flux_marshaller_register_many (key_marshaller, key_marshal_info);
  }
}

static FluxRSAPublicKey *
new_public (void)
{
  FluxRSAPublicKey *pkey;

  pkey = g_new0 (FluxRSAPublicKey, 1);
  flux_object_init              (FLUX_OBJECT (pkey));
  flux_object_set_type          (FLUX_OBJECT (pkey), FLUX_RSA_PUBLIC_KEY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (pkey), (FluxFinalizeFunc *) flux_rsa_finalize_public_key);

  init_marshaller ();

  return pkey;
}

static FluxRSASecretKey *
new_secret (void)
{
  FluxRSASecretKey *skey;

  skey = g_new0 (FluxRSASecretKey, 1);
  flux_object_init              (FLUX_OBJECT (skey));
  flux_object_set_type          (FLUX_OBJECT (skey), FLUX_RSA_SECRET_KEY_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (skey), (FluxFinalizeFunc *) flux_rsa_finalize_secret_key);

  init_marshaller ();

  return skey;
}

static FluxRSAPublicKey *
dup_public (FluxRSAPublicKey *old)
{
  FluxRSAPublicKey *new;

  new = new_public ();
  mpz_init_set (new->key.n, old->key.n);
  mpz_init_set (new->e,     old->e);

  return new;
}

static FluxRSASecretKey *
dup_secret (FluxRSASecretKey *old)
{
  FluxRSASecretKey *new;

  new = new_secret ();
  mpz_init_set (new->key.n, old->key.n);
  mpz_init_set (new->d,     old->d);

  return new;
}

static void
randomize_mpz_strong (mpz_t *n, guint bits_num, FluxRng *rng)
{
  guchar *buf;

  buf = g_malloc (bits_to_bytes (bits_num));

  flux_rng_get_bytes (rng, buf, bits_to_bytes (bits_num));
  buf [0] &= 0xff >> (8 - (bits_num % 8));  /* Trim MSB */
  raw_to_mpz         (*n,  buf, bits_to_bytes (bits_num));

  g_free (buf);
}

gboolean
flux_rsa_generate_keypair (guint key_size, FluxRSAPublicKey **public_key, FluxRSASecretKey **secret_key)
{
  FluxRng *rng;
  mpz_t    p, q, n, e, d, temp_a, temp_b;

  flux_init_mp ();

  *public_key = new_public ();
  *secret_key = new_secret ();

  rng = flux_rng_new ();

  mpz_init (p);
  mpz_init (q);
  mpz_init (n);
  mpz_init (e);
  mpz_init (d);
  mpz_init (temp_a);
  mpz_init (temp_b);

  do
  {
    randomize_mpz_strong (&p, key_size / 2, rng);
    randomize_mpz_strong (&q, key_size / 2, rng);
    randomize_mpz_strong (&d, key_size / 2, rng);

    mpz_setbit (p, (key_size / 2) - 1);
    mpz_setbit (q, (key_size / 2) - 1);

    mpz_add (d, d, p);
    mpz_add (d, d, q);

    mpz_nextprime (p, p);
    mpz_nextprime (q, q);
    mpz_nextprime (d, d);

    mpz_mul (n, p, q);

    mpz_sub_ui (temp_a, p, 1);
    mpz_sub_ui (temp_b, q, 1);

    mpz_mul (temp_a, temp_a, temp_b);
  }
  while (!mpz_invert (e, d, temp_a));

  mpz_init_set ((*public_key)->key.n, n);
  mpz_init_set ((*public_key)->e, e);

  mpz_init_set ((*secret_key)->key.n, n);
  mpz_init_set ((*secret_key)->d, d);

  mpz_clear (p);
  mpz_clear (q);
  mpz_clear (n);
  mpz_clear (e);
  mpz_clear (d);
  mpz_clear (temp_a);
  mpz_clear (temp_b);

  flux_object_unref (FLUX_OBJECT (rng));
  return TRUE;
}

FluxRSAKey *
flux_rsa_dup_key (FluxRSAKey *old)
{
  FluxRSAKey *new;

  flux_object_check (FLUX_OBJECT (old), FLUX_RSA_KEY_TYPE);

  if (flux_object_get_type (FLUX_OBJECT (old)) == FLUX_RSA_PUBLIC_KEY_TYPE)
    new = FLUX_RSA_KEY (dup_public (FLUX_RSA_PUBLIC_KEY (old)));
  else
    new = FLUX_RSA_KEY (dup_secret (FLUX_RSA_SECRET_KEY (old)));

  return new;
}

static void
finalize_common (FluxRSAKey *key)
{
  mpz_clear (key->n);
}

void
flux_rsa_finalize_public_key (FluxRSAPublicKey *public_key)
{
  finalize_common (FLUX_RSA_KEY (public_key));
  mpz_clear (public_key->e);
}

void
flux_rsa_finalize_secret_key (FluxRSASecretKey *secret_key)
{
  finalize_common (FLUX_RSA_KEY (secret_key));
  mpz_clear (secret_key->d);
}

guint
flux_rsa_get_key_length (FluxRSAKey *key)
{
  flux_object_check (FLUX_OBJECT (key), FLUX_RSA_KEY_TYPE);

  return bits_in_mpz (key->n);
}

static guint
get_ciphertext_block_size (mpz_t n)
{
  return bits_to_bytes (bits_in_mpz (n));
}

static guint
get_plaintext_block_size (mpz_t n)
{
  /* Reserve 1 bit for initial 0-byte guard, and 1 bit to make sure
   * input is lower than modulo */
  return (bits_in_mpz (n) - 2) >> 3;
}

guint
flux_rsa_get_block_size (FluxRSAKey *key)
{
  return get_plaintext_block_size (key->n);
}

static void
rsa_crypt (mpz_t n, mpz_t exp,
           guchar       **out, guint *out_len,
           const guchar  *in,  guint  in_len,
           gboolean plaintext_in)
{
  mpz_t   in_m, out_m;
  guchar *buf;
  gint    buf_len;

  mpz_init (in_m);
  mpz_init (out_m);

  raw_to_mpz (in_m, in, in_len);

  if (plaintext_in)
  {
    /* Set guard bit if plaintext in. This must be done before encryption,
     * so leading 0-bytes are preserved throughout the calculation. */
    mpz_setbit (in_m, bytes_to_bits (in_len));
  }

  mpz_powm (out_m, in_m, exp, n);

  if (plaintext_in)
  {
    /* Plaintext in, ciphertext out */
    buf_len = get_ciphertext_block_size (n);
  }
  else
  {
    /* Ciphertext in, plaintext out */
    buf_len = bits_to_bytes (bits_in_mpz (out_m) - 1);
    mpz_clrbit (out_m, bits_in_mpz (out_m) - 1);
  }

  buf = g_malloc (buf_len);
  mpz_to_raw (out_m, buf, buf_len);

  mpz_clear (in_m);
  mpz_clear (out_m);

  *out     = buf;
  *out_len = buf_len;
}

static void
rsa_crypt_multiple_blocks (mpz_t n, mpz_t exp,
                           guchar       **out, guint *out_len,
                           const guchar  *in,  guint  in_len,
                           gboolean plaintext_in)
{
  guchar *buf;
  guint   buf_len;
  gint    in_pos = 0, in_increment;

  if (plaintext_in)
    in_increment = get_plaintext_block_size (n);
  else
    in_increment = get_ciphertext_block_size (n);

  *out     = NULL;
  *out_len = 0;

  for ( ; in_pos < in_len; in_pos += in_increment)
  {
    rsa_crypt (n, exp,
               &buf,        &buf_len,
               in + in_pos, MIN (in_increment, in_len - in_pos),
               plaintext_in);

    if (*out == NULL)
    {
      *out     = buf;
      *out_len = buf_len;
    }
    else
    {
      *out = g_realloc (*out, *out_len + buf_len);
      memcpy (*out + *out_len, buf, buf_len);
      *out_len += buf_len;
      g_free (buf);
    }
  }
}

void
flux_rsa_sign (FluxRSASecretKey *secret_key,
               guchar       **signature, guint *signature_len,
               const guchar  *hash,      guint  hash_len)
{
  flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);

  rsa_crypt_multiple_blocks (secret_key->key.n, secret_key->d,
                             signature, signature_len,
                             hash,      hash_len,
                             TRUE);
}

gboolean
flux_rsa_verify (FluxRSAPublicKey *public_key,
                 const guchar *signature, guint signature_len,
                 const guchar *hash,      guint hash_len)
{
  guchar   *buf;
  gint      buf_len;
  gboolean  result = TRUE;

  flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);

  rsa_crypt_multiple_blocks (public_key->key.n, public_key->e,
                             &buf,       &buf_len,
                             signature,  signature_len,
                             FALSE);

  if (buf_len != hash_len)
    result = FALSE;

  if (memcmp (buf, hash, buf_len))
    result = FALSE;

  g_free (buf);
  return result;
}

void
flux_rsa_encrypt (FluxRSAPublicKey *public_key,
                  guchar       **ciphertext, guint *ciphertext_len,
                  const guchar  *plaintext,  guint  plaintext_len)
{
  flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);

  rsa_crypt_multiple_blocks (public_key->key.n, public_key->e,
                             ciphertext, ciphertext_len,
                             plaintext,  plaintext_len,
                             TRUE);
}

void
flux_rsa_decrypt (FluxRSASecretKey *secret_key,
                  guchar       **plaintext,  guint *plaintext_len,
                  const guchar  *ciphertext, guint  ciphertext_len)
{
  flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);

  rsa_crypt_multiple_blocks (secret_key->key.n, secret_key->d,
                             plaintext,  plaintext_len,
                             ciphertext, ciphertext_len,
                             FALSE);
}

FluxRSAPublicKey *
flux_rsa_import_public_key (const guchar *in, guint in_len)
{
  FluxRSAPublicKey *public_key;
  PublicKeyM       *public_key_m;
  mpz_t             temp_mpz;

  g_return_val_if_fail (in != NULL, NULL);
  g_return_val_if_fail (in_len > 1, NULL);

  flux_init_mp ();

  public_key = new_public ();

  if (!flux_marshaller_decode_pack (key_marshaller, "rsa-public-key", in, in_len, (gpointer) &public_key_m))
  {
    flux_object_unref (FLUX_OBJECT (public_key));
    return NULL;
  }

  if (public_key_m->n == NULL || public_key_m->e == NULL)
  {
    if (public_key_m->n)
      g_free (public_key_m->n);
    if (public_key_m->e)
      g_free (public_key_m->e);

    flux_object_unref (FLUX_OBJECT (public_key));
    g_free (public_key_m);
    return NULL;
  }

  mpz_init (temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->n, public_key_m->n_len);
  mpz_init_set (public_key->key.n, temp_mpz);

  raw_to_mpz (temp_mpz, public_key_m->e, public_key_m->e_len);
  mpz_init_set (public_key->e, temp_mpz);

  mpz_clear (temp_mpz);
  g_free (public_key_m->n);
  g_free (public_key_m->e);
  g_free (public_key_m);
  return public_key;
}

FluxRSASecretKey *
flux_rsa_import_secret_key (const guchar *in, guint in_len)
{
  FluxRSASecretKey *secret_key;
  SecretKeyM       *secret_key_m;
  mpz_t             temp_mpz;

  g_return_val_if_fail (in != NULL, NULL);
  g_return_val_if_fail (in_len > 1, NULL);

  flux_init_mp ();

  secret_key = new_secret ();

  if (!flux_marshaller_decode_pack (key_marshaller, "rsa-secret-key", in, in_len, (gpointer) &secret_key_m))
  {
    flux_object_unref (FLUX_OBJECT (secret_key));
    return NULL;
  }

  if (secret_key_m->n == NULL || secret_key_m->d == NULL)
  {
    if (secret_key_m->n)
      g_free (secret_key_m->n);
    if (secret_key_m->d)
      g_free (secret_key_m->d);

    flux_object_unref (FLUX_OBJECT (secret_key));
    g_free (secret_key_m);
    return NULL;
  }

  mpz_init (temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->n, secret_key_m->n_len);
  mpz_init_set (secret_key->key.n, temp_mpz);

  raw_to_mpz (temp_mpz, secret_key_m->d, secret_key_m->d_len);
  mpz_init_set (secret_key->d, temp_mpz);

  mpz_clear (temp_mpz);
  g_free (secret_key_m->n);
  g_free (secret_key_m->d);
  g_free (secret_key_m);
  return secret_key;
}

void
flux_rsa_export_public_key (FluxRSAPublicKey *public_key, guchar **out, guint *out_len)
{
  PublicKeyM public_key_m;

  flux_object_check (FLUX_OBJECT (public_key), FLUX_RSA_PUBLIC_KEY_TYPE);
  g_return_if_fail (out != NULL);
  g_return_if_fail (out_len != NULL);

  mpz_to_raw_alloc (public_key->key.n, &public_key_m.n, &public_key_m.n_len);
  mpz_to_raw_alloc (public_key->e, &public_key_m.e, &public_key_m.e_len);

  flux_marshaller_encode_pack (key_marshaller, "rsa-public-key", &public_key_m, (gpointer *) out, out_len);

  g_free (public_key_m.n);
  g_free (public_key_m.e);
}

void
flux_rsa_export_secret_key (FluxRSASecretKey *secret_key, guchar **out, guint *out_len)
{
  SecretKeyM secret_key_m;

  flux_object_check (FLUX_OBJECT (secret_key), FLUX_RSA_SECRET_KEY_TYPE);
  g_return_if_fail (out != NULL);
  g_return_if_fail (out_len != NULL);

  mpz_to_raw_alloc (secret_key->key.n, &secret_key_m.n, &secret_key_m.n_len);
  mpz_to_raw_alloc (secret_key->d, &secret_key_m.d, &secret_key_m.d_len);

  flux_marshaller_encode_pack (key_marshaller, "rsa-secret-key", &secret_key_m, (gpointer *) out, out_len);

  g_free (secret_key_m.n);
  g_free (secret_key_m.d);
}
