/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-socket-util.c - Socket utilities.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <fcntl.h>
# include <unistd.h>
#else
# include <io.h>
#endif

#include "flux-types.h"
#include "flux-socket-util.h"

#define DEBUG(x)

#define WAKEUP_READ  0
#define WAKEUP_WRITE 1

#ifdef HAVE_IPV6

/* IPv4 and IPv6 capable */

#define FLUX_SOCKADDR_GET_FAMILY(sa)                             \
  (((struct sockaddr *) (sa))->sa_family)

#define FLUX_SOCKADDR_GET_LEN(sa)                                \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   sizeof (struct sockaddr_in) : sizeof (struct sockaddr_in6))

#define FLUX_SOCKADDR_GET_ADDR(sa)                               \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   (gpointer) &((struct sockaddr_in *) (sa))->sin_addr :         \
   (gpointer) &((struct sockaddr_in6 *) (sa))->sin6_addr)

#define FLUX_SOCKADDR_GET_ADDR_LEN(sa)                           \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   sizeof (struct in_addr) : sizeof (struct in6_addr))

#define FLUX_SOCKADDR_GET_PORT(sa)                               \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   ((struct sockaddr_in *) (sa))->sin_port :                     \
   ((struct sockaddr_in6 *) (sa))->sin6_port)

#define preferred_ntop(family, addr, buf, buf_len) inet_ntop (family, addr, buf, buf_len)
#define preferred_pton(family, src, dest)          inet_pton (family, src, dest)

#else

/* IPv4 only */

#define FLUX_SOCKADDR_GET_FAMILY(sa)                             \
  (((struct sockaddr *) (sa))->sa_family)

#define FLUX_SOCKADDR_GET_LEN(sa)                                \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   sizeof (struct sockaddr_in) : 0)

#define FLUX_SOCKADDR_GET_ADDR(sa)                               \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   (gpointer) &((struct sockaddr_in *) (sa))->sin_addr :         \
   (gpointer) NULL)

#define FLUX_SOCKADDR_GET_ADDR_LEN(sa)                           \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   sizeof (struct in_addr) : 0)

#define FLUX_SOCKADDR_GET_PORT(sa)                               \
  ((((struct sockaddr *) (sa))->sa_family == AF_INET) ?          \
   ((struct sockaddr_in *) (sa))->sin_port : 0)

static gchar *
preferred_ntop (gint family, struct in_addr *sa_in_addr, gchar *buf, gint buf_len)
{
  static GStaticMutex  mutex = G_STATIC_MUTEX_INIT;
  const gchar         *addr_str = NULL;
  gchar               *result;

  if (family != AF_INET)
    return NULL;

  g_static_mutex_lock (&mutex);

  addr_str = inet_ntoa (*sa_in_addr);

  if (addr_str)
  {
    strncpy (buf, addr_str, buf_len);
    buf [buf_len - 1] = '\0';
    result = buf;
  }
  else
  {
    result = NULL;
  }

  g_static_mutex_unlock (&mutex);

  return result;
}

static gint
preferred_pton (gint family, const gchar *src, gpointer dest)
{
#ifndef G_PLATFORM_WIN32
  if (inet_aton (src, dest))
  {
    DEBUG (g_print ("inet_aton successful.\n"));
    return 1;
  }

  DEBUG (g_print ("inet_aton failed.\n"));
  return 0;
#else
  guint32 addr;

  addr = inet_addr (src);
  if (addr == INADDR_NONE && strcmp (src, "255.255.255.255"))
    return 0;

  *((guint32 *) dest) = addr;
  return 1;
#endif
}

#endif

gint
flux_sockaddr_get_family (struct sockaddr *sa)
{
  return FLUX_SOCKADDR_GET_FAMILY (sa);
}

gint
flux_sockaddr_get_len (struct sockaddr *sa)
{
  return FLUX_SOCKADDR_GET_LEN (sa);
}

gpointer
flux_sockaddr_get_addr (struct sockaddr *sa)
{
  return FLUX_SOCKADDR_GET_ADDR (sa);
}

gint
flux_sockaddr_get_addr_len (struct sockaddr *sa)
{
  return FLUX_SOCKADDR_GET_ADDR_LEN (sa);
}

gint
flux_sockaddr_get_port (struct sockaddr *sa)
{
  return FLUX_SOCKADDR_GET_PORT (sa);
}

struct sockaddr *
flux_sockaddr_new (FluxIPAddrType type, const gchar *addr_str, gint port)
{
  struct sockaddr *sa;

  if (type == FLUX_IP_ADDR_IPV4)
  {
    struct sockaddr_in in_sa;

    if (preferred_pton (AF_INET, addr_str, &in_sa.sin_addr) < 1)
      return NULL;

    in_sa.sin_family = AF_INET;
    in_sa.sin_port   = g_htons (port);
    sa = g_memdup (&in_sa, sizeof (struct sockaddr_in));
  }
#ifdef HAVE_IPV6
  else if (type == FLUX_IP_ADDR_IPV6)
  {
    struct sockaddr_in6 in6_sa;

    if (preferred_pton (AF_INET6, addr_str, &in6_sa.sin6_addr) < 1)
      return NULL;

    in6_sa.sin6_family = AF_INET6;
    in6_sa.sin6_port   = g_htons (port);
    sa = g_memdup (&in6_sa, sizeof (struct sockaddr_in6));
  }
#endif
  else
  {
    sa = NULL;
  }

  return sa;
}

gchar *
flux_sockaddr_to_string (struct sockaddr *sa)
{
  gchar buf [256];

  if (!flux_sockaddr_is_ip (sa))
    return NULL;

  if (!preferred_ntop (flux_sockaddr_get_family (sa), flux_sockaddr_get_addr (sa), buf, 256))
    return NULL;

  return g_strdup (buf);
}

gboolean
flux_sockaddr_is_ipv4 (struct sockaddr *sa)
{
  if (flux_sockaddr_get_family (sa) == AF_INET)
    return TRUE;

  return FALSE;
}

gboolean
flux_sockaddr_is_ipv6 (struct sockaddr *sa)
{
#ifdef HAVE_IPV6
  if (flux_sockaddr_get_family (sa) == AF_INET6)
    return TRUE;
#endif

  return FALSE;
}

gboolean
flux_sockaddr_is_ip (struct sockaddr *sa)
{
  if (flux_sockaddr_is_ipv4 (sa) || flux_sockaddr_is_ipv6 (sa))
    return TRUE;

  return FALSE;
}

gboolean
flux_addr_family_is_ipv4 (gint in_family)
{
  if (in_family == AF_INET)
    return TRUE;

  return FALSE;
}

gboolean
flux_addr_family_is_ipv6 (gint in_family)
{
#ifdef HAVE_IPV6
  if (in_family == AF_INET6)
    return TRUE;
#endif

  return FALSE;
}

gboolean
flux_addr_family_is_ip (gint in_family)
{
  if (flux_addr_family_is_ipv4 (in_family) || flux_addr_family_is_ipv6 (in_family))
    return TRUE;

  return FALSE;
}

gboolean
flux_socket_set_nonblock (gint fd, gboolean nonblock)
{
#ifndef G_PLATFORM_WIN32
  gint flags;

  flags = fcntl (fd, F_GETFL);
  if (flags < 0)
    return FALSE;

  if (fcntl (fd, F_SETFL, nonblock ? flags | O_NONBLOCK : flags & ~O_NONBLOCK) < 0)
    return FALSE;
#else
  u_long arg = nonblock ? 1 : 0;

  if (ioctlsocket (fd, FIONBIO, &arg))
    return FALSE;
#endif

  return TRUE;
}

gboolean
flux_wakeup_pipe_create (gint *fd)
{
#ifndef G_PLATFORM_WIN32
  pipe (fd);

  flux_socket_set_nonblock (fd [WAKEUP_READ], TRUE);
  flux_socket_set_nonblock (fd [WAKEUP_WRITE], TRUE);

  return TRUE;
#else
  struct sockaddr_in sa_in_1, sa_in_2;
  gint               my_fd [2] = { -1, -1 };
  gint               len;

  my_fd [WAKEUP_WRITE] = socket (AF_INET, SOCK_DGRAM, 0);
  if (my_fd [WAKEUP_WRITE] < 0)
    goto error;

  sa_in_1.sin_family = AF_INET;
  sa_in_1.sin_port = 0;
  sa_in_1.sin_addr.s_addr = htonl (INADDR_ANY);

  if (bind (my_fd [WAKEUP_WRITE], (struct sockaddr *) &sa_in_1, sizeof (sa_in_1)) != 0)
    goto error;

  sa_in_2.sin_family = AF_INET;
  sa_in_2.sin_port = 0;
  sa_in_2.sin_addr.s_addr = htonl (INADDR_ANY);

  my_fd [WAKEUP_READ] = socket (AF_INET, SOCK_DGRAM, 0);
  if (my_fd [WAKEUP_READ] < 0)
    goto error;

  if (bind (my_fd [WAKEUP_READ], (struct sockaddr *) &sa_in_2, sizeof (sa_in_2)) != 0)
    goto error;

  len = sizeof (sa_in_2);
  if (getsockname (my_fd [WAKEUP_READ], (struct sockaddr *) &sa_in_2, &len) != 0)
    goto error;

  sa_in_2.sin_addr.s_addr = htonl (INADDR_LOOPBACK);

  if (connect (my_fd [WAKEUP_WRITE], (struct sockaddr *) &sa_in_2, sizeof (sa_in_2)) < 0)
    goto error;

  flux_socket_set_nonblock (my_fd [WAKEUP_READ], TRUE);
  flux_socket_set_nonblock (my_fd [WAKEUP_WRITE], TRUE);

  fd [WAKEUP_READ]  = my_fd [WAKEUP_READ];
  fd [WAKEUP_WRITE] = my_fd [WAKEUP_WRITE];

  return TRUE;

error:
  if (my_fd [WAKEUP_READ] >= 0)
    flux_socket_close (my_fd [WAKEUP_READ]);
  if (my_fd [WAKEUP_WRITE] >= 0)
    flux_socket_close (my_fd [WAKEUP_WRITE]);

  fd [WAKEUP_READ]  = -1;
  fd [WAKEUP_WRITE] = -1;

  return FALSE;
#endif
}

void
flux_wakeup_pipe_close (gint *fd)
{
  flux_socket_close (fd [WAKEUP_READ]);
  flux_socket_close (fd [WAKEUP_WRITE]);

  fd [WAKEUP_READ]  = -1;
  fd [WAKEUP_WRITE] = -1;
}

gint
flux_wakeup_pipe_get_watch_fd (gint *fd)
{
  return fd [WAKEUP_READ];
}

void
flux_wakeup_pipe_wakeup (gint *fd)
{
  const gchar buf [1] = { '\0' };

  if (fd [WAKEUP_WRITE] < 0)
    return;

#ifndef G_PLATFORM_WIN32
  write (fd [WAKEUP_WRITE], buf, sizeof (buf));
#else
  send (fd [WAKEUP_WRITE], buf, sizeof (buf), 0);
#endif
}

void
flux_wakeup_pipe_handle_wakeup (gint *fd)
{
  gchar buf [256];

#ifndef G_PLATFORM_WIN32
  while (read (fd [WAKEUP_READ], buf, sizeof (buf)) > 0)
    ;
#else
  recv (fd [WAKEUP_READ], buf, sizeof (buf), 0);
#endif
}
