/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-stream-util.c - Utilities for block-oriented stream processing.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include <io.h>
# include "win32/win32-compat.h"
#endif

#include "flux-types.h"
#include "flux-stream-util.h"

/**
 * read_with_pos: Wrapper for read() that throws away the position.
 * 
 * Reads from a stream, at its current position. Throws away the
 * application-provided position.
 **/
static gint
read_with_pos (gpointer handle, guchar *block, guint pos, guint size)
{
  return read (GPOINTER_TO_INT (handle), block, size);
}

/**
 * stream_read: Read and process a stream piece-wise.
 * 
 * Reading starts at the current stream position and ends when at least
 * one of the following conditions is satisfied:
 * 
 * 1) The stream ends - returns -1.
 * 2) max_size bytes have been processed - returns 0.
 * 3) Processor function returns a non-zero value - returns that.
 * 
 * A NULL reader means use the system read() - and source is the filehandle
 * integer.
 **/
gint
flux_stream_process (gpointer source, guint block_size, guint max_size,
                     FluxStreamProcessorFunc *processor,
                     FluxStreamReaderFunc *reader,
                     gpointer data)
{
  guchar *block;
  guint count = 0;
  gint nr;
  gint ret = 0;

  if (!reader) reader = (FluxStreamReaderFunc *) read_with_pos;

  block = malloc (block_size);

  /* Do as many block_size blocks as possible */
  
  for ( ; count + block_size <= max_size; count += block_size)
  {
    nr = reader (source, block, count, block_size);
    if (nr <= 0) { ret = -1; break; }  /* Stream failure or early stop */
    ret = processor (block, count, nr, data);
    if (nr != block_size) ret = -1;  /* Stream stopped early */
    if (ret) break;
  }
  
  /* Do any remaining data */

  if (!ret && count < max_size)
  {
    nr = reader (source, block, count, max_size - count);
    if (nr <= 0) ret = -1;  /* Stream failure or early stop */
    else ret = processor (block, count, nr, data);
    if (nr != max_size - count) ret = -1;  /* Stream stopped early */
  }

  free (block);
  return ret;
}

/**
 * stream_read_dual: Read and process two streams piece-wise.
 * 
 * Reading starts at the current stream positions and ends when at least
 * one of the following conditions is satisfied:
 * 
 * 1) Stream 1 ends - returns -1.
 * 2) Stream 2 ends - returns 1.
 * 3) max_size bytes have been processed - returns 0.
 * 4) Processor function returns a non-zero value - returns that.
 * 
 * Useful when comparing two large blocks of data, or when generating a
 * third stream which is an orthogonal function of the first two.
 * 
 * A NULL reader means use the system read() - and source is the filehandle
 * integer.
 **/
gint
flux_stream_process_dual (gpointer source1, gpointer source2,
                          guint block_size, guint max_size,
                          FluxStreamDualProcessorFunc *processor,
                          FluxStreamReaderFunc *reader1,
                          FluxStreamReaderFunc *reader2,
                          gpointer data)
{
  guchar *block1, *block2;
  guint count = 0;
  gint nr1, nr2;
  gint ret = 0;

  if (!reader1) reader1 = (FluxStreamReaderFunc *) read_with_pos;
  if (!reader2) reader2 = (FluxStreamReaderFunc *) read_with_pos;

  block1 = malloc (block_size);
  block2 = malloc (block_size);

  /* Do as many block_size blocks as possible */

  for ( ; count + block_size <= max_size; count += block_size)
  {
    nr1 = reader1 (source1, block1, count, block_size);
    if (nr1 <= 0) { ret = -1; break; }  /* Stream failure or early stop */

    nr2 = reader2 (source2, block2, count, block_size);
    if (nr2 <= 0) { ret = 1; break; }  /* Stream failure or early stop */

    ret = processor (block1, block2, count, nr1 < nr2 ? nr1 : nr2, data);
    if (nr1 != block_size || nr2 != block_size) ret = -1;  /* Stream stopped early */
    if (ret) break;
  }

  /* Do any remaining data */

  if (!ret && count < max_size)
  {
    nr1 = reader1 (source1, block1, count, max_size - count);
    if (nr1 <= 0) ret = -1;  /* Stream failure or early stop */
    else
    {
      nr2 = reader2 (source2, block2, count, max_size - count);
      if (nr2 <= 0) ret = 1;  /* Stream failure or early stop */
      else
      {
        ret = processor (block1, block2, count, nr1 < nr2 ? nr1 : nr2, data);
        if (nr1 < nr2) ret = -1;  /* Stream stopped early */
        else if (nr1 > nr2) ret = 1;  /* Stream stopped early */
      }
    }
  }

  free (block1);
  free (block2);
  return ret;
}
