/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-tcp-io.c - TCP I/O channel.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/socket.h>
# include <netinet/in.h>
#else
# include <io.h>
# include <process.h>
# include <winsock2.h>
#endif

#include <time.h>

#include "flux-init.h"
#include "flux-fifo.h"
#include "flux-io.h"
#include "flux-ip-resolver.h"
#include "flux-util.h"
#include "flux-socket-util.h"
#include "flux-tcp-io.h"

#define DEBUG(x)

#define CONNECT_TIMEOUT_SEC    10
#define DISCONNECT_TIMEOUT_SEC 8

#define MAX_READ          4096
#define MAX_WRITE         4096
#define MAX_READ_BUFFER   8192
#define MAX_WRITE_BUFFER  8192
#define THREAD_STACK_SIZE (MAX_READ + MAX_WRITE + 16384)

#define assert_types(o) flux_object_check (FLUX_OBJECT (o), FLUX_TCP_IO_TYPE)

#define MIN3(a, b, c) MIN (MIN (a, b), c)

#define IS_CONNECTED(io)          ((io)->fd >= 0 && (io)->have_connected_fd)
#define IS_IN_CIPHER_EXCHANGE(io) ((io)->reading_cipher_invitation || (io)->reading_cipher_confirmation)

static void install_callback (FluxTcpIO *tcp_io);

static GStaticMutex  tcp_io_mutex         = G_STATIC_MUTEX_INIT;
static GThread      *tcp_io_thread        = NULL;
static GList        *tcp_io_list          = NULL;
static gint          tcp_io_wakeup_fd [2] = { -1, -1 };

/* -------------------- *
 * Forward declarations *
 * -------------------- */

void _flux_io_lock   (void);
void _flux_io_unlock (void);

static void set_disconnected (FluxTcpIO *tcp_io);

/* ------------- *
 * Debug helpers *
 * ------------- */

DEBUG (

static gint ios_open = 0;

static guchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static void
dump_hex (guchar *buf, guint len)
{
  gint i;

  for (i = 0; i < len; i++)
    g_print ("%c%c ", int_to_hex (buf [i] >> 4), int_to_hex (buf [i] & 0x0f));

  g_print ("\n");
}

)

/* --------------------------- *
 * Callback dispatch (to user) *
 * --------------------------- */

static void
dispatch_reads (FluxTcpIO *tcp_io)
{
  FluxFifo *fifo;
  gint      current_refs;
  gint      dispatched_bytes = 0;

  if (tcp_io->state != FLUX_TCP_IO_STATE_CONNECTED)
    return;

  fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (tcp_io));

  /* While we're at or above threshold (in held bytes), invoke read func */

  current_refs = flux_object_get_refs (FLUX_OBJECT (tcp_io));

  DEBUG (g_print ("begin dispatching reads\n"));

  _flux_io_lock ();

  while (FLUX_IO (tcp_io)->read_threshold &&
         FLUX_IO (tcp_io)->read_func &&
         flux_fifo_get_used (fifo) >= FLUX_IO (tcp_io)->read_threshold &&
         flux_object_get_refs (FLUX_OBJECT (tcp_io)) >= current_refs &&
         dispatched_bytes < MAX_READ)
  {
    dispatched_bytes += FLUX_IO (tcp_io)->read_threshold;

    _flux_io_unlock ();
    g_static_mutex_unlock (&tcp_io_mutex);
    DEBUG (g_print ("begin read callback\n"));
    FLUX_IO (tcp_io)->read_func (FLUX_IO (tcp_io), FLUX_IO (tcp_io)->read_threshold,
                                 FLUX_IO (tcp_io)->read_func_data);
    DEBUG (g_print ("end read callback\n"));
    g_static_mutex_lock (&tcp_io_mutex);
    _flux_io_lock ();
  }

  DEBUG (g_print ("end dispatching reads\n"));

  /* If we had to stop prematurely due to hitting the dispatched_bytes limit,
   * reinstall so we'll run again after other idle handlers */
  if (FLUX_IO (tcp_io)->read_threshold &&
      FLUX_IO (tcp_io)->read_func &&
      flux_fifo_get_used (fifo) >= FLUX_IO (tcp_io)->read_threshold &&
      flux_object_get_refs (FLUX_OBJECT (tcp_io)) >= current_refs)
  {
    install_callback (tcp_io);
  }

  _flux_io_unlock ();
}

static void
dispatch_writes (FluxTcpIO *tcp_io)
{
  FluxFifo *fifo;
  gint      current_refs;
  gint      dispatched_bytes = 0;

  if (tcp_io->state != FLUX_TCP_IO_STATE_CONNECTED)
    return;

  fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (tcp_io));

  /* While we're at or above threshold (in free bytes), invoke write func */

  current_refs = flux_object_get_refs (FLUX_OBJECT (tcp_io));

  DEBUG (g_print ("begin dispatching writes\n"));

  _flux_io_lock ();

  while (tcp_io->state == FLUX_TCP_IO_STATE_CONNECTED &&
         FLUX_IO (tcp_io)->write_threshold &&
         FLUX_IO (tcp_io)->write_func &&
         (flux_fifo_get_used (fifo) == 0 ||
          flux_fifo_get_used (fifo) + FLUX_IO (tcp_io)->write_threshold <= MAX_WRITE_BUFFER) &&
         flux_object_get_refs (FLUX_OBJECT (tcp_io)) >= current_refs &&
         dispatched_bytes < MAX_WRITE)
  {
    dispatched_bytes += FLUX_IO (tcp_io)->write_threshold;

    _flux_io_unlock ();
    g_static_mutex_unlock (&tcp_io_mutex);
    DEBUG (g_print ("begin write callback\n"));
    FLUX_IO (tcp_io)->write_func (FLUX_IO (tcp_io),
                                  FLUX_IO (tcp_io)->write_threshold,
                                  FLUX_IO (tcp_io)->write_func_data);
    DEBUG (g_print ("end write callback\n"));
    g_static_mutex_lock (&tcp_io_mutex);
    _flux_io_lock ();
  }

  /* If we had to stop prematurely due to hitting the dispatched_bytes limit,
   * reinstall so we'll run again after other idle handlers */
  if (tcp_io->state == FLUX_TCP_IO_STATE_CONNECTED &&
      FLUX_IO (tcp_io)->write_threshold &&
      FLUX_IO (tcp_io)->write_func &&
      (flux_fifo_get_used (fifo) == 0 ||
       flux_fifo_get_used (fifo) + FLUX_IO (tcp_io)->write_threshold <= MAX_WRITE_BUFFER) &&
      flux_object_get_refs (FLUX_OBJECT (tcp_io)) >= current_refs)
  {
    install_callback (tcp_io);
  }

  _flux_io_unlock ();

  DEBUG (g_print ("end dispatching writes\n"));
}

static void
dispatch_link_pre_data (FluxTcpIO *tcp_io)
{
  if (tcp_io->user_connected_flag)
  {
    g_assert (tcp_io->state != FLUX_TCP_IO_STATE_CONNECTED);

    tcp_io->state = FLUX_TCP_IO_STATE_CONNECTED;
    tcp_io->user_connected_flag = FALSE;

    if (tcp_io->link_func)
    {
      g_static_mutex_unlock (&tcp_io_mutex);
      tcp_io->link_func (tcp_io, FLUX_TCP_CONNECTION_ESTABLISHED, tcp_io->link_func_data);
      g_static_mutex_lock (&tcp_io_mutex);
    }
  }
}

static void
dispatch_link_post_data (FluxTcpIO *tcp_io)
{
  if (tcp_io->user_disconnected_flag)
  {
    g_assert (tcp_io->state != FLUX_TCP_IO_STATE_DISCONNECTED);

    if (tcp_io->link_func)
    {
      if (tcp_io->state == FLUX_TCP_IO_STATE_CONNECTED ||
          tcp_io->state == FLUX_TCP_IO_STATE_DISCONNECTING)
      {
        g_static_mutex_unlock (&tcp_io_mutex);
        tcp_io->link_func (tcp_io, FLUX_TCP_CONNECTION_CLOSED, tcp_io->link_func_data);
        g_static_mutex_lock (&tcp_io_mutex);
      }
      else if (tcp_io->state == FLUX_TCP_IO_STATE_CONNECTING)
      {
        g_static_mutex_unlock (&tcp_io_mutex);
        tcp_io->link_func (tcp_io, FLUX_TCP_CONNECTION_FAILED, tcp_io->link_func_data);
        g_static_mutex_lock (&tcp_io_mutex);
      }
    }
    else
    {
      /* This is ok because main_callback () holds an extra ref */
      flux_object_unref (FLUX_OBJECT (tcp_io));
    }

    tcp_io->state = FLUX_TCP_IO_STATE_DISCONNECTED;
    tcp_io->user_disconnected_flag = FALSE;
  }
}

/* -------------------------------------- *
 * User interaction callbacks (from user) *
 * -------------------------------------- */

static void
do_wakeup (void)
{
  flux_wakeup_pipe_wakeup (tcp_io_wakeup_fd);
}

static void
user_read_cb (gpointer data)
{
  FluxTcpIO *tcp_io = data;
  FluxFifo  *fifo;

  assert_types (tcp_io);

  do_wakeup ();
}

static void
user_write_cb (gpointer data)
{
  FluxTcpIO *tcp_io = data;
  FluxFifo  *fifo;

  assert_types (tcp_io);

  do_wakeup ();
}

static void
system_write_cb (gpointer data)
{
  FluxTcpIO *tcp_io = data;

  assert_types (tcp_io);

  do_wakeup ();
}

/* ------------ *
 * Cipher setup *
 * ------------ */

static void
write_cipher_element (FluxTcpIO *tcp_io, guchar *buf, guint len)
{
  guint16 len_pi;

  len_pi = len;
  len_pi = GUINT16_TO_BE (len_pi);

  flux_dual_fifo_set_write_hook (FLUX_DUAL_FIFO (tcp_io), system_write_cb, tcp_io);

  flux_dual_fifo_write (FLUX_DUAL_FIFO (tcp_io), &len_pi, sizeof (len_pi));
  flux_dual_fifo_write (FLUX_DUAL_FIFO (tcp_io), buf, len);
  flux_dual_fifo_flush_output (FLUX_DUAL_FIFO (tcp_io));

  flux_dual_fifo_set_write_hook (FLUX_DUAL_FIFO (tcp_io), user_write_cb, tcp_io);
}

static void
write_cipher_invitation (FluxTcpIO *tcp_io)
{
  guchar *buf;
  guint   len;

  DEBUG (g_print ("Sending cipher invitation.\n"));

  flux_cipher_export_invitation (tcp_io->cipher, &buf, &len);
  write_cipher_element (tcp_io, buf, len);
  g_free (buf);
}

static void
write_cipher_confirmation (FluxTcpIO *tcp_io)
{
  guchar *buf;
  guint   len;

  DEBUG (g_print ("Sending cipher confirmation.\n"));

  flux_cipher_export_confirmation (tcp_io->cipher, &buf, &len);
  DEBUG (dump_hex (buf, len));

  write_cipher_element (tcp_io, buf, len);
  g_free (buf);
}

static gboolean
try_read_cipher_element_size (FluxTcpIO *tcp_io)
{
  if (flux_dual_fifo_get_max_read (FLUX_DUAL_FIFO (tcp_io)) >= sizeof (tcp_io->cipher_element_size))
  {
    flux_dual_fifo_read (FLUX_DUAL_FIFO (tcp_io), &tcp_io->cipher_element_size,
                         sizeof (tcp_io->cipher_element_size));
    tcp_io->cipher_element_size = GUINT16_FROM_BE (tcp_io->cipher_element_size);

    DEBUG (g_print ("Incoming element size: %d\n", (int) tcp_io->cipher_element_size));
    return TRUE;
  }

  return FALSE;
}

static guchar *
try_read_cipher_element (FluxTcpIO *tcp_io)
{
  guchar *buf = NULL;

  if (tcp_io->cipher_element_size &&
      flux_dual_fifo_get_max_read (FLUX_DUAL_FIFO (tcp_io)) >= tcp_io->cipher_element_size)
  {
    buf = g_malloc (tcp_io->cipher_element_size);
    flux_dual_fifo_read (FLUX_DUAL_FIFO (tcp_io), buf, tcp_io->cipher_element_size);
  }

  return buf;
}

static gboolean
try_read_cipher_invitation (FluxTcpIO *tcp_io)
{
  guchar   *buf;
  gboolean  success;

  if (!tcp_io->cipher_element_size)
    try_read_cipher_element_size (tcp_io);

  buf = try_read_cipher_element (tcp_io);
  if (buf)
  {
    DEBUG (g_print ("Got cipher invitation.\n"));

    success = flux_cipher_import_invitation (tcp_io->cipher, buf, tcp_io->cipher_element_size);
    tcp_io->reading_cipher_invitation = FALSE;
    tcp_io->cipher_element_size = 0;
    g_free (buf);

    if (!success)
    {
      set_disconnected (tcp_io);
      return FALSE;
    }

    write_cipher_confirmation (tcp_io);

    flux_dual_fifo_set_cipher (FLUX_DUAL_FIFO (tcp_io), tcp_io->cipher);
    flux_dual_fifo_set_compression (FLUX_DUAL_FIFO (tcp_io), FLUX_IO (tcp_io)->read_compression_level,
                                    FLUX_IO (tcp_io)->write_compression_level);

    tcp_io->user_connected_flag = TRUE;
    return TRUE;
  }

  return FALSE;
}

static gboolean
try_read_cipher_confirmation (FluxTcpIO *tcp_io)
{
  guchar   *buf;
  gboolean  success;

  if (!tcp_io->cipher_element_size)
    try_read_cipher_element_size (tcp_io);

  buf = try_read_cipher_element (tcp_io);
  if (buf)
  {
    DEBUG (g_print ("Got cipher confirmation.\n"));
    DEBUG (dump_hex (buf, tcp_io->cipher_element_size));

    success = flux_cipher_import_confirmation (tcp_io->cipher, buf, tcp_io->cipher_element_size);

    tcp_io->reading_cipher_confirmation = FALSE;
    tcp_io->cipher_element_size = 0;
    g_free (buf);

    if (!success)
    {
      set_disconnected (tcp_io);
      return FALSE;
    }

    flux_dual_fifo_set_cipher (FLUX_DUAL_FIFO (tcp_io), tcp_io->cipher);
    flux_dual_fifo_set_compression (FLUX_DUAL_FIFO (tcp_io), FLUX_IO (tcp_io)->read_compression_level,
                                    FLUX_IO (tcp_io)->write_compression_level);

    tcp_io->user_connected_flag = TRUE;
    return TRUE;
  }

  return FALSE;
}

/**
 * enqueue_cipher_confirmation_bytes: Steal key bytes and install cipher when exact amount is read.
 *
 * The cipher confirmation block must be read with special care. The cipher must
 * be installed exactly after the last byte has been read, so decryption can start
 * at the next incoming byte.
 **/
static void
enqueue_cipher_confirmation_bytes (FluxTcpIO *tcp_io, FluxFifo *fifo, guchar *buf, guint buf_len)
{
  while (buf_len)
  {
    flux_fifo_enqueue (fifo, buf, 1);
    buf++;
    buf_len--;

    if (try_read_cipher_confirmation (tcp_io))
      break;
  }

  if (buf_len)
    flux_fifo_enqueue (fifo, buf, buf_len);
}

/* ------------------------- *
 * Socket session management *
 * ------------------------- */

/**
 * setup_connected_socket: Initialize a socket that was just connected on the lower layer.
 *
 * Returns: TRUE if the high-level connection init is finished, FALSE otherwise.
 **/
static gboolean
set_connected (FluxTcpIO *tcp_io, gboolean is_outgoing)
{
  /* Connect read/write callbacks */

  tcp_io->have_connected_fd = TRUE;

  if (!tcp_io->cipher)
  {
    /* Unciphered */

    flux_dual_fifo_set_compression (FLUX_DUAL_FIFO (tcp_io), FLUX_IO (tcp_io)->read_compression_level,
                                    FLUX_IO (tcp_io)->write_compression_level);
    tcp_io->user_connected_flag = TRUE;
    return TRUE;
  }

  if (is_outgoing)
  {
    /* Ciphered, outgoing */

    tcp_io->reading_cipher_invitation = TRUE;
  }
  else
  {
    /* Ciphered, incoming */

    write_cipher_invitation (tcp_io);
    tcp_io->reading_cipher_confirmation = TRUE;
  }

  tcp_io->timestamp = time (NULL);
  return FALSE;
}

static void
set_disconnected (FluxTcpIO *tcp_io)
{
  tcp_io->user_disconnected_flag = TRUE;
  tcp_io->have_connected_fd = FALSE;
  flux_socket_close (tcp_io->fd);
  tcp_io->fd = -1;
}

/**
 * reset_socket: Clear all data and metadata, useful before connecting.
 **/
static void
reset_socket (FluxTcpIO *tcp_io)
{
  flux_dual_fifo_reset (FLUX_DUAL_FIFO (tcp_io));
  flux_dual_fifo_set_cipher (FLUX_DUAL_FIFO (tcp_io), NULL);
}

static gint
network_max_read (FluxTcpIO *tcp_io)
{
  FluxFifo *fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (tcp_io));
  gint      max_len;

  /* FIXME: Respect max input buffer size? */

  _flux_io_lock ();

  if      (FLUX_IO (tcp_io)->read_cap == -1)
    max_len = MAX_READ;
  else if (FLUX_IO (tcp_io)->read_count < FLUX_IO (tcp_io)->read_slice_cap)
    max_len = MIN (MAX_READ, FLUX_IO (tcp_io)->read_slice_cap - FLUX_IO (tcp_io)->read_count);
  else
    max_len = 0;

  max_len = MIN (max_len, MAX_READ_BUFFER - flux_fifo_get_used (fifo));
  if (max_len < 0)
    max_len = 0;

  _flux_io_unlock ();

  return max_len;
}

static gint
network_max_write (FluxTcpIO *tcp_io)
{
  FluxFifo *fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (tcp_io));
  gint      max_len;

  _flux_io_lock ();

  if      (FLUX_IO (tcp_io)->write_cap == -1)
    max_len = MIN  (MAX_WRITE, flux_fifo_get_used (fifo));
  else if (FLUX_IO (tcp_io)->write_count < FLUX_IO (tcp_io)->write_slice_cap)
    max_len = MIN3 (MAX_WRITE, flux_fifo_get_used (fifo),
                    FLUX_IO (tcp_io)->write_slice_cap - FLUX_IO (tcp_io)->write_count);
  else
    max_len = 0;

  _flux_io_unlock ();

  return max_len;
}

static gboolean
main_callback (FluxTcpIO *tcp_io)
{
  flux_object_ref (FLUX_OBJECT (tcp_io));
  g_static_mutex_lock (&tcp_io_mutex);

  DEBUG (g_print ("main_callback()\n"));

  tcp_io->callback_id = 0;

  dispatch_link_pre_data (tcp_io);   /* Connected? */
  dispatch_reads (tcp_io);
  dispatch_link_post_data (tcp_io);  /* Disconnected? */
  dispatch_writes (tcp_io);

  DEBUG (g_print ("main_callback() done\n"));

  g_static_mutex_unlock (&tcp_io_mutex);
  flux_object_unref (FLUX_OBJECT (tcp_io));
  return FALSE;
}

static void
install_callback (FluxTcpIO *tcp_io)
{
  if (tcp_io->callback_id == 0)
    tcp_io->callback_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) main_callback, tcp_io, NULL);
}

static void
maybe_install_callback_for_write (FluxTcpIO *tcp_io)
{
  FluxFifo *fifo;
  gint      used;

  if (tcp_io->state != FLUX_TCP_IO_STATE_CONNECTED)
    return;

  fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (tcp_io));

  _flux_io_lock ();

  used = flux_fifo_get_used (fifo);

  if (FLUX_IO (tcp_io)->write_threshold > 0 && FLUX_IO (tcp_io)->write_func &&
      (used == 0 || used + FLUX_IO (tcp_io)->write_threshold <= MAX_WRITE_BUFFER))
  {
    install_callback (tcp_io);
  }

  _flux_io_unlock ();  
}

static void
worker_exception (FluxTcpIO *tcp_io)
{
  DEBUG (g_print ("worker_exception ()\n"));

  /* Connecting socket: Connection failed.
   * Connected socket:  Other error. */

  if (!tcp_io->have_connected_fd)
  {
    /* Connection failed */

    DEBUG (g_print ("Exception (connection failed?).\n"));
    set_disconnected (tcp_io);
  }
  else
  {
    /* Other error */

    DEBUG (g_print ("Exception (other error).\n"));
    set_disconnected (tcp_io);
  }

  install_callback (tcp_io);
}

static void
worker_write (FluxTcpIO *tcp_io)
{
  guchar buf [MAX_WRITE];
  gint  max_len;
  gint  result;

  DEBUG (g_print ("worker_write ()\n"));

  /* Connecting socket: Connection attempt finished. [WIN32]
   * Connected socket:  Can write. */

  if (!tcp_io->have_connected_fd)
  {
    gint err;
    gint err_len = sizeof (err);

    /* Connection attempt finished */

#ifndef G_PLATFORM_WIN32
    getsockopt (tcp_io->fd, SOL_SOCKET, SO_ERROR, &err, &err_len);
#else
    getsockopt (tcp_io->fd, SOL_SOCKET, SO_ERROR, (gchar *) &err, &err_len);
#endif

    if (err == 0)
    {
      DEBUG (g_print ("Got write! Connected.\n"));

      if (set_connected (tcp_io, TRUE))
        install_callback (tcp_io);
    }
    else
    {
      DEBUG (g_print ("Got write! Failed connect.\n"));

      set_disconnected (tcp_io);
      install_callback (tcp_io);
    }
  }
  else
  {
    FluxFifo *fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (tcp_io));

    /* Data */

    max_len = network_max_write (tcp_io);
    if (max_len < 1)
      return;

    flux_fifo_peek (fifo, buf, max_len);

    result = send (tcp_io->fd, buf, max_len, 0);

    if (result < 0)
    {
      /* Broken pipe */

      DEBUG (g_print ("send() failed - broken pipe.\n"));
      set_disconnected (tcp_io);
    }
    else
    {
      flux_fifo_drop (fifo, result);

      _flux_io_lock ();
      FLUX_IO (tcp_io)->write_count += result;
      _flux_io_unlock ();
    }

    install_callback (tcp_io);
  }
}

static void
worker_read (FluxTcpIO *tcp_io)
{
  guchar buf [MAX_READ];
  gint   max_len;
  gint   result;

  DEBUG (g_print ("worker_read ()\n"));

  /* Connecting socket: Connection attempt finished. [UNIX]
   * Connected socket:  Got data (n > 0) or EOF (n == 0). */

  if (!tcp_io->have_connected_fd)
  {
    if (set_connected (tcp_io, TRUE))
      install_callback (tcp_io);
  }

  max_len = network_max_read (tcp_io);
  if (max_len < 1)
    return;

  result = recv (tcp_io->fd, buf, max_len, 0);

  if (result < 1)
  {
    /* EOF */

    DEBUG (g_print ("Read failed! EOF.\n"));
    set_disconnected (tcp_io);
  }
  else
  {
    FluxFifo *fifo = flux_dual_fifo_get_read_fifo (FLUX_DUAL_FIFO (tcp_io));
    gboolean  enqueue_success;

    /* Data */

    if (tcp_io->reading_cipher_confirmation)
    {
      enqueue_cipher_confirmation_bytes (tcp_io, fifo, buf, result);
      enqueue_success = TRUE;
    }
    else
      enqueue_success = flux_fifo_enqueue (fifo, buf, result);

    if (enqueue_success && tcp_io->reading_cipher_invitation)
      try_read_cipher_invitation (tcp_io);

    _flux_io_lock ();
    FLUX_IO (tcp_io)->read_count += result;
    _flux_io_unlock ();

    if (!enqueue_success)
    {
      /* Decode error (typically in decompression) */

      DEBUG (g_print ("Decode failed! Closing TCP IO.\n"));
      set_disconnected (tcp_io);
    }
  }

  install_callback (tcp_io);
}

static void
worker_check_close (FluxTcpIO *tcp_io)
{
  FluxFifo *fifo = flux_dual_fifo_get_write_fifo (FLUX_DUAL_FIFO (tcp_io));
  guint     t;

  t = time (NULL);

  if (tcp_io->disconnect_request_flag &&
      (!flux_fifo_get_used (fifo) ||
       (tcp_io->timestamp && tcp_io->timestamp + DISCONNECT_TIMEOUT_SEC <= t)))
  {
    /* Stalled on drain, or fully drained */
    set_disconnected (tcp_io);
    install_callback (tcp_io);
  }
  else if (tcp_io->timestamp && tcp_io->timestamp + CONNECT_TIMEOUT_SEC <= t &&
           IS_IN_CIPHER_EXCHANGE (tcp_io))
  {
    /* Stalled on cipher exchange */
    set_disconnected (tcp_io);
    install_callback (tcp_io);
  }
}

static void
worker_wakeup (void)
{
  /* Empty the wakeup pipe */

  DEBUG (g_print ("worker_wakeup ()\n"));
  flux_wakeup_pipe_handle_wakeup (tcp_io_wakeup_fd);
}

static gboolean
worker_should_read (FluxTcpIO *tcp_io)
{
  if (!tcp_io->have_connected_fd)
#ifndef G_PLATFORM_WIN32
    return TRUE;
#else
    return FALSE;
#endif

  if (tcp_io->reading_cipher_invitation || tcp_io->reading_cipher_confirmation)
    return TRUE;

  return network_max_read (tcp_io) > 0 ? TRUE : FALSE;
}

static gboolean
worker_should_write (FluxTcpIO *tcp_io)
{
  if (!tcp_io->have_connected_fd)
#ifndef G_PLATFORM_WIN32
    return FALSE;
#else
    return TRUE;
#endif

  return network_max_write (tcp_io) > 0 ? TRUE : FALSE;
}

static gpointer
worker_main (void)
{
  g_static_mutex_lock (&tcp_io_mutex);

  while (tcp_io_list)
  {
    GList  *l, *next;
    fd_set  read_fds;
    fd_set  write_fds;
    fd_set  exception_fds;
    gint    fd_max = flux_wakeup_pipe_get_watch_fd (tcp_io_wakeup_fd);
    gint    result;

    /* Establish watch parameters */

    FD_ZERO (&read_fds);
    FD_ZERO (&write_fds);
    FD_ZERO (&exception_fds);

    FD_SET (flux_wakeup_pipe_get_watch_fd (tcp_io_wakeup_fd), &read_fds);

    for (l = tcp_io_list; l; l = g_list_next (l))
    {
      FluxTcpIO *tcp_io = l->data;

      if (tcp_io->fd < 0)
        continue;

      if (worker_should_read (tcp_io))
      {
        FD_SET (tcp_io->fd, &read_fds);
      }

      if (worker_should_write (tcp_io))
      {
        FD_SET (tcp_io->fd, &write_fds);
      }

      FD_SET (tcp_io->fd, &exception_fds);
      fd_max = MAX (fd_max, tcp_io->fd);

      maybe_install_callback_for_write (tcp_io);
    }

    /* Block on select () */

    g_static_mutex_unlock (&tcp_io_mutex);
    result = select (fd_max + 1, &read_fds, &write_fds, &exception_fds, NULL);
    g_static_mutex_lock (&tcp_io_mutex);

    /* Act on results */

    if (result < 1)
    {
#ifdef G_PLATFORM_WIN32
      DEBUG (g_print ("select() error: %s\n", g_win32_error_message (WSAGetLastError ())));
#endif
      continue;
    }

    if (FD_ISSET (flux_wakeup_pipe_get_watch_fd (tcp_io_wakeup_fd), &read_fds))
      worker_wakeup ();

    for (l = tcp_io_list; l; l = next)
    {
      FluxTcpIO *tcp_io = l->data;

      next = g_list_next (l);

      if (tcp_io->fd < 0)
        continue;

      if (FD_ISSET (tcp_io->fd, &exception_fds))
        worker_exception (tcp_io);

      if (tcp_io->fd < 0)
        continue;

      if (FD_ISSET (tcp_io->fd, &write_fds))
        worker_write (tcp_io);

      if (tcp_io->fd < 0)
        continue;

      if (FD_ISSET (tcp_io->fd, &read_fds))
        worker_read (tcp_io);

      if (tcp_io->fd < 0)
        continue;

      worker_check_close (tcp_io);
    }
  }

  tcp_io_thread = NULL;
  g_static_mutex_unlock (&tcp_io_mutex);
  return NULL;
}

static void
close_wakeup_pipe (void)
{
  flux_wakeup_pipe_close (tcp_io_wakeup_fd);
}

/* --- *
 * API *
 * --- */

void
flux_tcp_io_init (FluxTcpIO *tcp_io, gint fd, FluxCipher *cipher, gint read_compression_level,
                  gint write_compression_level)
{
  flux_io_init                  (FLUX_IO (tcp_io));
  flux_object_set_type          (FLUX_OBJECT (tcp_io), FLUX_TCP_IO_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (tcp_io), flux_tcp_io_finalize);

  flux_init_sockets ();

  if (!g_thread_supported ())
    g_thread_init (NULL);

  flux_dual_fifo_set_write_hook (FLUX_DUAL_FIFO (tcp_io), user_write_cb, tcp_io);
  flux_dual_fifo_set_read_hook  (FLUX_DUAL_FIFO (tcp_io), user_read_cb,  tcp_io);
  tcp_io->state = FLUX_TCP_IO_STATE_DISCONNECTED;

  flux_io_set_compression (FLUX_IO (tcp_io), read_compression_level, write_compression_level);

  if (cipher)
  {
    flux_object_ref (FLUX_OBJECT (cipher));
    tcp_io->cipher = cipher;
  }

  if (fd >= 0)
  {
    tcp_io->fd = fd;

    flux_socket_set_nonblock (tcp_io->fd, TRUE);

    set_connected (tcp_io, FALSE);
    tcp_io->state = FLUX_TCP_IO_STATE_CONNECTING;
  }
  else
    tcp_io->fd = -1;

  g_static_mutex_lock (&tcp_io_mutex);

  tcp_io_list = g_list_prepend (tcp_io_list, tcp_io);

  if (tcp_io_wakeup_fd [0] < 0)
  {
    flux_wakeup_pipe_create (tcp_io_wakeup_fd);
    g_atexit (close_wakeup_pipe);
  }

  if (!tcp_io_thread)
    tcp_io_thread = g_thread_create_full ((GThreadFunc) worker_main, NULL,
                                          THREAD_STACK_SIZE, FALSE, FALSE,
                                          G_THREAD_PRIORITY_NORMAL, NULL);
  else
    do_wakeup ();

  g_static_mutex_unlock (&tcp_io_mutex);

  DEBUG (ios_open++);
  DEBUG (g_print ("FluxTcpIO init. Open=%d\n", ios_open));
}

FluxTcpIO *
flux_tcp_io_new (void)
{
  FluxTcpIO *tcp_io;

  tcp_io = g_new0 (FluxTcpIO, 1);
  flux_tcp_io_init (tcp_io, -1, NULL, 0, 0);

  return tcp_io;
}

FluxTcpIO *
flux_tcp_io_new_with_fd (gint fd)
{
  FluxTcpIO *tcp_io;

  g_return_val_if_fail (fd >= 0, NULL);

  tcp_io = g_new0 (FluxTcpIO, 1);
  flux_tcp_io_init (tcp_io, fd, NULL, 0, 0);

  return tcp_io;
}

FluxTcpIO *
flux_tcp_io_new_with_cipher (FluxCipher *cipher)
{
  FluxTcpIO *tcp_io;

  g_return_val_if_fail (cipher != NULL, NULL);

  tcp_io = g_new0 (FluxTcpIO, 1);
  flux_tcp_io_init (tcp_io, -1, cipher, 0, 0);

  return tcp_io;
}

FluxTcpIO *
flux_tcp_io_new_with_fd_and_cipher (gint fd, FluxCipher *cipher)
{
  FluxTcpIO *tcp_io;

  g_return_val_if_fail (cipher != NULL, NULL);
  g_return_val_if_fail (fd >= 0, NULL);

  tcp_io = g_new0 (FluxTcpIO, 1);
  flux_tcp_io_init (tcp_io, fd, cipher, 0, 0);

  return tcp_io;
}

void
flux_tcp_io_finalize (FluxObject *object)
{
  FluxTcpIO      *tcp_io = FLUX_TCP_IO (object);
  FluxTcpIOState  state;

  assert_types (tcp_io);

  DEBUG (g_print ("Finalizing TcpIO."));

  g_static_mutex_lock (&tcp_io_mutex);

  state = flux_tcp_io_get_state (FLUX_TCP_IO (tcp_io));

  if (state != FLUX_TCP_IO_STATE_DISCONNECTED)
  {
    tcp_io->disconnect_request_flag = TRUE;
    tcp_io->state = FLUX_TCP_IO_STATE_DISCONNECTING;

    tcp_io->link_func = NULL;

    flux_io_set_read_func  (FLUX_IO (tcp_io), NULL, NULL);
    flux_io_set_write_func (FLUX_IO (tcp_io), NULL, NULL);

    /* Child finalizers have run, reducing this object to a FluxTcpIO */
    flux_object_set_type          (object, FLUX_TCP_IO_TYPE);
    flux_object_set_finalize_func (object, flux_tcp_io_finalize);

    /* Start drain phase */
    flux_object_ref (object);
    DEBUG (g_print ("FluxTcpIO finalization delayed. Open=%d\n", g_list_length (tcp_io_list)));

    do_wakeup ();
  }
  else
  {
    if (tcp_io->callback_id != 0)
    {
      g_source_remove (tcp_io->callback_id);
      tcp_io->callback_id = 0;
    }

    if (tcp_io->lookup_id != 0)
    {
      flux_ip_resolver_cancel_lookup (flux_ip_resolver_get_default (), tcp_io->lookup_id);
      tcp_io->lookup_id = 0;
    }

    tcp_io_list = g_list_remove (tcp_io_list, tcp_io);

    if (tcp_io->cipher)
      flux_object_unref (FLUX_OBJECT (tcp_io->cipher));

    if (tcp_io->remote_ip_addr)
      flux_object_unref (FLUX_OBJECT (tcp_io->remote_ip_addr));

    flux_io_finalize (object);

    DEBUG (ios_open--);
    DEBUG (g_print ("FluxTcpIO finalized. Open=%d\n", g_list_length (tcp_io_list)));
  }

  g_static_mutex_unlock (&tcp_io_mutex);
}

static void
lookup_cb (GList *ip_addr_list, GList *name_list, FluxTcpIO *tcp_io)
{
  FluxIPAddr      *ip_addr;
  struct sockaddr *sa = NULL;
  FluxIPAddrType   family;

  g_static_mutex_lock (&tcp_io_mutex);

  tcp_io->lookup_id = 0;

  if (!ip_addr_list)
  {
    DEBUG (g_print ("Name resolution failed.\n"));
    goto error;
  }

  ip_addr = ip_addr_list->data;
  DEBUG (g_print ("%p: Name resolved to: %s\n", tcp_io, flux_ip_addr_get_string (ip_addr)));

  family = flux_ip_addr_get_type (ip_addr);

  tcp_io->have_connected_fd = FALSE;

  tcp_io->fd = socket (family == FLUX_IP_ADDR_IPV6 ? AF_INET6 : AF_INET, SOCK_STREAM, 0);
  if (tcp_io->fd < 0)
  {
    DEBUG (g_print ("Could not open socket.\n"));
    goto error;
  }

  if (!flux_socket_set_nonblock (tcp_io->fd, TRUE))
  {
    DEBUG (g_print ("Could not set nonblock.\n"));
    goto error;
  }

  sa = flux_ip_addr_get_sockaddr (ip_addr, tcp_io->remote_port);
  if (!sa)
  {
    DEBUG (g_print ("Could not get sockaddr.\n"));
    goto error;
  }

  if (connect (tcp_io->fd, sa, flux_sockaddr_get_len (sa)) != 0 &&
#ifndef G_PLATFORM_WIN32
      errno != EINPROGRESS)
#else
      WSAGetLastError () != WSAEWOULDBLOCK)
#endif
  {
    DEBUG (g_print ("Got error from connect(): %d\n", errno));
    goto error;
  }

  flux_object_ref (FLUX_OBJECT (ip_addr));

  if (tcp_io->remote_ip_addr)
    flux_object_unref (FLUX_OBJECT (tcp_io->remote_ip_addr));

  tcp_io->remote_ip_addr = ip_addr;

  do_wakeup ();

  g_free (sa);
  g_static_mutex_unlock (&tcp_io_mutex);
  return;

error:
  if (tcp_io->fd >= 0)
  {
    flux_socket_close (tcp_io->fd);
    tcp_io->fd = -1;
  }

  g_free (sa);

  tcp_io->state = FLUX_TCP_IO_STATE_DISCONNECTED;

  if (tcp_io->link_func)
  {
    g_static_mutex_unlock (&tcp_io_mutex);
    tcp_io->link_func (tcp_io, FLUX_TCP_CONNECTION_FAILED, tcp_io->link_func_data);
 }
  else
  {
    g_static_mutex_unlock (&tcp_io_mutex);
  }
}

void
flux_tcp_io_connect_begin (FluxTcpIO *tcp_io, const gchar *host, guint port)
{
  assert_types (tcp_io);

  if (flux_tcp_io_get_state (tcp_io) != FLUX_TCP_IO_STATE_DISCONNECTED)
  {
    DEBUG (g_print ("Cannot connect - bad state.\n"));
    return;
  }

  g_static_mutex_lock (&tcp_io_mutex);

  reset_socket (tcp_io);
  tcp_io->state = FLUX_TCP_IO_STATE_CONNECTING;
  tcp_io->remote_port = port;

  DEBUG (g_print ("Resolving...\n"));

  tcp_io->lookup_id = flux_ip_resolver_lookup (flux_ip_resolver_get_default (), host,
                                               (FluxIPLookupFunc *) lookup_cb, tcp_io);

  g_static_mutex_unlock (&tcp_io_mutex);
}

void
flux_tcp_io_disconnect (FluxTcpIO *tcp_io)
{
  assert_types (tcp_io);

  g_static_mutex_lock (&tcp_io_mutex);
  tcp_io->disconnect_request_flag = TRUE;
  tcp_io->state = FLUX_TCP_IO_STATE_DISCONNECTING;
  g_static_mutex_unlock (&tcp_io_mutex);
}

void
flux_tcp_io_set_link_func (FluxTcpIO *tcp_io, FluxTcpLinkFunc *func, gpointer data)
{
  assert_types (tcp_io);

  g_static_mutex_lock (&tcp_io_mutex);

  tcp_io->link_func      = func;
  tcp_io->link_func_data = data;

  g_static_mutex_unlock (&tcp_io_mutex);
}

void
flux_tcp_io_set_cipher (FluxTcpIO *tcp_io, FluxCipher *cipher)
{
  assert_types (tcp_io);

  g_static_mutex_lock (&tcp_io_mutex);

  if (tcp_io->cipher == cipher)
  {
    g_static_mutex_unlock (&tcp_io_mutex);
    return;
  }

  if (tcp_io->cipher)
    flux_object_unref (FLUX_OBJECT (tcp_io->cipher));

  tcp_io->cipher = cipher;
  flux_object_ref (FLUX_OBJECT (tcp_io->cipher));

  g_static_mutex_unlock (&tcp_io_mutex);
}

FluxCipher *
flux_tcp_io_get_cipher (FluxTcpIO *tcp_io)
{
  assert_types (tcp_io);

  return tcp_io->cipher;
}

FluxTcpIOState
flux_tcp_io_get_state (FluxTcpIO *tcp_io)
{
  assert_types (tcp_io);

  return tcp_io->state;
}

gint
flux_tcp_io_get_fd (FluxTcpIO *tcp_io)
{
  assert_types (tcp_io);

  return tcp_io->fd;
}

FluxIPAddr *
flux_tcp_io_get_local_ip_addr (FluxTcpIO *tcp_io)
{
  gchar       sa [256];
  gint        len = 256;
  FluxIPAddr *ip_addr;

  assert_types (tcp_io);

  if (tcp_io->state == FLUX_TCP_IO_STATE_DISCONNECTED)
    return NULL;

  if (tcp_io->fd < 0)
    return NULL;

  if (getsockname (tcp_io->fd, (struct sockaddr *) sa, &len) != 0)
    return NULL;

  ip_addr = flux_ip_addr_new ();
  flux_ip_addr_set_sockaddr (ip_addr, (struct sockaddr *) sa);

  return ip_addr;
}

gint
flux_tcp_io_get_local_port (FluxTcpIO *tcp_io)
{
  gchar sa [256];
  gint  len = 256;

  assert_types (tcp_io);

  if (tcp_io->state == FLUX_TCP_IO_STATE_DISCONNECTED)
    return -1;

  if (tcp_io->fd < 0)
    return -1;

  if (getsockname (tcp_io->fd, (struct sockaddr *) sa, &len) != 0)
    return -1;

  return flux_sockaddr_get_port ((struct sockaddr *) sa);
}

FluxIPAddr *
flux_tcp_io_get_remote_ip_addr (FluxTcpIO *tcp_io)
{
  assert_types (tcp_io);

  if (tcp_io->state == FLUX_TCP_IO_STATE_DISCONNECTED)
    return NULL;

  if (tcp_io->remote_ip_addr)
    flux_object_ref (FLUX_OBJECT (tcp_io->remote_ip_addr));

  return tcp_io->remote_ip_addr;
}

gint
flux_tcp_io_get_remote_port (FluxTcpIO *tcp_io)
{
  assert_types (tcp_io);

  if (tcp_io->state == FLUX_TCP_IO_STATE_DISCONNECTED)
    return -1;

  return tcp_io->remote_port;
}

void
_flux_tcp_io_set_remote_info (FluxTcpIO *tcp_io, FluxIPAddr *ip_addr, gint port)
{
  if (tcp_io->remote_ip_addr)
    flux_object_unref (FLUX_OBJECT (tcp_io->remote_ip_addr));

  tcp_io->remote_ip_addr = ip_addr;
  flux_object_ref (FLUX_OBJECT (ip_addr));

  tcp_io->remote_port = port;
}
