/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-tcp-listener.c - Listener for TCP connections.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <sys/time.h>
#else
# include <io.h>
# include <process.h>
# include <winsock2.h>
# include <ws2tcpip.h>
# include <time.h>
# include "win32/win32-compat.h"
#endif

#include "flux-init.h"
#include "flux-tcp-io.h"
#include "flux-util.h"
#include "flux-socket-util.h"
#include "flux-tcp-listener.h"

#define DEBUG(x)

#define THREAD_STACK_SIZE 16384

void _flux_tcp_io_set_remote_info (FluxTcpIO *tcp_io, FluxIPAddr *ip_addr, gint port);

static GStaticMutex  tcp_listener_mutex         = G_STATIC_MUTEX_INIT;
static GThread      *tcp_listener_thread        = NULL;
static GList        *tcp_listener_list          = NULL;
static gint          tcp_listener_wakeup_fd [2] = { -1, -1 };

static void
do_wakeup (void)
{
  flux_wakeup_pipe_wakeup (tcp_listener_wakeup_fd);
}

static void
connect_ready_cb (FluxTcpIO *tio, FluxTcpLinkMessage message, gpointer data)
{
  FluxTcpListener *tl = data;

  DEBUG (g_print ("Got post-cipher-exchange connect callback.\n"));

  g_hash_table_remove (tl->wait_connect_table, tio);

  flux_tcp_io_set_link_func (FLUX_TCP_IO (tio), NULL, NULL);

  if (message == FLUX_TCP_CONNECTION_ESTABLISHED)
  {
    DEBUG (g_print ("Connection established, handing off to client.\n"));
    tl->connection_func (tl, FLUX_TCP_IO (tio), tl->connection_func_data);
  }
  else
  {
    /* Cipher negotiation failed, or disconnect */
    DEBUG (g_print ("No connection, terminating.\n"));
    flux_object_unref (FLUX_OBJECT (tio));
  }
}

static gboolean
connect_cb (FluxTcpListener *tcp_listener)
{
  gchar                    sa [256];
  gint                     sa_len = sizeof (sa);
  FluxTcpIO               *tio;
  FluxCipher              *tio_cipher;
  gint                     fd;

  g_static_mutex_lock (&tcp_listener_mutex);

  tcp_listener->callback_id = 0;

  DEBUG (g_print ("Doing accept ()...\n"));
  fd = accept (tcp_listener->fd, (struct sockaddr *) &sa, &sa_len);
  DEBUG (g_print ("Done.\n"));

  if (fd < 0)
  {
    DEBUG (g_print ("accept() yielded bad fd.\n"));
    do_wakeup ();
    g_static_mutex_unlock (&tcp_listener_mutex);
    return FALSE;
  }

  if (tcp_listener->connection_func)
  {
    FluxIPAddr *remote_ip_addr;
    gint        remote_port;

    remote_ip_addr = flux_ip_addr_new ();
    flux_ip_addr_set_sockaddr (remote_ip_addr, (struct sockaddr *) sa);
    remote_port = flux_sockaddr_get_port ((struct sockaddr *) sa);

    DEBUG (g_print ("remote addr = %s, port = %d\n", flux_ip_addr_get_string (remote_ip_addr), remote_port));

    tio = tcp_listener->spawned_object_size > sizeof (FluxTcpIO) ?
      g_malloc0 (tcp_listener->spawned_object_size) : g_new0 (FluxTcpIO, 1);

    if (tcp_listener->cipher)
    {
      DEBUG (g_print ("Starting cipher exchange.\n"));
      tio_cipher = flux_cipher_dup (tcp_listener->cipher);
      flux_cipher_randomize (tio_cipher);
      flux_tcp_io_init (tio, fd, tio_cipher, tcp_listener->read_compression_level,
                        tcp_listener->write_compression_level);
      flux_object_unref (FLUX_OBJECT (tio_cipher));
      flux_tcp_io_set_link_func (FLUX_TCP_IO (tio), connect_ready_cb, tcp_listener);

      _flux_tcp_io_set_remote_info (tio, remote_ip_addr, remote_port);
      flux_object_unref (FLUX_OBJECT (remote_ip_addr));

      g_hash_table_insert (tcp_listener->wait_connect_table, tio, tio);
    }
    else
    {
      DEBUG (g_print ("No cipher, handing off to client.\n"));

      flux_tcp_io_init (tio, fd, NULL, tcp_listener->read_compression_level,
                        tcp_listener->write_compression_level);

      _flux_tcp_io_set_remote_info (tio, remote_ip_addr, remote_port);
      flux_object_unref (FLUX_OBJECT (remote_ip_addr));

      g_static_mutex_unlock (&tcp_listener_mutex);
      tcp_listener->connection_func (tcp_listener, tio, tcp_listener->connection_func_data);
      g_static_mutex_lock (&tcp_listener_mutex);
    }
  }
  else
  {
    flux_socket_close (fd);
  }

  do_wakeup ();
  g_static_mutex_unlock (&tcp_listener_mutex);
  return FALSE;
}

static void
worker_accept (FluxTcpListener *tcp_listener)
{
  if (tcp_listener->callback_id != 0)
    return;

  tcp_listener->callback_id = g_idle_add_full (G_PRIORITY_LOW, (GSourceFunc) connect_cb, tcp_listener, NULL);
}

static void
worker_wakeup (void)
{
  /* Empty the wakeup pipe */

  flux_wakeup_pipe_handle_wakeup (tcp_listener_wakeup_fd);
}

static gpointer
worker_main (void)
{
  g_static_mutex_lock (&tcp_listener_mutex);

  while (tcp_listener_list)
  {
    GList  *l;
    fd_set  read_fds;
    gint    fd_max = flux_wakeup_pipe_get_watch_fd (tcp_listener_wakeup_fd);
    gint    result;

    DEBUG (g_print ("watch fd = %d\n", fd_max));

    FD_ZERO (&read_fds);
    FD_SET (flux_wakeup_pipe_get_watch_fd (tcp_listener_wakeup_fd), &read_fds);

    for (l = tcp_listener_list; l; l = g_list_next (l))
    {
      FluxTcpListener *tcp_listener = l->data;

      if (tcp_listener->fd < 0)
        continue;

      if (tcp_listener->callback_id)
        continue;

      FD_SET (tcp_listener->fd, &read_fds);
      fd_max = MAX (fd_max, tcp_listener->fd);
    }

    DEBUG (g_print ("FluxTcpListener select()\n"));

    g_static_mutex_unlock (&tcp_listener_mutex);
    result = select (fd_max + 1, &read_fds, NULL, NULL, NULL);
    g_static_mutex_lock (&tcp_listener_mutex);

    DEBUG (g_print ("FluxTcpListener select() done\n"));

    if (result < 1)
    {
#ifdef G_PLATFORM_WIN32
      DEBUG (g_print ("select() error %d: %s\n", result, g_win32_error_message (WSAGetLastError ())));
#endif
      continue;
    }

    if (FD_ISSET (flux_wakeup_pipe_get_watch_fd (tcp_listener_wakeup_fd), &read_fds))
      worker_wakeup ();

    for (l = tcp_listener_list; l; l = g_list_next (l))
    {
      FluxTcpListener *tcp_listener = l->data;

      if (tcp_listener->fd < 0)
        continue;

      if (FD_ISSET (tcp_listener->fd, &read_fds))
        worker_accept (tcp_listener);
    }
  }

  DEBUG (g_print ("FluxTcpListener worker quitting.\n"));

  tcp_listener_thread = NULL;
  g_static_mutex_unlock (&tcp_listener_mutex);
  return NULL;
}

static gint
create_fd (gint port)
{
  struct sockaddr_in sa_in;
  gint               on = 1;
  gint               fd;

  sa_in.sin_family = AF_INET;
  memset (&sa_in.sin_addr, 0, sizeof (sa_in.sin_addr));
  sa_in.sin_port = g_htons (port);

  fd = socket (AF_INET, SOCK_STREAM, 0);
  if (fd < 0)
  {
    return fd;
  }

#ifndef G_PLATFORM_WIN32
  setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on));
#else
  setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, (gchar *) &on, sizeof (on));
#endif

  if (bind (fd, (struct sockaddr *) &sa_in, sizeof (struct sockaddr_in)) != 0)
  {
    goto error;
  }

  if (listen (fd, 10) != 0)
  {
    goto error;
  }

  flux_socket_set_nonblock (fd, TRUE);
  return fd;

error:
  flux_socket_close (fd);
  return -1;
}

static void
close_wakeup_pipe (void)
{
  flux_wakeup_pipe_close (tcp_listener_wakeup_fd);
}

gboolean
flux_tcp_listener_init (FluxTcpListener *tl, gint port, FluxCipher *cipher)
{
  flux_object_init              (FLUX_OBJECT (tl));
  flux_object_set_type          (FLUX_OBJECT (tl), FLUX_TCP_LISTENER_TYPE);
  flux_object_set_finalize_func (FLUX_OBJECT (tl), flux_tcp_listener_finalize);

  flux_init_sockets ();

  if (!g_thread_supported ())
    g_thread_init (NULL);

  tl->fd              = create_fd (port);
  tl->port            = port;
  tl->cipher          = cipher;
  tl->connection_func = NULL;

  if (tl->fd < 0)
    return FALSE;

  tl->wait_connect_table = g_hash_table_new (g_direct_hash, g_direct_equal);

  if (tl->cipher)
    flux_object_ref (FLUX_OBJECT (tl->cipher));

  g_static_mutex_lock (&tcp_listener_mutex);

  tcp_listener_list = g_list_prepend (tcp_listener_list, tl);

  if (tcp_listener_wakeup_fd [0] < 0)
  {
    flux_wakeup_pipe_create (tcp_listener_wakeup_fd);
    g_atexit (close_wakeup_pipe);
  }

  if (!tcp_listener_thread)
    tcp_listener_thread = g_thread_create_full ((GThreadFunc) worker_main, NULL,
                                                THREAD_STACK_SIZE, FALSE, FALSE,
                                                G_THREAD_PRIORITY_NORMAL, NULL);
  else
    do_wakeup ();

  g_static_mutex_unlock (&tcp_listener_mutex);

  return TRUE;
}

FluxTcpListener *
flux_tcp_listener_new (gint port)
{
  FluxTcpListener *tl;

  tl = g_new0 (FluxTcpListener, 1);

  if (!flux_tcp_listener_init (tl, port, NULL))
  {
    g_free (tl);
    tl = NULL;
  }

  return tl;
}

FluxTcpListener *
flux_tcp_listener_new_with_cipher (gint port, FluxCipher *cipher)
{
  FluxTcpListener *tl;

  tl = flux_tcp_listener_new (port);
  if (tl)
    flux_tcp_listener_set_cipher (tl, cipher);

  return tl;
}

static void
drop_connect (FluxTcpIO *tio)
{
  flux_tcp_io_set_link_func (tio, NULL, NULL);
  flux_object_unref (FLUX_OBJECT (tio));
}

void
flux_tcp_listener_finalize (FluxObject *tl)
{
  DEBUG (g_print ("Finalizing listener.\n"));

  g_static_mutex_lock (&tcp_listener_mutex);

  do_wakeup ();
  tcp_listener_list = g_list_remove (tcp_listener_list, tl);

  if (FLUX_TCP_LISTENER (tl)->callback_id)
    g_source_remove (FLUX_TCP_LISTENER (tl)->callback_id);

  if (FLUX_TCP_LISTENER (tl)->cipher)
    flux_object_unref (FLUX_OBJECT (FLUX_TCP_LISTENER (tl)->cipher));

  /* FIXME: Do this in worker? */
  flux_socket_close (FLUX_TCP_LISTENER (tl)->fd);

  g_hash_table_foreach (FLUX_TCP_LISTENER (tl)->wait_connect_table, (GHFunc) drop_connect, NULL);
  g_hash_table_destroy (FLUX_TCP_LISTENER (tl)->wait_connect_table);

  g_static_mutex_unlock (&tcp_listener_mutex);
}

void
flux_tcp_listener_set_cipher (FluxTcpListener *tl, FluxCipher *cipher)
{
  flux_object_check (FLUX_OBJECT (tl), FLUX_TCP_LISTENER_TYPE);

  if (tl->cipher)
    flux_object_unref (FLUX_OBJECT (tl->cipher));

  tl->cipher = cipher;
  flux_object_ref (FLUX_OBJECT (tl->cipher));
}

void
flux_tcp_listener_get_compression (FluxTcpListener *tl, gint *read_level, gint *write_level)
{
  flux_object_check (FLUX_OBJECT (tl), FLUX_TCP_LISTENER_TYPE);

  if (read_level)
    *read_level = tl->read_compression_level;
  if (write_level)
    *write_level = tl->write_compression_level;
}

void
flux_tcp_listener_set_compression (FluxTcpListener *tl, gint read_level, gint write_level)
{
  flux_object_check (FLUX_OBJECT (tl), FLUX_TCP_LISTENER_TYPE);

  tl->read_compression_level  = read_level;
  tl->write_compression_level = write_level;
}

void
flux_tcp_listener_set_connection_func (FluxTcpListener *tl, FluxTcpConnectionFunc *func, gpointer data)
{
  flux_object_check (FLUX_OBJECT (tl), FLUX_TCP_LISTENER_TYPE);

  tl->connection_func      = func;
  tl->connection_func_data = data;
}

gint
flux_tcp_listener_get_port (FluxTcpListener *tl)
{
  flux_object_check (FLUX_OBJECT (tl), FLUX_TCP_LISTENER_TYPE);

  return tl->port;
}
