/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-tree.c - N-ary tree where nodes can hold any data.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>

#ifndef G_PLATFORM_WIN32
# include <unistd.h>
#else
# include <io.h>
# include <process.h>
#endif

#include "flux-types.h"
#include "flux-rmd160.h"
#include "flux-mem-chunk.h"
#include "flux-string-util.h"
#include "flux-stream-util.h"
#include "flux-tree.h"

#define BLOCK_INCREMENT 1024

static const char *tree_spaces            = " \t\r\n";
static const char *tree_data_terminators  = " \t\r\n\"\\{}";
static const char *tree_quoted_data_terminators = "\"\\";
static const char *tree_printables        =
  " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
  "0123456789()[]<>:;,./=+-_!?@#$%^&*~`";
static const char  tree_brace_open        = '{';
static const char  tree_brace_close       = '}';
static const char  tree_escape            = '\\';
static const char  tree_quote             = '"';

static FluxMemChunk *_flux_tree_mem_chunk = NULL;

static void
free_node (FluxTree *tree)
{
  flux_mem_chunk_free (_flux_tree_mem_chunk, tree);
}

static void
free_data (FluxTree *tree)
{
  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
      g_free (tree->data);

    case FLUX_TREE_DATA_IN_HEADER:
      tree->data_storage_type = FLUX_TREE_NO_DATA;
      tree->data = NULL;
      break;

    case FLUX_TREE_NO_DATA:
      break;

    default:
      g_assert_not_reached ();
  }

  tree->size = 0;
}

#define data_in_memory(tree) (tree->data_storage_type == FLUX_TREE_DATA_IN_MEMORY || \
                              tree->data_storage_type == FLUX_TREE_DATA_IN_HEADER)

static gpointer
pointer_to_data (FluxTree *tree)
{
  gpointer data;

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
      data = tree->data;
      break;

    case FLUX_TREE_DATA_IN_HEADER:
      data = &tree->data;
      break;

    case FLUX_TREE_NO_DATA:
      data = NULL;
      break;

    default:
      data = NULL;  /* Humour compiler */
      g_assert_not_reached ();
      break;
  }

  return data;
}

static void
reparent_siblings (FluxTree *parent_tree, FluxTree *tree)
{
  FluxTree *old_parent_tree, *next_tree;
  
  old_parent_tree = flux_tree_get_parent (tree);
  if (!old_parent_tree)
    return;

  for (tree = flux_tree_get_first_child (old_parent_tree); tree; tree = next_tree)
  {
    next_tree = flux_tree_get_next (tree);
    
    flux_tree_detach (tree);
    flux_tree_add_as_last_child (parent_tree, tree);
  }

  flux_tree_destroy (old_parent_tree);
}

FluxTree *
flux_tree_new (void)
{
  FluxTree *tree;

  if (!_flux_tree_mem_chunk)
  {
    _flux_tree_mem_chunk = flux_mem_chunk_new (64, sizeof (FluxTree));
  }

  tree = flux_mem_chunk_alloc0 (_flux_tree_mem_chunk);
  return tree;
}

FluxTree *
flux_tree_new_with_data (gconstpointer src, gint len)
{
  FluxTree *tree;

  tree = flux_tree_new ();

  flux_tree_data_set (tree, src, len);
  return tree;
}

FluxTree *
flux_tree_new_with_parent_and_data (FluxTree *parent, gconstpointer src, gint len)
{
  FluxTree *tree;

  tree = flux_tree_new_with_data (src, len);
  if (parent)
    flux_tree_add (parent, tree);

  return tree;
}

/* #warning This func can be sped up immensely by not detaching every node before deletion. */

void
flux_tree_destroy (FluxTree *tree)
{
  FluxTree *child, *child_next;

  /* Delete children */
  for (child = flux_tree_get_first_child (tree); child; child = child_next)
  {
    child_next = flux_tree_get_next (child);
    flux_tree_destroy (child);
  }

  /* Unlink */
  flux_tree_detach (tree);

  /* Free self */
  free_data (tree);
  free_node (tree);
}

/**
 * flux_tree_reclaim:
 * @tree: root of the tree to reclaim
 *
 * Free a tree from memory, and all data associated with it.
 * This function will stop at the first tree that has external
 * references, as determined by flux_tree_has_external_ref().
 */
void
flux_tree_reclaim (FluxTree *tree)
{
  FluxTree *child, *child_next;

  if (flux_tree_has_external_ref (tree)) {
    tree->parent = NULL; /* See tree_dealloc in fluxpy.c for why this is ok. */
    return;
  }

  /* Delete children */
  for (child = flux_tree_get_first_child (tree); child; child = child_next)
  {
    child_next = flux_tree_get_next (child);
    flux_tree_reclaim (child);
  }

  /* Unlink */
  flux_tree_detach (tree);

  /* Free self */
  free_data (tree);
  free_node (tree);
}

FluxTree *
flux_tree_dup (FluxTree *tree)
{
  FluxTree *new_node;

  new_node = flux_tree_new ();
  new_node->is_auto_root = tree->is_auto_root;
  new_node->data_storage_type = tree->data_storage_type;
  new_node->size = tree->size;

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
      new_node->data = g_malloc (tree->size);
      memcpy (new_node->data, tree->data, tree->size);
      break;

    case FLUX_TREE_DATA_IN_HEADER:
      new_node->data = tree->data;
      break;

    case FLUX_TREE_NO_DATA:
      break;

    default:
      g_assert_not_reached ();
  }

  return new_node;
}

FluxTree *
flux_tree_dup_all (FluxTree *tree)
{
  FluxTree *new_node, *child, *new_child;

  /* Self */
  new_node = flux_tree_dup (tree);

  /* Children */
  for (child = flux_tree_get_first_child (tree); child; child = flux_tree_get_next (child))
  {
    new_child = flux_tree_dup_all (child);
    flux_tree_add_as_last_child (new_node, new_child);
  }

  return new_node;
}

FluxTree *
flux_tree_get_root (FluxTree *tree)
{
  FluxTree *flux_tree_last;
  
  for (flux_tree_last = tree; tree; tree = flux_tree_get_parent (tree))
    flux_tree_last = tree;

  return flux_tree_last;
}

FluxTree *
flux_tree_get_parent (FluxTree *tree)
{
  return tree->parent;
}

FluxTree *
flux_tree_get_next (FluxTree *tree)
{
  return tree->next;
}

FluxTree *
flux_tree_get_prev (FluxTree *tree)
{
  return tree->prev;
}

FluxTree *
flux_tree_get_first_child (FluxTree *tree)
{
  return tree->child_first;
}

FluxTree *
flux_tree_get_last_child (FluxTree *tree)
{
  return tree->child_last;
}

FluxTree *
flux_tree_get_nth_child (FluxTree *tree, gint n)
{
  FluxTree *child;
  gint      i;

  if (n >= 0)
  {
    /* Forwards */

    child = flux_tree_get_first_child (tree);

    for (i = 0; i < n; i++)
    {
      if (!child)
        break;

      child = flux_tree_get_next (child);
    }
  }
  else
  {
    /* Backwards */

    n = abs (n) - 1;
    child = flux_tree_get_last_child (tree);

    for (i = 0; i < n; i++)
    {
      if (!child)
        break;

      child = flux_tree_get_prev (child);
    }
  }

  return child;
}

FluxTree *
flux_tree_get_first_sibling (FluxTree *tree)
{
  FluxTree *flux_tree_ret;
  
  flux_tree_ret = flux_tree_get_parent (tree);
  if (flux_tree_ret)
    return flux_tree_get_first_child (flux_tree_ret);
  
  return tree;
}

FluxTree *
flux_tree_get_last_sibling (FluxTree *tree)
{
  FluxTree *flux_tree_ret;

  flux_tree_ret = flux_tree_get_parent (tree);
  if (flux_tree_ret)
    return flux_tree_get_last_child (flux_tree_ret);
  
  return tree;
}

FluxTree *
flux_tree_get_nth_sibling (FluxTree *tree, gint n)
{
  FluxTree *parent, *ret;

  parent = flux_tree_get_parent (tree);

  if (parent)
    ret = flux_tree_get_last_child (parent);
  else if (n == 0 || n == -1)
    ret = tree;
  else
    ret = NULL;

  return ret;
}

FluxTree *
flux_tree_get_common_parent (FluxTree *tree_0, FluxTree *tree_1)
{
  gint depth_0, depth_1;
  
  depth_0 = flux_tree_get_depth (tree_0);
  depth_1 = flux_tree_get_depth (tree_1);
  
  while (depth_0 > depth_1)
  {
    tree_0 = flux_tree_get_parent (tree_0);
    depth_0--;
  }

  while (depth_1 > depth_0)
  {
    tree_1 = flux_tree_get_parent (tree_1);
    depth_1--;
  }
  
  /* Depths are now equal */
  
  while (tree_0 != tree_1)
  {
    tree_0 = flux_tree_get_parent (tree_0);
    tree_1 = flux_tree_get_parent (tree_1);
  }

  return tree_0;  /* If NULL, nodes belong to different trees */
}

void
flux_tree_add_as_first_child (FluxTree *parent_tree, FluxTree *tree)
{
  FluxTree *first_tree;
  
  if (flux_tree_get_parent (tree))
  {
    /* FIXME: Looks like it takes all its siblings with it. Why? Stupid me. */
    reparent_siblings (parent_tree, tree);
  }
  else
  {
    first_tree = flux_tree_get_first_child (parent_tree);

    parent_tree->child_first = tree;

    tree->prev   = NULL;
    tree->next   = first_tree;
    tree->parent = parent_tree;

    if (first_tree)
      first_tree->prev = tree;
    else
      parent_tree->child_last = tree;
  }
}

void
flux_tree_add_as_last_child (FluxTree *parent_tree, FluxTree *tree)
{
  FluxTree *last_tree;

  if (flux_tree_get_parent (tree))
  {
    reparent_siblings (parent_tree, tree);
  }
  else
  {
    last_tree = flux_tree_get_last_child(parent_tree);

    parent_tree->child_last = tree;
    tree->prev = last_tree;
    tree->next = 0;
    tree->parent = parent_tree;

    if (last_tree)
      last_tree->next = tree;
    else
      parent_tree->child_first = tree;
  }
}

void
flux_tree_add_as_first_sibling (FluxTree *sibling_tree, FluxTree *tree)
{
  FluxTree *parent_tree;

  parent_tree = flux_tree_get_parent (sibling_tree);

  if (!parent_tree)
  {
    /* Need a new root */
    parent_tree = flux_tree_new ();
    flux_tree_set_auto_root (parent_tree, TRUE);
    flux_tree_add_as_first_child (parent_tree, sibling_tree);
  }

  flux_tree_add_as_first_child (parent_tree, tree);
}

void
flux_tree_add_as_last_sibling (FluxTree *sibling_tree, FluxTree *tree)
{
  FluxTree *parent_tree;
  
  parent_tree = flux_tree_get_parent (sibling_tree);
  
  if (!parent_tree)
  {
    /* Need a new root */
    parent_tree = flux_tree_new ();
    flux_tree_set_auto_root (parent_tree, TRUE);
    flux_tree_add_as_first_child (parent_tree, sibling_tree);
  }

  flux_tree_add_as_last_child (parent_tree, tree);
}

void
flux_tree_add_before (FluxTree *next_tree, FluxTree *tree)
{
  FluxTree *parent_tree;

  parent_tree = flux_tree_get_parent (next_tree);

  if (!parent_tree)
  {
    /* Need a new root */
    parent_tree = flux_tree_new ();
    flux_tree_set_auto_root (parent_tree, TRUE);
    flux_tree_add_as_first_child (parent_tree, tree);
    flux_tree_add_as_last_child (parent_tree, next_tree);
  }
  else
  {
    if (!next_tree->prev)
      parent_tree->child_first = tree;
    else
      next_tree->prev->next = tree;

    tree->next = next_tree;
    tree->prev = next_tree->prev;
    tree->parent = parent_tree;
    next_tree->prev = tree;
  }
}

void
flux_tree_add_after (FluxTree *prev_tree, FluxTree *tree)
{
  FluxTree *parent_tree;
  
  parent_tree = flux_tree_get_parent (prev_tree);
  
  if (!parent_tree)
  {
    /* Need a new root */
    parent_tree = flux_tree_new ();
    flux_tree_set_auto_root (parent_tree, TRUE);
    flux_tree_add_as_first_child (parent_tree, prev_tree);
    flux_tree_add_as_last_child (parent_tree, tree);
  }
  else
  {
    if (!prev_tree->next)
      parent_tree->child_last = tree;
    else
      prev_tree->next->prev = tree;

    tree->prev = prev_tree;
    tree->next = prev_tree->next;
    tree->parent = parent_tree;
    prev_tree->next = tree;
  }
}

void
flux_tree_detach (FluxTree *tree)
{
  if (tree->prev)
    tree->prev->next = tree->next;
  else if (tree->parent)
    tree->parent->child_first = tree->next;
  
  if (tree->next)
    tree->next->prev = tree->prev;
  else if (tree->parent)
    tree->parent->child_last = tree->prev;
  
  tree->prev = tree->next = tree->parent = NULL;
}

/* TODO: Check if this works when nodes already reference each other */

void
flux_tree_swap (FluxTree *tree_0, FluxTree *tree_1)
{
  FluxTree flux_tree_buf;
  
  memcpy (&flux_tree_buf, tree_1, sizeof (flux_tree_buf));
  memcpy (tree_1, tree_0, sizeof (*tree_1));
  memcpy (tree_0, &flux_tree_buf, sizeof (*tree_0));
  
  if (tree_1->prev)
    tree_1->prev->next = tree_1;
  else if (tree_1->parent)
    tree_1->parent->child_first = tree_1;
  
  if (tree_1->next)
    tree_1->next->prev = tree_1;
  else if (tree_1->parent)
    tree_1->parent->child_last = tree_1;  /* FIXME: Optimize */
  
  /* - */
  
  if (tree_0->prev)
    tree_0->prev->next = tree_0;
  else if (tree_0->parent)
    tree_0->parent->child_first = tree_0;
  
  if (tree_0->next)
    tree_0->next->prev = tree_0;
  else if (tree_0->parent)
    tree_0->parent->child_last = tree_0;  /* FIXME: Optimize */
}

gboolean
flux_tree_is_ancestor (FluxTree *ancestor, FluxTree *tree)
{
  for (; tree && ancestor != tree; tree = flux_tree_get_parent (tree))
    ;

  return tree ? TRUE : FALSE;
}

guint32
flux_tree_get_depth (FluxTree *tree)
{
  gint depth;
  
  for (depth = 0; tree; tree = flux_tree_get_parent (tree), depth++)
    ;
  
  return depth - 1;
}

FluxTree *
flux_tree_get_next_in_breadth_with_level (FluxTree *tree, gint depth, gint level)
{
  FluxTree *tree_0;

  tree_0 = tree;

  /* If we're too shallow, try to go down first */

  while (level < depth && flux_tree_get_first_child (tree_0))
  {
    tree_0 = flux_tree_get_first_child (tree_0);
    level++;

    if (level == depth)
      return tree_0;
  }

  for (;;)
  {
    if (level == depth)
    {
      if (flux_tree_get_next (tree_0))
      {
        /* Proceed to next node on this branch/right level */

        tree_0 = flux_tree_get_next (tree_0);
        return tree_0;
      }
      else
      {
        /* Take us out of the local minimum */
        
        while (tree_0 && !flux_tree_get_next (tree_0))
        {
          tree_0 = flux_tree_get_parent (tree_0);
          level--;
        }
        
        if (!tree_0)
          break;

        tree_0 = flux_tree_get_next (tree_0);
      }
    }
    else if (level < depth && flux_tree_get_first_child (tree_0))
    {
      /* Try to go deeper if we have to */
      
      while (flux_tree_get_first_child(tree_0))
      {
        tree_0 = flux_tree_get_first_child (tree_0);
        level++;

        if (level == depth)
          return tree_0;
      }
    }
    else if (level < depth && flux_tree_get_next (tree_0))
    {
      tree_0 = flux_tree_get_next (tree_0);
    }
    else
    {
      /* Back up, this branch doesn't go deep enough 
       * (or we're too deep) */

      while (!flux_tree_get_next (tree_0))
      {
        tree_0 = flux_tree_get_parent (tree_0);

        if (!tree_0)
          break;  /* We hit root, thus no match */

        level--;
      }
      
      if (!tree_0)
        break;  /* Somewhat ugly, but gets the job done */

      tree_0 = flux_tree_get_next (tree_0);
    }
  }

  return NULL;  /* No more at this depth */
}

FluxTree *
flux_tree_get_next_in_same_depth (FluxTree *tree)
{
  gint level;

  level = flux_tree_get_depth (tree);
  return flux_tree_get_next_in_breadth_with_level (tree, level, level);
}

FluxTree *
flux_tree_get_next_infix (FluxTree *tree, FluxTree *top)
{
  if (flux_tree_get_first_child (tree))
    return flux_tree_get_first_child (tree);

  if (tree == top)
    return NULL;

  if (flux_tree_get_next (tree))
    return flux_tree_get_next (tree);

  while (flux_tree_get_parent (tree) && !flux_tree_get_next (flux_tree_get_parent (tree)))
  {
    tree = flux_tree_get_parent (tree);
    if (tree == top)
      return NULL;
  }

  tree = flux_tree_get_parent (tree);
  if (!tree)
    return NULL;
  if (tree == top)
    return NULL;

  return flux_tree_get_next (tree);
}

gboolean
flux_tree_has_children (FluxTree *tree)
{
  return flux_tree_get_first_child (tree) ? TRUE : FALSE;
}

gboolean
flux_tree_is_root (FluxTree *tree)
{
  return tree->parent ? FALSE : TRUE;
}

gboolean
flux_tree_is_leaf (FluxTree *tree)
{
  return tree->child_first ? FALSE : TRUE;
}

gboolean
flux_tree_is_first (FluxTree *tree)
{
  return tree->prev ? FALSE : TRUE;
}

gboolean
flux_tree_is_last (FluxTree *tree)
{
  return tree->next ? FALSE : TRUE;
}

gboolean
flux_tree_is_sibling (FluxTree *tree_0, FluxTree *tree_1)
{
  return tree_0->parent == tree_1->parent ? TRUE : FALSE;
}

gboolean
flux_tree_is_parent_of (FluxTree *parent, FluxTree *child)
{
  FluxTree *tree;
  
  for (tree = flux_tree_get_first_child (parent); tree; tree = flux_tree_get_next (tree))
    if (tree == child)
      return TRUE;
  
  return FALSE;
}

gboolean
flux_tree_has_data (FluxTree *tree)
{
  gboolean result;

  if (flux_tree_data_get_size (tree) > 0)
    result = TRUE;
  else
    result = FALSE;

  return result;
}

/* Shrink or grow area occupied by data, preserving the data that fits.
 * If area grows, it is padded with zero. */
void
flux_tree_data_set_size (FluxTree *tree, gint len)
{
  gint old_len;

  old_len = tree->size;

  if (!len)
  {
    free_data (tree);
    return;
  }

  if (len == old_len)
  {
    return;
  }

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
      if (len <= sizeof (gpointer))
      {
        /* Move to header */
        gpointer data;

        data = tree->data;

        if (len)
          memcpy (&tree->data, data, len);

        g_free (data);
        tree->data_storage_type = FLUX_TREE_DATA_IN_HEADER;
      }
      else
      {
        /* Resize memory block */
        tree->data = g_realloc (tree->data, len);
      }
      break;

    case FLUX_TREE_DATA_IN_HEADER:
      if (len > sizeof (gpointer))
      {
        /* Move out of header */
        gpointer data;

        data = tree->data;
        tree->data = g_malloc (len);
        memcpy (tree->data, &data, old_len);
        tree->data_storage_type = FLUX_TREE_DATA_IN_MEMORY;
      }
      break;

    case FLUX_TREE_NO_DATA:
      if (len > sizeof (gpointer))
      {
        /* Create memory block */
        tree->data = g_malloc (len);
        tree->data_storage_type = FLUX_TREE_DATA_IN_MEMORY;
      }
      else
      {
        /* Do nothing, but indicate we are storing data in header */
        tree->data_storage_type = FLUX_TREE_DATA_IN_HEADER;
      }
      break;

    default:
      g_assert_not_reached ();
  }

  if (len > old_len)
    memset ((gchar *) pointer_to_data (tree) + old_len, 0, len - old_len);

  tree->size = len;
}

void
flux_tree_data_set_bytes (FluxTree *tree, gconstpointer src, guint32 start, guint32 len)
{
  g_return_if_fail (len != 0);

  if (start + len > tree->size)
  {
    flux_tree_data_set_size (tree, start + len);
  }

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
    case FLUX_TREE_DATA_IN_HEADER:
      memcpy ((gchar *) pointer_to_data (tree) + start, src, len);
      break;

    case FLUX_TREE_NO_DATA:
    default:
      g_assert_not_reached ();
  }
}

guint32
flux_tree_data_get_bytes (FluxTree *tree, gpointer dest, guint32 start, guint32 len)
{
  /* This also handles the case where the tree has no data */
  if (start >= tree->size)
    return 0;

  if (start + len >= tree->size)
    len = tree->size - start;

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
    case FLUX_TREE_DATA_IN_HEADER:
      memcpy (dest, (gchar *) pointer_to_data (tree) + start, len);
      break;

    case FLUX_TREE_NO_DATA:
    default:
      g_assert_not_reached ();
  }

  return len;
}

void
flux_tree_data_append_bytes (FluxTree *tree, gconstpointer src, guint32 len)
{
  g_return_if_fail (len != 0);

  flux_tree_data_set_bytes (tree, src, tree->size, len);
}

void
flux_tree_data_prepend_bytes (FluxTree *tree, gconstpointer src, guint32 len)
{
  gchar   *node_data;
  guint32  old_size;

  g_return_if_fail (len != 0);

  old_size = tree->size;
  flux_tree_data_set_size (tree, old_size + len);
  node_data = pointer_to_data (tree);

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
    case FLUX_TREE_DATA_IN_HEADER:
      memmove (node_data + len, node_data, old_size);
      memcpy (node_data, src, len);
      break;

    case FLUX_TREE_NO_DATA:
    default:
      g_assert_not_reached ();
  }
}

void
flux_tree_data_set (FluxTree *tree, gconstpointer src, guint32 len)
{
  g_return_if_fail (src != NULL);

  flux_tree_data_set_size (tree, len);
  if (!len)
    return;

  flux_tree_data_set_bytes (tree, src, 0, len);
}

gchar *
flux_tree_data_get_str (FluxTree *tree)
{
  gchar *str;

  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
    case FLUX_TREE_DATA_IN_HEADER:
      str = g_malloc (tree->size + 1);
      memcpy (str, pointer_to_data (tree), tree->size);
      *(str + tree->size) = '\0';
      break;

    case FLUX_TREE_NO_DATA:
      str = NULL;
      break;

    default:
      str = NULL;
      g_assert_not_reached ();
      break;
  }

  return str;
}

void
flux_tree_data_set_str (FluxTree *tree, const gchar *src)
{
  flux_tree_data_set (tree, src, strlen (src));
}

void
flux_tree_data_delete (FluxTree *tree)
{
  switch (tree->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
    case FLUX_TREE_DATA_IN_HEADER:
      free_data (tree);
      break;

    case FLUX_TREE_NO_DATA:
      break;

    default:
      g_assert_not_reached ();
  }
}

void
flux_tree_data_swap (FluxTree *tree_0, FluxTree *tree_1)
{
  gpointer data;
  gboolean flag;
  guint32  size;

  data                     = tree_0->data;
  tree_0->data             = tree_1->data;
  tree_1->data             = data;

  size                     = tree_0->size;
  tree_0->size             = tree_1->size;
  tree_1->size             = size;

  flag                     = tree_0->data_is_ready;
  tree_0->data_is_ready    = tree_1->data_is_ready;
  tree_1->data_is_ready    = flag;
}

FluxTree *
flux_tree_split (FluxTree *tree, guint32 pos)
{
  FluxTree *new_node;
  guint len, len_first, len_last;

  len = flux_tree_data_get_size (tree);
  g_return_val_if_fail (pos <= len, NULL);

  len_first = pos;
  len_last  = len - pos;

  new_node = flux_tree_new ();
  flux_tree_add_after (tree, new_node);

  if (len_last)
  {
    guchar *ptr;

    ptr = pointer_to_data (tree);
    g_assert (ptr != NULL);
    ptr += len_first;

    flux_tree_data_set (new_node, ptr, len_last);
  }

  /* OPTIMIZE: Can save some time by implementing data_shrink and data_grow.
   * We know we're shrinking here. */
  flux_tree_data_set_size (tree, len_first);

  return new_node;
}

void
flux_tree_join (FluxTree *target, FluxTree *source)
{
  guint     len, pos, increment;
  guchar   *buf;
  FluxTree *child;

  flux_tree_detach (source);

  switch (source->data_storage_type)
  {
    case FLUX_TREE_DATA_IN_MEMORY:
    case FLUX_TREE_DATA_IN_HEADER:
      flux_tree_data_append_bytes (target, pointer_to_data (source), source->size);
      break;

    case FLUX_TREE_NO_DATA:
      break;

    default:
      increment = BLOCK_INCREMENT;
      buf = g_malloc (increment);

      for (len = source->size, pos = 0; len > 0; len -= increment)
      {
        if (pos + len > increment)
          increment = len - pos;

        flux_tree_data_get_bytes (source, buf, pos, increment);
        flux_tree_data_append_bytes (target, buf, increment);
      }

      g_free (buf);
      break;
  }

  while ((child = flux_tree_get_first_child (source)))
  {
    flux_tree_detach (child);
    flux_tree_add_as_last_child (target, child);
  }

  flux_tree_destroy (source);
}

static gint
cmp_processor (guchar *block1, guchar *block2,
               gint pos, gint size,
               gconstpointer data)
{
  return memcmp (block1, block2, size);
}

gint
flux_tree_cmp (FluxTree *tree_0, FluxTree *tree_1)
{
  gint ret;
  gint len;

  if (flux_tree_data_get_size (tree_0) < flux_tree_data_get_size (tree_1))
    return -1;
  else if (flux_tree_data_get_size (tree_0) > flux_tree_data_get_size (tree_1))
    return 1;

  len = flux_tree_data_get_size (tree_0);
  if (len == 0)
    return 0;

  if (data_in_memory (tree_0) && data_in_memory (tree_1))
  {
    ret = memcmp (pointer_to_data (tree_0), pointer_to_data (tree_1), len);
  }
  else
  {
    ret = flux_stream_process_dual (tree_0, tree_1, BLOCK_INCREMENT, len,
                                    (FluxStreamDualProcessorFunc *) cmp_processor,
                                    (FluxStreamReaderFunc *) flux_tree_data_get_bytes,
                                    (FluxStreamReaderFunc *) flux_tree_data_get_bytes,
                                    NULL);
  }
  
  return ret;
}

static gint
casecmp_processor (guchar *block1, guchar *block2,
                   gint pos, gint size,
                   gconstpointer data)
{
  return flux_memcasecmp (block1, block2, size);
}

gint
flux_tree_casecmp (FluxTree *tree_0, FluxTree *tree_1)
{
  gint ret;
  gint len;

  if (flux_tree_data_get_size (tree_0) < flux_tree_data_get_size (tree_1))
    return -1;
  else if (flux_tree_data_get_size (tree_0) > flux_tree_data_get_size (tree_1))
    return 1;

  len = flux_tree_data_get_size (tree_0);
  if (len == 0)
    return 0;

  if (data_in_memory (tree_0) && data_in_memory (tree_1))
  {
    ret = flux_memcasecmp (pointer_to_data (tree_0), pointer_to_data (tree_1), len);
  }
  else
  {
    ret = flux_stream_process_dual (tree_0, tree_1, BLOCK_INCREMENT, len,
                                    (FluxStreamDualProcessorFunc *) casecmp_processor,
                                    (FluxStreamReaderFunc *) flux_tree_data_get_bytes,
                                    (FluxStreamReaderFunc *) flux_tree_data_get_bytes,
                                    NULL);
  }

  return ret;
}

static gint
memcmp_processor (guchar *block,
                  gint pos, gint size,
                  gconstpointer data)
{
  return memcmp (block, (guchar *) data + pos, size);
}

gint
flux_tree_memcmp (FluxTree *tree, gconstpointer p, guint32 len)
{
  gint ret;

  if (flux_tree_data_get_size (tree) < len)
    return -1;
  else if (flux_tree_data_get_size (tree) > len)
    return 1;

  if (len == 0)
    return 0;

  if (data_in_memory (tree))
  {
    ret = memcmp (pointer_to_data (tree), p, len);
  }
  else
  {
    ret = flux_stream_process (tree, BLOCK_INCREMENT, len,
                               (FluxStreamProcessorFunc *) memcmp_processor,
                               (FluxStreamReaderFunc *) flux_tree_data_get_bytes,
                               (gpointer) p);
  }

  return ret;
}

static gint
memcasecmp_processor (guchar *block,
                      gint pos, gint size,
                      gconstpointer data)
{
  return flux_memcasecmp (block, (guchar *) data + pos, size);
}

gint
flux_tree_memcasecmp (FluxTree *tree, gconstpointer p, guint32 len)
{
  gint ret;

  if (flux_tree_data_get_size (tree) < len)
    return -1;
  else if (flux_tree_data_get_size (tree) > len)
    return 1;

  if (len == 0)
    return 0;

  if (data_in_memory (tree))
  {
    ret = flux_memcasecmp (pointer_to_data (tree), p, len);
  }
  else
  {
    ret = flux_stream_process (tree, BLOCK_INCREMENT, len,
                               (FluxStreamProcessorFunc *) memcasecmp_processor,
                               (FluxStreamReaderFunc *) flux_tree_data_get_bytes,
                               (gpointer) p);
  }

  return ret;
}

/* FIXME: Awkward function */
size_t
flux_tree_spn_from (FluxTree *tree, const gchar *accept, size_t start)
{
  gint    ret = -1;
  guint32 n;
  gint    count;
  guint32 len = flux_tree_data_get_size (tree);
  guchar *block;

  g_return_val_if_fail (start < len, 0);

  if (data_in_memory (tree))
    return flux_memspn ((guchar *) pointer_to_data (tree) + start, len - start, accept);

  block = g_malloc (BLOCK_INCREMENT);

  for (count = start; count + BLOCK_INCREMENT <= len; count += BLOCK_INCREMENT)
  {
    flux_tree_data_get_bytes (tree, block, count, BLOCK_INCREMENT);
    n = flux_memspn (block, BLOCK_INCREMENT, accept);

    if (n < BLOCK_INCREMENT)
    {
      ret = count + n;
      break;
    }
  }

  if (ret == -1)
  {
    if (count < len)
    {
      flux_tree_data_get_bytes (tree, block, count, len - count);
      n = flux_memspn (block, len - count, accept);
    
      count += n;
    }
    
    ret = count;
  }
  
  g_free (block);
  return ret;
}

/* FIXME: Awkward function */
size_t
flux_tree_cspn_from (FluxTree *tree, const gchar *reject, size_t start)
{
  gint     ret = -1;
  guint32  n;
  gint     count;
  guint32  len = flux_tree_data_get_size (tree);
  guchar  *block;

  g_return_val_if_fail (start < len, 0);

  if (data_in_memory (tree))
    return flux_memcspn ((guchar *) pointer_to_data (tree) + start, len - start, reject);

  block = g_malloc (BLOCK_INCREMENT);

  for (count = start; count + BLOCK_INCREMENT <= len; count += BLOCK_INCREMENT)
  {
    flux_tree_data_get_bytes (tree, block, count, BLOCK_INCREMENT);
    n = flux_memcspn (block, BLOCK_INCREMENT, reject);

    if (n < BLOCK_INCREMENT)
    {
      ret = count + n;
      break;
    }
  }

  if (ret == -1)
  {
    if (count < len)
    {
      flux_tree_data_get_bytes (tree, block, count, len - count);
      n = flux_memcspn (block, len - count, reject);

      count += n;
    }
    
    ret = count;
  }
  
  g_free (block);
  return ret;
}

/* FIXME: Awkward function */
gint
flux_tree_chr (FluxTree *tree, gint c)
{
  gint      ret = -1;
  gpointer  p;
  gint      count;
  guint32   len = flux_tree_data_get_size (tree);
  guchar   *block;

  if (!flux_tree_has_data (tree))
    return -1;

  if (data_in_memory (tree))
  {
    p = memchr (pointer_to_data (tree), c, len);
    if (p)
      return (gchar *) p - (gchar *) pointer_to_data (tree);
    else
      return -1;
  }

  block = g_malloc (BLOCK_INCREMENT);

  for (count = 0; count + BLOCK_INCREMENT <= len; count += BLOCK_INCREMENT)
  {
    flux_tree_data_get_bytes (tree, block, count, BLOCK_INCREMENT);
    p = memchr (block, c, BLOCK_INCREMENT);

    if (p)
    {
      ret = count + ((guint) p - (guint) block);
      break;
    }
  }

  if (ret == -1 && count < len)
  {
    flux_tree_data_get_bytes (tree, block, count, len - count);
    p = memchr (block, c, len - count);

    if (p)
      ret = count + ((guint) p - (guint) block);
  }

  g_free (block);
  return ret;
}

/* FIXME: Awkward function */
gint
flux_tree_rchr (FluxTree *tree, gint c)
{
  gint     ret = -1;
  gpointer p;
  gint     count;
  gint     len = flux_tree_data_get_size (tree);
  guchar  *block;

  if (!flux_tree_has_data (tree))
    return -1;
  
  if (data_in_memory (tree))
  {
    p = flux_memrchr (pointer_to_data (tree), c, len);
    if (p)
      return (guint) p - (guint) pointer_to_data (tree);
    else
      return -1;
  }

  block = g_malloc (BLOCK_INCREMENT);
  
  for (count = len - BLOCK_INCREMENT; count >= 0; count -= BLOCK_INCREMENT)
  {
    flux_tree_data_get_bytes (tree, block, count, BLOCK_INCREMENT);
    p = flux_memrchr (block, c, BLOCK_INCREMENT);

    if (p)
    {
      ret = count + ((guint) p - (guint) block);
      break;
    }
  }

  if (ret == -1 && count < 0)
  {
    flux_tree_data_get_bytes (tree, block, 0, count + BLOCK_INCREMENT);
    p = flux_memrchr (block, c, count + BLOCK_INCREMENT);

    if (p)
      ret = (guint) p - (guint) block;
  }

  g_free (block);
  return ret;
}

FluxTree *
flux_tree_find_first_sibling (FluxTree *tree, gconstpointer data, guint32 len)
{
  for (tree = flux_tree_get_first_sibling (tree); tree; tree = flux_tree_get_next (tree))
    if (!flux_tree_memcmp (tree, data, len))
      break;

  return tree;
}

FluxTree *
flux_tree_find_last_sibling (FluxTree *tree, gconstpointer data, guint32 len)
{
  for (tree = flux_tree_get_last_sibling (tree); tree; tree = flux_tree_get_prev (tree))
    if (!flux_tree_memcmp (tree, data, len))
      break;

  return tree;
}

FluxTree *
flux_tree_find_next (FluxTree *tree, gconstpointer data, guint32 len)
{
  for (tree = flux_tree_get_next (tree); tree; tree = flux_tree_get_next (tree))
    if (!flux_tree_memcmp (tree, data, len))
      break;

  return tree;
}

FluxTree *
flux_tree_find_prev (FluxTree *tree, gconstpointer data, guint32 len)
{
  for (tree = flux_tree_get_prev (tree); tree; tree = flux_tree_get_prev (tree))
    if (!flux_tree_memcmp (tree, data, len))
      break;

  return tree;
}

FluxTree *
flux_tree_find_first_child (FluxTree *tree, gconstpointer data, guint32 len)
{
  tree = flux_tree_get_first_child (tree);
  if (!tree)
    return NULL;
  
  return flux_tree_find_first_sibling (tree, data, len);
}

FluxTree *
flux_tree_find_last_child (FluxTree *tree, gconstpointer data, guint32 len)
{
  tree = flux_tree_get_last_child (tree);
  if (!tree)
    return NULL;
  
  return flux_tree_find_last_sibling (tree, data, len);
}

void
flux_tree_stat_all (FluxTree *root, guint32 *count, guint32 *size)
{
  FluxTree *tree_0;
  
  *count = 1;
  *size  = flux_tree_data_get_size (root);

  FLUX_TREE_FOR_ALL (root, tree_0)
  {
    (*count)++;
    *size += flux_tree_data_get_size (tree_0);
  }
}

guint32
flux_tree_data_get_size_all (FluxTree *root)
{
  FluxTree *tree_0;
  guint32   size;

  size = flux_tree_data_get_size (root);

  FLUX_TREE_FOR_ALL (root, tree_0)
  {
    size += flux_tree_data_get_size (tree_0);
  }

  return size;
}

guint32
flux_tree_count_all (FluxTree *tree)
{
  FluxTree *next_tree;
  gint      num;

  for (num = 1, tree = flux_tree_get_first_child (tree); tree; tree = next_tree, num++)
  {
    if (flux_tree_get_first_child (tree))
      num += flux_tree_count_all (tree) - 1;

    next_tree = flux_tree_get_next (tree);
  }

  return num;
}

guint32
flux_tree_get_size_children (FluxTree *root)
{
  FluxTree *tree_0;
  guint32   size;

  size = 0;

  FLUX_TREE_FOR_EACH (root, tree_0)
  {
    size += flux_tree_data_get_size (tree_0);
  }

  return size;
}

guint32
flux_tree_count_children (FluxTree *tree)
{
  gint num;

  for (num = 0, tree = flux_tree_get_first_child (tree); tree;
       tree = flux_tree_get_next (tree), num++)
    ;

  return num;
}

guint32
flux_tree_count_siblings (FluxTree *tree)
{
  FluxTree *parent;
  guint32   r;

  parent = flux_tree_get_parent (tree);

  if (parent)
  {
    r = flux_tree_count_children (parent);
  }
  else
    r = 1;

  return r;  /* Only self (am root node) */
}

static gboolean
print_indent (FluxTree *tree, FILE *out)
{
  gboolean result = TRUE;
  gint     i;

  i = flux_tree_get_depth (tree);
  if (!i)
    g_error("Loop in tree structure!");

  for (i -= 1; i && result; i--)
    result = fwrite ("  ", 2, 1, out);

  return result;
}

static gboolean
print_data (FluxTree *tree, FILE *out)
{
  gboolean result = TRUE;
  gboolean quoted = FALSE;
  guint    i;
  guchar   c;

  if (flux_tree_has_data (tree))
  {
    if (flux_tree_chr (tree, ' ') != -1)
    {
      result = fwrite ("\"", 1, 1, out);
      quoted = TRUE;
    }

    if (!result)
      goto out;

    for (i = 0; ; i++)
    {
      if (!flux_tree_data_get_bytes (tree, &c, i, 1))
        break;
      
      switch (c)
      {
        case '{':
        case '}':
        case '"':
        case '\'':
        case '\\':
          result = fwrite ("\\", 1, 1, out);
          if (result)
            result = fwrite (&c, 1, 1, out);
          break;
        
        case '\n':
          result = fwrite ("\\n", 2, 1, out);
          break;
        
        case '\r':
          result = fwrite ("\\r", 2, 1, out);
          break;
        
        case '\t':
          result = fwrite ("\\t", 2, 1, out);
          break;

        case '\0':
          result = fwrite ("\\00", 3, 1, out);
          break;

        default:
          if (strchr (tree_printables, c))
            result = fwrite (&c, 1, 1, out);
          else
          {
            result = fwrite ("\\", 1, 1, out);
            if (!result)
              goto out;

            if ((c >> 4) < 10)
              result = (fputc ('0' + (c >> 4), out) != EOF);
            else
              result = (fputc ('a' + (c >> 4) - 10, out) != EOF);

            if ((c & 0x0f) < 10)
              result = (fputc ('0' + (c & 0x0f), out) != EOF);
            else
              result = (fputc ('a' + (c & 0x0f) - 10, out) != EOF);
          }
          break;
      }
    }

    if (result && quoted)
      result = fwrite ("\"", 1, 1, out);
  }
  else
    result = fwrite ("\"\"", 2, 1, out);

out:
  return result;
}

static gboolean
print_token (FluxTree *tree, FILE *out, FluxPrintMode mode)
{
  gboolean result = TRUE;

  if (mode == FLUX_PRINT_COMPACT)
  {
    if (flux_tree_get_prev (tree))
      result = fwrite (" ", 1, 1, out);
  }
  else
  {
    if (flux_tree_get_prev (tree))
      result = fwrite("\n", 1, 1, out);

    if (result)
      result = print_indent (tree, out);
  }

  if (result)
    result = print_data (tree, out);

  return result;
}

static gboolean
print_start_brace (FluxTree *tree, FILE *out, FluxPrintMode mode)
{
  gboolean result = TRUE;

  /* Pre-sub-branch space/brace/space */
      
  if (mode == FLUX_PRINT_COMPACT)
  {
    result = fwrite (" { ", 3, 1, out);
  }
  else if (mode == FLUX_PRINT_KNR)
  {
    result = fwrite (" {\n", 3, 1, out);
  }
  else if (mode == FLUX_PRINT_ALLMAN)
  {
    result = fwrite ("\n", 1, 1, out);
    if (result)
      result = print_indent (tree, out);
    if (result)
      result = fwrite ("{\n", 2, 1, out);
  }

  return result;
}

static gboolean
print_end_brace (FluxTree *tree, FILE *out, FluxPrintMode mode)
{
  gboolean result = TRUE;

  /* Post-sub-branch brace */

  if (mode == FLUX_PRINT_COMPACT)
  {
    result = fwrite (" }", 2, 1, out);
  }
  else if (mode == FLUX_PRINT_KNR)
  {
    result = fwrite ("\n", 1, 1, out);
    if (result)
      result = print_indent (tree, out);
    if (result)
      result = fwrite ("}", 1, 1, out);
  }
  else if (mode == FLUX_PRINT_ALLMAN)
  {
    result = fwrite ("\n", 1, 1, out);
    if (result)
      result = print_indent (tree, out);
    if (result)
      result = fwrite ("}", 1, 1, out);
  }

  return result;
}

static gboolean
print_branch (FluxTree *tree, FILE *out, FluxPrintMode mode)
{
  gboolean result = TRUE;

  for (; tree; tree = flux_tree_get_next (tree))
  {
    result = print_token (tree, out, mode);
    if (!result)
      break;

    if (!flux_tree_is_leaf (tree))
    {
      result = print_start_brace (tree, out, mode);
      if (!result)
        break;

      result = print_branch (flux_tree_get_first_child (tree), out, mode);
      if (!result)
        break;

      result = print_end_brace (tree, out, mode);
      if (!result)
        break;
    }
  }

  return result;
}

gboolean
flux_tree_print_to_file (FluxTree *tree, FILE *out, FluxPrintMode mode)
{
  gboolean result;

  result = print_branch (flux_tree_get_first_child (tree), out, mode);

  if (result)
    result = fwrite ("\n", 1, 1, out);

  return result;
}

typedef enum
{
  SCAN_MODE_SPACE,
  SCAN_MODE_DATA,
  SCAN_MODE_ESCAPED_DATA,
  SCAN_MODE_ESCAPED_LAST_DIGIT,
  SCAN_MODE_QUOTED_DATA,
  SCAN_MODE_QUOTED_ESCAPED_DATA,
  SCAN_MODE_QUOTED_ESCAPED_LAST_DIGIT
}
ScanState;

typedef struct
{
  ScanState      state;
  FluxTreeError  error;

  gchar         *buf;
  guint          len;
  guint          pos;

  FluxTree      *parent;
  FluxTree      *current;
  guchar         escaped_char;
}
ScanContext;

/* State machine diagram
 * =====================
 *
 *           quoted-data <-> quoted-escaped-data -> quoted-escaped-last-digit -.
 *           ^  v      ^-------------------------------------------------------'
 * space <-> data
 *           ^  ^-------------------------------------------------------.
 *           `-------------> escaped-data --------> escaped-last-digit -'
 */

static gboolean
scan_check_brace (ScanContext *context, gchar c)
{
  if (c == tree_brace_open)
  {
    if (!context->current)
    {
      /* A brace was opened with no preceding data. Just create
       * an empty node to use as the parent. */
      context->current = flux_tree_new ();
      flux_tree_add_as_last_child (context->parent, context->current);
    }

    context->parent = context->current;
    context->current = NULL;
    return TRUE;
  }
  else if (c == tree_brace_close)
  {
    context->parent = flux_tree_get_parent (context->parent);
    context->current = NULL;

    if (!context->parent)
    {
      /* We have one close brace too many. Return an error. */
      context->error = FLUX_TREE_ERROR_TOO_MANY_CLOSE_BRACES;
    }

    return TRUE;
  }

  return FALSE;
}

static gint
scan_check_hexdigit (gchar c)
{
  gint d;

  d = c - '0';
  if (d >= 0 && d <= 9)
    return d;

  d = c - 'a';
  if (d >= 0 && d <= 5)
    return d + 10;

  d = c - 'A';
  if (d >= 0 && d <= 5)
    return d + 10;

  return -1;
}

static void
scan_space (ScanContext *context)
{
  guint seg_len = strspn (context->buf + context->pos, tree_spaces);

  if (!seg_len)
  {
    if (scan_check_brace (context, *(context->buf + context->pos)))
    {
      context->pos++;
    }
    else
    {
      context->current = flux_tree_new ();
      flux_tree_add_as_last_child (context->parent, context->current);
      context->state = SCAN_MODE_DATA;
    }
  }
  else
  {
    context->pos += seg_len;
  }
}

static void
scan_data (ScanContext *context)
{
  guint seg_len = strcspn (context->buf + context->pos, tree_data_terminators);

  if (!seg_len)
  {
    gchar c = *(context->buf + context->pos);

    if (scan_check_brace (context, c))
    {
      context->state = SCAN_MODE_SPACE;
      context->pos++;
    }
    else if (c == tree_quote)
    {
      context->state = SCAN_MODE_QUOTED_DATA;
      context->pos++;
    }
    else if (c == tree_escape)
    {
      context->state = SCAN_MODE_ESCAPED_DATA;
      context->pos++;
    }
    else
    {
      context->state = SCAN_MODE_SPACE;
    }
  }
  else
  {
    g_assert (context->current);
    flux_tree_data_append_bytes (context->current, context->buf + context->pos, seg_len);
    context->pos += seg_len;
  }
}

static void
scan_quoted_data (ScanContext *context)
{
  guint seg_len = strcspn (context->buf + context->pos, tree_quoted_data_terminators);

  if (!seg_len)
  {
    gchar c = *(context->buf + context->pos);

    if (c == tree_quote)
    {
      context->state = SCAN_MODE_DATA;
      context->pos++;
    }
    else if (c == tree_escape)
    {
      context->state = SCAN_MODE_QUOTED_ESCAPED_DATA;
      context->pos++;
    }
  }
  else
  {
    g_assert (context->current);
    flux_tree_data_append_bytes (context->current, context->buf + context->pos, seg_len);
    context->pos += seg_len;
  }
}

static void
scan_escaped_data (ScanContext *context)
{
  gchar c = *(context->buf + context->pos);
  gint  d = scan_check_hexdigit (c);

  if (d >= 0)
  {
    context->escaped_char = d << 4;
    context->state = SCAN_MODE_ESCAPED_LAST_DIGIT;
  }
  else if (c == 'n')
  {
    flux_tree_data_append_bytes (context->current, "\n", 1);
    context->state = SCAN_MODE_DATA;
  }
  else if (c == 'r')
  {
    flux_tree_data_append_bytes (context->current, "\r", 1);
    context->state = SCAN_MODE_DATA;
  }
  else if (c == 't')
  {
    flux_tree_data_append_bytes (context->current, "\t", 1);
    context->state = SCAN_MODE_DATA;
  }
  else
  {
    flux_tree_data_append_bytes (context->current, &c, 1);
    context->state = SCAN_MODE_DATA;
  }

  context->pos++;
}

static void
scan_escaped_last_digit (ScanContext *context)
{
  gchar c = *(context->buf + context->pos);
  gint  d = scan_check_hexdigit (c);

  if (d >= 0)
  {
    context->escaped_char |= d;
    flux_tree_data_append_bytes (context->current, &context->escaped_char, 1);
    context->pos++;
  }
  else
  {
    /* There was only one digit! Should emit a warning here. */
    context->escaped_char >>= 4;
    flux_tree_data_append_bytes (context->current, &context->escaped_char, 1);
  }

  context->state = SCAN_MODE_DATA;
}

static void
scan_from_buffer (ScanContext *context, gchar *buf, guint len)
{
  context->buf = buf;
  context->len = len;
  context->pos = 0;

  while (context->pos < context->len)
  {
    if (context->error != FLUX_TREE_ERROR_NONE)
    {
      return;
    }

    switch (context->state)
    {
      case SCAN_MODE_SPACE:
        scan_space (context);
        break;

      case SCAN_MODE_DATA:
        scan_data (context);
        break;

      case SCAN_MODE_ESCAPED_DATA:
        scan_escaped_data (context);
        break;

      case SCAN_MODE_ESCAPED_LAST_DIGIT:
        scan_escaped_last_digit (context);
        break;

      case SCAN_MODE_QUOTED_ESCAPED_LAST_DIGIT:
        scan_escaped_last_digit (context);
        context->state = SCAN_MODE_QUOTED_DATA;
        break;

      case SCAN_MODE_QUOTED_DATA:
        scan_quoted_data (context);
        break;

      case SCAN_MODE_QUOTED_ESCAPED_DATA:
        scan_escaped_data (context);
        if (context->state == SCAN_MODE_DATA)
          context->state = SCAN_MODE_QUOTED_DATA;
        else if (context->state == SCAN_MODE_ESCAPED_LAST_DIGIT)
          context->state = SCAN_MODE_QUOTED_ESCAPED_LAST_DIGIT;
        else
          g_assert_not_reached ();
        break;

      default:
        g_assert_not_reached ();
    }
  }
}

FluxTree *
flux_tree_scan_from_file (FILE *in)
{
  FluxTree    *tree;
  ScanContext *context;
  gchar        buf [BLOCK_INCREMENT + 1];
  guint        len;

  context         = g_new0 (ScanContext, 1);
  context->state  = SCAN_MODE_SPACE;
  context->parent = flux_tree_new ();
  context->error  = FLUX_TREE_ERROR_NONE;

  while ((len = fread (buf, 1, BLOCK_INCREMENT, in)) > 0)
  {
    *(buf + len) = '\0';
    scan_from_buffer (context, buf, len);
    if (context->error != FLUX_TREE_ERROR_NONE)
    {
      if (context->parent)
      {
        flux_tree_destroy (flux_tree_get_root (context->parent));
        context->parent = NULL;
      }
      break;
    }
  }

  if (context->parent)
    tree = flux_tree_get_root (context->parent);
  else
    tree = NULL;

  g_free (context);
  return tree;
}


