/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-tree.h - N-ary tree where nodes can hold any data.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#ifndef _FLUX_TREE_H
#define _FLUX_TREE_H

#ifdef __FLUX_BUILD
# include "flux-types.h"
#else
# include <flux/flux-types.h>
#endif

#include <stdio.h>

#define FLUX_TREE(x) ((FluxTree *) x)

#define FLUX_TREE_FOR_EACH(tree, c) \
  for ((c) = flux_tree_get_first_child (tree); (c); (c) = flux_tree_get_next (c))

#define FLUX_TREE_FOR_ALL(tree, c) \
  for ((c) = flux_tree_get_first_child (tree); (c); c = flux_tree_get_next_infix (c, tree))

typedef enum
{
  FLUX_TREE_NO_DATA,
  FLUX_TREE_DATA_IN_HEADER,
  FLUX_TREE_DATA_IN_MEMORY
}
FluxTreeDataStorageType;

struct _FluxTree
{
  FluxTree *parent;
  FluxTree *child_first, *child_last;
  FluxTree *next, *prev;

  guint16   data_is_ready    : 1;  /* Data is fully loaded (for constructors) */
  guint16   is_auto_root     : 1;  /* Node was created to maintain a single root */
  guint16   has_external_ref : 1;  /* Node is referenced by external wrappers */

  guint16   data_storage_type;

  guint32   size;
  gpointer  data;  /* Pointer to metadata, pointer to data, or data embedded in field */
};

/* Allocation */

FluxTree *flux_tree_new                      (void);
FluxTree *flux_tree_new_with_data            (gconstpointer src, gint len);
FluxTree *flux_tree_new_with_parent_and_data (FluxTree *parent, gconstpointer src, gint len);
void      flux_tree_destroy                  (FluxTree *tree);
void      flux_tree_reclaim                  (FluxTree *tree);
FluxTree *flux_tree_dup                      (FluxTree *tree);
FluxTree *flux_tree_dup_all                  (FluxTree *tree);

FluxTree *flux_tree_split                    (FluxTree *tree, guint32 pos);
void      flux_tree_join                     (FluxTree *target, FluxTree *source);

/* XXX Status set */

#define   flux_tree_set_ready(tree, ready)         ((tree)->data_is_ready = (ready))
#define   flux_tree_set_auto_root(tree, auto_root) ((tree)->is_auto_root = (auto_root))
#define   flux_tree_set_external_ref(tree, ref)    ((tree)->has_external_ref = (ref))

/* XXX Status get */

gboolean  flux_tree_has_data               (FluxTree *tree);
gboolean  flux_tree_has_children           (FluxTree *tree);
#define   flux_tree_is_ready(tree)         ((tree)->data_is_ready    ? TRUE : FALSE)
#define   flux_tree_is_auto_root(tree)     ((tree)->is_auto_root     ? TRUE : FALSE)
#define   flux_tree_has_external_ref(tree) ((tree)->has_external_ref ? TRUE : FALSE)

/* Statistics */

guint32   flux_tree_get_depth              (FluxTree *tree);  /* Root is at level 0 */
void      flux_tree_stat_all               (FluxTree *root, guint32 *count, guint32 *size);

guint32   flux_tree_get_size_children      (FluxTree *root);
guint32   flux_tree_get_size_all           (FluxTree *root);

guint32   flux_tree_count_children         (FluxTree *tree);
guint32   flux_tree_count_all              (FluxTree *tree);
guint32   flux_tree_count_siblings         (FluxTree *tree);

/* Navigation */

gboolean  flux_tree_is_parent_of           (FluxTree *parent, FluxTree *child);
gboolean  flux_tree_is_root                (FluxTree *tree);
gboolean  flux_tree_is_leaf                (FluxTree *tree);
gboolean  flux_tree_is_first               (FluxTree *tree);
gboolean  flux_tree_is_last                (FluxTree *tree);
gboolean  flux_tree_is_sibling             (FluxTree *tree_1, FluxTree *tree_2);

FluxTree *flux_tree_get_root               (FluxTree *tree);
FluxTree *flux_tree_get_parent             (FluxTree *tree);
FluxTree *flux_tree_get_next               (FluxTree *tree);
FluxTree *flux_tree_get_prev               (FluxTree *tree);
FluxTree *flux_tree_get_first_child        (FluxTree *tree);
FluxTree *flux_tree_get_last_child         (FluxTree *tree);
FluxTree *flux_tree_get_nth_child          (FluxTree *tree, gint n);
FluxTree *flux_tree_get_first_sibling      (FluxTree *tree);
FluxTree *flux_tree_get_last_sibling       (FluxTree *tree);
FluxTree *flux_tree_get_nth_sibling        (FluxTree *tree, gint n);

FluxTree *flux_tree_get_next_infix         (FluxTree *tree, FluxTree *top);
FluxTree *flux_tree_get_common_parent      (FluxTree *tree_0, FluxTree *tree_1);
FluxTree *flux_tree_get_next_in_breadth_with_level (FluxTree *tree, int depth, int level);
FluxTree *flux_tree_get_next_in_same_depth (FluxTree *tree);
#define   flux_tree_get_next_in_breadth(tree, depth) \
  flux_tree_get_next_in_breadth_with_level(tree, depth, flux_tree_get_depth (tree))

/* Linking */

void      flux_tree_add_as_first_child     (FluxTree *parent_tree, FluxTree *tree);
void      flux_tree_add_as_last_child      (FluxTree *parent_tree, FluxTree *tree);
void      flux_tree_add_as_first_sibling   (FluxTree *sibling_tree, FluxTree *tree);
void      flux_tree_add_as_last_sibling    (FluxTree *sibling_tree, FluxTree *tree);
void      flux_tree_add_before             (FluxTree *next_tree, FluxTree *tree);
void      flux_tree_add_after              (FluxTree *prev_tree, FluxTree *tree);
#define   flux_tree_add(parent_tree, tree) flux_tree_add_as_last_child (parent_tree, tree)

void      flux_tree_swap                   (FluxTree *tree_0, FluxTree *tree_1);
void      flux_tree_detach                 (FluxTree *tree);

gboolean  flux_tree_is_ancestor            (FluxTree *ancestor, FluxTree *tree);

/* Data */

void      flux_tree_data_swap              (FluxTree *tree_0, FluxTree *tree_1);
void      flux_tree_data_delete            (FluxTree *tree);

#define   flux_tree_data_get_size(tree)    ((tree)->size)
void      flux_tree_data_set_size          (FluxTree *tree, gint len);

guint32   flux_tree_data_get_size_all      (FluxTree *tree);

void      flux_tree_data_set               (FluxTree *tree, gconstpointer src, guint32 len);
void      flux_tree_data_set_bytes         (FluxTree *tree, gconstpointer src, guint32 start, guint32 len);
void      flux_tree_data_set_str           (FluxTree *tree, const gchar *str);
void      flux_tree_data_append_bytes      (FluxTree *tree, gconstpointer src, guint32 len);
void      flux_tree_data_prepend_bytes     (FluxTree *tree, gconstpointer src, guint32 len);

guint32   flux_tree_data_get_bytes         (FluxTree *tree, gpointer dest, guint32 start, guint32 len);
gchar    *flux_tree_data_get_str           (FluxTree *tree);

/* Matching */

int       flux_tree_cmp                    (FluxTree *tree_0, FluxTree *tree_1);
int       flux_tree_casecmp                (FluxTree *tree_0, FluxTree *tree_1);
int       flux_tree_memcmp                 (FluxTree *tree, gconstpointer p, guint32 len);
int       flux_tree_memcasecmp             (FluxTree *tree, gconstpointer p, guint32 len);
#define   flux_tree_strcmp(tree, s)        (flux_tree_memcmp ((tree), (s), strlen (s)))
#define   flux_tree_strcasecmp(tree, s)    (flux_tree_memcasecmp ((tree), (s), strlen (s)))

size_t    flux_tree_spn_from               (FluxTree *tree, const gchar *accept, size_t start);
size_t    flux_tree_cspn_from              (FluxTree *tree, const gchar *reject, size_t start);
#define   flux_tree_spn(tree, accept)      flux_tree_spn_from(tree, accept, 0)
#define   flux_tree_cspn(tree, reject)     flux_tree_cspn_from(tree, reject, 0)
int       flux_tree_chr                    (FluxTree *tree, int c);
int       flux_tree_rchr                   (FluxTree *tree, int c);

/* Searching */

FluxTree *flux_tree_find_first_sibling     (FluxTree *tree, gconstpointer data, guint32 len);
FluxTree *flux_tree_find_last_sibling      (FluxTree *tree, gconstpointer data, guint32 len);
FluxTree *flux_tree_find_next              (FluxTree *tree, gconstpointer data, guint32 len);
FluxTree *flux_tree_find_prev              (FluxTree *tree, gconstpointer data, guint32 len);

FluxTree *flux_tree_find_first_child       (FluxTree *tree, gconstpointer data, guint32 len);
FluxTree *flux_tree_find_last_child        (FluxTree *tree, gconstpointer data, guint32 len);

/* XXX Hashing */

guint32   flux_tree_hash                   (FluxTree *tree);
guint32   flux_tree_hash_all               (FluxTree *tree);

/* Scanning/printing */

FluxTree *flux_tree_scan_from_file         (FILE *in);
gboolean  flux_tree_print_to_file          (FluxTree *tree, FILE *out, FluxPrintMode mode);

#endif /* _FLUX_TREE_H */
