/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* flux-twofish.h - The Twofish symmetric cipher.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Matthew Skala       <mskala@ansuz.sooke.bc.ca>
 *          Hans Petter Jansson <hpj@copyleft.no>
 *
 * The original author, Matthew Skala, has disclaimed all copyright
 * interest in this code, thus putting it in the public domain. This
 * version, modified for Flux, is released under the above license.
 */

#ifndef _FLUX_TWOFISH_H
#define _FLUX_TWOFISH_H

#ifdef __FLUX_BUILD
# include "flux-types.h"
# include "flux-object.h"
#else
# include <flux/flux-types.h>
# include <flux/flux-object.h>
#endif

#define FLUX_TWOFISH_KEY(key) ((FluxTwofishKey *) key)

/* Structure for an expanded Twofish key. s contains the key-dependent
 * S-boxes composed with the MDS matrix; w contains the eight "whitening"
 * subkeys, K[0] through K[7]. k holds the remaining, "round" subkeys.  Note
 * that k[i] corresponds to what the Twofish paper calls K[i+8]. */

struct _FluxTwofishKey
{
  FluxObject object;

  guint key_len;       /* Key length in bytes */
  guint v_pos;

  union
  {
    guchar key [32];   /* Unexpanded key */
    unsigned long __align;
  }
  key;

  union
  {
    guchar v [16];     /* Vectors (not secret) */
    unsigned long __align;
  }
  v;

  guint32 s [4] [256],
          w [8],
          k [32];
};

#define flux_twofish_bits(key) ((key)->key_len << 3)  /* Returns bits */
#define flux_twofish_len(key)  ((key)->key_len)       /* Returns bytes */

FluxTwofishKey *flux_twofish_generate_key (guint nbits);
FluxTwofishKey *flux_twofish_dup_key      (FluxTwofishKey *key);
void            flux_twofish_finalize_key (FluxTwofishKey *key);

FluxTwofishKey *flux_twofish_import_key   (const guchar *key, guint keylen);
void            flux_twofish_export_key   (FluxTwofishKey *key, guchar **out, guint *len);

void            flux_twofish_import_iv    (FluxTwofishKey *key, const guchar *iv);
void            flux_twofish_export_iv    (FluxTwofishKey *key, guchar **out, guint *len);
void            flux_twofish_permute_iv   (FluxTwofishKey *key);

void            flux_twofish_encrypt_ecb  (FluxTwofishKey *ctx, guchar *out, const guchar *in);
void            flux_twofish_decrypt_ecb  (FluxTwofishKey *ctx, guchar *out, const guchar *in);

void            flux_twofish_encrypt_cbc  (FluxTwofishKey *key, guchar *out, const guchar *in);
void            flux_twofish_decrypt_cbc  (FluxTwofishKey *key, guchar *out, const guchar *in);

void            flux_twofish_encrypt_cfb  (FluxTwofishKey *key, guchar *out, const guchar *in, guint len);
void            flux_twofish_decrypt_cfb  (FluxTwofishKey *key, guchar *out, const guchar *in, guint len);

void            flux_twofish_encrypt_ofb  (FluxTwofishKey *key, guchar *out, const guchar *in, guint len);
void            flux_twofish_decrypt_ofb  (FluxTwofishKey *key, guchar *out, const guchar *in, guint len);

#endif  /* _FLUX_TWOFISH_H */
