/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* fifo-test.c - Tests for FluxFifo.
 *
 * Copyright (C) 2005 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#define __FLUX_BUILD

#define BUF_SIZE 4096

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "flux-types.h"
#include "flux-fifo.h"

static gboolean
compress_file (const gchar *file_name, gint compress_level, gint decompress_level)
{
  FluxFifo *reference_fifo;
  FluxFifo *compress_fifo;
  FluxFifo *decompress_fifo;
  gint      orig_len       = 0;
  gint      compressed_len = 0;
  FILE     *fhd;
  gchar     buf [BUF_SIZE];

  fhd = fopen (file_name, "rb");
  if (!fhd)
  {
    g_print ("Could not open '%s'.\n", file_name);
    return FALSE;
  }

  reference_fifo  = flux_fifo_new (4, 256, 1024);
  compress_fifo   = flux_fifo_new (4, 256, 1024);
  decompress_fifo = flux_fifo_new (4, 256, 1024);

  flux_fifo_set_mode (reference_fifo, FLUX_FIFO_MODE_SIMPLE);
  flux_fifo_set_mode (compress_fifo, FLUX_FIFO_MODE_ENCODE);
  flux_fifo_set_mode (decompress_fifo, FLUX_FIFO_MODE_DECODE);

  flux_fifo_set_compression (compress_fifo, compress_level);
  flux_fifo_set_compression (decompress_fifo, decompress_level);

  for (;;)
  {
    gint m, n;

    m = fread (buf, 1, BUF_SIZE, fhd);
    orig_len += m;

    if (m)
    {
      flux_fifo_enqueue (reference_fifo, buf, m);
      flux_fifo_enqueue (compress_fifo, buf, m);
    }
    else
    {
      flux_fifo_flush (compress_fifo);
    }

    while ((n = flux_fifo_get_used (compress_fifo)))
    {
      gchar *compress_buf;

      compressed_len += n;

      compress_buf = g_malloc (n);
      flux_fifo_dequeue (compress_fifo, compress_buf, n);
      flux_fifo_enqueue (decompress_fifo, compress_buf, n);
      g_free (compress_buf);
    }

    while ((n = flux_fifo_get_used (decompress_fifo)))
    {
      gchar *decompress_buf;
      gchar *reference_buf;

      decompress_buf = g_malloc (n);
      reference_buf = g_malloc (n);

      flux_fifo_dequeue (decompress_fifo, decompress_buf, n);
      flux_fifo_dequeue (reference_fifo, reference_buf, n);

      if (memcmp (decompress_buf, reference_buf, n))
      {
        g_print ("Error: Deflated + inflated output does not match original.\n");
        fclose (fhd);
        g_free (reference_buf);
        g_free (decompress_buf);
        return FALSE;
      }

      g_free (reference_buf);
      g_free (decompress_buf);
    }

    if (!m)
      break;
  }

  fclose (fhd);

  flux_object_unref (FLUX_OBJECT (compress_fifo));
  flux_object_unref (FLUX_OBJECT (decompress_fifo));

  if (flux_fifo_get_used (reference_fifo))
  {
    g_print ("Error: Deflated + inflated output is shorter than original.\n");
    return FALSE;
  }

  flux_object_unref (FLUX_OBJECT (reference_fifo));

  g_print ("Size: %d  Compressed: %d  Reduced to: %f\n", orig_len, compressed_len,
           ((gdouble) compressed_len * 100.0) / (gdouble) orig_len);

  return TRUE;
}

int
main (int argc, char *argv [])
{
  gint i;

  if (argc < 2)
  {
    g_print ("Usage: %s <input-file>\n", argv [0]);
    return 1;
  }

  for (i = 0; i < 10; i++)
  {
    g_print ("Running compression level %d...\n", i);
    compress_file (argv [1], i, i);
  }

  for (i = 1; i < 10; i++)
  {
    g_print ("Running compression level %d with decompression level %d...\n", i, 6);
    compress_file (argv [1], i, 6);
  }

  return 0;
}
