/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* packed-bin-tree-test.c - Tests packed binary trees.
 *
 * Copyright (C) 2004 Hans Petter Jansson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-types.h"
#include "flux-packed-bin-tree.h"
#include <string.h>         /* memcpy, memset */

#define N_LEAVES 10000

static void
test_insert (FluxPackedBinTree *tree, FluxItem item)
{
  gint i;

  for (i = 0; i < N_LEAVES; i++)
  {
    FluxBinPath *path;
    gint         index;
    guint        value;

    /* Tests */

    path = flux_packed_bin_tree_leaf_to_path (tree, item, i);
    g_assert (path);

    index = flux_packed_bin_tree_path_to_leaf (tree, item, path);
    g_assert (index == i);

    flux_bin_path_up (path);
    index = flux_packed_bin_tree_path_to_leaf (tree, item, path);
    g_assert (index < 0);

    index = flux_packed_bin_tree_path_to_index (tree, item, path);
    g_assert (index >= 0);

    flux_object_unref (FLUX_OBJECT (path));

    path = flux_packed_bin_tree_index_to_path (tree, item, index);
    g_assert (path);
    g_assert (flux_packed_bin_tree_path_to_index (tree, item, path) == index);

    flux_object_unref (FLUX_OBJECT (path));

    /* Insert value */

    path = flux_packed_bin_tree_leaf_to_path (tree, item, i);
    index = flux_packed_bin_tree_path_to_index (tree, item, path);

    value = i + 1;
    flux_packed_bin_tree_write (tree, item, index, &value);

    /* Propagate up */

    do
    {
      gconstpointer p;

      flux_bin_path_up (path);

      flux_bin_path_left (path);
      flux_packed_bin_tree_read (tree, item, flux_packed_bin_tree_path_to_index (tree, item, path),
                                 &value);
      flux_bin_path_up (path);

      flux_bin_path_right (path);
      p = flux_packed_bin_tree_peek (tree, item, flux_packed_bin_tree_path_to_index (tree, item, path));
      g_assert (p);
      value += *((guint *) p);
      flux_bin_path_up (path);

      flux_packed_bin_tree_write (tree, item, flux_packed_bin_tree_path_to_index (tree, item, path),
                                  &value);
    }
    while (flux_bin_path_get_depth (path) > 0);

    flux_object_unref (FLUX_OBJECT (path));
  }
}

static guint
test_validate (FluxPackedBinTree *tree, FluxItem item, FluxBinPath *path)
{
  guint value;
  guint left_value;
  guint right_value;
  gint  index;

  index = flux_packed_bin_tree_path_to_index (tree, item, path);
  g_assert (index >= 0);

  if (!flux_packed_bin_tree_read (tree, item, index, &value))
    g_error ("Read failed!");

  if (flux_packed_bin_tree_path_to_leaf (tree, item, path) < 0)
  {
    flux_bin_path_left (path);
    left_value = test_validate (tree, item, path);
    flux_bin_path_up (path);

    flux_bin_path_right (path);
    right_value = test_validate (tree, item, path);
    flux_bin_path_up (path);

    g_assert (value == left_value + right_value);
  }

  return value;
}

gint
main (gint argc, gchar *argv [])
{
  FluxPackedBinTree *tree;
  FluxItem           item;
  FluxBinPath       *path;

  tree = flux_packed_bin_tree_new (NULL, sizeof (guint));
  item = flux_packed_bin_tree_create_tree (tree, N_LEAVES);

  test_insert (tree, item);

  path = flux_bin_path_new ();
  test_validate (tree, item, path);
  flux_object_unref (FLUX_OBJECT (path));

  flux_object_unref (FLUX_OBJECT (tree));
  return 0;
}
