/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* pat-tree-benchmark.c - Benchmarks Patricia tree using various storage backends.
 *
 * Copyright (C) 2001-2003 Hans Petter Jansson
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#include "config.h"
#include "flux-mem-heap.h"
#include "flux-mem-chunk.h"
#include "flux-disk-heap.h"
#include "flux-disk-chunk.h"
#include "flux-pat-tree.h"
#include <string.h>     /* memset, memcpy */
#include <sys/types.h>  /* open */
#include <sys/stat.h>   /* open */
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>      /* open */
#include <stdio.h>

#define KEY_LEN   8
#define VALUE_LEN 8

static guint  intervals_done  = 0;
static guint  total_records   = 0;
static guint  operations_done = 0;

static FILE  *log_fhd;

void
next_interval (gint sig)
{
  intervals_done++;

  fprintf (log_fhd, "%d %d\n", total_records, operations_done / 5);

  operations_done = 0;
}

gboolean
run_benchmark (const gchar *log_name, FluxStorage *storage)
{
  FluxPatTree      *pat_tree;
  gchar             key [KEY_LEN];
  gchar             value [VALUE_LEN];
  struct itimerval  itv;

  intervals_done  = 0;
  total_records   = 0;
  operations_done = 0;

  memset (value, 0, VALUE_LEN);

  log_fhd = fopen (log_name, "wt");

  pat_tree = flux_pat_tree_new (storage, "pat-test", KEY_LEN, VALUE_LEN);

  for ( ; total_records < 400000; )
  {
    gint i, last_interval;

    for (i = 0; i < 5000; i++)
    {
      ((guint32 *) key) [0] = g_random_int ();
      ((guint32 *) key) [1] = g_random_int ();

      if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, value, VALUE_LEN))
        g_error ("Failed to insert.");

      total_records++;
    }

    itv.it_interval.tv_sec  = itv.it_value.tv_sec  = 0;
    itv.it_interval.tv_usec = itv.it_value.tv_usec = 5000000;
    signal (SIGALRM, next_interval);
    setitimer (ITIMER_REAL, &itv, NULL);

    last_interval = intervals_done;
    while (last_interval == intervals_done)
    {
      gpointer found;
      guint    found_len;

      ((guint32 *) key) [0] = g_random_int ();
      ((guint32 *) key) [1] = g_random_int ();

      if (flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN,
                              &found, &found_len))
        g_free (found);

      operations_done++;
    }

    itv.it_interval.tv_sec  = itv.it_value.tv_sec  = 0;
    itv.it_interval.tv_usec = itv.it_value.tv_usec = 0;
    signal (SIGALRM, next_interval);
    setitimer (ITIMER_REAL, &itv, NULL);
  }

  fclose (log_fhd);
  flux_object_unref (FLUX_OBJECT (pat_tree));
  return TRUE;
}

gint
main (gint argc, gchar *argv [])
{
  FluxStorage      *storage;
  gint              fd;

  g_log_set_always_fatal (G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL);

  storage = FLUX_STORAGE (flux_mem_chunk_new (256, flux_pat_tree_calc_node_size (KEY_LEN, VALUE_LEN)));
  run_benchmark ("pat_benchmark-mem_chunk.log", storage);
  flux_object_unref (FLUX_OBJECT (storage));

  storage = FLUX_STORAGE (flux_mem_heap_new ());
  run_benchmark ("pat_benchmark-mem_heap.log", storage);
  flux_object_unref (FLUX_OBJECT (storage));

  fd = open ("pat_benchmark-disk_chunk.db", O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
  storage = FLUX_STORAGE (flux_disk_chunk_new (fd, flux_pat_tree_calc_node_size (KEY_LEN, VALUE_LEN)));
  run_benchmark ("pat_benchmark-disk_chunk.log", storage);
  flux_object_unref (FLUX_OBJECT (storage));

  fd = open ("pat_benchmark-disk_heap.db", O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
  storage = FLUX_STORAGE (flux_disk_heap_new (fd));
  run_benchmark ("pat_benchmark-disk_heap.log", storage);
  flux_object_unref (FLUX_OBJECT (storage));

  return 0;
}
