#include "config.h"
#include "flux-mem-heap.h"
#include "flux-disk-heap.h"
#include "flux-pat-tree.h"
#include <string.h>     /* memset, memcpy */
#include <sys/types.h>  /* open */
#include <sys/stat.h>   /* open */
#include <fcntl.h>      /* open */

#define KEY_LEN     64
#define VALUE_LEN   sizeof (gint32)

#define NUM_NUMBERS (65536 * 2)
#define ITERATIONS  300000

static void
stress (FluxPatTree *pat_tree)
{
  gint32   num_tab [NUM_NUMBERS];
  gchar    key [KEY_LEN];
  guint32  value = 1;
  gpointer found_data;
  guint    found_data_len;
  gint     i, j;

  memset (key, 0, KEY_LEN);

  for (i = 0; i < NUM_NUMBERS; i++)
  {
    num_tab [i] = i;
    memcpy (key, &num_tab [i], sizeof (gint32));
    if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &num_tab [i], VALUE_LEN))
      g_error ("Insert failed!");
  }

  for (i = 0; i < ITERATIONS; i++)
  {
    gint index;

    index = g_random_int_range (0, NUM_NUMBERS);
    if (num_tab [index] != -1)
    {
      memcpy (key, &num_tab [index], sizeof (gint32));

      flux_pat_tree_remove (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN);

      if (flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			      key, KEY_LEN, &found_data, &found_data_len))
	g_error ("Key removal failed! [%02x%02x]",
		 ((guchar *) &index)[3], ((guchar *) &index)[2]);

      num_tab [index] = -1;
    }
    else
    {
      num_tab [index] = index;

      memcpy (key, &num_tab [index], sizeof (gint32));

      if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &num_tab [index], VALUE_LEN))
        g_error ("Insertion function failed!");

      if (!flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			       key, KEY_LEN, &found_data, &found_data_len))
	g_error ("Key insertion succeeded, but key could not be found! [%02x%02x]",
		 ((guchar *) &index)[3], ((guchar *) &index)[2]);

      g_free (found_data);
    }

    for (j = 0; j < NUM_NUMBERS; j++)
    {
      if (num_tab [j] == -1)
      {
        memcpy (key, &j, sizeof (gint32));

        if (flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
      		                key, KEY_LEN, &found_data, &found_data_len))
        {
	  g_error ("Removed key could still be found! [%02x%02x]",
	           ((guchar *) &j)[3], ((guchar *) &j)[2]);
        }
      }
      else
      {
        memcpy (key, &num_tab [j], sizeof (gint32));

        if (!flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
      			         key, KEY_LEN, &found_data, &found_data_len))
        {
	  g_error ("Key disappeared! [%02x%02x]",
	           ((guchar *) &j)[3], ((guchar *) &j)[2]);
        }
        else if (memcmp (found_data, &num_tab [j], VALUE_LEN))
        {
	  g_error ("Corrupt data! [%02x%02x]",
	           ((guchar *) &j)[3], ((guchar *) &j)[2]);
        }
        else
        {
	  g_free (found_data);
        }
      }
    }
  }

  g_print ("Passed stress test!\n");
}

static void
stress_cursor (FluxPatTree *pat_tree)
{
  FluxPatTreeCursor *cursor;
  gint32   num_tab [NUM_NUMBERS];
  gchar    key [KEY_LEN];
  guint32  value = 1;
  gpointer found_data;
  guint    found_data_len;
  gint32   i, j;

  /* Populate */

  memset (key, 0, KEY_LEN);

  for (i = 0; i < NUM_NUMBERS; i++)
  {
    num_tab [i] = 1;
    memcpy (key, &i, sizeof (gint32));
    if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &i, VALUE_LEN))
      g_error ("Insert failed!\n");
  }

  /* Removals/insertions with running cursor */

  /* 0 = removed, 1 = initial population, 2 = re-inserted, 3 = initial population (seen) */

  cursor = flux_pat_tree_traverse (pat_tree, FLUX_ITEM_NULL);

  for (i = 0; i < ITERATIONS; i++)
  {
    gint32 index;

    for (j = 0; j < 8; j++)
    {
      index = g_random_int_range (0, NUM_NUMBERS);
      if (num_tab [index] != 0)
      {
        /* Remove */

        memcpy (key, &index, sizeof (gint32));

        flux_pat_tree_remove (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN);

        if (flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			        key, KEY_LEN, &found_data, &found_data_len))
	  g_error ("Key removal failed! [%02x%02x]",
		   ((guchar *) &index)[3], ((guchar *) &index)[2]);

        num_tab [index] = 0;
      }
#if 1
      else
      {
        /* Re-insert */

        memcpy (key, &index, sizeof (gint32));
        if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &index, VALUE_LEN))
          g_error ("Insert failed!\n");

	if (!flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN,
				 &found_data, &found_data_len))
	  g_error ("Concurrent key insertion failed! [%02x%02x]",
		   ((guchar *) &index)[3], ((guchar *) &index)[2]);

        if (!found_data || !found_data_len)
	  g_print ("Inserted key did not get data assigned.\n");

        if (memcmp (found_data, &index, VALUE_LEN))
	  g_print ("Inserted data is corrupt.\n");

	g_free (found_data);

        num_tab [index] = 2;
      }
#endif
    }

    if (!flux_cursor_next (FLUX_CURSOR (cursor)))
      break;

    if (!flux_pat_tree_cursor_read (cursor, NULL, NULL, &found_data, &found_data_len))
    {
      g_warning ("Go to next ok, but read failed.");
      continue;
    }

    if (found_data_len != VALUE_LEN)
      g_error ("Cursor encountered data with bad length.");

    memcpy (&j, found_data, VALUE_LEN);
    if (j < 0 || j >= NUM_NUMBERS)
      g_error ("Cursor encountered bad data.");

    if (num_tab [j] == 0)
      g_warning ("Cursor encountered removed data.");
    else if (num_tab [j] == 3)
      g_print ("Cursor encountered already seen data (1).\n");
    else if (num_tab [j] != 1 && num_tab [j] != 2)
      g_error ("Mangled num_tab.");

    num_tab [j] = 3;

    g_free (found_data);
  }

  flux_object_unref (FLUX_OBJECT (cursor));

  /* See if cursor missed something that's still there */

  {
    gint missed = 0, removed = 0, caught = 0;

    for (i = 0; i < NUM_NUMBERS; i++)
    {
      if (num_tab [i] == 0)
      {
        removed++;
      }
      else if (num_tab [i] == 1)
      {
        g_print ("Missed number %d.\n", i);
        missed++;
      }
      else if (num_tab [i] == 3)
      {
        caught++;
      }
    }

    g_print ("Concurrency test: missed=%d, removed=%d, caught=%d.\n", missed, removed, caught);
  }

  /* See if tree structure is still ok */

  cursor = flux_pat_tree_traverse (pat_tree, FLUX_ITEM_NULL);

  for (i = 0; i < ITERATIONS; i++)
  {
    gint32 index;

    if (!flux_cursor_next (FLUX_CURSOR (cursor)))
      break;

    if (!flux_pat_tree_cursor_read (cursor, NULL, NULL, &found_data, &found_data_len))
    {
      g_warning ("Go to next ok, but read failed.");
      continue;
    }

    if (found_data_len != VALUE_LEN)
      g_error ("Cursor encountered data with bad length.");

    memcpy (&j, found_data, VALUE_LEN);
    if (j < 0 || j >= NUM_NUMBERS)
      g_error ("Cursor encountered bad data.");

    if (num_tab [j] == 0)
      g_warning ("Cursor encountered removed data.");
    else if (num_tab [j] == 4)
      g_warning ("Cursor encountered already seen data (2).");
    else if (num_tab [j] == 1)
      g_print ("Found missed number %d.\n", j);
    else if (num_tab [j] != 2 && num_tab [j] != 3)
      g_error ("Mangled num_tab.");

    if (num_tab [j] == 2 || num_tab [j] == 3)
      num_tab [j] = 4;

    g_free (found_data);
  }

  /* See if second cursor missed something that's still there */

  {
    gint missed = 0, removed = 0, caught = 0, newmissed = 0;

    for (i = 0; i < NUM_NUMBERS; i++)
    {
      if (num_tab [i] == 0)
      {
        removed++;
      }
      else if (num_tab [i] == 1)
      {
        missed++;
      }
      else if (num_tab [i] == 2 || num_tab [i] == 3)
      {
        newmissed++;
      }
      else if (num_tab [i] == 4)
      {
        caught++;
      }
    }

    g_print ("Post-scan test: missed=%d, removed=%d, caught=%d, newmisses=%d.\n",
             missed, removed, caught, newmissed);
  }
}

static void
stress_fast (FluxPatTree *pat_tree)
{
  gint32   num_tab [NUM_NUMBERS];
  gchar    key [KEY_LEN];
  guint32  value = 1;
  gpointer found_data;
  guint    found_data_len;
  gint     i;

  memset (key, 0, KEY_LEN);

  for (i = 0; i < NUM_NUMBERS; i++)
  {
    num_tab [i] = i;
    memcpy (key, &num_tab [i], sizeof (gint32));
    if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &value, VALUE_LEN))
      g_error ("Insert failed!\n");
  }

  for (i = 0; i < ITERATIONS; i++)
  {
    gint index;

    index = g_random_int_range (0, NUM_NUMBERS);
    if (num_tab [index] != -1)
    {
      memcpy (key, &num_tab [index], sizeof (gint32));

      if (!flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			       key, KEY_LEN, &found_data, &found_data_len))
	g_error ("Key disappeared! [%02x%02x]\n",
		 ((guchar *) &index)[0], ((guchar *) &index)[1]);

      g_free (found_data);

      flux_pat_tree_remove (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN);

      if (flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			      key, KEY_LEN, &found_data, &found_data_len))
	g_error ("Key removal failed! [%02x%02x]",
		 ((guchar *) &index)[0], ((guchar *) &index)[1]);

      num_tab [index] = -1;
    }
    else
    {
      num_tab [index] = index;

      memcpy (key, &num_tab [index], sizeof (gint32));

      if (!flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &num_tab [index], VALUE_LEN))
        g_error ("Insertion function failed!");

      if (!flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			       key, KEY_LEN, &found_data, &found_data_len))
	g_error ("Key insertion succeeded, but key could not be found! [%02x%02x]",
		 ((guchar *) &index)[3], ((guchar *) &index)[2]);

      g_free (found_data);
    }
  }

  g_print ("Passed fast stress test!\n");
  g_print ("Cache hits/misses: %ld/%ld (%.1f%% hit rate)\n", FLUX_DISK_CHUNK (pat_tree)->data_cache->hits,
	   FLUX_DISK_CHUNK (pat_tree)->data_cache->misses,
	   (gdouble) (FLUX_DISK_CHUNK (pat_tree)->data_cache->hits * 100) /
	   (gdouble) (FLUX_DISK_CHUNK (pat_tree)->data_cache->hits + FLUX_DISK_CHUNK (pat_tree)->data_cache->misses));
}

static guchar
reverse_bits (guchar c)
{
  guchar d = 0;
  gint   i;

  for (i = 0; i < 8; i++)
  {
    d >>= 1;
    d |= (c << i) & 0x80;
  }

  return d;
}

static void
traverse (FluxPatTree *pat_tree)
{
  FluxPatTreeCursor *cursor;
  gpointer           found_key, found_value;
  guint              found_key_len, found_value_len;

  cursor = flux_pat_tree_traverse (pat_tree, FLUX_ITEM_NULL);
  if (!cursor)
  {
    g_print ("Couldn't start traversal.\n");
    return;
  }

  for (;;)
  {
    guchar  key [KEY_LEN];
    guint32 value;

    if (!flux_pat_tree_cursor_read (cursor, &found_key, &found_key_len, &found_value, &found_value_len))
      g_print ("Couldn't read.\n");

    value = *((guint32 *) found_value);
    memcpy (key, found_key, KEY_LEN);
    g_free (found_key);
    g_free (found_value);

    g_print ("[%02x%02x]: %u\n", reverse_bits (key [1]), reverse_bits (key [0]), value);

    if (!flux_cursor_next (FLUX_CURSOR (cursor)))
      break;
  }

  flux_object_unref (FLUX_OBJECT (cursor));
}

static void
find_with_mask (FluxPatTree *pat_tree, gconstpointer key, gconstpointer mask)
{
  FluxPatTreeCursor *cursor;
  gpointer           found_key, found_value;
  guint              found_key_len, found_value_len;

  cursor = flux_pat_tree_find_with_mask (pat_tree, FLUX_ITEM_NULL, key, mask, KEY_LEN);
  if (!cursor)
  {
    g_print ("No matches.\n");
    return;
  }

  for (;;)
  {
    guchar  key [KEY_LEN];
    guint32 value;

    if (!flux_pat_tree_cursor_read (cursor, &found_key, &found_key_len, &found_value, &found_value_len))
      g_print ("Couldn't read.\n");

    value = *((guint32 *) found_value);
    memcpy (key, found_key, KEY_LEN);
    g_free (found_key);
    g_free (found_value);
    g_print ("%s\n", key);

    if (!flux_cursor_next (FLUX_CURSOR (cursor)))
      break;
  }

  flux_object_unref (FLUX_OBJECT (cursor));
}

static void
test_traverse (FluxPatTree *pat_tree)
{
  FluxPatTreeCursor *cursor;
  gint32  num_tab [NUM_NUMBERS];
  gchar   key [KEY_LEN];
  guint32 value = 1;
  gint    i;

  memset (key, 0, KEY_LEN);

  for (i = 0; i < NUM_NUMBERS; i++)
  {
    num_tab [i] = i;
    memcpy (key, &num_tab [i], sizeof (gint32));
    flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &value, VALUE_LEN);
  }

  memset (num_tab, 0, sizeof (num_tab));

  cursor = flux_pat_tree_traverse (pat_tree, FLUX_ITEM_NULL);
  if (!cursor)
  {
    g_print ("Couldn't start traversal.\n");
    return;
  }

  for (;;)
  {
    gpointer found_key, found_value;
    guint    found_key_len, found_value_len;
    guchar   key [KEY_LEN];
    guint32  value;

    if (!flux_pat_tree_cursor_read (cursor, &found_key, &found_key_len, &found_value, &found_value_len))
      g_print ("Couldn't read.\n");

    value = *((guint32 *) found_value);
    memcpy (key, found_key, KEY_LEN);
    g_free (found_key);
    g_free (found_value);

    g_print ("[%02x%02x]: %u\n", reverse_bits (key [1]), reverse_bits (key [0]), value);

    if (num_tab [*((guint32 *) key)] != 0)
      g_error ("Got node twice! [%02x%02x]", reverse_bits (key [1]), reverse_bits (key [0]));

    num_tab [*((guint32 *) key)] = 1;

    if (!flux_cursor_next (FLUX_CURSOR (cursor)))
      break;
  }

  if (flux_cursor_next (FLUX_CURSOR (cursor)))
    g_error ("Re-next succeeded, while it should not!");

  for (i = 0; i < NUM_NUMBERS; i++)
  {
    if (num_tab [i] != 1)
      g_error ("Missing index: [%04x]", i);
  }

  flux_object_unref (FLUX_OBJECT (cursor));
  g_print ("Passed traversal!\n");
}

gint
main (gint argc, gchar *argv [])
{
  FluxStorage *storage;
  FluxPatTree *pat_tree;
  gint        fd;
  gchar       key [KEY_LEN];
  gchar       mask [KEY_LEN];
  guint32     value = 1;
  gpointer    found_value;
  guint       found_value_len;
  gint        i;

  memset (key, 0, sizeof (key));

  g_log_set_always_fatal (G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL);

#if 1
  fd = open ("pat-tree-test.db", O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
  storage = FLUX_STORAGE (flux_disk_heap_new (fd));
#else
  storage = FLUX_STORAGE (flux_mem_heap_new ());
#endif

  g_random_set_seed (time (NULL));

  pat_tree = flux_pat_tree_new (storage, "pat-test", KEY_LEN, VALUE_LEN);

  if (!strcmp (argv [1], "insert"))
  {
    strcpy (key, argv [2]);
    if (flux_pat_tree_insert (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN, &value, VALUE_LEN))
      g_print ("Ok.\n");
    else
      g_print ("Failed.\n");
  }
  else if (!strcmp (argv [1], "find"))
  {
    strcpy (key, argv [2]);

    if (flux_pat_tree_find (pat_tree, FLUX_ITEM_NULL,
			    key, KEY_LEN, &found_value, &found_value_len))
    {
      value = *((guint32 *) found_value);
      g_free (found_value);
      g_print ("Found, value = %d.\n", value);
    }
    else
      g_print ("Not found.\n");
  }
  else if (!strcmp (argv [1], "remove"))
  {
    strcpy (key, argv [2]);

    if (flux_pat_tree_remove (pat_tree, FLUX_ITEM_NULL, key, KEY_LEN))
      g_print ("Ok.\n");
    else
      g_print ("Failed.\n");
  }
  else if (!strcmp (argv [1], "stress"))
  {
    stress (pat_tree);
  }
  else if (!strcmp (argv [1], "fast-stress"))
  {
    stress_fast (pat_tree);
  }
  else if (!strcmp (argv [1], "stress-cursor"))
  {
    stress_cursor (pat_tree);
  }
  else if (!strcmp (argv [1], "traverse"))
  {
    traverse (pat_tree);
  }
  else if (!strcmp (argv [1], "test-traverse"))
  {
    test_traverse (pat_tree);
  }
  else if (!strcmp (argv [1], "find-mask"))
  {
    strcpy (key, argv [2]);
    memset (mask, 0xff, sizeof (mask));
    for (i = 0; i < KEY_LEN; i++)
    {
      if (((guchar *) key) [i] == '.')
	mask [i] = 0;
    }

    find_with_mask (pat_tree, key, mask);
  }

  flux_object_unref (FLUX_OBJECT (pat_tree));
  return 0;
}
