/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

#define __FLUX_BUILD

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include "flux-types.h"
#include "flux-rng.h"
#include "flux-rmd160.h"

#define HASH_SIZE 20
#define DATA_SIZE 20
#define FOLD_SIZE 8

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

static gint
compare_bits (guchar *a, guchar *b, gint len)
{
  guint count = 0;
  gint  i;

  for (i = 0; i < len; i++)
  {
    gint   j;
    guint8 a_d = a [i];
    guint8 b_d = b [i];

    for (j = 7; j >= 0; j--)
    {
      if (((a_d >> j) & 1) == ((b_d >> j) & 1))
        count++;
    }
  }

  return count;
}

static void
fold_mem (gconstpointer src, guint srclen, gpointer dest, guint destlen)
{
  gint i, j = 0;

  memset (dest, 0, destlen);

  for (i = 0; i < srclen; i++)
  {
    ((guchar *) dest) [j++] ^= ((guchar *) src) [i];
    if (j == destlen)
      j = 0;
  }
}

int
main (int argc, char *argv [])
{
  FluxRmd160 *rmd160;
  FluxRng    *rng;
  gint        len;
  guchar      data [DATA_SIZE];
  gint        data_len;
  guchar      orig [HASH_SIZE];
  guchar      orig_fold [FOLD_SIZE];
  guchar      hash [HASH_SIZE];
  guchar      hash_fold [FOLD_SIZE];
  gint        best_n_matched = 0;
  gint        i;
  guint       time_sec;

  g_random_set_seed (42);

  rng = flux_rng_new ();
  flux_rng_get_bytes (rng, orig, HASH_SIZE);
  fold_mem (orig, HASH_SIZE, orig_fold, FOLD_SIZE);

  rmd160 = flux_rmd160_new ();
  flux_rmd160_begin (rmd160);

  time_sec = time (NULL);

  for (data_len = 20; data_len <= DATA_SIZE; data_len++)
  {
    memset (data, 0, data_len);

    for (i = 0; i < data_len; )
    {
      gint n_matched;

      flux_rmd160_begin (rmd160);
      flux_rmd160_add (rmd160, data, data_len);
      flux_rmd160_end (rmd160);
      flux_rmd160_get_hash (rmd160, hash);

      fold_mem (hash, HASH_SIZE, hash_fold, FOLD_SIZE);

      n_matched = compare_bits (orig_fold, hash_fold, FOLD_SIZE);

      if (n_matched > best_n_matched)
      {
        guint time_delta = time (NULL) - time_sec;

        g_print ("%6ds: Matched %d bits at data len %d.\n",
                 time_delta, n_matched, data_len);

        best_n_matched = n_matched;
      }

#if 0
      /* Increment data */
      for (i = 0; i < data_len; i++)
      {
        if (data [i] == 255)
        {
          data [i] = 0;
        }
        else
        {
          (data [i])++;
          break;
        }
      }
#else
#if 0
      flux_rng_get_bytes (rng, data, data_len);
#endif
      *((guint32 *) &data [0]) = g_random_int ();
      *((guint32 *) &data [4]) = g_random_int ();
      *((guint32 *) &data [8]) = g_random_int ();
      *((guint32 *) &data [12]) = g_random_int ();
      *((guint32 *) &data [16]) = g_random_int ();
#endif
    }
  }

#if 0
  for (i = 0; i < 20; i++)
  {
    guchar b;

    b = hash [i];
    fputc (int_to_hex (b >> 4), stdout);
    fputc (int_to_hex (b & 0x0f), stdout);
  }

  g_print ("\n");
#endif

  flux_object_unref (FLUX_OBJECT (rmd160));
  flux_object_unref (FLUX_OBJECT (rng));
  return 0;
}
