/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

/* rsa-test.c - Tests for the RSA asymmetric cipher.
 *
 * Copyright (C) 2003 Hans Petter Jansson.
 *
 * Flux is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * Authors: Hans Petter Jansson <hpj@copyleft.no>
 */

#define __FLUX_BUILD

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "flux-types.h"
#include "flux-rng.h"
#include "flux-rsa.h"

#define BUF_SIZE 1024

#define b2B(b)           (((b) + 7) / 8)
#define bits_in_mpz(n)   (mpz_sizeinbase (n, 2))
#define bytes_in_mpz(n)  (mpz_sizeinbase (n, 256))

static gchar
int_to_hex (int n)
{
  if (n < 0 || n > 0x0f)
    g_error ("Integer not within base16 digit range.");

  if (n < 0x0a)
    return ('0' + n);

  return ('a' + n - 0x0a);
}

#define PRIVACY_ITERATIONS      200
#define AUTHENTICITY_ITERATIONS 200

FluxRng *rng = NULL;

gboolean
test_privacy (FluxRSAPublicKey *public_key, FluxRSASecretKey *secret_key, gboolean do_halt)
{
  guchar   *pre_plaintext,     *ciphertext,     *post_plaintext;
  guint     pre_plaintext_len,  ciphertext_len,  post_plaintext_len;
  gboolean  result = TRUE;

  pre_plaintext_len = g_random_int_range (1, flux_rsa_get_block_size (FLUX_RSA_KEY (public_key)) * 3 + 1);
  pre_plaintext = g_malloc (pre_plaintext_len);

  flux_rng_get_bytes (rng, pre_plaintext, pre_plaintext_len);

  flux_rsa_encrypt (public_key, &ciphertext, &ciphertext_len, pre_plaintext, pre_plaintext_len);
  flux_rsa_decrypt (secret_key, &post_plaintext, &post_plaintext_len, ciphertext, ciphertext_len);

  if (pre_plaintext_len != post_plaintext_len)
  {
    result = FALSE;
    if (do_halt)
      g_error ("Plaintext lengths differ!");
  }
  else if (memcmp (pre_plaintext, post_plaintext, pre_plaintext_len))
  {
    result = FALSE;
    if (do_halt)
      g_error ("Plaintexts differ!");
  }

  g_free (pre_plaintext);
  g_free (ciphertext);
  g_free (post_plaintext);

  return result;
}

gboolean
test_authenticity (FluxRSAPublicKey *public_key, FluxRSASecretKey *secret_key, gboolean do_halt)
{
  guchar   *authentic_plaintext,     *signature,      *fake_plaintext;
  guint     authentic_plaintext_len,  signature_len,   fake_plaintext_len;
  gboolean  result = TRUE;

  authentic_plaintext_len = g_random_int_range (1, flux_rsa_get_block_size (FLUX_RSA_KEY (public_key)) * 3 + 1);
  authentic_plaintext = g_malloc (authentic_plaintext_len);
  flux_rng_get_bytes (rng, authentic_plaintext, authentic_plaintext_len);

  fake_plaintext_len = g_random_int_range (1, flux_rsa_get_block_size (FLUX_RSA_KEY (public_key)) * 3 + 1);
  fake_plaintext = g_malloc (fake_plaintext_len);
  do
  {
    flux_rng_get_bytes (rng, fake_plaintext, fake_plaintext_len);
  }
  while (fake_plaintext_len == authentic_plaintext_len &&
         !memcmp (fake_plaintext, authentic_plaintext, fake_plaintext_len));

  flux_rsa_sign (secret_key, &signature, &signature_len, authentic_plaintext, authentic_plaintext_len);

  if (!flux_rsa_verify (public_key, signature, signature_len, authentic_plaintext, authentic_plaintext_len))
  {
    result = FALSE;
    if (do_halt)
      g_error ("Authentic message didn't verify!");
  }

  if (flux_rsa_verify (public_key, signature, signature_len, fake_plaintext, fake_plaintext_len))
  {
    result = FALSE;
    if (do_halt)
      g_error ("Fake message verified!");
  }

  g_free (authentic_plaintext);
  g_free (signature);
  g_free (fake_plaintext);
  return result;
}

gboolean
test_import_export (FluxRSAPublicKey *public_key, FluxRSASecretKey *secret_key, gboolean do_halt)
{
  FluxRSAPublicKey *imported_public_key;
  FluxRSASecretKey *imported_secret_key;
  guchar           *buf;
  guint             buf_len;
  gboolean          result = TRUE;

  flux_rsa_export_public_key (public_key, &buf, &buf_len);
  imported_public_key = flux_rsa_import_public_key (buf, buf_len);
  g_free (buf);

  flux_rsa_export_secret_key (secret_key, &buf, &buf_len);
  imported_secret_key = flux_rsa_import_secret_key (buf, buf_len);
  g_free (buf);

  if (!test_privacy      (imported_public_key, secret_key, FALSE) ||
      !test_privacy      (public_key, imported_secret_key, FALSE) ||
      !test_privacy      (imported_public_key, imported_secret_key, FALSE) ||
      !test_authenticity (imported_public_key, secret_key, FALSE) ||
      !test_authenticity (public_key, imported_secret_key, FALSE) ||
      !test_authenticity (imported_public_key, imported_secret_key, FALSE))
  {
    result = FALSE;
    if (do_halt)
      g_error ("Import/export broke key.");
  }

  flux_object_unref (FLUX_OBJECT (imported_public_key));
  flux_object_unref (FLUX_OBJECT (imported_secret_key));
  return result;
}

int
main (int argc, char *argv [])
{
  FluxRSAPublicKey *public_key;
  FluxRSASecretKey *secret_key;

  rng = flux_rng_new ();

  for (;;)
  {
    gint i;

    g_print ("Generating key... ");
    fflush (stdout);

    if (!flux_rsa_generate_keypair (g_random_int_range (64, 2048), &public_key, &secret_key))
      g_error ("Couldn't generate keypair!");

    g_print ("ok, %d bits, block size %d bytes\n",
             flux_rsa_get_key_length (FLUX_RSA_KEY (public_key)),
             flux_rsa_get_block_size (FLUX_RSA_KEY (public_key)));

    g_print ("Testing en/decryption... ");
    fflush (stdout);
    for (i = 0; i < PRIVACY_ITERATIONS; i++)
      test_privacy (public_key, secret_key, TRUE);
    g_print ("ok\n");

    g_print ("Testing sign/verify... ");
    fflush (stdout);
    for (i = 0; i < AUTHENTICITY_ITERATIONS; i++)
      test_authenticity (public_key, secret_key, TRUE);
    g_print ("ok\n");

    g_print ("Testing import/export... ");
    fflush (stdout);
    test_import_export (public_key, secret_key, TRUE);
    g_print ("ok\n");

    flux_object_unref (FLUX_OBJECT (public_key));
    flux_object_unref (FLUX_OBJECT (secret_key));
  }

  return 0;
}
