/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */

#define __FLUX_BUILD

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "flux-types.h"
#include "flux-tree.h"

#define ITER   100000000
#define DATA_SIZE_MAX 64

typedef FluxTree *(*TestFunc)(FluxTree *);

FluxTree *
set_data_random (FluxTree *tree)
{
  gint   len;
  gchar *buf;
  gint   i;

  len = random () % (DATA_SIZE_MAX - 1) + 1;
  buf = g_malloc (len);

  for (i = 0; i < len; i++)
  {
    buf [i] = random ();
  }

  flux_tree_data_set (tree, buf, len);
  g_free (buf);
  return tree;
}

FluxTree *
set_data_random_str (FluxTree *tree)
{
  gchar *string;
  gint   len;
  gint   i;

  len = random () % (DATA_SIZE_MAX - 1) + 1;
  string = g_malloc (len + 1);

  for (i = 0; i < len; i++)
  {
    string [i] = 'A';
  }

  string [i] = '\0';
  flux_tree_data_set_str (tree, string);
  g_free (string);
  return tree;
}

FluxTree *
prepend_data_random (FluxTree *tree)
{
  gint   len;
  gchar *buf;
  gint   i;

  len = random () % (DATA_SIZE_MAX - 1) + 1;
  buf = g_malloc (len);

  for (i = 0; i < len; i++)
  {
    buf [i] = random ();
  }

  flux_tree_data_prepend_bytes (tree, buf, len);
  g_free (buf);
  return tree;
}

FluxTree *
append_data_random (FluxTree *tree)
{
  gint   len;
  gchar *buf;
  gint   i;

  len = random () % (DATA_SIZE_MAX - 1) + 1;
  buf = g_malloc (len);

  for (i = 0; i < len; i++)
  {
    buf [i] = random ();
  }

  flux_tree_data_append_bytes (tree, buf, len);
  g_free (buf);
  return tree;
}

FluxTree *
set_data_length (FluxTree *tree)
{
  gint len;

  len = random () % (DATA_SIZE_MAX - 1);
  flux_tree_data_set_size (tree, len);
  return tree;
}

FluxTree *
get_next (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_get_next (tree);
  if (other)
    return other;
  else
    return tree;
}

FluxTree *
get_prev (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_get_prev (tree);
  if (other)
    return other;
  else
    return tree;
}

FluxTree *
get_first_child (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_get_first_child (tree);
  if (other)
    return other;
  else
    return tree;
}

FluxTree *
get_last_child (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_get_last_child (tree);
  if (other)
    return other;
  else
    return tree;
}

FluxTree *
get_parent (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_get_parent (tree);
  if (other)
    return other;
  else
    return tree;
}

FluxTree *
new_first_child (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_new ();
  flux_tree_add_as_first_child (tree, other);
  return tree;
}

FluxTree *
new_last_child (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_new ();
  flux_tree_add_as_last_child (tree, other);
  return tree;
}

FluxTree *
new_first_sibling (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_new ();
  flux_tree_add_as_first_sibling (tree, other);
  return tree;
}

FluxTree *
new_last_sibling (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_new ();
  flux_tree_add_as_last_sibling (tree, other);
  return tree;
}

FluxTree *
destroy_branch (FluxTree *tree)
{
  FluxTree *other;

  other = flux_tree_get_first_sibling (tree);
  if (!other || other == tree)
    other = flux_tree_get_last_sibling (tree);
  if (!other || other == tree)
    other = flux_tree_get_parent (tree);
  if (!other || other == tree)
    return tree;

  flux_tree_destroy (tree);
  return other;
}

TestFunc funcs [] =
{
  new_first_child,
  new_last_child,
  new_first_sibling,
  new_last_sibling,
  destroy_branch,
  get_next,
  get_prev,
  get_first_child,
  get_last_child,
  get_parent,
  set_data_length,
  set_data_random,
  set_data_random_str,
  prepend_data_random,
  append_data_random,
  NULL
};

void
print_scan_test (FluxTree *tree)
{
  FILE *file;

  file = fopen ("test.tree.1", "w");
  flux_tree_print_to_file (tree, file, FLUX_PRINT_ALLMAN);
  fclose (file);

  flux_tree_destroy (tree);

  file = fopen ("test.tree.1", "r");
  tree = flux_tree_scan_from_file (file);
  fclose (file);

  if (tree)
  {
    g_print ("Print/scan test succeeded.\n");
    file = fopen ("test.tree.2", "w");
    flux_tree_print_to_file (tree, file, FLUX_PRINT_ALLMAN);
    fclose (file);
  }
  else
  {
    g_print ("Print/scan test failed.\n");
  }
}

int
main (int argc, char *argv [])
{
  FluxTree *tree;
  gint      test_func_max;
  gint      i;

  srandom (time (NULL));
  tree = flux_tree_new ();

  for (test_func_max = 0; funcs [test_func_max]; test_func_max++)
    ;

  g_print ("Have %d test functions.\n", test_func_max);

  for (i = 0; i < ITER; i++)
  {
    TestFunc action;

    action = funcs [random () % test_func_max];
    tree = action (tree);
  }

  tree = flux_tree_get_root (tree);
  print_scan_test (tree);

#if 0
  flux_tree_print_to_file (tree, stdout, FLUX_PRINT_ALLMAN);
#endif
  return 0;
}
